/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  ObjectTreeNode.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2012.
 *
 *  Description: Tree nodes of the shape and the group trees.
 *
 *  -------------------------------------------------------------------------*/

#include "ObjectTreeNode.h"

#include <algorithm>
#include <list>

#include "ObjectWrapper.h"
#include "ProgressReporter.h"
#include "R3SurfelClassifier.h"
#include "UnionFind.h"

using namespace std;

namespace igra {

extern int print_verbose; // = 0

////////////////////////////////////////////////////////////////////////
// ShapeTreeNode class implementation
////////////////////////////////////////////////////////////////////////


ShapeTreeNode::
ShapeTreeNode(ShapeTreeNode *parent,
		list<ShapeTreeNode*> *children,
    ShapeTreeNode *vertex1, ShapeTreeNode *vertex2,
    ObjectWrapper *object,
    RNScalar edge_length, int size) :
    		ShapeTreeTraitsType(parent, children, object, edge_length, size),
        vertex_(vertex1, vertex2),
        avr_subtree_link_cost_(edge_length),
        voted_valid_(true),
        voted_in_tree_()
{
  set_avr_link_cost();
}

ShapeTreeNode::
ShapeTreeNode(const ShapeTreeNode &o) :
				ShapeTreeTraitsType(o),
        vertex_(o.vertex_),
        avr_subtree_link_cost_(o.avr_subtree_link_cost_),
        voted_valid_(o.voted_valid_),
        voted_in_tree_(o.voted_in_tree_)
{
	set_avr_link_cost();
}


ShapeTreeNode::
~ShapeTreeNode() {
}


void ShapeTreeNode::
reset(ShapeTreeNode *parent, ShapeTreeNode *child1, ShapeTreeNode *child2,
    ShapeTreeNode *vertex1, ShapeTreeNode *vertex2, ObjectWrapper *object,
    RNScalar edge_length) {

	list<ShapeTreeNode*> tmp;
	if (child1) tmp.push_back(child1);
	if (child2) tmp.push_back(child2);
	ShapeTreeTraitsType::reset(parent, &tmp, object, edge_length);
  vertex_.first = vertex1;
  vertex_.second = vertex2;
  set_avr_link_cost();
}


void ShapeTreeNode::
FillSpecificStats(list<string> &data) {
	// Useful variables.
	char buffer[1024];

	// Fill in stats pertaining to the node.
	sprintf(buffer, "  ---Shape node--");
	data.push_back(buffer);
	sprintf(buffer, "   Linking vertices:");
	data.push_back(buffer);
	sprintf(buffer, "    1: %d",
			vertex_.first ? vertex_.first->object()->object()->SceneIndex() : -1);
	data.push_back(buffer);
	sprintf(buffer, "    2: %d",
			vertex_.second ? vertex_.second->object()->object()->SceneIndex() : -1);
	data.push_back(buffer);
	sprintf(buffer, "   average subtree length: %g", avr_subtree_link_cost_);
	data.push_back(buffer);
	sprintf(buffer, "   voted stats validated: %s",
			voted_valid_ ? "true" : "false");
	data.push_back(buffer);
	sprintf(buffer, "   Labeled in tree: %zu", voted_in_tree_.size());
	data.push_back(buffer);
}


RNScalar ShapeTreeNode::
non_zero_link_cost() const {
  ShapeTreeNode *node = (ShapeTreeNode*) this;
  while (node) {
    if (node->node_weight_ > 0.) return node->node_weight_;
    node = node->parent_;
  }
  return 1.;
}


void ShapeTreeNode::
set_avr_link_cost() {
  avr_subtree_link_cost_ = node_weight_;
  int denom = 1;
	for (ShapeTreeNode::ChildrenIterator chit = children_.begin();
			chit != children_.end(); ++chit) {
		ShapeTreeNode *child = *chit;
    int num_links = child->size_ - 1;
    denom += num_links;
    avr_subtree_link_cost_ += num_links * child->avr_subtree_link_cost_;
	}

  avr_subtree_link_cost_ /= denom;
}


////////////////////////////////////////////////////////////////////////
// GroupTreeNode class definition
////////////////////////////////////////////////////////////////////////


GTNTuple::
GTNTuple(GroupTreeNode *n1, GroupTreeNode *n2) :
        i(n1->object()->index()), j(n2->object()->index()),
        ufpi(NULL), ufpj(NULL),
        gtni(n1), gtnj(n2),
        p_sh(0.0), p_rp(0.0), p_d(0.0), p_patt(0.0),
        dist(RN_INFINITY), p_lab(1.0) {
}


GroupTreeNode::
GroupTreeNode(R3SurfelClassifier *classifier,
		ObjectWrapper *object, GroupTreeNode *parent) :
    	GroupTreeTraitsType (parent, NULL, object, 0, 0),
    	classifier_(classifier),
    	rp_mst_(),
    	p_label_(1.0),
    	t_singular_(0.0),
    	t_group_(0.0),
    	valid_(false)
{
	assert(object);
	size_ = 1;
	set<ObjectWrapper*> group;
	group.insert(object_);
}


GroupTreeNode::
GroupTreeNode(R3SurfelClassifier *classifier, GroupTreeNode *parent) :
		GroupTreeTraitsType(parent, NULL, NULL, 0, 0),
		classifier_(classifier),
		rp_mst_(),
		p_label_(1.0),
		t_singular_(0.0),
		t_group_(0.0),
		valid_(false)
{
}


GroupTreeNode::
GroupTreeNode(const GroupTreeNode &orig) :
			GroupTreeTraitsType(orig),
			classifier_(orig.classifier_),
			rp_mst_(orig.rp_mst_),
			p_label_(orig.p_label_),
			t_singular_(orig.t_singular_),
			t_group_(orig.t_group_),
			valid_(orig.valid_)
{
}


GroupTreeNode::
~GroupTreeNode() {
}


void GroupTreeNode::FillSpecificStats(list<string> &data) {
	// Useful variables.
	char buffer[1024];

	// Fill in stats pertaining to the node.
	data.push_back("  ---Group node--");
	data.push_back("");

	sprintf(buffer, "    Benefit = %g", benefit());
	data.push_back(buffer);
	data.push_back("     Benefit = p_lab * T_single - T_group");
	data.push_back("");

	sprintf(buffer, "    T_singular = %g", t_singular());
	data.push_back(buffer);
	sprintf(buffer, "    T_group = %g", t_group());
	data.push_back(buffer);
	sprintf(buffer, "    P_label = %g", p_label());
	data.push_back(buffer);
	sprintf(buffer, "    *Valid: %c", valid_ ? '+' : '-');
	data.push_back(buffer);
}


void GroupTreeNode::
add_child(GroupTreeNode *child) {
	if (!child) return;
	GroupTreeTraitsType::add_child(child);
	invalidate();
}


void GroupTreeNode::
remove_child(GroupTreeNode *child) {
	if (!child) return;
	GroupTreeTraitsType::remove_child(child);
	invalidate();
}


void GroupTreeNode::
print_group(FILE *fp) {
	if (!fp) return;

	fprintf(fp, "\t\t\tSize: %d\n\t\t\t\t", size_);
	set<ObjectWrapper*> own_group = GetLeaves();
	for (set<ObjectWrapper*>::const_iterator it = own_group.begin();
			it != own_group.end(); ++it) {
		fprintf(fp, "%d\t", (*it)->index());
	}
	fprintf(fp, "\n");
}


void GroupTreeNode::
print_group_with_eval(FILE *fp) {
	if (!fp) return;

	fprintf(fp, "\t\t\tBenefit: %lf\n", benefit());
	fprintf(fp, "\t\t\tT_singular: %lf\n", t_singular());
	fprintf(fp, "\t\t\tT_group: %lf\n", t_group());
	fprintf(fp, "\t\t\tP_label: %lf\n", p_label());
	fprintf(fp, "\t\t\tValid: %c\n", valid_?'+': '-');

	print_group(fp);
}


void GroupTreeNode::
invalidate() {
	valid_ = false;
	if (parent_) parent_->invalidate();
}


void GroupTreeNode::
validate() {
	const RNScalar single_t = classifier_->LabelChoiceTime();

	// Get objects in the group.
	set<ObjectWrapper*> own_group_set = GetLeaves();
	assert(own_group_set.size() == (size_t )size_);

	// Compute t_singular.
	t_singular_ = 0.0;
	for (set<ObjectWrapper*>::const_iterator it = own_group_set.begin();
			it != own_group_set.end(); ++it) {
		ObjectWrapper *ow = *it;
		if (!ow || !ow->object()) continue;
		if (!ow->labeled()) t_singular_+= single_t;
	}
	int temp = 0;

	// Compute t_group_ and p_label_.
	if ((size_t) size_ - 1 != rp_mst_.size())
	        PopulateSupportData(own_group_set, temp);
	t_group_ = 0.0;
	for (int i = 0; i < size_ - 1; ++i) {
		const GTNTuple& t = rp_mst_[i];
		// All other decisions are binary: same label or not.
		// Regular pattern and proximity can make it easier or harder.
		t_group_ += t.TGroupD();
	}
	t_group_ *= t_g_marginal_alpha;

    // There is always label identification present in the decision making.
    if (size_)
        t_group_ += classifier_->LabelChoiceTime();

    // Validation complete.
	valid_ = true;
}


set<ObjectWrapper*> GroupTreeNode::
GetLeaves() const {
	set<ObjectWrapper*> group;

	RNArray<GroupTreeNode*> stack;
	stack.InsertTail((GroupTreeNode*) this);
	while(stack.NEntries()) {
		GroupTreeNode *node = stack.Tail();
		stack.RemoveTail();

		if (node->is_leaf()) {
			group.insert(node->object());
		} else {
			for (GroupTreeNode::ChildrenConstIterator chit = node->children_.begin();
					chit != node->children_.end(); ++chit) {
				stack.InsertTail(*chit);
			}
		}
	}

	return group;
}


struct Accelerator{
	RNScalar p_lab, Ts, Tg;
	RNScalar b() const { return p_lab * Ts - Tg;}
};


void GroupTreeNode::
GrowForestGlobally(R3SurfelClassifier *classifier,
		set<GroupTreeNode*> &forest) {
	//Useful variables.
	const int N = forest.size();
	if (N < 2) return;

	// Useful constants.
	const RNScalar single_t = classifier->LabelChoiceTime();
	// p_lab smaller than threshold1 will definitely not form a beneficial group
	const RNScalar threshold1 = 1.0 / single_t;

	// Useful variables.
	set<GroupTreeNode*>::const_iterator it1, it2;
	set<GroupTreeNode*>::const_iterator f_begin = forest.begin();
	set<GroupTreeNode*>::const_iterator f_end = forest.end();
	UnionFind<int> gaze_tree;
	map<GroupTreeNode*, Accelerator> gtn_to_accl;
	vector<void*> tmp_gaze_ptrs(forest.size(), NULL);

	// Initialize.
	for (it1 = f_begin; it1 != f_end; ++it1) {
		GroupTreeNode *node = *it1;
		assert(node->is_leaf());
		Accelerator &acc = gtn_to_accl[node];
		acc.p_lab = 1.0;
		acc.Ts = node->object()->labeled() ? 0.0 : single_t;
		acc.Tg = single_t;

		int i = node->object()->index();
		tmp_gaze_ptrs[i] = gaze_tree.add(i);
	}

	ProgressReporter progress("Producing tree", print_verbose);

	// Produce all pair tuples.
    list<GTNTuple> q;
    for (it1 = f_begin; it1 != f_end; ++it1) {
        ObjectWrapper *owi = (*it1)->object();
        int i = owi->index();
        void *ufpi = tmp_gaze_ptrs[i];

        for (it2 = f_begin; it2 != it1; ++it2) {
            ObjectWrapper *owj = (*it2)->object();
            int j = owj->index();
            void *ufpj = tmp_gaze_ptrs[j];

            RNScalar p_sh = classifier->GetInducedAffinity(i, j);
            if (RNIsLessOrEqual(p_sh, threshold1))
                continue;

            GTNTuple t = classifier->GetBestRPinSet(owi, owj);
            t.ufpi = ufpi;
            t.ufpj = ufpj;
            t.p_sh = p_sh;
            t.p_lab = p_sh;
            q.push_back(t);
        }
    }

    // Have tuples sorted in most promising to least promising order.
    //q.sort(GTNTuple::CmpPat1);
    q.sort();
    q.reverse();

	// Group until not groupable.
	while (gaze_tree.size() > 1 && !q.empty()) {
#if 1
		list<GTNTuple>::iterator it = q.begin();
		list<GTNTuple>::const_iterator q_end = q.end();
		//printf("\nq.size = %zu, gaze_tree.size() = %d", q.size(), gaze_tree.size());
		while(it != q_end) {
			GTNTuple &t = *it;

			// Remove tuple if nodes already joined through a different path.
			void *ptr_i = gaze_tree.find(t.ufpi);
			void *ptr_j = gaze_tree.find(t.ufpj);
			if (ptr_i == ptr_j) {
				it = q.erase(it);
				continue;
			}

			// Pre-evaluate using approximation.
			GroupTreeNode *gtni = t.gtni->root();
			GroupTreeNode *gtnj = t.gtnj->root();

			Accelerator acci = gtn_to_accl[gtni];
			Accelerator accj = gtn_to_accl[gtnj];
			Accelerator acc;
			acc.p_lab = min(t.p_lab, min(acci.p_lab, accj.p_lab));
            acc.Ts = acci.Ts + accj.Ts;
            acc.Tg = acci.Tg + accj.Tg + t.TGroupD() - single_t;
			if (acc.b() < max(acci.b(), accj.b())) {
				++it;
				continue;
			}

			// If promising - join.
			gaze_tree.merge(t.ufpi, t.ufpj);
			GroupTreeNode *newGTN = new GroupTreeNode(classifier);
			newGTN->add_child(gtni);
			newGTN->add_child(gtnj);
			newGTN->link_children();

			forest.insert(newGTN);
			forest.erase(gtni);
			forest.erase(gtnj);

			// Update accelerator information.
			gtn_to_accl.erase(gtni);
			gtn_to_accl.erase(gtnj);
			gtn_to_accl[newGTN] = acc;

			it = q.erase(it);

			break;
		}

		// If nothing was joined => we're done.
		if (it == q_end)
			break;

#else
		// The only difference from above is that connections are discarded
		// if they don't create a good link the first time, so they don't
		// have a chance of being reused in the future linking.

		// Pop from priority queue.
		GTNTuple t = q.back();
		q.pop_back();

		// Check if not already in the same tree.
		void *ptr_i = gaze_tree.find(t.i);
		void *ptr_j = gaze_tree.find(t.j);
		if (ptr_i == ptr_j) continue;

		// Pre-evaluate using approximation.
		GroupTreeNode *gtni = ind_to_gtn[t.i]->root();
		GroupTreeNode *gtnj = ind_to_gtn[t.j]->root();

		Accelerator acci = gtn_to_accl[gtni];
		Accelerator accj = gtn_to_accl[gtnj];
		Accelerator acc;
		acc.p_lab = min(t.p_lab, min(acci.p_lab, accj.p_lab));
		acc.Ts = acci.Ts + accj.Ts;
		acc.Tg = acci.Tg + accj.Tg + t.TGroupD() - single_t;
		if (acc.b() < max(acci.b(), accj.b())) continue;

		// If promising - join.
		gaze_tree.merge(t.i, t.j);
		GroupTreeNode *newGTN = new GroupTreeNode(classifier);
		newGTN->add_child(gtni);
		newGTN->add_child(gtnj);
		newGTN->link_children();

		forest.insert(newGTN);
		forest.erase(gtni);
		forest.erase(gtnj);

		// Update accelerator information.
		gtn_to_accl.erase(gtni);
		gtn_to_accl.erase(gtnj);
		gtn_to_accl[newGTN] = acc;
#endif
	}
}


void GroupTreeNode::
GrowForest(R3SurfelClassifier *classifier,
		set<GroupTreeNode*> &forest, GroupTreeNode *seed) {
	if (!seed) return GrowForestGlobally(classifier, forest);

	//Useful variables.
	const int seed_index = seed->object()->index();
	const int N = forest.size();
	if (N < 2) return;

	// Useful constants.
	const RNScalar single_t = classifier->LabelChoiceTime();
	// p_lab smaller than threshold1 will definitely not form a beneficial group
	const RNScalar threshold1 = 1.0 / single_t;

	// Useful variables.
	set<GroupTreeNode*>::const_iterator it1, it2;
	set<GroupTreeNode*>::const_iterator f_begin = forest.begin();
	set<GroupTreeNode*>::const_iterator f_end = forest.end();
	UnionFind<int> gaze_tree;
	vector<void*> tmp_gaze_ptrs(forest.size(), NULL);
	map<GroupTreeNode*, Accelerator> gtn_to_accl;
	for (it1 = f_begin; it1 != f_end; ++it1) {
		GroupTreeNode *node = *it1;
		assert(node->is_leaf());
		Accelerator &acc = gtn_to_accl[node];
		acc.p_lab = 1.0;
		acc.Ts = node->object()->labeled() ? 0.0 : single_t;
		acc.Tg = single_t;

        int i = node->object()->index();
        tmp_gaze_ptrs[i] = gaze_tree.add(i);
	}

    ProgressReporter progress("Producing tree", print_verbose);

    // Produce all pair tuples.
    list<GTNTuple> q;
    for (it1 = f_begin; it1 != f_end; ++it1) {
        ObjectWrapper *owi = (*it1)->object();
        int i = owi->index();
        void *ufpi = tmp_gaze_ptrs[i];

        for (it2 = f_begin; it2 != it1; ++it2) {
            ObjectWrapper *owj = (*it2)->object();
            int j = owj->index();
            void *ufpj = tmp_gaze_ptrs[j];

            RNScalar p_sh = classifier->GetInducedAffinity(i, j);
            if (RNIsLessOrEqual(p_sh, threshold1))
                continue;

            GTNTuple t = classifier->GetBestRPinSet(owi, owj);
            t.ufpi = ufpi;
            t.ufpj = ufpj;
            t.p_sh = p_sh;
            t.p_lab = p_sh;
            q.push_back(t);
        }
    }

	// Have tuples sorted in most promising to least promising order.
	//q.sort(GTNTuple::CmpPat1);
    q.sort();
	q.reverse();

	// Group until not groupable.
	while (gaze_tree.size() > 1 && !q.empty()) {
#if 1
		list<GTNTuple>::iterator it = q.begin();
		list<GTNTuple>::const_iterator q_end = q.end();
		while(it != q_end) {
			GTNTuple &t = *it;

			// Remove tuple if nodes already joined through a different path.
			void *ptr_i = gaze_tree.find(t.ufpi);
			void *ptr_j = gaze_tree.find(t.ufpj);
			if (ptr_i == ptr_j) {
				it = q.erase(it);
				continue;
			}

			// Only join around seed tree.
			void *ptr_seed = gaze_tree.find(seed_index);
			if (ptr_i != ptr_seed && ptr_j != ptr_seed) {
				++it;
				continue;
			}

			// Pre-evaluate using approximation.
			GroupTreeNode *gtni = t.gtni->root();
			GroupTreeNode *gtnj = t.gtnj->root();

			Accelerator acci = gtn_to_accl[gtni];
			Accelerator accj = gtn_to_accl[gtnj];
			Accelerator acc;
			acc.p_lab = min(t.p_lab, min(acci.p_lab, accj.p_lab));
            acc.Ts = acci.Ts + accj.Ts;
            acc.Tg = acci.Tg + accj.Tg + t.TGroupD() - single_t;
			if (acc.b() < max(acci.b(), accj.b())) {
				++it;
				continue;
			}

			// If promising - join.
			gaze_tree.merge(t.ufpi, t.ufpj);
			GroupTreeNode *newGTN = new GroupTreeNode(classifier);
			newGTN->add_child(gtni);
			newGTN->add_child(gtnj);
			newGTN->link_children();

			forest.insert(newGTN);
			forest.erase(gtni);
			forest.erase(gtnj);

			// Update accelerator information.
			gtn_to_accl.erase(gtni);
			gtn_to_accl.erase(gtnj);
			gtn_to_accl[newGTN] = acc;

			break;
		}

		// If nothing was joined => we're done.
		if (it == q_end)
			break;

#else
		// The only difference from above is that connections are discarded
		// if they don't create a good link the first time, so they don't
		// have a chance of being reused in the future linking.

		// Pop from priority queue.
		GTNTuple t = q.back();
		q.pop_back();

		// Check if not already in the same tree.
		void *ptr_i = gaze_tree.find(t.i);
		void *ptr_j = gaze_tree.find(t.j);
		if (ptr_i == ptr_j) continue;

		// Pre-evaluate using approximation.
		GroupTreeNode *gtni = ind_to_gtn[t.i]->root();
		GroupTreeNode *gtnj = ind_to_gtn[t.j]->root();

		Accelerator acci = gtn_to_accl[gtni];
		Accelerator accj = gtn_to_accl[gtnj];
		Accelerator acc;
		acc.p_lab = min(t.p_lab, min(acci.p_lab, accj.p_lab));
		acc.Ts = acci.Ts + accj.Ts;
		acc.Tg = acci.Tg + accj.Tg + t.TGroupD() - single_t;
		if (acc.b() < max(acci.b(), accj.b())) continue;

		// If promising - join.
		gaze_tree.merge(t.i, t.j);
		GroupTreeNode *newGTN = new GroupTreeNode(classifier);
		newGTN->add_child(gtni);
		newGTN->add_child(gtnj);
		newGTN->link_children();

		forest.insert(newGTN);
		forest.erase(gtni);
		forest.erase(gtnj);

		// Update accelerator information.
		gtn_to_accl.erase(gtni);
		gtn_to_accl.erase(gtnj);
		gtn_to_accl[newGTN] = acc;
#endif
	}
}


void GroupTreeNode::
GreedyTreeAroundSeed(R3SurfelClassifier *classifier,
		GroupTreeNode *seed, set<GroupTreeNode*> &forest) {
	GroupTreeNode *root = seed;
	forest.erase(seed);

	RNTime timer; timer.Read();
	const RNScalar time_limit = 10;

	// Group until not groupable.
	do {
		GroupTreeNode *best_next = NULL;
		GroupTreeNode *new_root = NULL;

		for (set<GroupTreeNode*>::const_iterator it = forest.begin();
				it != forest.end(); ++it) {
			GroupTreeNode *cand = *it;

			GroupTreeNode *temp = new GroupTreeNode(classifier);
			temp->add_child(root);
			temp->add_child(cand);
			temp->link_children();

			if (temp->benefit() > 0.0
					&& (!new_root
							|| new_root->benefit() < temp->benefit())) {
				best_next = cand;
				if (new_root) delete new_root;
				new_root = temp;
			} else {
				delete temp;
			}
		}

		if (best_next) {

			forest.erase(best_next);

			root = new_root;

			root->link_children();
		}
	} while (forest.size() && root->benefit() > 0.0
			&& timer.Elapsed() < time_limit);

	forest.clear();
	forest.insert(root);
}


void GroupTreeNode::
PopulateSupportData(const set<ObjectWrapper*> &g, int &temp) {
	// Check everything.
	assert(size_ > 0);

	// Clear everything first.
	rp_mst_.clear();
	invalidate();
	const int N = g.size();
	if (N < 2) return;
	rp_mst_.reserve(N - 1);

	// Useful variables.
	set<ObjectWrapper*>::const_iterator it1, it2, it3;
	set<ObjectWrapper*>::const_iterator g_begin = g.begin();
	set<ObjectWrapper*>::const_iterator g_end = g.end();
	set<int> group;
    map<int, void*> tmp_gaze_ptrs;
	UnionFind<int> gaze_tree;
	for (it1 = g_begin; it1 != g_end; ++it1) {
		int i = (*it1)->index();
		group.insert(i);
		tmp_gaze_ptrs[i] = gaze_tree.add(i);
    }

    // Produce MST construction priority queue.
    list<GTNTuple> q;
    for (it1 = g_begin; it1 != g_end; ++it1) {
        ObjectWrapper *owi = *it1;
        int i = owi->index();
        void *ufpi = tmp_gaze_ptrs[i];

        for (it2 = g_begin; it2 != it1; ++it2) {
            ObjectWrapper *owj = *it2;
            int j = owj->index();
            void *ufpj = tmp_gaze_ptrs[j];

            GTNTuple t = classifier_->GetBestRPinSet(owi, owj, group);
            t.p_sh = classifier_->GetInducedAffinity(i, j);
            t.ufpi = ufpi;
            t.ufpj = ufpj;
            q.push_back(t);
        }
    }

    // Build RP+distance based MST.
    q.sort(GTNTuple::CmpPat);
    for (list<GTNTuple>::reverse_iterator rit = q.rbegin();
            gaze_tree.size() > 1 && rit != q.rend(); ++rit) {
        GTNTuple t = *rit;
        if (gaze_tree.merge(t.ufpi, t.ufpj))
            rp_mst_.push_back(t);
    }

    temp++;

	// Update p_label_.
    p_label_ = 1.0;
#if 1
    // Build induced shape affinity based MST.
    q.sort(GTNTuple::CmpSh);
    gaze_tree.unlink();
    for (list<GTNTuple>::reverse_iterator rit = q.rbegin();
            gaze_tree.size() > 1 && rit != q.rend(); ++rit) {
        GTNTuple t = *rit;
        if (gaze_tree.merge(t.ufpi, t.ufpj))
            p_label_ *= t.p_sh;
    }
#else
    for (int i = 0; i < size_ - 1; ++i) {
        const GTNTuple& t = rp_mst_[i];
        //p_label_ = min(p_label_, t.p_sh);
        p_label_ *= t.p_sh;
    }
#endif

    temp++;
}


#include "DrawText.cpp"


void GroupTreeNode::
DrawDebug(bool show_more) {
	if (!valid_) validate();

	glDisable(GL_DEPTH_TEST);

	// Draw MST.
	glColor3d(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
	for (int i = 0; (size_t) i < rp_mst_.size(); i++) {
		const GTNTuple& t = rp_mst_[i];

		ObjectWrapper *owi = classifier_->Object(t.i);
		ObjectWrapper *owj = classifier_->Object(t.j);

		R3LoadPoint(owi->object()->Centroid());
		R3LoadPoint(owj->object()->Centroid());
	}
	glEnd();

	// Draw edge costs.
	for (int i = 0; (size_t)i < rp_mst_.size(); i++) {
		const GTNTuple& t = rp_mst_[i];

		ObjectWrapper *owi = classifier_->Object(t.i);
		ObjectWrapper *owj = classifier_->Object(t.j);
		if (!owi || !owj) continue;
		R3Point c = 0.5 * (owi->object()->Centroid() + owj->object()->Centroid());
		c[2] += 2;

		char score[50];
		if (show_more)
            sprintf(score,
                    "tg= %.2f\n"
                    "pat=%.2f=%.2f(rp)*%.2f(p_d)\n"
                    "sh=%.2f, dist = %.2f",
                    t_g_marginal_alpha * t.TGroupD(),
                    t.p_patt, t.p_rp, t.p_d,
                    t.p_sh, t.dist);
		else
			sprintf(score, "%.2f", t.p_patt);

		igra::DrawText(c, score);
		igra::DrawText(owi->object()->Centroid(), "O");
		igra::DrawText(owj->object()->Centroid(), "X");
	}

	glEnable(GL_DEPTH_TEST);
}


RNScalar GTNTuple::
THicksBinary(RNScalar p) {
	RNScalar t = 0.0;
	if (RNIsPositive(p)) t += p * log2(1 + 1 / p);
	RNScalar not_p = 1.0 - p;
	if (RNIsPositive(not_p)) t += not_p * log2(1 + 1 / not_p);
	return t;
}

} //namespace igra
