/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  ObjectTreeNode.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2012.
 *
 *  Description: Tree nodes of the shape and the group trees.
 *
 *  -------------------------------------------------------------------------*/


#ifndef OBJECTTREENODE_H_
#define OBJECTTREENODE_H_

#include "R2Shapes/R2Shapes.h"
#include "TreeTraits.h"

#include <map>
#include <set>
#include <vector>

namespace igra {

extern RNScalar t_g_marginal_alpha; // = 1.0;

///////////////////////////////////////////////
/// ObjectWrapper tree node class
///////////////////////////////////////////////

class ObjectWrapper;

template<class Node>
class ObjectTreeNode: public TreeTraits<Node, ObjectWrapper> {
};

///////////////////////////////////////////////
/// Shape tree node class
///////////////////////////////////////////////

class ShapeTreeNode: public TreeTraits<ShapeTreeNode, ObjectWrapper> {
public:
    // Constructor/destructor.
    explicit ShapeTreeNode(ShapeTreeNode *parent = NULL,
            list<ShapeTreeNode*> *children = NULL, ShapeTreeNode *leaf1 = NULL,
            ShapeTreeNode *leaf2 = NULL, ObjectWrapper *object = NULL,
            RNScalar link_cost = 0, int tree_size = 0);
    explicit ShapeTreeNode(const ShapeTreeNode &o);
    ~ShapeTreeNode();

    // Reset node.
    virtual void reset(ShapeTreeNode *parent, ShapeTreeNode *child1,
            ShapeTreeNode *child2, ShapeTreeNode *vertex1,
            ShapeTreeNode *vertex2, ObjectWrapper *object, RNScalar link_cost);

    // Node properties
    RNScalar non_zero_link_cost() const;

protected:
    typedef TreeTraits<ShapeTreeNode, ObjectWrapper> ShapeTreeTraitsType;
    void set_avr_link_cost();

    // Debug information
    virtual void FillSpecificStats(list<string> &data);

public:
    // Node's feature-spatial properties.
    pair<ShapeTreeNode*, ShapeTreeNode*> vertex_; // leaf nodes linked here
    RNScalar avr_subtree_link_cost_;

    // Optimization fields.
    bool voted_valid_;
    set<ObjectWrapper*> voted_in_tree_;
};

///////////////////////////////////////////////
/// Group tree node class
///////////////////////////////////////////////

class R3SurfelClassifier;

class GroupTreeNode;

struct GTNTuple {
    int i, j;
    void *ufpi, *ufpj;
    GroupTreeNode *gtni, *gtnj;
    RNScalar p_sh, p_rp, p_d, p_patt, dist;
    RNScalar p_lab;

    GTNTuple(GroupTreeNode *n1, GroupTreeNode *n2);

    inline void print(FILE *file = stdout) const {
        fprintf(file,
                "Tuple: (%d->%d; sh:%f; rp:%f; pd:%f; dist:%f; patt: %f)\n", i,
                j, p_sh, p_rp, p_d, dist, p_patt);
    }

    static RNScalar THicksBinary(RNScalar p);

    inline RNScalar complexity() const {
        return p_patt + (1.0 - p_patt) * dist;
    }

    inline RNScalar TGroupD() const {
        return complexity() * THicksBinary(p_sh);
    }

    inline bool operator<(const GTNTuple &o) const {
        return RNIsEqual(p_patt, o.p_patt) ? dist > o.dist : p_patt < o.p_patt;
        //return RNIsEqual(alpha(), o.alpha()) ?
        //        dist > o.dist : alpha() < o.alpha();

        //return RNIsEqual(dist, o.dist) ? p_patt < o.p_patt : dist > o.dist;
        //return TGroupD() < o.TGroupD();
    }

    static bool CmpPat(const GTNTuple &a, const GTNTuple &b) {
        return RNIsEqual(a.p_patt, b.p_patt) ?
                a.dist > b.dist : a.p_patt < b.p_patt;
    }

    static bool CmpDist(const GTNTuple &a, const GTNTuple &b) {
        return RNIsEqual(a.dist, b.dist) ? a.p_patt < b.p_patt : a.dist > b.dist;
    }

    static bool CmpSh(const GTNTuple &a, const GTNTuple &b) {
        return a.p_sh < b.p_sh;
    }
};

struct GroupTreeNode: public TreeTraits<GroupTreeNode, ObjectWrapper> {
public:
    // Constructor/destructor.
    GroupTreeNode(R3SurfelClassifier *classifier, ObjectWrapper *object,
            GroupTreeNode *parent = NULL);
    GroupTreeNode(R3SurfelClassifier *classifier, GroupTreeNode *parent = NULL);
    GroupTreeNode(const GroupTreeNode &orig);
    ~GroupTreeNode();

    static void GrowForestGlobally(R3SurfelClassifier *classifier,
            set<GroupTreeNode*> &forest);
    static void GrowForest(R3SurfelClassifier *classifier,
            set<GroupTreeNode*> &forest, GroupTreeNode *seed = NULL);
    static void GreedyTreeAroundSeed(R3SurfelClassifier *classifier,
            GroupTreeNode *seed, set<GroupTreeNode*> &forest);

    set<ObjectWrapper*> GetLeaves() const;

    // Accessors.
    RNScalar p_label(void);
    RNScalar t_singular(void);
    RNScalar t_group(void);
    bool valid(void) const;

    // Benefit estimates.
    RNScalar benefit(void);

    // Print group contents.
    void print_group(FILE *fp = stdout);
    void print_group_with_eval(FILE *fp = stdout);

    // Mutators.

    // Children manipulations.
    virtual void add_child(GroupTreeNode *child);
    virtual void remove_child(GroupTreeNode *child);

    // Invalidate precomputations.
    void invalidate();

    // Drawing.
    void DrawDebug(bool show_more);

protected:
    void PopulateSupportData(const set<ObjectWrapper*> &group, int &temp);

    typedef TreeTraits<GroupTreeNode, ObjectWrapper> GroupTreeTraitsType;

    // Debug information
    virtual void FillSpecificStats(list<string> &data);

    // Lazy computing helper.
    void validate(void);

protected:
    // Classifier pointer.
    R3SurfelClassifier *classifier_;

    // Support data.
    vector<GTNTuple> rp_mst_;

    // Lazily computed properties.
    RNScalar p_label_;
    RNScalar t_singular_;
    RNScalar t_group_;

    bool valid_;
};

////////////////////////////////////////////////////////////////
// Inline methods.
////////////////////////////////////////////////////////////////

inline RNScalar GroupTreeNode::p_label(void) {
    if (!valid_) validate();
    return p_label_;
}

inline RNScalar GroupTreeNode::t_singular(void) {
    if (!valid_) validate();
    return t_singular_;
}

inline RNScalar GroupTreeNode::t_group(void) {
    if (!valid_) validate();
    return t_group_;
}

inline RNScalar GroupTreeNode::benefit(void) {
    return p_label() * t_singular() - t_group();
}

inline bool GroupTreeNode::valid(void) const {
    return valid_;
}

} //namespace igra

#endif /* OBJECTTREENODE_H_ */
