/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  ObjectWrapper.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2012.
 *
 *  Description: Convenience wrapper for a R3SurfelObject with relavant
 *               information for classification and group suggestion.
 *
 *  -------------------------------------------------------------------------*/

#include "ObjectWrapper.h"

#include <limits>

#include "ObjectTreeNode.h"
#include "R3SurfelClassifier.h"
#include "R3Surfels/R3Surfels.h"

using namespace std;

namespace igra {

const RNScalar kMaxScalar = numeric_limits<RNScalar>::max();
extern R3SurfelLabel *unknown_label;

ObjectWrapper::ObjectWrapper(R3SurfelObject *object, int index,
        R3SurfelClassifier *classifier) :
                object_(object),
                features_(NULL),
                shp_node_(
                        new ShapeTreeNode(NULL, NULL, NULL, NULL, this, 0, 1)),
                grp_node_(new GroupTreeNode(classifier, this)),
                location_2D_(object_->BBox().XCenter(),
                        object_->BBox().YCenter()),
                box_2D_(object_->BBox().XMin(), object_->BBox().YMin(),
                        object_->BBox().XMax(), object_->BBox().YMax()),
                linear_size_(
                        max(object_->BBox().XLength(),
                                max(object_->BBox().YLength(),
                                        object_->BBox().ZLength()))),
                votes_(),
                best_vote_label_(NULL),
                best_vote_conf_(0.0),
                votes_valid_(false),
                pred_(),
                best_pred_label_(NULL),
                best_pred_conf_(0.0),
                best_pred_aff_(0.0),
                pred_entropy_(0.0),
                pred_valid_(false),
                index_(index) {
}

ObjectWrapper::ObjectWrapper(const ObjectWrapper &orig) :
                object_(orig.object_),
                features_(orig.features_),
                shp_node_(orig.shp_node_),
                grp_node_(orig.grp_node_),
                location_2D_(orig.location_2D_),
                box_2D_(orig.box_2D_),
                linear_size_(orig.linear_size_),
                votes_(orig.votes_),
                best_vote_label_(orig.best_vote_label_),
                best_vote_conf_(orig.best_vote_conf_),
                votes_valid_(orig.votes_valid_),
                pred_(orig.pred_),
                best_pred_label_(orig.best_pred_label_),
                best_pred_conf_(orig.best_pred_conf_),
                best_pred_aff_(orig.best_pred_aff_),
                pred_entropy_(orig.pred_entropy_),
                pred_valid_(orig.pred_valid_),
                index_(orig.index_) {
}

ObjectWrapper::~ObjectWrapper() {
    // Wrapper does not own the object, however owns everything else.
    if (features_) delete features_;
    if (shp_node_) delete shp_node_;
    if (grp_node_) delete grp_node_;
}

const R2Point& ObjectWrapper::location2D(ObjectWrapper *ow, void *data) {
    return ow ? ow->location2D() : R2null_point;
}

void ObjectWrapper::reset_object(R3SurfelObject *object) {
    assert(object);
    object_ = object;
    if (features_) {
        delete features_;
        features_ = NULL;
    }
}

void ObjectWrapper::UnlinkObject() {
    // object_ should be unlinked (before deletion).
    // Preserve the feature vector for the future and reset the object pointer.
    features_ = new R3SurfelFeatureVector(object_->FeatureVector());
    object_ = NULL;
}

const R3SurfelFeatureVector& ObjectWrapper::FeatureVector() const {
    if (object_) return object_->FeatureVector();
    return *features_;
}

void ObjectWrapper::validate_votes() {
    if (votes_valid_) return;

    best_vote_label_ = NULL;
    best_vote_conf_ = 0.0;

    // Confidence is a simple fraction of votes for max votes label to all votes.
    RNScalar total = 0.0;
    VotesType::const_iterator it_end = votes_.end();
    for (VotesType::const_iterator it = votes_.begin(); it != it_end; ++it) {
        RNScalar vote = it->second;
        total += vote;
        if (vote > best_vote_conf_) {
            best_vote_conf_ = vote;
            best_vote_label_ = it->first;
        }
    }

    if (total > 0.0) best_vote_conf_ /= total;

    votes_valid_ = true;
}

void ObjectWrapper::validate_predictions() {
    if (pred_valid_) return;

    // Reset precomp prediction fields.
    best_pred_label_ = unknown_label;
    best_pred_conf_ = 0.0;
    best_pred_aff_ = 0.0;
    pred_entropy_ = 0.0;

    // Compute factors to normalize all contributions to sum to one.
    RNScalar total = 0.0;
    PredType::const_iterator it_end = pred_.end();
    for (PredType::const_iterator it = pred_.begin(); it != it_end; ++it) {
        RNScalar weight = it->second.second;
        total += weight;
        if (weight > best_pred_aff_) {
            best_pred_aff_ = weight;
            best_pred_label_ = it->first;
        }
    }

    // Useful normalization constants.
    const RNScalar weight_factor = (total > 1.0 ? 1.0 / total : 1.0);
    const RNScalar p_unknown = max(0.0, 1.0 - total);

    // Compute entropy.
    for (PredType::const_iterator it = pred_.begin(); it != it_end; ++it) {
        RNScalar p = weight_factor * it->second.second;
        assert(0.0 <= p && p <= 1.0);
        if (p > 0.0) pred_entropy_ -= p * log2(p);
    }
    if (p_unknown > 0.0) pred_entropy_ -= p_unknown * log2(p_unknown);

    // Set the confidence.
    best_pred_conf_ = weight_factor * best_pred_aff_;

    pred_valid_ = true;
}

void ObjectWrapper::cast_vote(R3SurfelLabel* label, RNScalar weight) {
    RNScalar &entry = votes_[label];
    entry += weight;
    if (entry <= 0.0) votes_.erase(label);
    clear_pred();

    votes_valid_ = false;
}

void ObjectWrapper::replace_votes(R3SurfelLabel* label, RNScalar weight) {
    clear_vote();
    cast_vote(label, weight);
}

int ObjectWrapper::try_update_pred(R3SurfelLabel *label, int index,
        RNScalar score) {
    PredType::iterator it = pred_.find(label);

    // Nothing changed? Return 0;
    if (it != pred_.end() && it->second.second >= score) return 0;

    // Need update.
    pair<int, RNScalar> &entry = pred_[label];
    entry.first = index;
    entry.second = score;

    pred_valid_ = false;

    // Return 1 change occurance.
    return 1;
}

void ObjectWrapper::FillStats(list<string> &data) const {
    // Useful variables.
    char buffer[1024];

    // Fill in stats pertaining to the object.
    sprintf(buffer, "-----------------");
    data.push_back(buffer);
    sprintf(buffer, "  Wrapper %d (%p)", index_, this);
    data.push_back(buffer);
    sprintf(buffer, "  Object %p", object_);
    data.push_back(buffer);
    if (object_) {
        sprintf(buffer, "   Index: %d", object_->SceneIndex());
        data.push_back(buffer);
        sprintf(buffer, "   Name: %s", object_->Name());
        data.push_back(buffer);
        data.push_back("");

        data.push_back("   Assignments");
        for (int i = 0; i < object_->NLabelAssignments(); ++i) {
            R3SurfelLabelAssignment *a = object_->LabelAssignment(i);
            string originator;
            switch (a->Originator()) {
                case R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR:
                    originator = "Human";
                    break;
                case R3_SURFEL_LABEL_ASSIGNMENT_MACHINE_ORIGINATOR:
                    originator = "Machine";
                    break;
                case R3_SURFEL_LABEL_ASSIGNMENT_GROUND_TRUTH_ORIGINATOR:
                    originator = "Ground truth";
                    break;
                default:
                    originator = "Unidentified origin";
                    break;
            }
            sprintf(buffer, "    %s: %s (%g)", originator.c_str(),
                    ((a->Label() && a->Label()->Name()) ?
                            a->Label()->Name() : "No name"), a->Confidence());
            data.push_back(buffer);
        }
        data.push_back("");

        data.push_back("   Votes");
        for (map<R3SurfelLabel*, RNScalar>::const_iterator it = votes_.begin();
                it != votes_.end(); ++it) {
            sprintf(buffer, "    %s: %f", it->first->Name(), it->second);
            data.push_back(buffer);
        }
        data.push_back("");

        data.push_back("   Predictions");
        PredType::const_iterator pred_end = pred_.end();
        for (PredType::const_iterator it = pred_.begin(); it != pred_end;
                ++it) {
            sprintf(buffer, "    %s: %f from %d", it->first->Name(),
                    it->second.second, it->second.first);
            data.push_back(buffer);
        }
        data.push_back("");
    }
}

R3SurfelLabel* ObjectWrapper::prev_contribution(int index) const {
    PredType::const_iterator pred_end = pred_.end();
    for (PredType::const_iterator it = pred_.begin(); it != pred_end; ++it) {
        if (it->second.first == index) return it->first;
    }

    return NULL;
}

} // namespace igra
