/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  ObjectWrapper.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2012.
 *
 *  Description: Convenience wrapper for a R3SurfelObject with relavant
 *               information for classification and group suggestion.
 *
 *  -------------------------------------------------------------------------*/


#ifndef OBJECTWRAPPER_H_
#define OBJECTWRAPPER_H_

#include <list>
#include <map>
#include <string>

#include "R3Graphics/R3Graphics.h"
#include "R3Surfels/R3Surfels.h"

using namespace std;

namespace igra {

///////////////////////////////////////////////
/// Forward declarations
///////////////////////////////////////////////

class ShapeTreeNode;
class GroupTreeNode;
class R3SurfelClassifier;

///////////////////////////////////////////////
/// ObjectWrapper
///////////////////////////////////////////////


class ObjectWrapper {
public:
    ObjectWrapper(R3SurfelObject *object, int index,
            R3SurfelClassifier *classifier);
    explicit ObjectWrapper(const ObjectWrapper &orig);
    ~ObjectWrapper();

    // Accessors.
    int index() const;
    R3SurfelObject* object() const;
    ShapeTreeNode* shape_node() const;
    GroupTreeNode* group_node() const;
    static const R2Point& location2D(ObjectWrapper *ow, void *data);
    const R2Box& bbox2D() const;
    const R2Point& location2D() const;
    R2Vector vector2D() const;
    RNLength get_size() const;

    // Mutators.
    void reset_object(R3SurfelObject *object);

    // Feature vector.
    void UnlinkObject();
    const R3SurfelFeatureVector& FeatureVector() const;

    // Classification accessors.
    bool labeled();

    R3SurfelLabel* best_vote_label();
    R3SurfelLabel* best_pred_label();
    R3SurfelLabel* best_asgn_label();

    RNScalar best_vote_conf();
    RNScalar best_pred_conf();
    RNScalar best_asgn_conf();

    RNScalar best_pred_aff();
    RNScalar pred_entropy();

    // Classification mutators.
    void clear_vote();
    void cast_vote(R3SurfelLabel* label, RNScalar weight);
    void replace_votes(R3SurfelLabel* label, RNScalar weight);
    int clear_pred();
    int clear_pred(R3SurfelLabel* label);
    int try_update_pred(R3SurfelLabel *label, int index, RNScalar score);

    // Serialize data into a string.
    void FillStats(list<string> &data) const;

    // Helpers.
    R3SurfelLabel* need_reset_on_increase(ObjectWrapper *contender) const;
    R3SurfelLabel* need_reset_on_decrease(ObjectWrapper *contender) const;

protected:
    // Validators.
    void validate_votes();
    void validate_predictions();

    // Helpers' helpers.
    R3SurfelLabel* prev_contribution(int index) const;

    // Object description.
    R3SurfelObject *object_;
    R3SurfelFeatureVector *features_;

    // Tree structure fields.
    ShapeTreeNode *shp_node_;
    GroupTreeNode *grp_node_;

    // 2D representation.
    R2Point location_2D_;
    R2Box box_2D_;
    RNLength linear_size_;


    // Labeling votes.
    typedef map<R3SurfelLabel*, RNScalar> VotesType;
    VotesType votes_;
    R3SurfelLabel *best_vote_label_;
    RNScalar best_vote_conf_;
    bool votes_valid_;

    // Predictions.
    typedef map<R3SurfelLabel*, pair<int, RNScalar> > PredType;
    PredType pred_;
    R3SurfelLabel *best_pred_label_;
    RNScalar best_pred_conf_;
    RNScalar best_pred_aff_;
    RNScalar pred_entropy_;
    bool pred_valid_;

    // Own index in the array.
    int index_;
};

// Inline methods.


inline int ObjectWrapper::
index() const {
	return index_;
}


inline R3SurfelObject* ObjectWrapper::
object() const {
	return object_;
}


inline ShapeTreeNode* ObjectWrapper::
shape_node() const {
	return shp_node_;
}


inline GroupTreeNode* ObjectWrapper::
group_node() const {
	return grp_node_;
}


inline const R2Box& ObjectWrapper::
bbox2D() const {
	return box_2D_;
}


inline const R2Point& ObjectWrapper::
location2D() const {
	return location_2D_;
}


inline R2Vector ObjectWrapper::
vector2D() const {
	return R2Vector(object_->BBox().XCenter(), object_->BBox().YCenter());
}


inline RNLength ObjectWrapper::
get_size() const {
    return linear_size_;
}


inline bool ObjectWrapper::
labeled() {
	return !votes_.empty() && best_vote_label();
}


inline R3SurfelLabel* ObjectWrapper::
best_vote_label(){
	validate_votes();
	return best_vote_label_;
}


inline R3SurfelLabel* ObjectWrapper::
best_pred_label() {
	validate_predictions();
	return best_pred_label_;
}


inline R3SurfelLabel* ObjectWrapper::
best_asgn_label() {
	return labeled() ? best_vote_label() : best_pred_label();
}


inline RNScalar ObjectWrapper::
best_vote_conf() {
	validate_votes();
	return best_vote_conf_;
}


inline RNScalar ObjectWrapper::
best_pred_conf() {
	validate_predictions();
	return best_pred_conf_;
}


inline RNScalar ObjectWrapper::
best_asgn_conf() {
	return labeled() ? best_vote_conf() : best_pred_conf();
}


inline RNScalar ObjectWrapper::
best_pred_aff() {
	validate_predictions();
	return labeled() ? 1.0 : best_pred_aff_;
}


inline RNScalar ObjectWrapper::
pred_entropy() {
	validate_predictions();
	return labeled() ? 0.0 : pred_entropy_;
}


inline void ObjectWrapper::
clear_vote() {
	votes_.clear();
	votes_valid_ = false;
}


inline int ObjectWrapper::
clear_pred() {
	int old_size = pred_.size();
	pred_.clear();
	pred_valid_ = false;
	return old_size;
}


inline int ObjectWrapper::
clear_pred(R3SurfelLabel* label) {
	PredType::iterator it = pred_.find(label);
	if (it == pred_.end()) return 0;	// nothing to clear, 0 changes.

	// Clearing needs to happen, 1 change.
	pred_.erase(it);
	pred_valid_ = false;
	return 1;
}


inline R3SurfelLabel* ObjectWrapper::
need_reset_on_increase(ObjectWrapper *contender) const {
	R3SurfelLabel* prev = prev_contribution(contender->index());
	if (prev != contender->best_vote_label())
		return prev;
	return NULL;
}


inline R3SurfelLabel* ObjectWrapper::
need_reset_on_decrease(ObjectWrapper *contender) const {
	return prev_contribution(contender->index());
}

} // namespace igra

#endif /* OBJECTWRAPPER_H_ */
