/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  Plotter.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2012.
 *
 *  Description: Classes that facilitate packaging and serializing data into
 *               gnuplot files and sripts as well as executing plotting scripts.
 *
 *  -------------------------------------------------------------------------*/


#include "Plotter.h"

#include <limits>
#include <vector>
#include <cstring>

namespace igra {

const double kMaxDouble = numeric_limits<double>::max();

const string kPlotterDataFileExt = ".dat";
const string kPlotterScriptFileExt = ".gplot";
const string kPlotterImgFormat = "png";
const string kPlotterImgFileExt = "." + kPlotterImgFormat;
const string kGnuplot = "gnuplot";
const string kPlotterDataMissing = "-";

Point Point::merge(const Point &a, const Point &b) {
    Point result;

    result.n = a.n + b.n;

    if (result.n > 0) {
        result.mean = (a.n * a.mean + b.n * b.mean) / result.n;

        double sum1 = (a.n * a.dev * a.dev + b.n * b.dev * b.dev) / result.n;
        double diff = a.mean - b.mean;
        double sum2 = a.n * b.n * diff * diff / (result.n * result.n);
        result.dev = sqrt(sum1 + sum2);
    }

    return result;
}

////////////////////////////////////////////////////////////////////////
// Sequence methods.
////////////////////////////////////////////////////////////////////////

Sequence::Sequence(const char *title) :
        title(title), values() {
}

Sequence::Sequence(const string &title) :
        title(title), values() {
}

Sequence::Sequence(const Sequence &orig) :
        title(orig.title), values(orig.values) {
}

void Sequence::reset(const char *new_title) {
    title = new_title;
    values.clear();
}

////////////////////////////////////////////////////////////////////////
// Params methods.
////////////////////////////////////////////////////////////////////////

Params::Params() :
                x_size(1920),
                y_size(1080),
                marker_size(1),
                x_ticks(true),
                y_ticks(true),
                x_unit_limit(false),
                y_unit_limit(false),
                x_log_scale(false),
                y_log_scale(false),
                legend(ON),
                plot_type(LINE) {
}

Params::Params(const Params &o) {
    memcpy(this, &o, sizeof(*this));
}

Params& Params::operator=(const Params &o) {
    memcpy(this, &o, sizeof(*this));
    return *this;
}

int Params::WriteGnuplotHeader(FILE *fp) const {
    if (!fp) return 0;

    fprintf(fp, "reset\n");
    fprintf(fp, "set term %s size %d, %d font \"verdana.ttf,16\"\n",
            kPlotterImgFormat.c_str(), x_size, y_size);

    if (x_ticks) fprintf(fp, "set grid xtics\n");
    if (y_ticks) fprintf(fp, "set grid ytics\n");

    if (x_unit_limit) fprintf(fp, "set xrange [0:1]\n");
    else if (x_log_scale) fprintf(fp, "set xrange [0.1:]\n");
    else fprintf(fp, "set auto x\n");

    if (y_unit_limit) fprintf(fp, "set yrange [0:1]\n");
    else if (y_log_scale) fprintf(fp, "set yrange [0.1:]\n");
    else fprintf(fp, "set auto y\n");

    if (x_log_scale) fprintf(fp, "set logscale x\n");
    if (y_log_scale) fprintf(fp, "set logscale y\n");

    fprintf(fp, "set key %s %s %s %s %s %s %s\n", legend[ON] ? "on" : "off",
            legend[LEFT] ? "left" : "", legend[RIGHT] ? "right" : "",
            legend[TOP] ? "top" : "", legend[BOTTOM] ? "bottom" : "",
            legend[OUTSIDE] ? "outside" : "", legend[BELOW] ? "below" : "");

    switch (plot_type) {
        case POINTS:
            fprintf(fp, "set style data points\n");
            break;
        case LINE:
            fprintf(fp, "set style data linespoints\n");
            break;
        case ERROR_BAR:
            fprintf(fp, "set style data errorbars\n");
            break;
        case HISTOGRAM:
            fprintf(fp, "set style fill solid 1.00 border 0\n");
            fprintf(fp, "set style histogram errorbars\n");
            fprintf(fp, "set style data histogram\n");
            break;
        case STACKED:
            fprintf(fp, "set style fill solid 1.00 border 0\n");
            fprintf(fp, "set style histogram rowstacked\n");
            fprintf(fp, "set style data histogram\n");
            break;
        default:
            return 0;
    }

    // Return success.
    return 1;
}

string Params::using_line(int index) const {
    char temp[255];
    temp[0] = 0;
    int col_ind = column_index(index);
    switch (plot_type) {
        case POINTS:
            sprintf(temp, "using 1:%d:%d", col_ind, col_ind + 1);
            break;
        case LINE:
            sprintf(temp, "using 1:%d:%d", col_ind, col_ind + 1);
            break;
        case ERROR_BAR:
            sprintf(temp, "using 1:%d:%d", col_ind, col_ind + 1);
            break;
        case HISTOGRAM:
            sprintf(temp, "using %d:%d:xtic(1)", col_ind, col_ind + 1);
            break;
        case STACKED:
            sprintf(temp, "using %d:xtic(1)", col_ind);
            break;
        default:
            break;
    }

    // Create and return result.
    string result(temp);
    return result;
}

string Params::size_line() const {
    char temp[255];
    temp[0] = 0;
    sprintf(temp, "lw %f", marker_size);
    string result(temp);
    return result;
}

int Params::column_index(int index) const {
    return 2 * (index + 1);
}

////////////////////////////////////////////////////////////////////////
// Methods for writing data.
////////////////////////////////////////////////////////////////////////

Plotter::Plotter(const char* title, const char* x_label, const char* y_label) :
        title_(title), x_label_(x_label), y_label_(y_label), data() {
}

int Plotter::Insert(Sequence *sequence, bool overwrite) {
    //Check if sequence with same name already exists.
    for (list<Sequence*>::iterator it = data.begin(); it != data.end(); ++it) {
        Sequence *seq = *it;
        if (!sequence->title.compare(seq->title)) {
            if (overwrite) {
                printf("Overwriting sequence %s\n", sequence->title.c_str());
                it = data.erase(it);
                if (it != data.begin()) --it;
            } else {
                fprintf(stderr, "Cannot overwrite sequence %s.\n",
                        sequence->title.c_str());
                return 0;
            }
        }
    }

    // Insert the sequence.
    data.push_back(sequence);
    return data.size();
}

int Plotter::Plot(const char *target_file_prefix, const Params &params) const {
    // Check everything.
    if (!target_file_prefix) return 0;

    //No need to create empty files if no data present.
    bool empty = true;
    for (list<Sequence*>::const_iterator it = data.begin(); it != data.end();
            ++it) {
        Sequence * seq = *it;
        empty &= seq->values.empty();
    }
    if (empty) return 1;

    // Produce paths.
    string data_fname = target_file_prefix + kPlotterDataFileExt;
    string script_fname = target_file_prefix + kPlotterScriptFileExt;
    string plot_fname = target_file_prefix + kPlotterImgFileExt;

    // Write data to files.
    if (!WriteDataFile(data_fname.c_str())) return 0;

    // Write script file.
    if (!WriteScriptFile(script_fname.c_str(), data_fname.c_str(),
            plot_fname.c_str(), params)) {
        return 0;
    }

    // Execute script file.
    if (!ExecuteGnuplot(script_fname.c_str())) return 0;

    // Return success.
    return 1;
}

int Plotter::ExecuteGnuplot(const char *script_fname) const {
    if (!system(NULL)) {
        fprintf(stderr, "Command processor is unavailable.");
        return 0;
    }

    string cmd_line = kGnuplot + " " + script_fname;
    if (system(cmd_line.c_str())) {
        fprintf(stderr, "System call failed:\n\t%s\n", cmd_line.c_str());
        return 0;
    }

    // Return success.
    return 1;
}

int Plotter::WriteCommentHeader(FILE *fp) const {
    if (!fp) return 0;

    fprintf(fp, "#\n# Plotting\n");
    fprintf(fp, "#  %s\n", title_.c_str());
    fprintf(fp, "#  %s(%s)\n#\n", y_label_.c_str(), x_label_.c_str());

    return 1;
}

int Plotter::WriteDataFile(const char* fname) const {
    //Check everything.
    if (!fname) return 0;

    // Open file.
    FILE *fp = fopen(fname, "wt");
    if (!fp) return 0;

    // Write header.
    WriteCommentHeader(fp);
    fprintf(fp, "# Data in columns represent:\n");
    fprintf(fp, "x");
    list<Sequence*>::const_iterator dit;
    for (dit = data.begin(); dit != data.end(); ++dit) {
        Sequence *seq = *dit;
        fprintf(fp, "\t%s\t%s_dev", seq->title.c_str(), seq->title.c_str());
    }
    fprintf(fp, "\n");

    // Initialize for writing.
    const int N = data.size();
    vector<list<Point>::const_iterator> begins(N);
    vector<list<Point>::const_iterator> ends(N);
    vector<list<Point>::const_iterator> currs(N);
    int index = 0;
    for (dit = data.begin(); dit != data.end(); ++dit, ++index) {
        Sequence *seq = *dit;
        begins[index] = seq->values.begin();
        ends[index] = seq->values.end();
        currs[index] = seq->values.begin();
    }

    // Write data.
    while (true) {
        // Find the smallest x of non-printed.
        double min_x = kMaxDouble;
        int sum = 0;
        for (int i = 0; i < N; ++i) {
            if (currs[i] == ends[i]) continue;
            min_x = min(min_x, currs[i]->x);
            ++sum;
        }

        // Is everyone at the end?
        if (!sum) break;

        fprintf(fp, "%g\t", min_x);
        for (int i = 0; i < N; ++i) {
            if (currs[i] == ends[i] || currs[i]->x > min_x) {
                fprintf(fp, "%s\t%s\t", kPlotterDataMissing.c_str(),
                        kPlotterDataMissing.c_str());
            } else {
                fprintf(fp, "%g\t%g\t", currs[i]->mean, currs[i]->dev);
                ++currs[i];
            }
        }
        fprintf(fp, "\n");
    }

    // Clean up.s
    fclose(fp);

    // Return success.
    return 1;
}

////////////////////////////////////////////////////////////////////////
// Methods for writing scripts.
////////////////////////////////////////////////////////////////////////

int Plotter::WriteScriptFile(const char *script_fname, const char *data_fname,
        const char *img_fname, const Params &params) const {
    // Check everything.
    if (!script_fname || !data_fname || !img_fname) return 0;

    // Open file.
    FILE *fp = fopen(script_fname, "wt");
    if (!fp) return 0;

    // Write header.
    WriteCommentHeader(fp);

    // Write specific plot parameters ettings.
    if (!params.WriteGnuplotHeader(fp)) return 0;

    // Write common properties settings.
    fprintf(fp, "set output \"%s\"\n", img_fname);
    fprintf(fp, "set title \"%s\"\n", title_.c_str());
    fprintf(fp, "set xlabel \"%s\"\n", x_label_.c_str());
    fprintf(fp, "set ylabel \"%s\"\n", y_label_.c_str());
    fprintf(fp, "set datafile missing \"%s\"\n", kPlotterDataMissing.c_str());
    fprintf(fp, "set xtics rotate by -45\n");

    // Write the plot command.
    fprintf(fp, "plot ");
    list<Sequence*>::const_iterator dit;
    int index = 0;
    for (dit = data.begin(); dit != data.end(); ++dit, ++index) {
        // New line if not first series.
        if (dit != data.begin()) fprintf(fp, ", \\\n\t");

        fprintf(fp, "\"%s\" %s %s title columnheader(%d)", data_fname,
                params.using_line(index).c_str(), params.size_line().c_str(),
                params.column_index(index));
    }

    //Clean up.
    fclose(fp);

    //Return success.
    return 1;
}

} // namespace igra
