/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  Plotter.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2012.
 *
 *  Description: Classes that facilitate packaging and serializing data into
 *               gnuplot files and sripts as well as executing plotting scripts.
 *
 *  -------------------------------------------------------------------------*/


#ifndef PLOTTER_H_
#define PLOTTER_H_

#include <cstdio>
#include <list>
#include <string>

#include "RNBasics/RNBasics.h"

using namespace std;

namespace igra {

////////////////////////////////////////////////////////////////////////
// Plotting structures.
////////////////////////////////////////////////////////////////////////

// Data point.
struct Point {
    double x;
    double mean;
    double dev;
    int n;  // number of samples used to compute.

    Point(): x(0), mean(0), dev(0), n(0){};
    Point(double x, double mean, int n = 1, double dev = 0):
        x(x), mean(mean), dev(dev), n(n){};
    Point(const Point &o):
            x(o.x), mean(o.mean), dev(o.dev), n(o.n){};

    static Point merge(const Point &a, const Point &b);
};

// Set of data points belonging to same data set (e.g. plot line).
struct Sequence {
    Sequence(const char *title);
    Sequence(const string &title);
    Sequence(const Sequence &orig);
    void reset(const char *new_title);

    string title;
    list<Point> values;
};


// Plot params.
struct Params {
    // Image resolution
    int x_size;
    int y_size;

    float marker_size;

    // Show ticks on plots
    bool x_ticks;
    bool y_ticks;

    // Normalize respective range to [0,1]
    bool x_unit_limit;
    bool y_unit_limit;

    // Use logscale along respective axes.
    bool x_log_scale;
    bool y_log_scale;

    // Where to show legend.
    enum {
        ON = 0x01,
        LEFT = 0x02,
        RIGHT = 0x04,
        TOP = 0x08,
        BOTTOM = 0x10,
        OUTSIDE = 0x20,
        BELOW = 0x40
    };
    RNFlags legend;

    //Type of the plot.
    enum {
        POINTS,
        LINE,
        ERROR_BAR,
        HISTOGRAM,
        STACKED
    };
    int plot_type;

    Params();
    Params(const Params &o);
    Params& operator=(const Params &o);

    int WriteGnuplotHeader(FILE *fp) const;
    string using_line(int index) const;
    string size_line() const;
    int column_index(int index) const;
};


// Plot class.
class Plotter {
public:
    Plotter(const char* title, const char* x_label, const char* y_label);

    int Insert(Sequence *sequence, bool overwrite = true);

    int Plot(const char* target_file_prefix, const Params &params) const;

protected:
    int WriteDataFile(const char *fname) const;
    int WriteScriptFile(const char *script_fname, const char *data_fname,
            const char *img_fname, const Params &params) const;
    int ExecuteGnuplot(const char *script_fname) const;

    int WriteCommentHeader(FILE *fp) const;

    // fields.
    string title_;
    string x_label_;
    string y_label_;
    list<Sequence*> data;
};

} // namespace igra

#endif /* PLOTTER_H_ */
