/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  R3SurfelClassifier.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2011.
 *
 *  Description: Classifier class that also implements selection suggestion
 *               functionality.
 *
 *  -------------------------------------------------------------------------*/

#include <algorithm>
#include <limits>
#include <iostream>
#include <queue>
#include <string>
#include <sstream>

#include <stdio.h>

#include "R3Surfels/R3Surfels.h"
#include "R3SurfelClassifier.h"
#include "R3SurfelViewer.h"
#include "R3SurfelLabeler.h"
#include "DrawText.cpp"
#include "ProgressReporter.h"

namespace igra {

// Useful constants.
const RNScalar kMaxScalar = numeric_limits<RNScalar>::max();

// Debugging mode constants. Normally should be zero.
int print_verbose = 0;

// Limit in time the most tolling procedure - tree node validation.
bool limit_validation_time = false;
RNScalar time_limit = 0.0;

// Useful global variables.
R3SurfelLabel *unknown_label = NULL;
//	- debug visualization.
static R3SurfelClassifier *visualized_classifier = NULL;
bool update_main_window = false;

const char* inliers_precomp_fname = "inliers.bin";
const char* rp_precomp_fname = "rp.bin";

///////////////////////////////////////
//// USEFUL AUXILIARY FUNCTIONS    ////
///////////////////////////////////////

void SetSymMatrixValue(vector<vector<float> > *matrix, int i, int j,
        float value) {
    // Check everything.
    if (i == j || !matrix) return;

    // Retrieve the value.
    int i_ind = max(i, j);
    int j_ind = min(i, j);
    if (j_ind < 0 || matrix->size() <= (size_t) i_ind) return;

    (*matrix)[i_ind][j_ind] = value;
}

float GetSymMatrixValue(const vector<vector<float> > *matrix, int i, int j,
        float bad_value = 0.0) {
    // Check everything.
    if (i == j || !matrix) return 0.0;

    // Retrieve the value.
    int i_ind = max(i, j);
    int j_ind = min(i, j);
    if (j_ind < 0 || matrix->size() <= (size_t) i_ind) return bad_value;

    return (*matrix)[i_ind][j_ind];
}

template<typename MatrixType, typename HeaderType>
int ReadSymmetricMatrix(const char* filename,
        vector<vector<MatrixType> > &matrix,
        vector<HeaderType> *header = NULL) {

    int matrix_entries_read = 0;

    // Open file.
    FILE *fp = fopen(filename, "rb");
    if (!fp) return 0;

    // Read header if such is expected.
    if (header) {
        vector<HeaderType> &v = *header;
        fread(&v[0], sizeof(HeaderType), v.size(), fp);
    }

    // Read the matrix.
    for (int i = 0; (size_t) i < matrix.size(); ++i) {
        vector<MatrixType> &v = matrix[i];
        fread(&v[0], sizeof(MatrixType), v.size(), fp);
        if (ferror(fp)) break;
        matrix_entries_read += v.size();
    }

    if (print_verbose) {
        printf("\t\tReading from %s: %s\n", filename,
                (ferror(fp) ? "failed" : "successful"));
    }

    fclose(fp);

    return matrix_entries_read;
}

template<typename MatrixType, typename HeaderType>
int WriteSymmetricMatrix(const char* filename,
        const vector<vector<MatrixType> > &matrix,
        const vector<HeaderType> *header = NULL) {

    int matrix_entries_written = 0;

    // Open file.
    FILE *fp = fopen(filename, "wb");
    if (!fp) return 0;

    // Read header if such is expected.
    if (header) {
        const vector<HeaderType> &v = *header;
        fwrite(&v[0], sizeof(HeaderType), v.size(), fp);
    }

    // Read the matrix.
    for (int i = 0; (size_t) i < matrix.size(); ++i) {
        const vector<MatrixType> &v = matrix[i];
        fwrite(&v[0], sizeof(MatrixType), v.size(), fp);
        if (ferror(fp)) break;
        matrix_entries_written += v.size();
    }

    if (print_verbose) {
        printf("\t\tWriting to %s: %s\n", filename,
                (ferror(fp) ? "failed" : "successful"));
    }

    fclose(fp);

    return matrix_entries_written;
}

////////////////////////////////////////////////////////////////////////
// Classifier constructor/destructor
////////////////////////////////////////////////////////////////////////

R3SurfelClassifier::R3SurfelClassifier(R3SurfelScene *scene,
        R3SurfelLabeler* labeler, VisualizationStyle style) :
                scene_(scene),

                objects_(),
                object_leaf_(),

                labeled_(),
                p_increased_(),
                p_decreased_(),

                group_tree_root_(NULL),
                label_choice_time_(0),
                max_benefit_cut_(),
                last_suggestion_node_(NULL),
                group_traversal_(),
                group_traversal_position_(0),
                induced_aff_changed_(true),

                feature_dist_(new vector<vector<float> >),
                spatial_dist_(new vector<vector<float> >),
                shape_aff_(new vector<vector<float> >),
                rp_aff_(new vector<vector<map<int, float> > >),
                rp_aff_max_(new vector<vector<float> >),
                non_ctr_p_(new vector<vector<float> >),
                density_sum_(),
                density_num_(0),

                labeler_(labeler),
                vis_window_(0),
                vis_window_width_(1024),
                vis_window_height_(768),
                vis_style_(style),
                vis_space_box_(),
                vis_clipping_box_(),
                vis_selection_box_(),
                vis_stats_box_(),
                mouse_drag_distance_squared_(0),
                last_labeled(),
                draw_purpose_(DRAW_PURPOSE_DRAW),
                color_scheme_(COLORING_SCHEME_ASSIGNMENT),
                selection_container_(),
                refit_data_to_screen_(true),
                update_mds_fs_(true),
                mds_fs_locations_(),
                update_mds_aff_(true),
                mds_aff_locations_(),
                update_mds_induced_(true),
                mds_induced_locations_(),
                evaluation_group_(
                NULL) {

    // Set unknown label - has a special meaning.
    unknown_label = scene_->FindLabelByName("Unknown");
    if (!unknown_label) unknown_label = scene_->RootLabel();

    // Set UI initial values.
    memset(mouse_button_, 0, 3 * sizeof(int));
    memset(mouse_screen_coords_, -1, 2 * sizeof(int));

    // Set default label choice time.
    label_choice_time_ = LabelChoiceTime(scene_);

    // Initialize object containers.
    const int N = scene_->NObjects();
    objects_.reserve(N);
    int object_index = 0;
    for (int i = 0; i < N; ++i) {
        R3SurfelObject *object = scene_->Object(i);
        if (!object || object->NParts() > 0) continue;

        ObjectWrapper *ow = new ObjectWrapper(object, object_index++, this);
        objects_.push_back(ow);
        object_leaf_[object] = ow;
    }

    // Update predictions.
    PopulateDistancesAffinities();
    for (int i = 0; i < N; ++i) {
        R3SurfelObject *object = scene_->Object(i);
        if (!object || object->NParts() > 0) continue;

        R3SurfelLabelAssignment *ha = object->HumanLabelAssignment();
        if (!ha) continue;

        UpdateAfterInsertLabelAssignment(ha);
    }
    PredictLabelAssignments();
}

R3SurfelClassifier::~R3SurfelClassifier(void) {
    // Release non-leaf group nodes.
    ReleaseGroupTree();
    // Release leaf-nodes.
    for (int i = 0; (size_t) i < objects_.size(); ++i)
        delete objects_[i];

    if (feature_dist_) delete feature_dist_;
    if (spatial_dist_) delete spatial_dist_;
    if (shape_aff_) delete shape_aff_;
    if (rp_aff_) delete rp_aff_;
    if (rp_aff_max_) delete rp_aff_max_;
    if (non_ctr_p_) delete non_ctr_p_;
    if (evaluation_group_) delete evaluation_group_;

    Stop();
}

void R3SurfelClassifier::ReleaseGroupTree(void) {
    GroupTreeNode *root = GetGroupRootNode();
    if (!root) return;

    RNArray<GroupTreeNode*> stack;
    stack.InsertTail(root);
    while (stack.NEntries()) {
        // Pop stack.
        GroupTreeNode *node = stack.Tail();
        stack.RemoveTail();

        if (node->is_leaf()) continue;

        // Put children on stack for future purging.
        for (GroupTreeNode::ChildrenIterator it = node->children_.begin();
                it != node->children_.end(); ++it) {
            GroupTreeNode *child = *it;
            if (!child->is_leaf()) stack.InsertTail(child);
        }

        // Delete the node.
        node->unlink_children();
        delete node;
    }

    last_suggestion_node_ = NULL;
    max_benefit_cut_.reset(NULL);
}

////////////////////////////////////
//// LABEL ASSIGNMENT FUNCTIONS ////
////////////////////////////////////

void R3SurfelClassifier::UpdateAfterInsertLabelAssignment(
        R3SurfelLabelAssignment *assignment) {
    UpdateObjectLabeling(assignment, true);
}

void R3SurfelClassifier::UpdateBeforeRemoveLabelAssignment(
        R3SurfelLabelAssignment *assignment) {
    UpdateObjectLabeling(assignment, false);
}

void R3SurfelClassifier::UpdateObjectLabeling(
        R3SurfelLabelAssignment *assignment, bool insert) {
    // Check everything.
    if (assignment->Originator() != R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR)
        return;
    map<R3SurfelObject *, ObjectWrapper *>::iterator it = object_leaf_.find(
            assignment->Object());
    if (it == object_leaf_.end()) return;

    ObjectWrapper *ow = it->second;
    R3SurfelLabel *label = assignment->Label();
    RNScalar weight =
            insert ? assignment->Confidence() : -assignment->Confidence();

    assert(ow);
    assert(label);

    UpdateObjectLabeling(ow, label, weight);
}

void R3SurfelClassifier::UpdateObjectLabeling(ObjectWrapper *ow,
        R3SurfelLabel *label, RNScalar weight) {
    if (!ow || !label) return;

    // Before.
    bool lab_before = ow->labeled();
    RNScalar best_conf_before = ow->best_vote_conf();
    R3SurfelLabel* best_label_before = ow->best_vote_label();
    if (best_label_before) {
        map<R3SurfelLabel*, set<int> >::iterator temp_it = labeled_.find(
                best_label_before);
        if (temp_it != labeled_.end()) {
            temp_it->second.erase(ow->index());
            if (temp_it->second.empty()) labeled_.erase(temp_it);
        }
    }

    // Cast.
    ow->replace_votes(label, weight);

    // After.
    RNScalar best_conf_after = ow->best_vote_conf();
    R3SurfelLabel *best_label_after = ow->best_vote_label();
    ReassignObjectBestAsgn(ow);

    bool lab_after = ow->labeled();
    if (lab_after) labeled_[best_label_after].insert(ow->index());

    // Update density weights.
    if (lab_before != lab_after) {
        const int N = objects_.size();
        const int index = ow->index();
        if (lab_after) {
            for (int i = 0; i < N; ++i)
                if (i != index) density_sum_[i] -= GetShapeAffinity(i, index);
            density_num_--;
        } else {
            for (int i = 0; i < N; ++i)
                if (i != index) density_sum_[i] += GetShapeAffinity(i, index);
            density_num_++;
        }
    }

    // Put into respective alteration bin.
    bool vote_changed = (best_label_before != best_label_after);

    if (best_label_after
            && (vote_changed || best_conf_after > best_conf_before))
        p_increased_.insert(ow);

    if (best_label_before
            && (vote_changed || best_conf_after < best_conf_before))
        p_decreased_.insert(ow);
}

int R3SurfelClassifier::PredictLabelAssignments(void) {
    // Check if something needs to be done.
    if ((p_increased_.size() + p_decreased_.size()) == 0) return 1;

    bool something_changed = false;
    bool revalidate_group_tree = false;

    // Update predictions for other objects.
    for (size_t i = 0; i < objects_.size(); ++i) {
        ObjectWrapper *ow = objects_[i];
        if (!ow->group_node()->valid()) revalidate_group_tree = true;
        if (ow->labeled()) continue;

        int changes_num = 0;

        if (p_decreased_.count(ow)) {
            changes_num += GlobalResetPrediction(ow);
        } else {

            // Useful variables.
            set<R3SurfelLabel*> need_global_reset;
            map<R3SurfelLabel*, set<ObjectWrapper*> > competitors;
            set<ObjectWrapper*>::const_iterator it;
            set<ObjectWrapper*>::const_iterator it_end;

            // Check for ones who increased vote score.
            it_end = p_increased_.end();
            for (it = p_increased_.begin(); it != it_end; ++it) {
                ObjectWrapper *ow_cont = *it;
                assert(ow_cont->labeled());

                competitors[ow_cont->best_asgn_label()].insert(ow_cont);

                R3SurfelLabel *reassess_label = ow->need_reset_on_increase(
                        ow_cont);
                if (reassess_label) need_global_reset.insert(reassess_label);
            }

            // Check for ones who lost voting score.
            it_end = p_decreased_.end();
            for (it = p_decreased_.begin(); it != it_end; ++it) {
                ObjectWrapper *ow_cont = *it;
                if (ow_cont->labeled())
                    competitors[ow_cont->best_asgn_label()].insert(ow_cont);

                R3SurfelLabel *reassess_label = ow->need_reset_on_decrease(
                        ow_cont);
                if (reassess_label) need_global_reset.insert(reassess_label);
            }

            // Filter to avoid redundant work.
            for (set<R3SurfelLabel*>::const_iterator itl =
                    need_global_reset.begin(); itl != need_global_reset.end();
                    ++itl) {
                R3SurfelLabel *l = *itl;
                competitors.erase(l);
            }

            // Reset globally for labels.
            changes_num += GlobalResetPrediction(ow, need_global_reset);

            // Attempt to bump contributions for the rest.
            for (map<R3SurfelLabel*, set<ObjectWrapper*> >::const_iterator ito =
                    competitors.begin(); ito != competitors.end(); ++ito) {

                RNScalar best_score = 0.0;
                int best_contender = -1;
                it_end = ito->second.end();
                for (it = ito->second.begin(); it != it_end; ++it) {
                    ObjectWrapper *ow_cont = *it;
                    RNScalar shaff = GetShapeAffinity(ow->index(),
                            ow_cont->index());
                    RNScalar ow_cont_score = ow_cont->best_vote_conf();
                    RNScalar cont_score = shaff * ow_cont_score;

                    if (cont_score > best_score) {
                        best_score = cont_score;
                        best_contender = ow_cont->index();
                    }
                }

                if (best_contender < 0) continue;

                changes_num += ow->try_update_pred(ito->first, best_contender,
                        best_score);
            }
        }

        if (changes_num > 0) {
            ReassignObjectBestAsgn(ow);
            something_changed = true;
        }

    } // for i - end.

    // clean-up prediction container.
    p_increased_.clear();
    p_decreased_.clear();

    if (something_changed || revalidate_group_tree) {
        update_mds_induced_ = true;
        induced_aff_changed_ = true;
        if (vis_window_) glutPostWindowRedisplay(vis_window_);
    }

    // Return success value.
    return 1;
}

int R3SurfelClassifier::GlobalResetPrediction(ObjectWrapper *ow) {
    if (!ow) return 0;

    int changed_num = ow->clear_pred();

    for (map<R3SurfelLabel*, set<int> >::const_iterator lit = labeled_.begin();
            lit != labeled_.end(); ++lit) {
        R3SurfelLabel *curr_label = lit->first;

        const set<int> &pool = lit->second;

        RNScalar best_score = 0.0;
        int best_contender = -1;
        for (set<int>::const_iterator its = pool.begin(); its != pool.end();
                ++its) {
            ObjectWrapper *ow_cont = objects_[*its];
            assert(ow_cont->labeled());
            RNScalar shaff = GetShapeAffinity(ow->index(), ow_cont->index());
            RNScalar ow_cont_score = ow_cont->best_vote_conf();
            RNScalar cont_score = shaff * ow_cont_score;

            if (cont_score > best_score) {
                best_score = cont_score;
                best_contender = ow_cont->index();
            }
        }

        if (best_contender < 0) continue;

        changed_num += ow->try_update_pred(curr_label, best_contender,
                best_score);
    }

    return changed_num;
}

int R3SurfelClassifier::GlobalResetPrediction(ObjectWrapper *ow,
        set<R3SurfelLabel*> &filter) {
    if (!ow) return 0;

    int changed_num = 0;

    for (set<R3SurfelLabel*>::const_iterator its = filter.begin();
            its != filter.end(); ++its) {
        R3SurfelLabel *curr_label = *its;

        changed_num += ow->clear_pred(curr_label);

        map<R3SurfelLabel*, set<int> >::const_iterator lit = labeled_.find(
                curr_label);
        if (lit == labeled_.end()) continue;

        const set<int> &pool = lit->second;
        RNScalar best_score = 0.0;
        int best_contender = -1;
        for (set<int>::const_iterator its = pool.begin(); its != pool.end();
                ++its) {
            ObjectWrapper *ow_cont = objects_[*its];
            assert(ow_cont->labeled());
            RNScalar shaff = GetShapeAffinity(ow->index(), ow_cont->index());
            RNScalar ow_cont_score = ow_cont->best_vote_conf();
            RNScalar cont_score = shaff * ow_cont_score;

            if (cont_score > best_score) {
                best_score = cont_score;
                best_contender = ow_cont->index();
            }
        }

        if (best_contender < 0) continue;

        changed_num += ow->try_update_pred(curr_label, best_contender,
                best_score);
    }

    return changed_num;
}

void R3SurfelClassifier::ReassignObjectBestAsgn(ObjectWrapper* ow) {
    ReassignObject(ow, ow->best_asgn_label(), ow->best_asgn_conf());
}

void R3SurfelClassifier::ReassignObject(ObjectWrapper *ow, R3SurfelLabel *label,
        RNScalar confidence) {
    if (!ow || !label) return;

    R3SurfelObject *object = ow->object();
    if (!object) return;

    R3SurfelLabel *best_prediction_prev = NULL;
    RNScalar best_conf_prev = -kMaxScalar;

    // Try to reuse existing assignments.
    RNArray<R3SurfelLabelAssignment *> excessive_assignments;
    R3SurfelLabelAssignment *useful_assignment = NULL;
    for (int i = 0; i < object->NLabelAssignments(); i++) {
        R3SurfelLabelAssignment *a = object->LabelAssignment(i);
        if (a->Originator() != R3_SURFEL_LABEL_ASSIGNMENT_MACHINE_ORIGINATOR)
            continue;
        if (a->Confidence() > best_conf_prev) {
            best_conf_prev = a->Confidence();
            best_prediction_prev = a->Label();
        }
        if (!useful_assignment && a->Label() == label) useful_assignment = a;
        else excessive_assignments.Insert(a);
    }

    // Remove excessive assignments.
    for (int i = 0; i < excessive_assignments.NEntries(); ++i) {
        R3SurfelLabelAssignment *a = excessive_assignments.Kth(i);
        scene_->RemoveLabelAssignment(a);
        delete a;
    }

    if (useful_assignment) {
        if (confidence != useful_assignment->Confidence())
            useful_assignment->SetConfidence(confidence);
    } else {
        scene_->InsertLabelAssignment(
                new R3SurfelLabelAssignment(object, label, confidence,
                        R3_SURFEL_LABEL_ASSIGNMENT_MACHINE_ORIGINATOR));
    }

    if (best_prediction_prev != label) ow->group_node()->invalidate();
}

///////////////////////////////////
//// QUERY OBJECT FUNCTIONS    ////
///////////////////////////////////

RNScalar R3SurfelClassifier::SuggestSingleObject(R3SurfelObject* &object,
        const RNFlags mode) {
    // Prepare.
    object = NULL;
    if (!PredictLabelAssignments()) return 0.0;

    // Assemble a candidate pool.
    RNArray<ObjectWrapper*> pool;
    RNScalar pool_value = 0.0;

    const int N = objects_.size();
    for (int i = 0; i < N; ++i) {
        ObjectWrapper *ow = objects_[i];
        if (!ow || ow->labeled()) continue;
        R3SurfelObject *o = ow->object();
        if (!o || o->NParts() > 0) continue;

        // Acquire object's value depending on the mode.
        RNScalar conf =
                mode[SUGGEST_SINGLE_CONFIDENCE] ?
                        1.0 - ow->best_pred_conf() : 1.0;
        RNScalar density_weight =
                mode[SUGGEST_SINGLE_DENSITY_WEIGHT] ? DensityWeight(i) : 1.0;
        RNScalar entropy =
                mode[SUGGEST_SINGLE_ENTROPY] ? ow->pred_entropy() : 1.0;
        RNScalar value = conf * density_weight * entropy;

        // Reset the pool if a better pool value is reached.
        if (value > pool_value) {
            pool.Empty();
            pool_value = value;
        }

        // Add to pool if value is the same as the pool value.
        if (RNIsEqual(value, pool_value)) pool.Insert(ow);
    }

    if (!pool.NEntries()) return 0.0;

    // Choose randomly from the pool.
    int rand_index = RNRandomScalar() * pool.NEntries();
    ObjectWrapper *ow = pool.Kth(rand_index);
    object = ow->object();
    last_suggestion_node_ = ow->group_node();
    group_traversal_.clear();
    group_traversal_position_ = 0;
    return last_suggestion_node_->t_group();
}

RNScalar R3SurfelClassifier::SuggestBestGroupForSeed(R3SurfelObject *seed,
        RNArray<R3SurfelObject*> &group) {
    group.Empty();
    if (!seed) return 0.0;

    // Prepare.
    if (!PredictLabelAssignments()) return 0.0;
    CreateGroupTree();

    map<R3SurfelObject*, ObjectWrapper*>::const_iterator it = object_leaf_.find(
            seed);
    if (it == object_leaf_.end()) return 0.0;

    // Find the best benefit node upstream.
    GroupTreeNode *node = it->second->group_node();
    GroupTreeNode *best_node = node;
    RNScalar best_b = node->benefit();
    GroupTreeNode *root = GetGroupRootNode();
    while (node != root) {
        RNScalar b = node->benefit();
        if (b >= best_b) {
            best_b = b;
            best_node = node;
        }
        node = node->parent_;
    }

    ResetGroupTraversal(best_node);

    last_suggestion_node_ = NULL;
    if (group_traversal_.size() == 0) return 0.0;

    // Return the next suggested group.
    last_suggestion_node_ = group_traversal_[group_traversal_position_];
    PopulateGroupWithObjects(last_suggestion_node_, group);
    return last_suggestion_node_->t_group();
}

RNScalar R3SurfelClassifier::SuggestBestGroup(RNArray<R3SurfelObject*> &group) {
    // Prepare.
    group.Empty();
    RNScalar cost = 0.0;
    if (!PredictLabelAssignments()) return cost;
    CreateGroupTree();

    // Assemble a candidate pool.
    RNArray<GroupTreeNode*> pool;
    RNScalar pool_value = 0.0;
    list<GroupTreeNode*>::iterator it, cut_end = max_benefit_cut_.cut.end();

    for (it = max_benefit_cut_.cut.begin(); it != cut_end; ++it) {
        GroupTreeNode *node = *it;
        RNScalar value = node->benefit();

        // Reset the pool if a better pool value is reached.
        if (value > pool_value) {
            pool.Empty();
            pool_value = value;
        }

        // Add to pool if value is the same as the pool value.
        if (RNIsEqual(value, pool_value)) pool.Insert(node);
    }

    if (!pool.NEntries()) return cost;

    // Choose randomly from the pool.
    int rand_index = RNRandomScalar() * pool.NEntries();
    last_suggestion_node_ = pool.Kth(rand_index);
    ResetGroupTraversal(last_suggestion_node_);
    PopulateGroupWithObjects(last_suggestion_node_, group);
    return last_suggestion_node_->t_group();
}

RNScalar R3SurfelClassifier::SuggestExpansion(RNArray<R3SurfelObject*> &group) {
    // Prepare.
    if (!PredictLabelAssignments()) return 0.0;
    CreateGroupTree();

    GroupTreeNode *node = CommonAncestor(group);

    // Adjust traversal path.
    if (node != last_suggestion_node_) ResetGroupTraversal(node);

    last_suggestion_node_ = NULL;
    group_traversal_position_ = min(group_traversal_position_ + 1,
            (int) group_traversal_.size() - 1);
    if (group_traversal_.size() == 0) return 0.0;

    // Return the next suggested group.
    last_suggestion_node_ = group_traversal_[group_traversal_position_];
    // Account for a forced expansion from the seed.
    if (node->size_ <= 1) ResetGroupTraversal(last_suggestion_node_, node);
    int old_group_size = group.NEntries();
    PopulateGroupWithObjects(last_suggestion_node_, group);
    if (group.NEntries() != old_group_size) UpdateNonCtrP(group, false);
    return last_suggestion_node_->t_group();
}

RNScalar R3SurfelClassifier::SuggestContraction(
        RNArray<R3SurfelObject*> &group) {
    // Prepare.
    if (!PredictLabelAssignments()) return 0.0;
    CreateGroupTree();

    GroupTreeNode *node = CommonAncestor(group);

    // Adjust traversal path.
    if (node != last_suggestion_node_) ResetGroupTraversal(node);

    last_suggestion_node_ = NULL;
    group_traversal_position_ = max(group_traversal_position_ - 1, 0);
    if (group_traversal_.size() == 0) return 0.0;

    // Return the next suggested group.
    last_suggestion_node_ = group_traversal_[group_traversal_position_];
    if (last_suggestion_node_->size_ != group.NEntries())
        UpdateNonCtrP(group, true);
    PopulateGroupWithObjects(last_suggestion_node_, group);
    return last_suggestion_node_->t_group();
}

void R3SurfelClassifier::CreateGroupTree(void) {
    if (!induced_aff_changed_) return;
    induced_aff_changed_ = false;

    ReleaseGroupTree();

    // Useful variables.
    set<GroupTreeNode*> forest;
    for (int i = 0; (size_t) i < objects_.size(); i++) {
        ObjectWrapper *ow = objects_[i];
        if (!ow || !ow->object() || !ow->group_node()) continue;
        forest.insert(ow->group_node());
    }

    GroupTreeNode::GrowForestGlobally(this, forest);

    ForestToTree(forest);

    ProgressReporter progress("Updating best cut", print_verbose);
    ValidateBestFirst(GetGroupRootNode());
    UpdateBestGroupsCuts(GetGroupRootNode(), max_benefit_cut_);

    refit_data_to_screen_ = true;
    RequestRedraw();
}

void R3SurfelClassifier::GrowGroupGreedy(const R3SurfelObject *object) {
    if (!object) return;

    map<R3SurfelObject*, ObjectWrapper*>::const_iterator it = object_leaf_.find(
            (R3SurfelObject*) object);
    if (it == object_leaf_.end()) return;

    ReleaseGroupTree();

    set<GroupTreeNode*> forest;
    for (int i = 0; (size_t) i < objects_.size(); i++) {
        ObjectWrapper *ow = objects_[i];
        if (!ow || !ow->object() || !ow->group_node()) continue;
        forest.insert(ow->group_node());
    }

    //GroupTreeNode::GreedyTreeAroundSeed(this, it->second->group_node(), forest);
    GroupTreeNode::GrowForest(this, forest, it->second->group_node());

    ForestToTree(forest);

    ProgressReporter progress("Updating best cut", print_verbose);
    ValidateBestFirst(GetGroupRootNode());
    UpdateBestGroupsCuts(GetGroupRootNode(), max_benefit_cut_);

    refit_data_to_screen_ = true;
    RequestRedraw();
}

void R3SurfelClassifier::ForestToTree(set<GroupTreeNode*> &forest) {
    ProgressReporter progress("Forest to tree", print_verbose);

    // Join forest in a tree rooted at group_tree_root_.

    group_tree_root_ = NULL;
    if (forest.size()) {
        // Acquire the root.
        if (forest.size() > 1) {
            group_tree_root_ = new GroupTreeNode(this);
            for (set<GroupTreeNode*>::iterator fit = forest.begin();
                    fit != forest.end(); ++fit) {
                GroupTreeNode *child = *fit;
                group_tree_root_->add_child(child);
            }
        } else group_tree_root_ = *(forest.begin());
        assert(group_tree_root_);

        // Verify tree structure.
        RNArray<GroupTreeNode*> stack;
        stack.InsertTail(group_tree_root_);
        while (stack.NEntries()) {
            GroupTreeNode *node = stack.Tail();
            stack.RemoveTail();

            if (node->is_leaf()) continue;

            node->link_children();
            for (GroupTreeNode::ChildrenIterator chit = node->children_.begin();
                    chit != node->children_.end(); ++chit) {
                stack.InsertTail(*chit);
            }
        }
    }
}

class VBFCompare {
public:
    bool operator()(GroupTreeNode *a, GroupTreeNode *b) {
        return a->benefit() < b->benefit();
    }
};

void R3SurfelClassifier::ValidateBestFirst(GroupTreeNode *root) {
    // Populate priority queue with immediate parents of leaves.
    priority_queue<GroupTreeNode*, vector<GroupTreeNode*>, VBFCompare> q;
    RNArray<GroupTreeNode *> stack;
    stack.InsertTail(root);
    while (stack.NEntries()) {
        GroupTreeNode *node = stack.Tail();
        stack.RemoveTail();

        if (node->is_leaf()) {
            GroupTreeNode * parent = node->parent_;
            bool push_parent = parent && (parent != root) && !parent->valid();
            for (GroupTreeNode::ChildrenConstIterator it =
                    parent->children_.begin(); it != parent->children_.end();
                    ++it) {
                GroupTreeNode *other_children = *it;
                push_parent &= other_children->is_leaf();
            }
            if (push_parent) q.push(parent);
        }

        for (GroupTreeNode::ChildrenConstIterator it = node->children_.begin();
                it != node->children_.end(); ++it) {
            stack.InsertTail(*it);
        }
    }

    RNTime timer;
    timer.Read();
    // While queue is not empty or the time is not up.
    while (!(q.empty()
            || (limit_validation_time && timer.Elapsed() > time_limit))) {
        // Pop off the top of the queue.
        GroupTreeNode *node = q.top();
        q.pop();

        if (node->size_ > 500) continue;

        // Make sure that all objects that were on the queue are validated.
        if (q.empty()) node->benefit();

        // Put parent and children on the queue iff not valid.
        GroupTreeNode * parent = node->parent_;
        if (parent && parent != root && !parent->valid()
                && parent->size_ <= 500) q.push(parent);

        for (GroupTreeNode::ChildrenConstIterator it = node->children_.begin();
                it != node->children_.end(); ++it) {
            GroupTreeNode *child = *it;
            if (child && child != root && !child->valid()
                    && child->size_ <= 500) q.push(child);
        }
    }
}

void R3SurfelClassifier::UpdateBestGroupsCuts(GroupTreeNode *node,
        TreeCut &cut) {
    if (!node) return;

    if (!node->is_leaf()) {
        for (GroupTreeNode::ChildrenIterator chit = node->children_.begin();
                chit != node->children_.end(); ++chit) {
            GroupTreeNode *child = *chit;

            // Update cut for the child.
            TreeCut sub_cut;
            UpdateBestGroupsCuts(child, sub_cut);

            cut.merge(sub_cut);
        }

        // Reset cuts if necessary.
        if (node == GetGroupRootNode() && !node->is_leaf()) return;
        if (!node->valid() || cut.benefit_expected_max > node->benefit())
            return;

        cut.reset(node);
    } else if (!node->object()->labeled()) cut.reset(node);
    else cut.reset(NULL);
}

int R3SurfelClassifier::PopulateGroupWithObjects(const GroupTreeNode *node,
        RNArray<R3SurfelObject*> &group) const {
    group.Empty();
    set<ObjectWrapper*> leaves = node->GetLeaves();

    set<ObjectWrapper*>::const_iterator l_end = leaves.end();
    for (set<ObjectWrapper*>::const_iterator it = leaves.begin(); it != l_end;
            ++it) {
        ObjectWrapper *ow = *it;
        if (!ow) continue;
        R3SurfelObject *o = ow->object();
        if (!o) continue;
        group.Insert(o);
    }

    return group.NEntries();
}

void R3SurfelClassifier::ResetGroupTraversal(GroupTreeNode *center,
        GroupTreeNode *tgt_leaf) {
    // Check everything.
    group_traversal_.clear();
    group_traversal_position_ = 0;
    GroupTreeNode *root = GetGroupRootNode();
    if (!center || (!root->is_leaf() && center == root)) return;

    // Useful variables.
    RNArray<GroupTreeNode*> stack, pool, down_path, up_path;

    set<GroupTreeNode*> directed_path;
    if (tgt_leaf) {
        // Prepare for directed path, if such has been requested.
        GroupTreeNode *node = tgt_leaf;
        while (node != center) {
            directed_path.insert(node);
            node = node->parent_;
        }
    }

    // Collect downward path.
    down_path.InsertTail(center);
    while (true) {
        GroupTreeNode *last_entry = down_path.Tail();
        stack.Empty();
        stack.InsertTail(last_entry);
        int size_threshold = last_entry->size_ / 2;

        RNScalar best_b = -RN_INFINITY;
        pool.Empty();
        while (stack.NEntries()) {
            GroupTreeNode *node = stack.Tail();
            stack.RemoveTail();
            for (GroupTreeNode::ChildrenConstIterator it =
                    node->children_.begin(); it != node->children_.end();
                    ++it) {
                GroupTreeNode *child = *it;
                // If not on directed path - don't follow the tree.
                if (tgt_leaf && !directed_path.count(child)) continue;
                if (child->size_ <= size_threshold) {
                    if (!child->valid()) continue;
                    RNScalar ch_b = child->benefit();
                    if (ch_b > best_b) {
                        best_b = ch_b;
                        pool.Empty();
                    }
                    if (ch_b == best_b) pool.Insert(child);
                }
                if (!child->is_leaf()) stack.InsertTail(child);
            }
        }

        if (pool.NEntries() > 0) {
            int rand_index = RNRandomScalar() * pool.NEntries();
            down_path.InsertTail(pool.Kth(rand_index));
        } else break;
    }

    // Collect upward path.
    up_path.Insert(center);
    while (true) {
        GroupTreeNode *best_node = NULL;
        RNScalar best_b = -RN_INFINITY;
        for (GroupTreeNode *node = up_path.Tail()->parent_; node != root; node =
                node->parent_) {
            if (!node->valid()) continue;
            RNScalar b = node->benefit();
            if (b > best_b) {
                best_b = b;
                best_node = node;
            }
        }

        if (best_node) up_path.InsertTail(best_node);
        else break;
    }

    // Assemble the traversal list.
    group_traversal_.reserve(down_path.NEntries() + up_path.NEntries());
    for (int i = down_path.NEntries() - 1; i > 0; --i)
        group_traversal_.push_back(down_path.Kth(i));
    for (int i = 0; i < up_path.NEntries(); ++i)
        group_traversal_.push_back(up_path.Kth(i));
    group_traversal_position_ = down_path.NEntries() - 1;
}

GroupTreeNode* R3SurfelClassifier::CommonAncestor(
        const RNArray<R3SurfelObject*> &group) const {
    const int N = group.NEntries();
    if (N == 0) return NULL;

    // Initialize common path.
    list<GroupTreeNode*> common_path;
    int i = 0;
    for (; i < N; ++i) {
        R3SurfelObject *o = group.Kth(i);
        map<R3SurfelObject*, ObjectWrapper*>::const_iterator it =
                object_leaf_.find(o);
        if (it == object_leaf_.end()) continue;

        GroupTreeNode *node = it->second->group_node();
        while (node) {
            common_path.push_front(node);
            node = node->parent_;
        }
        break;
    }

    // Finish constructing common path.
    for (; i < N; ++i) {
        list<GroupTreeNode*> path;
        R3SurfelObject *o = group.Kth(i);
        map<R3SurfelObject*, ObjectWrapper*>::const_iterator it =
                object_leaf_.find(o);
        if (it == object_leaf_.end()) continue;

        GroupTreeNode *node = it->second->group_node();
        while (node) {
            path.push_front(node);
            node = node->parent_;
        }

        list<GroupTreeNode*>::iterator cp_it = common_path.begin();
        list<GroupTreeNode*>::iterator cp_end = common_path.end();
        list<GroupTreeNode*>::iterator p_it = path.begin();
        list<GroupTreeNode*>::iterator p_end = path.end();

        for (; cp_it != cp_end && p_it != p_end && *cp_it == *p_it;
                ++cp_it, ++p_it)
            ;

        common_path.erase(cp_it, cp_end);
    }

    return common_path.size() ? common_path.back() : NULL;
}

void R3SurfelClassifier::UpdateNonCtrP(RNArray<R3SurfelObject*> &group,
        bool contraction) {
    set<int> indexes;

    for (int i = 0; i < group.NEntries(); ++i) {
        R3SurfelObject *o = group.Kth(i);
        if (!o) continue;
        map<R3SurfelObject*, ObjectWrapper*>::const_iterator it =
                object_leaf_.find(o);
        if (it != object_leaf_.end()) indexes.insert(it->second->index());
    }

    if (indexes.size() <= 1) return;

    RNScalar factor = 1.0 - 1.0 / (indexes.size());
    if (!contraction) factor = 1.0 / factor;

    set<int>::const_iterator b_it = indexes.begin(), e_it = indexes.end();
    for (set<int>::const_iterator iti = b_it; iti != e_it; ++iti) {
        for (set<int>::const_iterator itj = b_it; itj != iti; ++itj) {
            float v = GetSymMatrixValue(non_ctr_p_, *iti, *itj);
            v = min(1.0, factor * v);
            SetSymMatrixValue(non_ctr_p_, *iti, *itj, v);
        }
    }
}

/////////////////////////////////////////////////
//// COMPUTING VARIOUS PROPERTIES            ////
/////////////////////////////////////////////////

bool R3SurfelClassifier::Labeled(const R3SurfelObject *object) const {
    if (!object) return false;

    map<R3SurfelObject*, ObjectWrapper*>::const_iterator it = object_leaf_.find(
            (R3SurfelObject*) object);
    if (it == object_leaf_.end()) return false;

    return Labeled(it->second);
}

RNScalar R3SurfelClassifier::LabelChoiceTime(R3SurfelScene *scene) {
    int lc = 0;
    for (int i = 0; i < scene->NLabels(); ++i) {
        R3SurfelLabel *label = scene->Label(i);
        if (label && label->NParts() == 0) lc++;
    }
    return log2(lc + 1);
}

void R3SurfelClassifier::UpdateLabelMenuSequence(
        RNArray<R3SurfelObject*> object_set,
        RNArray<R3SurfelLabel*> &ranked_menu_sequence) {
    // Rank labels by likelihood of predictions for the object_set.

    ranked_menu_sequence.Empty();

    // Accumulate the votes from the distribution tables for all the objects.
    set<R3SurfelLabel*> voted_labels;
    for (int i = 0; i < object_set.NEntries(); ++i) {
        R3SurfelObject *object = object_set.Kth(i);
        if (!object || object->NParts()) continue;

        ObjectWrapper *ow = object_leaf_[object];
        assert(ow);

        R3SurfelLabel *label = ow->best_asgn_label();
        voted_labels.insert(label);
    }

    // Sort labels alphabetically by name (decrease search time through menu)
    map<string, R3SurfelLabel *> label_names;
    for (int i = 0; i < scene_->NLabels(); i++) {
        R3SurfelLabel *label = scene_->Label(i);
        if (!label || !label->Name()) continue;
        label_names[label->Name()] = label;
    }

    // Produce label list.
    map<string, R3SurfelLabel*>::const_iterator ln_it;
    map<string, R3SurfelLabel*>::const_iterator ln_end = label_names.end();
    for (ln_it = label_names.begin(); ln_it != ln_end; ++ln_it) {
        R3SurfelLabel *label = ln_it->second;
        if (voted_labels.count(label)) {
            ranked_menu_sequence.InsertTail(NULL);
            ranked_menu_sequence.InsertTail(label);
            ranked_menu_sequence.InsertTail(NULL);
        } else {
            ranked_menu_sequence.InsertTail(label);
        }
    }
}

RNScalar R3SurfelClassifier::GetCost(const RNArray<R3SurfelObject*> &group) {
    GroupTreeNode* node = new GroupTreeNode(this);
    for (int i = 0; i < group.NEntries(); ++i) {
        ObjectWrapper *ow = object_leaf_[group.Kth(i)];
        if (!ow) continue;
        node->add_child(ow->group_node());
    }

    RNScalar cost = node->t_group();
    delete node;
    return cost;
}

////////////////////////////////////////////////
//// AUXILARY GETTER FUNCTIONS FUNCTIONS    ////
////////////////////////////////////////////////

RNScalar R3SurfelClassifier::GetAffinity(const vector<vector<float> > *matrix,
        const R3SurfelObject *from, const R3SurfelObject *to) const {
    // Check everything.
    if (!from || !to) return 0.0;
    if (from == to || !matrix) return 1.0;

    map<R3SurfelObject*, ObjectWrapper*>::const_iterator it;

    // Get object 1 index.
    it = object_leaf_.find((R3SurfelObject*) from);
    if (it == object_leaf_.end()) return 0.0;
    int i = it->second->index();

    // Get object 2 index.
    it = object_leaf_.find((R3SurfelObject*) to);
    if (it == object_leaf_.end()) return 0.0;
    int j = it->second->index();

    return GetSymMatrixValue(matrix, i, j);
}

RNScalar R3SurfelClassifier::DensityWeight(const int i) {
    return density_num_ > 1 ? density_sum_[i] / (density_num_ - 1) : 1.0;
}

RNScalar R3SurfelClassifier::GetRPAffinity(int i, int j, int k) const {
    // Check everything.
    if (!rp_aff_) return 0.0;
    if (i == j) return 0.0;
    if (k == i || k == j) return 1.0;

    int i_ind = max(i, j);
    int j_ind = min(i, j);

    map<int, float> &t = (*rp_aff_)[i_ind][j_ind];
    map<int, float>::iterator it = t.find(k);

    if (it == t.end()) return 0.0;

    return it->second;
}

RNScalar R3SurfelClassifier::GetShapeAffinity(R3SurfelObject *from,
        R3SurfelObject *to) const {
    if (from == to) return 1.0;
    return GetAffinity(shape_aff_, from, to);
}

RNScalar R3SurfelClassifier::GetShapeAffinity(const int i, const int j) const {
    if (i == j) return 1.0;
    return GetSymMatrixValue(shape_aff_, i, j);
}

RNScalar R3SurfelClassifier::GetInducedAffinity(R3SurfelObject *from,
        R3SurfelObject *to) const {
    if (!from || !to) return 0.0;
    if (from == to) return 1.0;

    map<R3SurfelObject*, ObjectWrapper*>::const_iterator it;

    // Get object 1 index.
    it = object_leaf_.find((R3SurfelObject*) from);
    if (it == object_leaf_.end()) return 0.0;
    int i = it->second->index();

    // Get object 2 index.
    it = object_leaf_.find((R3SurfelObject*) to);
    if (it == object_leaf_.end()) return 0.0;
    int j = it->second->index();

    return GetInducedAffinity(i, j);
}

RNScalar R3SurfelClassifier::GetInducedAffinity(const int i,
        const int j) const {
    if (i == j) return 1.0;
    ObjectWrapper *owi = objects_[i];
    ObjectWrapper *owj = objects_[j];

    RNScalar affinity = 0.0;

    if (owi->best_asgn_label() == owj->best_asgn_label()) {

        // Prediction-induced affinity.
        if (owi->labeled() && owj->labeled()) {
            affinity = 1.0;
        } else {
            // Wormhole throw same label confirmations.
            affinity = max((RNScalar) GetSymMatrixValue(shape_aff_, i, j),
                    min(owi->best_pred_aff(), owj->best_pred_aff()));
        }

        // Account for prior contractions.
        affinity *= GetSymMatrixValue(non_ctr_p_, i, j);
    }

    return affinity;
}

RNScalar R3SurfelClassifier::GetFeatureDistance(R3SurfelObject *from,
        R3SurfelObject *to) const {
    if (from == to) return 0.0;
    return GetAffinity(feature_dist_, from, to);
}

RNScalar R3SurfelClassifier::GetFeatureDistance(const int i,
        const int j) const {
    if (i == j) return 0.0;
    return GetSymMatrixValue(feature_dist_, i, j);
}

RNScalar R3SurfelClassifier::GetSpatialDistance(R3SurfelObject *from,
        R3SurfelObject *to) const {
    if (from == to) return 0.0;
    return GetAffinity(spatial_dist_, from, to);
}

RNScalar R3SurfelClassifier::GetSpatialDistance(const int i,
        const int j) const {
    if (i == j) return 0.0;
    return GetSymMatrixValue(spatial_dist_, i, j, RN_INFINITY);
}

RNScalar R3SurfelClassifier::GetScreenSizeComfortScore(R3SurfelObject *from,
        R3SurfelObject *to) const {
    if (!from || !to) return 0.0;

    map<R3SurfelObject*, ObjectWrapper*>::const_iterator it;

    // Get object 1 index.
    it = object_leaf_.find((R3SurfelObject*) from);
    if (it == object_leaf_.end()) return 0.0;
    ObjectWrapper *owi = it->second;

    // Get object 2 index.
    it = object_leaf_.find((R3SurfelObject*) to);
    if (it == object_leaf_.end()) return 0.0;
    ObjectWrapper *owj = it->second;

    RNLength dist = GetSpatialDistance(owi->index(), owj->index());
    RNLength min_size = min(owi->get_size(), owj->get_size());
    return GetScreenSizeComfortScore(dist, min_size);
}

RNScalar R3SurfelClassifier::GetScreenSizeComfortScore(ObjectWrapper *from,
        ObjectWrapper *to) const {
    if (!from || !to) return 0.0;
    RNLength dist = GetSpatialDistance(from->index(), to->index());
    RNLength min_size = min(from->get_size(), to->get_size());
    return GetScreenSizeComfortScore(dist, min_size);
}

RNScalar R3SurfelClassifier::GetScreenSizeComfortScore(int from, int to) const {
    RNLength dist = GetSpatialDistance(from, to);
    RNLength min_size = min(objects_[from]->get_size(),
            objects_[to]->get_size());
    return GetScreenSizeComfortScore(dist, min_size);
}

RNScalar R3SurfelClassifier::GetScreenSizeComfortScore(const RNLength dist,
        const RNLength size) const {
    return 1.0 / (1.0 + GetScreenSizeDiscomfortScore(dist, size));
}

RNScalar R3SurfelClassifier::GetScreenSizeDiscomfortScore(const RNLength dist,
        const RNLength size) const {
    const RNScalar kDistToDiagRatio = 5.0;
    RNScalar relative_dist = (size > 0.0) ? dist / size : RN_INFINITY;
    return max(0.0, relative_dist - kDistToDiagRatio);
}

RNScalar R3SurfelClassifier::GetDistanceInSizeUnits(int from, int to) const {
    RNLength dist = GetSpatialDistance(from, to);
    RNLength min_size = min(objects_[from]->get_size(),
            objects_[to]->get_size());
    return min_size > 0.0 ? dist / min_size : RN_INFINITY;
}

GTNTuple R3SurfelClassifier::GetBestRPinSet(ObjectWrapper *owi,
        ObjectWrapper *owj, const set<int> &pool) const {
    GTNTuple result(owi->group_node(), owj->group_node());

    int i = owi->index();
    int j = owj->index();

    // Degenerate case.
    if (i == j) {
        result.p_d = 1.0;
        result.p_rp = 1.0;
        result.p_sh = 1.0;
        result.p_patt = 1.0;
        return result;
    }
    result.p_d = GetScreenSizeComfortScore(owi, owj);
    result.dist = GetDistanceInSizeUnits(i, j);

    if (pool.size() > 2) {
        float max_possible = GetSymMatrixValue(rp_aff_max_, i, j);

        // Useful variables.
        set<int>::const_iterator it_pool = pool.begin();
        set<int>::const_iterator pool_end = pool.end();
        const map<int, float> &aff = (*rp_aff_)[max(i, j)][min(i, j)];
        map<int, float>::const_iterator it_aff = aff.lower_bound(*it_pool);
        map<int, float>::const_iterator aff_end = aff.upper_bound(
                *pool.rbegin());

        for (it_pool = pool.begin(); it_pool != pool_end; it_pool++) {
            int k = *it_pool;

            while (it_aff != aff_end && it_aff->first < k)
                ++it_aff;

            if (it_aff == aff_end) break; // all other entries are zeros.
            if (it_aff->first != k) continue; // zero entry

            RNScalar p_rp = it_aff->second;
            result.p_rp = max(result.p_rp, p_rp);

            if (RNIsEqual(result.p_rp, max_possible)) break;
        }
    } else result.p_rp = 1.0;

    result.p_patt = result.p_rp * result.p_d;

    return result;
}

GTNTuple R3SurfelClassifier::GetBestRPinSet(ObjectWrapper *owi,
        ObjectWrapper *owj) const {
    GTNTuple result(owi->group_node(), owj->group_node());

    int i = owi->index();
    int j = owj->index();

    // Degenerate case.
    if (i == j) {
        result.p_d = 1.0;
        result.p_rp = 1.0;
        result.p_sh = 1.0;
        result.p_patt = 1.0;
        return result;
    }
    result.p_d = GetScreenSizeComfortScore(owi, owj);
    result.dist = GetDistanceInSizeUnits(i, j);

    if (objects_.size() > 2) {
        result.p_rp = GetSymMatrixValue(rp_aff_max_, i, j);
        /*
         // Useful variables.
         const map<int, float> &aff = (*rp_aff_)[max(i, j)][min(i, j)];
         map<int, float>::const_iterator it_aff;
         map<int, float>::const_iterator aff_end = aff.end();

         for (it_aff = aff.begin(); it_aff != aff_end; ++it_aff) {
         RNScalar p_rp = it_aff->second;
         result.p_rp = max(result.p_rp, p_rp);
         }
         */
    } else result.p_rp = 1.0;

    result.p_patt = result.p_rp * result.p_d;

    return result;
}

//////////////////////////////////////////////////////
//// POPULATING AFFINITIES FUNCTIONS FUNCTIONS    ////
//////////////////////////////////////////////////////

void R3SurfelClassifier::PopulateDistancesAffinities(void) {
    // Useful constants.
    const int N = objects_.size();

    RNTime timer;
    timer.Read();

    // Set no buffering for stdout for debugging information.
    if (print_verbose) {
        setbuf(stdout, NULL);
        printf("\tPopulating distance/affinity matrices\n");
    }

    PopulateFeatureDistances();
    PopulateSpatialDistances();
    PopulateShapeAffinities();
    PopulateNonCtrP();
    PopulateDensityWeights();
    PopulateRPAffinities();

    if (print_verbose) {
        printf("\tDone populating distance/affinity matrices"
                "(for %d objects in %g sec)\n", N, timer.Elapsed());
    }
}

void R3SurfelClassifier::PopulateFeatureDistances(void) {
    // Populate feature_dist_, a symmetric matrix, with feature space distances.
    ProgressReporter progress("Populating feature distance", print_verbose);
    vector<vector<float> > &distances = *feature_dist_;

    // 1. Initialize.
    const int N = objects_.size();
    distances.clear();
    for (int i = 0; i < N; ++i) {
        vector<float> v(i);
        distances.push_back(v);
    }
    if (N < 2) return;

    // 2. Populate.
    for (int i = 0; i < N; ++i) {
        progress.update((100.0 * i) / N);

        for (int j = 0; j < i; ++j) {
            ObjectWrapper *ow_i = objects_[i];
            ObjectWrapper *ow_j = objects_[j];
            distances[i][j] = ow_i->FeatureVector().EuclideanDistance(
                    ow_j->FeatureVector());
        }
    }
}

void R3SurfelClassifier::PopulateSpatialDistances(void) {
    // Populate spatial_dist_, a symmetric matrix, with physical space distances.
    ProgressReporter progress("Populating spatial distance", print_verbose);
    vector<vector<float> > &distances = *spatial_dist_;

    // 1. Initialize.
    const int N = objects_.size();
    distances.clear();
    for (int i = 0; i < N; ++i) {
        vector<float> v(i);
        distances.push_back(v);
    }

    // 2. Populate.
    for (int i = 0; i < N; ++i) {
        progress.update((100.0 * i) / N);

        R3SurfelObject *object_i = objects_[i]->object();
        if (!object_i) continue;
        R2Point p_i(object_i->Centroid()[0], object_i->Centroid()[1]);

        for (int j = 0; j < i; ++j) {
            R3SurfelObject *object_j = objects_[j]->object();
            if (!object_i) continue;
            R2Point p_j(object_j->Centroid()[0], object_j->Centroid()[1]);

            distances[i][j] = R2Distance(p_i, p_j);
        }
    }
}

void R3SurfelClassifier::PopulateShapeAffinities(void) {
    // Populate shape_aff_, a symmetric matrix, with density-sensitive affinities.
    vector<vector<float> > &affinities = *shape_aff_;

    // 1. Initialize.
    const int N = objects_.size();
    affinities.clear();
    for (int i = 0; i < N; ++i) {
        vector<float> v(i, 1.0);
        affinities.push_back(v);
    }
    if (N < 2) return;

    vector<RNScalar> ranked_affs(N);

#if 1
    // Density-sensitive implementation of metric in
    // http://www.gatsby.ucl.ac.uk/aistats/fullpapers/198.pdf
    // [Chapelle, O., & Zien, A. (2004).
    //  Semi-supervised classification by low density separation.]
    // Implementation similar to
    // http://link.springer.com/content/pdf/10.1007%2F978-90-481-9794-1_28.pdf
    // [Qingsheng Zhu, Peng Yang (2010).
    //  Density Sensitive Based Spectral Clustering]

    const int kMinClusterSize = 5;
    const RNScalar rho = 1.0;

    // 3b. Exponentiate.
    vector<vector<float> > exp_dist;
    for (int i = 0; i < N; ++i) {
        vector<float> v(i);
        exp_dist.push_back(v);
    }
    const RNScalar factor = rho;
    const RNScalar inv_factor = 1.0 / rho;
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < i; ++j)
            exp_dist[i][j] = exp(factor * GetFeatureDistance(i, j)) - 1.0;
    }

    // 3c. Build ranks
    vector<vector<int> > ranks;
    PopulateRanks(exp_dist, ranks);

    // 3d. Build kNN graph.
    map<int, set<int> > knn_graph;
    PopulateKNNGraph(ranks, kMinClusterSize, knn_graph);

    // 3e. Populate density-sensitive distance matrix.
    vector<vector<float> > ds_distances;
    PopulateDensitySensitiveDistances(exp_dist, knn_graph, ds_distances);

    // 3f. De-exponentiate.
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < i; ++j) {
            float &entry = ds_distances[i][j];
            entry = inv_factor * log(1.0 + entry);
        }
    }

    // 3g. Convert to affinities.
    DistancesToAffinities(ds_distances, ranks, kMinClusterSize, affinities);

#elif 0
    // Perfect affinity (1 for same gt label, 0 for different).
    ProgressReporter progress("Populating shape affinity", print_verbose);
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < i; ++j) {
            progress.update(100 * (0.5 * i * (i - 1) + j) / unique_entry_num);

            float &entry = shape_aff[i][j];
            entry = 0.0;

            ObjectWrapper *ow_i = objects_[i];
            ObjectWrapper *ow_j = objects_[j];
            assert(ow_i && ow_j);
            R3SurfelObject *o_i = ow_i->object();
            R3SurfelObject *o_j = ow_j->object();
            if (!o_i || !o_j) continue;
            if (o_i->GroundTruthLabel() == o_j->GroundTruthLabel()) entry = 1;
        }
    }
#endif	// similarity implementation selection - end
}

    void R3SurfelClassifier::PopulateNonCtrP(void) {
        // Populate non_ctr_p_, a symmetric matrix, with non contraction likelihoods.
        // Initial values - 1.
        vector<vector<float> > &non_ctr_p = *non_ctr_p_;

        // 1. Initialize.
        const int N = objects_.size();
        non_ctr_p.clear();
        for (int i = 0; i < N; ++i) {
            vector<float> v(i, 1.0);
            non_ctr_p.push_back(v);
        }
    }

    void R3SurfelClassifier::PopulateDensityWeights(void) {
        ProgressReporter progress("Populating density weights", print_verbose);

        // Initialize.
        const int N = objects_.size();
        density_sum_.clear();
        density_sum_.resize(N, 0.0);
        density_num_ = 0;

        for (int i = 0; i < N; ++i) {
            progress.update((100.0 * i) / N);

            ObjectWrapper * owi = objects_[i];

            density_num_++;
            for (int j = 0; j < i; ++j) {
                ObjectWrapper * owj = objects_[j];

                RNScalar aff = GetShapeAffinity(i, j);

                if (!owj->labeled()) density_sum_[i] += aff;
                if (!owi->labeled()) density_sum_[j] += aff;
            }
        }
    }

    void R3SurfelClassifier::PopulateRanks(const vector<vector<float> > &distances,
            vector<vector<int> > &ranks) {
        // Rank objects in terms of their mutual proximities with respect to distance.

        // Check everything.
        ranks.clear();
        const int N = int(objects_.size());
        if (N < 2) return;
        assert(distances.size() == size_t(N));

        ProgressReporter progress("Populating ranks", print_verbose);

        vector<multimap<RNScalar, int> > dist_ranks(N);
        for (int i = 0; i < N; ++i) {
            multimap<RNScalar, int> &ranks_i = dist_ranks[i];
            for (int j = 0; j < i; ++j) {
                multimap<RNScalar, int> &ranks_j = dist_ranks[j];

                RNScalar d = GetSymMatrixValue(&distances, i, j);
                ranks_i.insert(pair<RNScalar, int>(d, j));
                ranks_j.insert(pair<RNScalar, int>(d, i));
            }
            progress.update((50.0 * i) / N);
        }

        vector<int> row(N - 1);
        ranks.resize(N, row);
        for (int i = 0; i < N; ++i) {
            multimap<RNScalar, int> &ranks_i = dist_ranks[i];
            multimap<RNScalar, int>::const_iterator iti = ranks_i.begin();
            multimap<RNScalar, int>::const_iterator iti_end = ranks_i.end();

            vector<int> &row_entry = ranks[i];
            for (int rank = 0; rank < N - 1 && iti != iti_end; ++rank, ++iti) {
                int j = iti->second;
                assert(i != j);
                row_entry[rank] = j;
            }
            progress.update(50.0 + (50.0 * i) / N);
        }
    }

    void R3SurfelClassifier::PopulateKNNGraph(const vector<vector<int> > &ranks,
            int k0, map<int, set<int> > &knn_graph) {
        // Populates the kNN graph starting k0 and growing until connected.

        // Check everything.
        knn_graph.clear();
        const int N = int(objects_.size());
        if (N < 2) return;
        assert(ranks.size() == size_t(N));

        RNTime timer;
        timer.Read();

        int k_prev = 0;
        int k = k0;
        bool disconnected = true;
        while (disconnected) {
            if (k >= N) break;

            if (print_verbose) {
                printf("\r\t\tPopulating %dNN graph..                  ", k);
            }

            set<int> outsiders; // used for later verification of connectivity.

            // connect kNN graph.
            for (int i = 0; i < N; ++i) {
                outsiders.insert(outsiders.end(), i);
                const vector<int> &ranks_row = ranks[i];
                knn_graph[i].insert(ranks_row.begin() + k_prev,
                        ranks_row.begin() + k);

                // Make sure the graph is undirected.
                for (int rank = k_prev; rank < k; ++rank) {
                    knn_graph[ranks_row[rank]].insert(i);
                }
            }

            if (print_verbose) printf("%-16s", " Verifying..");

            // verify connectivity.
            list<int> stack;
            int first = knn_graph.begin()->first;
            stack.push_back(first);
            while (!stack.empty() && !outsiders.empty()) {
                // pop off the stack.
                int curr = stack.back();
                stack.pop_back();

                // remove from outsiders.
                outsiders.erase(curr);

                const set<int> &currNN = knn_graph[curr];
                for (set<int>::const_iterator it = currNN.begin();
                        it != currNN.end(); ++it) {
                    int currN = *it;
                    if (outsiders.count(currN)) stack.push_back(currN);
                }
            }

            disconnected = !outsiders.empty();

            // Make k larger if still disconnected.
            if (disconnected) {
                k_prev = k;
                ++k;
            }
        }	// disconnected loop - end.

        if (print_verbose) printf("\tDone in %.2f seconds\n", timer.Elapsed());
    }

    void R3SurfelClassifier::PopulateDensitySensitiveDistances(
            const vector<vector<float> > &distances,
            const map<int, set<int> > &knn_graph,
            vector<vector<float> > &ds_distances) {
        // Populate symmetrically the distance matrix with density-sensitive distances.

        // Check everything.
        ds_distances.clear();
        const int N = int(objects_.size());
        if (N < 2) return;
        assert(distances.size() == size_t(N));
        assert(knn_graph.size() == size_t(N));

        ProgressReporter progress("Populating density-sensitive distance",
                print_verbose);

        // 1. Initialize.
        for (int i = 0; i < N; ++i) {
            vector<float> v(i, 1.0);
            ds_distances.push_back(v);
        }

        for (int source = 0; source < (N - 1); ++source) {
            progress.update((100.0 * source) / (N - 1));

            // Dijkstra's algorithm to compute best path from source to the objects
            // of higher index (no redundancy!)

            // Initialize Dijkstra.
            vector<bool> visited(N, false);
            set<int> front;
            vector<RNScalar> d(N, kMaxScalar);
            d[source] = 0.0;
            front.insert(source);

            // Do Dijkstra.
            while (!front.empty()) {

                // find best sink: with smallest tentative distance.
                int sink = -1;
                RNScalar d_sink = kMaxScalar;
                set<int>::const_iterator sit;
                set<int>::const_iterator sit_end = front.end();
                for (sit = front.begin(); sit != sit_end; ++sit) {
                    int other = *sit;
                    if (!visited[other] && ((sink == -1) || (d[other] < d_sink))) {
                        sink = other;
                        d_sink = d[other];
                    }
                }
                assert(sink != -1);
                visited[sink] = true;
                front.erase(sink);

                // update distances for connected nodes.
                map<int, set<int> >::const_iterator knnit = knn_graph.find(sink);
                assert(knnit != knn_graph.end());
                const set<int> &neighbors = knnit->second;
                sit_end = neighbors.end();
                for (sit = neighbors.begin(); sit != sit_end; ++sit) {
                    int n = *sit;
                    assert(sink != n);
                    if (visited[n]) continue;

                    float d_sn = GetSymMatrixValue(&distances, sink, n);
                    RNScalar &d_n = d[n];
                    d_n = min(d_n, d_sink + d_sn);

                    front.insert(n);
                }
            }

            // Contribute to the output matrix.
            for (int sink = source + 1; sink < N; ++sink) {
                ds_distances[sink][source] = d[sink];
            }
        }
    }

    void R3SurfelClassifier::DistancesToAffinities(
            const vector<vector<float> > &distances,
            const vector<vector<int> > &ranks, int kmin,
            vector<vector<float> > &affinities) {
        // Convert distances to affinities.

        // Check everything.
        const int N = int(objects_.size());
        if (N < 2) return;
        assert(distances.size() == size_t(N));
        assert(affinities.size() == size_t(N));

        ProgressReporter progress("Converting to affinities", print_verbose);

        // Useful containers.
        vector<float> dkmin(N);
        for (int i = 0; i < N; ++i)
        dkmin[i] = GetSymMatrixValue(&distances, i, ranks[i][kmin]);

        // Compute affinities.
        for (int i = 0; i < N; ++i) {
            progress.update((100.0 * i) / N);
            for (int j = 0; j < i; ++j) {
                // Useful variables.
                const float &Dij = distances[i][j];
                const float &dkmin_i = dkmin[i];
                const float &dkmin_j = dkmin[j];
                float &s = affinities[i][j];
                // Pull up so that kmin NN have distance of 0.0.
                const float min_v = min(dkmin_i, dkmin_j);
                const float adj_v = Dij - min_v;
                s = 1.0 / (1.0 + max(float(0.0), adj_v));
            }
        }
    }

    void R3SurfelClassifier::PopulateRPAffinities(void) {
        // Populate rp_aff_, a symmetric 3D matrix,
        // with triplet regular pattern scores.
        vector<vector<map<int, float> > > &affinities = *rp_aff_;
        vector<vector<float> > &aff_max = *rp_aff_max_;

        // 1. Initialize.
        const int N = objects_.size();
        affinities.clear();
        aff_max.clear();
        for (int i = 0; i < N; ++i) {
            vector<map<int, float> > temp(i);
            affinities.push_back(temp);
            vector<float> tmp(i, 0.0);
            aff_max.push_back(tmp);
        }
        if (N < 2) return;

        // 2. Try reading from a precomp file.
        if (ReadRPAffinities()) return;

        // 3. Could not - populate anew.
        ProgressReporter progress("Populating RP affinities", print_verbose);
        for (int i = 0; i < N; ++i) {
            progress.update((100.0 * i) / N);

            ObjectWrapper *owi = objects_[i];
            const R2Point& pi = owi->location2D();
            const R2Box& bi = owi->bbox2D();
            const RNScalar sizei = owi->get_size();

            for (int j = 0; j < i; ++j) {
                ObjectWrapper *owj = objects_[j];
                const R2Point& pj = owj->location2D();
                const R2Box& bj = owj->bbox2D();
                const RNScalar sizej = owj->get_size();

                R2Vector vij(pj - pi);
                R2Vector vji(pi - pj);

                const RNScalar sizeij = min(sizei, sizej);

                map<int, float>& v = affinities[i][j];
                RNScalar max_v = 0.0;

                for (int k = 0; k < N; ++k) {
                    if (i == k || j == k) continue;

                    ObjectWrapper *owk = objects_[k];
                    const RNScalar sizeijk = min(sizeij, owk->get_size());

                    R2Box bijk(owk->bbox2D());
                    R2Box bjik(bijk);
                    bijk.Translate(vij);
                    bjik.Translate(vji);

                    RNScalar value1 = R2Distance(bijk, bi);
                    RNScalar value2 = R2Distance(bjik, bj);
                    RNScalar value = exp(-min(value1, value2) / sizeijk);

                    max_v = max(max_v, value);

                    if (!RNIsPositive(value)) continue;

                    v.insert(v.end(), pair<int, float>(k, value));
                }

                float &curr_max_v = aff_max[i][j];
                curr_max_v = max(curr_max_v, (float) max_v);
            }
        }

        //4. Write to precomp file.
        WriteRPAffinities();
    }

    int R3SurfelClassifier::ReadRPAffinities(void) {
        vector<vector<map<int, float> > > &affinities = *rp_aff_;
        vector<vector<float> > &aff_max = *rp_aff_max_;
        int matrix_entries_read = 0;
        bool failed = false;

        // Open file.
        const int N = objects_.size();
        stringstream fname;
        fname << N << "_" << rp_precomp_fname;
        char status[255];
        sprintf(status, "Reading RP affinities from %s", fname.str().c_str());
        ProgressReporter progress(status, print_verbose);
        progress.set_term_status("failed");
        FILE *fp = fopen(fname.str().c_str(), "rb");
        if (!fp) return 0;

        // Read and check header.
        int read_size;
        fread(&read_size, sizeof(int), 1, fp);
        if (read_size != N) return 0;

        // Read the entries.
        for (int i = 0; i < N && !failed; ++i) {
            progress.update((100.0 * i) / N);
            for (int j = 0; j < i && !failed; ++j) {
                int size;
                fread(&size, sizeof(int), 1, fp);
                map<int, float> &v = affinities[i][j];
                float max_v = 0.0;
                for (int k = 0; k < size && !failed; ++k) {
                    int tail;
                    float score;
                    fread(&tail, sizeof(int), 1, fp);
                    fread(&score, sizeof(float), 1, fp);
                    failed |= ferror(fp);
                    if (failed) break;

                    max_v = max(max_v, score);
                    matrix_entries_read++;
                    v.insert(pair<int, float>(tail, score));
                }

                float &curr_max_v = aff_max[i][j];
                curr_max_v = max(curr_max_v, max_v);
            }
        }

        fclose(fp);

        if (failed) {
            for (int i = 0; i < N && !failed; ++i) {
                for (int j = 0; j < i && !failed; ++j) {
                    map<int, float> &v = affinities[i][j];
                    v.clear();
                }
            }
            matrix_entries_read = 0;
        } else progress.set_term_status("success");

        return matrix_entries_read;
    }

    int R3SurfelClassifier::WriteRPAffinities(void) {
        vector<vector<map<int, float> > > &affinities = *rp_aff_;
        int matrix_entries_written = 0;
        bool failed = false;

        // Open file.
        const int N = objects_.size();
        stringstream fname;
        fname << N << "_" << rp_precomp_fname;
        char status[255];
        sprintf(status, "Writing RP affinities to %s", fname.str().c_str());
        ProgressReporter progress(status, print_verbose);
        progress.set_term_status("failed");
        FILE *fp = fopen(fname.str().c_str(), "wb");
        if (!fp) return 0;

        // Write header.
        fwrite(&N, sizeof(int), 1, fp);

        // Read the entries.
        for (int i = 0; i < N && !failed; ++i) {
            progress.update((100.0 * i) / N);
            for (int j = 0; j < i && !failed; ++j) {
                map<int, float> &v = affinities[i][j];
                int size = v.size();
                fwrite(&size, sizeof(int), 1, fp);

                for (map<int, float>::const_iterator it = v.begin();
                        it != v.end() && !failed; ++it) {
                    fwrite(&it->first, sizeof(int), 1, fp);
                    fwrite(&it->second, sizeof(float), 1, fp);
                    failed |= ferror(fp);
                    if (failed) break;

                    matrix_entries_written++;
                }
            }
        }

        fclose(fp);

        if (failed) matrix_entries_written = 0;
        else progress.set_term_status("success");

        return matrix_entries_written;
    }

//////////////////////////
//// VISUALIZATION    ////
//////////////////////////

    void GLUTClassifierRedraw(void) {
        if (visualized_classifier && visualized_classifier->Redraw())
        glutPostRedisplay();

        glutSwapBuffers();
    }

    void GLUTClassifierResize(int w, int h) {
        if (visualized_classifier && visualized_classifier->Resize(w, h))
        glutPostRedisplay();
    }

    void GLUTClassifierKeyboard(unsigned char key, int x, int y) {
        // Determine modifiers
        int modifiers = glutGetModifiers();
        int shift = (modifiers & GLUT_ACTIVE_SHIFT);
        // int ctrl = (modifiers & GLUT_ACTIVE_CTRL);
        // int alt = (modifiers & GLUT_ACTIVE_ALT);

        // Convert control keys.
        int translated_key = key;
        if ((key >= 1) && (key <= 26)) {
            if (shift) translated_key += 64;
            else translated_key += 96;
        }

        if (visualized_classifier
                && visualized_classifier->Keyboard(translated_key, x, y)) {
            glutPostRedisplay();
        }
    }

    void GLUTClassifierSpecial(int key, int x, int y) {
        if (visualized_classifier
                && visualized_classifier->KeyboardSpecial(key, x, y)) {
            glutPostRedisplay();
        }
    }

    void GLUTClassifierMouseButton(int button, int state, int x, int y) {
        if (visualized_classifier
                && visualized_classifier->MouseButton(button, state, x, y)) {
            glutPostRedisplay();
        }
    }

    void GLUTClassifierMouseMotion(int x, int y) {
        if (visualized_classifier && visualized_classifier->MouseMotion(x, y))
        glutPostRedisplay();
    }

    void GLUTClassifierPassiveMouseMotion(int x, int y) {
        if (visualized_classifier
                && visualized_classifier->PassiveMouseMotion(x, y))
        glutPostRedisplay();
    }

    void GLUTClassifierStop(void) {
        if (visualized_classifier) visualized_classifier->Stop();
    }

    int R3SurfelClassifier::AttemptGLUTInit(VisualizationStyle style) {
        // Only one instance can be visualized.
        if (style == R3_CLASSIFIER_VISUAL_STYLE_NONE
                || style >= R3_CLASSIFIER_VISUAL_STYLE_NUM) {
            return 0;
        }
        if (visualized_classifier) {
            fprintf(stderr,
                    "Only one classifier can be visualized. Attempt failed.\n");
            return 0;
        }

        // Open window
        // GLUT should already be initialized in the labeler
        glutInitWindowPosition(200, 200);
        glutInitWindowSize(vis_window_width_, vis_window_height_);
        vis_window_ = glutCreateWindow("UGOR: Classifier visualization");
        if (!vis_window_) return 0;

        // Initialize GLUT callback functions for this window
        glutDisplayFunc(GLUTClassifierRedraw);
        glutReshapeFunc(GLUTClassifierResize);
        glutKeyboardFunc(GLUTClassifierKeyboard);
        glutSpecialFunc(GLUTClassifierSpecial);
        glutMouseFunc(GLUTClassifierMouseButton);
        glutMotionFunc(GLUTClassifierMouseMotion);
        glutPassiveMotionFunc(GLUTClassifierPassiveMouseMotion);
        atexit(GLUTClassifierStop);

        visualized_classifier = this;
        vis_style_ = style;

        // Return success.
        return 1;
    }

    int R3SurfelClassifier::RequestRedraw(void) {
        if (!vis_window_) return 0;

        glutPostWindowRedisplay(vis_window_);

        // Return success.
        return 1;
    }

    void R3SurfelClassifier::DrawNoDataScreen(void) const {
        // Check everything
        if (!vis_window_) return;
        if (draw_purpose_ != DRAW_PURPOSE_DRAW) return;

        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINES);
        glVertex2f(vis_clipping_box_.XMin(), vis_clipping_box_.YMin());
        glVertex2f(vis_clipping_box_.XMax(), vis_clipping_box_.YMax());
        glVertex2f(vis_clipping_box_.XMax(), vis_clipping_box_.YMin());
        glVertex2f(vis_clipping_box_.XMin(), vis_clipping_box_.YMax());
        glEnd();
    }

    int R3SurfelClassifier::Redraw(void) {
        // Check everything.
        if (!vis_window_) return 0;

        RefitDataToScreen();

        glClear(GL_COLOR_BUFFER_BIT);

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        gluOrtho2D(vis_clipping_box_.XMin(), vis_clipping_box_.XMax(),
                vis_clipping_box_.YMin(), vis_clipping_box_.YMax());

        DrawTicks();

        switch (vis_style_) {
            case R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE:
            RedrawTree(static_cast<ShapeTreeNode*>(NULL));
            DrawStats();
            break;
            case R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE:
            RedrawTree(GetGroupRootNode());
            DrawStats();
            break;
            case R3_CLASSIFIER_VISUAL_STYLE_MDS:
            case R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF:
            case R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED:
            RedrawMDS();
            DrawStats();
            break;
            default:
            DrawNoDataScreen();
            break;
        }
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();

        return 0;
    }

    void R3SurfelClassifier::RefitDataToScreen(void) {
        // Check everything.
        if (!vis_window_) return;
        if (!refit_data_to_screen_) return;

        refit_data_to_screen_ = false;
        selection_container_.Reset();

        // Update vis_space_box_.
        vis_space_box_ = R2null_box;
        draw_purpose_ = DRAW_PURPOSE_FIT;
        Redraw();
        draw_purpose_ = DRAW_PURPOSE_DRAW;

        // Update clipping_window_.
        vis_clipping_box_ = vis_space_box_;
        R2Affine transform = R2identity_affine;
        transform.Translate(vis_clipping_box_.Centroid() - R2null_point);
        transform.Scale(1.1);
        transform.Translate(R2null_point - vis_clipping_box_.Centroid());
        vis_clipping_box_.Transform(transform);
    }

    void R3SurfelClassifier::RecenterOnSelection(void) {
        // Check everything.
        if (!vis_window_) return;

        // If already seen, don't alter anything.
        if (labeler_->NObjectSelections() != 1) return;

        // Update vis_space_box_.
        draw_purpose_ = DRAW_PURPOSE_FIT_SELECTION;
        vis_selection_box_ = R2null_box;
        Redraw();
        draw_purpose_ = DRAW_PURPOSE_DRAW;

        if (vis_selection_box_.IsEmpty()) return;

        // Update clipping_window_.
        if (vis_selection_box_.Area()) {
            RNScalar x_ratio = vis_selection_box_.XLength()
            / vis_space_box_.XLength();
            RNScalar y_ratio = vis_selection_box_.YLength()
            / vis_space_box_.YLength();
            RNScalar scale = 1.1 * max(x_ratio, y_ratio);
            RNScalar width = scale * vis_space_box_.XLength() / 2;
            RNScalar height = scale * vis_space_box_.YLength() / 2;
            vis_clipping_box_ = R2Box(vis_selection_box_.Centroid(), width, height);
        } else if (!R2Contains(vis_clipping_box_, vis_selection_box_)) {
            RNScalar width = 0.1 * vis_space_box_.XLength();
            RNScalar height = 0.1 * vis_space_box_.YLength();
            vis_clipping_box_ = R2Box(vis_selection_box_.Centroid(), width, height);
        }
    }

    void R3SurfelClassifier::SetEvaluationGroup(
            const RNArray<R3SurfelObject*> &group) {
        if (evaluation_group_) ClearEvaluationGroup();

        evaluation_group_ = new GroupTreeNode(this);
        for (int i = 0; i < group.NEntries(); ++i) {
            ObjectWrapper *ow = object_leaf_[group.Kth(i)];
            if (!ow) continue;
            evaluation_group_->add_child(ow->group_node());
        }
    }

    void R3SurfelClassifier::ClearEvaluationGroup(void) {
        if (!evaluation_group_) return;
        delete evaluation_group_;
        evaluation_group_ = NULL;
    }

    void R3SurfelClassifier::LoadColorSizeByScheme(R3SurfelObject *object) {
        if (!object) return;

        switch (color_scheme_) {
            case COLORING_SCHEME_CONFIDENCE: {
                R3SurfelLabelAssignment *a = object->PredictedLabelAssignment();
                R3SurfelLabel *label = a->Label();
                RNScalar conf = a->Confidence();

                // Don't draw unselectible labels.
                if (label && !labeler_->LabelVisible(label)) label = NULL;

                if (label) {
                    glPointSize(4);
                    RNScalar val = 2 * conf;
                    if (val > 1) glColor3f(1.0, val - 1.0, 0.0);
                    else glColor3f(val, 0.0, 1.0 - val);

                } else {
                    glPointSize(2);
                    glColor4f(0.5, 0.5, 0.5, 0.5);
                }
            }
            break;
            case COLORING_SCHEME_AFFINITY:
            glPointSize(4);

            if (labeler_ && labeler_->NObjectSelections() == 1) {
                R3SurfelObject *o = labeler_->ObjectSelection(0);
                RNScalar value = 1.0 / (1.0 + GetFeatureDistance(o, object));
                RNScalar val = 2 * value;
                if (val > 1) glColor3f(1.0, val - 1.0, 0.0);
                else glColor3f(val, 0.0, 1.0 - val);
            } else {
                glColor4f(0.0, 0.0, 1.0, 1.0);
            }
            break;

            case COLORING_SCHEME_INDUCED_AFFINITY:
            glPointSize(4);

            if (labeler_ && labeler_->NObjectSelections() == 1) {
                R3SurfelObject *o = labeler_->ObjectSelection(0);
                RNScalar value = GetInducedAffinity(o, object);
                RNScalar val = 2 * value;
                if (val > 1) glColor3f(1.0, val - 1.0, 0.0);
                else glColor3f(val, 0.0, 1.0 - val);
            } else {
                glColor4f(0.0, 0.0, 1.0, 1.0);
            }
            break;

            case COLORING_SCHEME_GT: {
                R3SurfelLabel *label = object->GroundTruthLabel();
                // Don't draw unselectible labels.
                if (label && !labeler_->LabelVisible(label)) label = NULL;

                if (label) {
                    glPointSize(4);
                    RNLoadRgb(label->Color());
                } else {
                    glPointSize(2);
                    glColor4f(0.5, 0.5, 0.5, 0.5);
                }
                break;
            }

            case COLORING_SCHEME_ASSIGNMENT:
            default: {
                // Set point size
                if (Labeled(object)) glPointSize(4);
                else glPointSize(2);

                R3SurfelLabel *label = object->PredictedLabel();

                // Don't draw unselectible labels.
                if (label && !labeler_->LabelVisible(label)) label = NULL;

                // Set point color
                if (label) RNLoadRgb(label->Color());
                else glColor4f(0.5, 0.5, 0.5, 0.5);
            }
            break;
        }
    }

    void R3SurfelClassifier::CustomPutVertex(RNScalar x, RNScalar y) {
        switch (draw_purpose_) {
            case DRAW_PURPOSE_FIT:
            vis_space_box_.Union(R2Point(x, y));
            break;
            case DRAW_PURPOSE_FIT_SELECTION:
            if (labeler_) {
                R3SurfelObject *o = selection_container_.current_object();
                if (o && labeler_->selection_objects.FindEntry(o))
                vis_selection_box_.Union(R2Point(x, y));
            }
            break;
            case DRAW_PURPOSE_SELECT:
            selection_container_.try_update_with_current(R2Point(x, y));
            break;
            case DRAW_PURPOSE_DRAW:
            default:
            glBegin(GL_POINTS);
            glVertex2f(x, y);
            glEnd();
            break;
        }
    }

    bool R3SurfelClassifier::DrawLeafNodeAtPoint(ObjectWrapper* ow,
            const RNScalar x, const RNScalar y, const RNRgb *override_color) {
        bool highlight = false;

        // Check everything.
        if (!ow) return highlight;
        R3SurfelObject *object = ow->object();
        if (!object) return highlight;

        if (draw_purpose_ == DRAW_PURPOSE_DRAW) {
            if (last_labeled.count(ow)) {
                // Highlight last labeled.
                glColor4f(1.0, 0.0, 1.0, 0.7);// cyan
                glPointSize(8);
                glBegin(GL_POINTS);
                glVertex2f(x, y);
                glEnd();

                glColor4f(0.0, 0.0, 0.0, 1.0);// black
                glPointSize(6);
                glBegin(GL_POINTS);
                glVertex2f(x, y);
                glEnd();
            }

            if (labeler_ && labeler_->NObjectSelections()
                    && labeler_->selection_objects.FindEntry(object)) {
                highlight = true;

                // Highlight selected object.
                glColor4f(1.0, 1.0, 0.0, 0.7);// solid yellow
                glPointSize(8);
                glBegin(GL_POINTS);
                glVertex2f(x, y);
                glEnd();

                glColor4f(0.0, 0.0, 0.0, 1.0);// black
                glPointSize(6);
                glBegin(GL_POINTS);
                glVertex2f(x, y);
                glEnd();
            }
        }

        // Draw point
        if (override_color) {
            glPointSize(4);
            RNLoadRgb(*override_color);
        } else LoadColorSizeByScheme(object);
        CustomPutVertex(x, y);

        return highlight;
    }

    void R3SurfelClassifier::DrawStats() {
        // Check everything
        if (!vis_window_) return;
        if (draw_purpose_ != DRAW_PURPOSE_DRAW) return;

        // Acquire information about a selected node.
        list<string> data;
        char buffer[1024];

        int selection_num = labeler_ ? labeler_->NObjectSelections() : 0;
        sprintf(buffer, " Selected %d objects.", selection_num);
        data.push_back(buffer);
        data.push_back("-----------------");
        data.push_back("");

        if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE) {
            if (selection_container_.something_selected()) {
                data.push_back("	-----------------");
                data.push_back("	Selection container node");
                selection_container_.FillStats(data);
            }
            if (last_suggestion_node_) {
                data.push_back("	-----------------");
                data.push_back("	Query group node");
                last_suggestion_node_->FillStats(data);
            }
            if (evaluation_group_) {
                data.push_back("	-----------------");
                data.push_back("	Evaluation group node");
                evaluation_group_->FillStats(data);
            }
            if (labeler_ && labeler_->NObjectSelections() == 1) {
                data.push_back("	-----------------");
                data.push_back("	Selected object node");
                object_leaf_[labeler_->ObjectSelection(0)]->group_node()->FillStats(
                        data);
            }

            data.push_back("	-----------------");
            data.push_back("	Max benefit cut (red)");
            max_benefit_cut_.FillStats(data);
        } else {
            if (labeler_ && labeler_->NObjectSelections() == 1) object_leaf_[labeler_->ObjectSelection(
                    0)]->shape_node()->FillStats(data);
            else selection_container_.FillStats(data);
        }

        void *font = GLUT_BITMAP_HELVETICA_10;
        int longest_line_length = 0;
        for (list<string>::iterator it = data.begin(); it != data.end(); ++it)
        longest_line_length = max(longest_line_length,
                glutBitmapLength(font, (const unsigned char *) it->c_str()));

        // Get convenient variables
        int width = vis_window_width_;
        int height = vis_window_height_;
        int line_height = 11;

        // upper right corner - less likely interference with the tree.
        int x = width - 16 - (longest_line_length + 8);
        int y = height - 16;
        if (0 && vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE) {
            // upper left corner - less likely interference with the plots.
            x = 16;
            y = height - 16;
        }

        // Set OpenGL modes
        glDisable(GL_LIGHTING);
        glColor3f(0, 0, 0);
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        gluOrtho2D(0, width, 0, height);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);

        // Draw framing box.
        const int y_min_drawable = 8;
        vis_stats_box_ = R2Box(x - 8,
                max(y - (int) data.size() * line_height - 8, y_min_drawable),
                x + longest_line_length + 8, y + 8);

        RNScalar fade_alpha =
        R2Contains(vis_stats_box_,
                R2Point(mouse_screen_coords_[0], mouse_screen_coords_[1])) ?
        0.2 : 1.0;

        glColor4d(0, 0, 0, 0.7 * fade_alpha);
        vis_stats_box_.Draw();
        glColor4d(1, 0, 0, 1 * fade_alpha);
        vis_stats_box_.Outline();

        // Draw data.
        glColor4d(1, 1, 1, 1 * fade_alpha);
        for (list<string>::const_iterator it = data.begin(); it != data.end();
                ++it) {
            y -= line_height;
            R2Point name_origin(x, y);
            if ((y - line_height) < y_min_drawable) {
                DrawText(name_origin, "   ...   ", font);
                break;
            } else {
                DrawText(name_origin, it->c_str(), font);
            }
        }

        // Reset OpenGL modes
        glDisable(GL_BLEND);
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
    }

    void R3SurfelClassifier::DrawEvaluationGroup(bool more_info) {
        if (!evaluation_group_) return;

        evaluation_group_->DrawDebug(more_info);
    }

    void R3SurfelClassifier::DrawTicks() {
        if (draw_purpose_ != DRAW_PURPOSE_DRAW) return;

        RNLoadRgb(RNwhite_rgb);

        RNScalar vcb_width = vis_clipping_box_.XLength();
        RNScalar vcb_height = vis_clipping_box_.YLength();
        RNScalar x_pad = 0.05 * vcb_width;
        RNScalar y_pad = 0.05 * vcb_height;
        RNScalar x_half_pad = 0.5 * x_pad;
        RNScalar y_half_pad = 0.5 * y_pad;
        RNScalar x_min = vis_clipping_box_.XMin() + x_pad;
        RNScalar y_min = vis_clipping_box_.YMin() + y_pad;
        RNScalar x_max = vis_clipping_box_.XMax() - x_pad;
        RNScalar y_max = vis_clipping_box_.YMax() - y_pad;
        RNScalar x_mid = 0.5 * (x_min + x_max);
        RNScalar y_mid = 0.5 * (y_min + y_max);

        glLineWidth(1);
        glBegin(GL_LINES);

        // left vertical bar.
        glVertex2d(x_min - x_half_pad, y_min);
        glVertex2d(x_min - x_pad, y_min);

        glVertex2d(x_min - x_half_pad, y_mid);
        glVertex2d(x_min - x_pad, y_mid);

        glVertex2d(x_min - x_half_pad, y_max);
        glVertex2d(x_min - x_pad, y_max);

        // right vertical bar
        glVertex2d(x_max + x_half_pad, y_min);
        glVertex2d(x_max + x_pad, y_min);

        glVertex2d(x_max + x_half_pad, y_mid);
        glVertex2d(x_max + x_pad, y_mid);

        glVertex2d(x_max + x_half_pad, y_max);
        glVertex2d(x_max + x_pad, y_max);

        // bottom horizontal bar
        glVertex2d(x_min, y_min - y_half_pad);
        glVertex2d(x_min, y_min - y_pad);

        glVertex2d(x_mid, y_min - y_half_pad);
        glVertex2d(x_mid, y_min - y_pad);

        glVertex2d(x_max, y_min - y_half_pad);
        glVertex2d(x_max, y_min - y_pad);

        // top horizontal bar
        glVertex2d(x_min, y_max + y_half_pad);
        glVertex2d(x_min, y_max + y_pad);

        glVertex2d(x_mid, y_max + y_half_pad);
        glVertex2d(x_mid, y_max + y_pad);

        glVertex2d(x_max, y_max + y_half_pad);
        glVertex2d(x_max, y_max + y_pad);

        glEnd();

        RNScalar pix_w = 2 * vis_clipping_box_.XLength() / vis_window_width_;
        RNScalar pix_h = 2 * vis_clipping_box_.YLength() / vis_window_height_;

        char value[100];
        sprintf(value, "%.2g", y_min);
        DrawText(R2Point(x_min - x_pad + pix_w, y_min + pix_h), value);
        sprintf(value, "%.2g", 0.5 * (y_min + y_max));
        DrawText(R2Point(x_min - x_pad + pix_w, y_mid + pix_h), value);
        sprintf(value, "%.2g", y_max);
        DrawText(R2Point(x_min - x_pad + pix_w, y_max + pix_h), value);

        sprintf(value, "%.2g", x_min);
        DrawText(R2Point(x_min + pix_w, y_min - y_pad + pix_h), value);
        sprintf(value, "%.2g", 0.5 * (x_min + x_max));
        DrawText(R2Point(x_mid + pix_w, y_min - y_pad + pix_h), value);
        sprintf(value, "%.2g", x_max);
        DrawText(R2Point(x_max + pix_w, y_min - y_pad + pix_h), value);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        glColor4f(1.0, 1.0, 1.0, 0.25);
        glBegin(GL_LINES);

        // Horizontal lines.
        glVertex2d(x_min - x_half_pad, y_min);
        glVertex2d(x_max + x_half_pad, y_min);

        glVertex2d(x_min - x_half_pad, y_mid);
        glVertex2d(x_max + x_half_pad, y_mid);

        glVertex2d(x_min - x_half_pad, y_max);
        glVertex2d(x_max + x_half_pad, y_max);

        // Vertical lines.
        glVertex2d(x_min, y_min - y_half_pad);
        glVertex2d(x_min, y_max + y_half_pad);

        glVertex2d(x_mid, y_min - y_half_pad);
        glVertex2d(x_mid, y_max + y_half_pad);

        glVertex2d(x_max, y_min - y_half_pad);
        glVertex2d(x_max, y_max + y_half_pad);

        glEnd();

        glDisable(GL_BLEND);
    }

    void R3SurfelClassifier::ForceBlockingMessage(const char* msg) {
        // For lengthy blocking procedures, draw something distracting (a message).

        if (!vis_window_) return;

        glClear(GL_COLOR_BUFFER_BIT);

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        gluOrtho2D(0, 1, 0, 1);

        const char* message = msg ? msg : "Wait. Magic is happening..";

        RNScalar x_shift = (10 * strlen(message) / 2)
        / (RNScalar) vis_window_width_;

        glColor3f(1, 1, 1);
        DrawText(R2Point(0.5 - x_shift, 0.5), message, GLUT_BITMAP_HELVETICA_18);

        glMatrixMode(GL_PROJECTION);
        glPopMatrix();

        glutSwapBuffers();
    }

    void R3SurfelClassifier::RedrawMDS(void) {
        if (vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS
                && vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF
                && vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED) {
            DrawNoDataScreen();
            return;
        }

        if (!objects_.size()) {
            DrawNoDataScreen();
            return;
        }

        UpdateMDSifNecessary();

        vector<R2Point> &points =
        (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS) ? mds_fs_locations_ :
        (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF) ?
        mds_aff_locations_ : mds_induced_locations_;

        if (points.size() != objects_.size()) return;

        for (int i = 0; (size_t) i < objects_.size(); ++i) {
            ShapeTreeNode *node = objects_[i]->shape_node();
            R2Point &p = points[i];
            selection_container_.set_current(node);
            DrawLeafNodeAtPoint(node->object_, p[0], p[1]);
        }
    }

    double **
    MultiDimensionalScaling(float *distance_matrix, int nmodels, int ndimensions) {
        // Borrowed from GAPS psbMDS. Performs a multi-dimensional scaling given the
        // distance matrix.

        // Copy matrix
        RNScalar *matrix = new RNScalar[nmodels * nmodels];
        RNScalar *matrixp = matrix;
        float *distance_matrixp = distance_matrix;
        for (int i = 0; i < nmodels; i++) {
            for (int j = 0; j < nmodels; j++) {
                RNScalar d = *(distance_matrixp++);
                *(matrixp++) = d * d;
            }
        }

        // Compute sums (first step towards means)
        double overall_mean = 0;
        RNScalar *row_mean = new RNScalar[nmodels];
        RNScalar *column_mean = new RNScalar[nmodels];
        for (int i = 0; i < nmodels; i++) {
            row_mean[i] = 0;
            column_mean[i] = 0;
        }
        matrixp = matrix;
        for (int i = 0; i < nmodels; i++) {
            for (int j = 0; j < nmodels; j++) {
                double distance = *(matrixp++);
                row_mean[i] += distance;
                column_mean[j] += distance;
                overall_mean += distance;
            }
        }

        // Divide by nmodels to get mean
        overall_mean /= nmodels * nmodels;
        for (int i = 0; i < nmodels; i++) {
            row_mean[i] /= nmodels;
            column_mean[i] /= nmodels;
        }

        // Double center matrix
        for (int i = 0; i < nmodels; i++) {
            for (int j = 0; j < nmodels; j++) {
                matrix[i * nmodels + j] = -0.5F
                * (matrix[i * nmodels + j] - row_mean[i] - column_mean[j]
                        + overall_mean);
            }
        }

        // Singular value decomposition
        RNScalar *U = new RNScalar[nmodels * nmodels];
        RNScalar *W = new RNScalar[nmodels];
        RNScalar *V = new RNScalar[nmodels * nmodels];
        RNSvdDecompose(nmodels, nmodels, matrix, U, W, V);

        // Compute square root of W
        for (int i = 0; i < nmodels; i++) {
            if (W[i] < 0) W[i] = 0;
            else W[i] = sqrt(W[i]);
        }

        // Compute coordinates: X = U * W;
        RNScalar *X = new RNScalar[nmodels * nmodels];
        RNScalar *Xp = X, *Up = U;
        for (int i = 0; i < nmodels; i++) {
            for (int j = 0; j < nmodels; j++) {
                *(Xp++) = *(Up++) * W[j];
            }
        }

        // Fill arrays of coordinates
        double **coordinates = new double *[nmodels];
        assert(coordinates);
        for (int i = 0; i < nmodels; i++) {
            coordinates[i] = new double[ndimensions];
            assert(coordinates[i]);
            for (int j = 0; j < ndimensions; j++) {
                coordinates[i][j] = X[i * nmodels + j];
            }
        }

        // Delete arrays
        delete[] U;
        delete[] W;
        delete[] V;
        delete[] X;

        // Return coordinates
        return coordinates;
    }

    void R3SurfelClassifier::UpdateMDSifNecessary(void) {
        // Avoid unnecessary updates.
        if ((vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS || !update_mds_fs_)
                && (vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF
                        || !update_mds_aff_)
                && (vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED
                        || !update_mds_induced_)) {
            return;
        }

        // Tell user it's gonna take a while.
        ForceBlockingMessage("Computing MDS..");

        // Update respective MDS conatiner.
        vector<R2Point> &points =
        (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS) ? mds_fs_locations_ :
        (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF) ?
        mds_aff_locations_ : mds_induced_locations_;

        // Produce distance matrix.
        const int nmodels = objects_.size();
        const int dim = 2;
        float* distance_matrix = new float[nmodels * nmodels];
        for (int i = 0; i < nmodels; ++i) {
            distance_matrix[i * nmodels + i] = 0;
            for (int j = 0; j < i; ++j) {
                RNScalar distance = kMaxScalar;

                if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS) {
                    distance = GetFeatureDistance(i, j);
                } else if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF) {
                    RNScalar v = GetShapeAffinity(i, j);
                    distance = (v > 0.0 ? 1.0 / v - 1.0 : RN_INFINITY);
                } else if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED) {
                    RNScalar v = GetInducedAffinity(i, j);
                    distance = (v > 0.0 ? 1.0 / v - 1.0 : RN_INFINITY);
                }

                distance_matrix[i * nmodels + j] = distance;
                distance_matrix[j * nmodels + i] = distance_matrix[i * nmodels + j];
            }
        }

        // Update MDS coordinates.
        double **coordinates = MultiDimensionalScaling(distance_matrix, nmodels,
                dim);
        if (!coordinates) {
            fprintf(stderr, "MDS coordinates could not be computed.");
            points.clear();
            return;
        }

        // Store locations.
        points.resize(nmodels);
        for (int i = 0; i < nmodels; ++i) {
            for (int j = 0; j < dim; ++j) {
                points[i][j] = coordinates[i][j];
            }
        }

        // Uncheck update request flags.
        if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS) update_mds_fs_ = false;
        else if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF) update_mds_aff_ =
        false;
        else if (vis_style_ == R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED)
        update_mds_induced_ = false;

        // Release resources.
        delete[] distance_matrix;
        for (int i = 0; i < nmodels; ++i)
        delete[] coordinates[i];
        delete[] coordinates;
    }

    template<>
    void R3SurfelClassifier::TryDrawCuts<GroupTreeNode>(GroupTreeNode *node,
            RNScalar x, RNScalar y) {
        if (!node) return;

        glLineStipple(1, 1);
        glLineWidth(4.0);
        glColor4f(1.0, 0.0, 1.0, 0.5);	// dim magenta
        max_benefit_cut_.TryDraw(node, x, y);
    }

    void R3SurfelClassifier::DrawBranch(RNScalar tgt_x, RNScalar tgt_y,
            RNScalar src_x, RNScalar src_y, const bool highlight_downstream,
            const bool highlight_upstream) {
        if (highlight_upstream) {
            glColor4f(1.0, 1.0, 0.0, 0.3);  //  dim yellow
            glLineWidth(2);
        } else if (highlight_downstream) {
            glColor4f(0.0, 1.0, 0.0, 0.3);  //  dim green
            glLineWidth(2);
        } else {
            glColor4f(0.5, 0.5, 0.5, 0.3);  //  grey
            glLineWidth(1);
        }
        glBegin(GL_LINE_STRIP);
        glVertex2f(src_x, src_y);
        glVertex2f(src_x, tgt_y);
        glVertex2f(tgt_x, tgt_y);
        glEnd();
    }

    int R3SurfelClassifier::Resize(int w, int h) {
        // Resize viewport.
        glViewport(0, 0, w, h);

        // Resize clipping window.
        RNScalar width = vis_clipping_box_.XLength() * w / vis_window_width_;
        RNScalar height = vis_clipping_box_.YLength() * h / vis_window_height_;
        vis_clipping_box_ = R2Box(vis_clipping_box_.Centroid(), width / 2,
                height / 2);

        // Remember window dimensions
        vis_window_width_ = w;
        vis_window_height_ = h;

        return 1;
    }

    int R3SurfelClassifier::Keyboard(unsigned char key, int x, int y) {
        // Invert y coordinate
        y = vis_window_height_ - y;

        // Determine modifiers
        int modifiers = glutGetModifiers();
        // int shift = (modifiers & GLUT_ACTIVE_SHIFT);
        int ctrl = (modifiers & GLUT_ACTIVE_CTRL);
        int alt = (modifiers & GLUT_ACTIVE_ALT);

        int redraw = 0;

        if (ctrl) {
            switch (key) {
                case 'C':
                case 'c':
                // Change coloring scheme.
                color_scheme_ = (ColoringScheme) ((color_scheme_ + 1)
                        % COLORING_SCHEME_NUM);
                redraw = 1;
                break;

                case 'E':
                case 'e':
                SetEvaluationGroup(labeler_->selection_objects);
                labeler_->ZoomToObjectSet(labeler_->selection_objects, true);
                redraw = 1;
                break;

                case 'S':
                case 's':
                if (labeler_->highlight_group_.NEntries()) {
                    // Simulate selection in the labeler_.
                    labeler_->BeginCommand(6);
                    labeler_->EmptyObjectSelections();
                    for (int i = 0; i < labeler_->highlight_group_.NEntries();
                            ++i) {
                        R3SurfelObject *o = labeler_->highlight_group_.Kth(i);
                        labeler_->InsertObjectSelection(o);
                    }
                    labeler_->EndCommand();
                    labeler_->ZoomToObjectSet(labeler_->highlight_group_, true);
                    update_main_window = true;
                    redraw = 1;
                }
                break;

                case 'V':
                case 'v':
                // Change visualization type.
                if (vis_window_) {
                    vis_style_ = (VisualizationStyle) (vis_style_
                            % (R3_CLASSIFIER_VISUAL_STYLE_NUM - 1) + 1);
                    refit_data_to_screen_ = true;
                    redraw = 1;
                }
                break;
            }
        } else if (alt) {
            switch (key) {
                case '1':
                if (vis_window_
                        && vis_style_
                        != R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE) {
                    vis_style_ = R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE;
                    refit_data_to_screen_ = true;
                    redraw = 1;
                }
                break;
                case '2':
                if (vis_window_
                        && vis_style_
                        != R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE) {
                    vis_style_ = R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE;
                    refit_data_to_screen_ = true;
                    redraw = 1;
                }
                break;
                case '3':
                if (vis_window_
                        && vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS) {
                    vis_style_ = R3_CLASSIFIER_VISUAL_STYLE_MDS;
                    refit_data_to_screen_ = true;
                    redraw = 1;
                }
                break;

                case '4':
                if (vis_window_
                        && vis_style_ != R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF) {
                    vis_style_ = R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF;
                    refit_data_to_screen_ = true;
                    redraw = 1;
                }
                break;

                case '5':
                if (vis_window_
                        && vis_style_
                        != R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED) {
                    vis_style_ = R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED;
                    refit_data_to_screen_ = true;
                    redraw = 1;
                }
                break;
            }
        } else {
            switch (key) {
                case 32:	// space
                refit_data_to_screen_ = true;
                redraw = 1;
                break;
            }
        }

        return redraw;
    }

    int R3SurfelClassifier::KeyboardSpecial(int key, int x, int y) {
        // Invert y coordinate
        y = vis_window_height_ - y;

        // Determine modifiers
        int modifiers = glutGetModifiers();
        int shift = (modifiers & GLUT_ACTIVE_SHIFT);
        // int ctrl = (modifiers & GLUT_ACTIVE_CTRL);
        // int alt = (modifiers & GLUT_ACTIVE_ALT);

        int redraw = 0;

        R2Affine transform = R2identity_affine;
        switch (key) {
            case GLUT_KEY_UP:
            if (shift) {
                // Scale up vertically.
                transform.Translate(
                        vis_clipping_box_.Centroid() - R2null_point);
                transform.Scale(RN_Y, 0.5);
                transform.Translate(
                        R2null_point - vis_clipping_box_.Centroid());
            } else {
                // Move up.
                transform.YTranslate(0.05 * vis_clipping_box_.YLength());
            }
            break;
            case GLUT_KEY_DOWN:
            if (shift) {
                // Scale down vertically.
                transform.Translate(
                        vis_clipping_box_.Centroid() - R2null_point);
                transform.Scale(RN_Y, 2.0);
                transform.Translate(
                        R2null_point - vis_clipping_box_.Centroid());
            } else {
                // Move down.
                transform.YTranslate(-0.05 * vis_clipping_box_.YLength());
            }
            break;
            case GLUT_KEY_LEFT:
            if (shift) {
                // Scale down horizontally.
                transform.Translate(
                        vis_clipping_box_.Centroid() - R2null_point);
                transform.Scale(RN_X, 2.0);
                transform.Translate(
                        R2null_point - vis_clipping_box_.Centroid());
            } else {
                // Shift left.
                transform.XTranslate(-0.05 * vis_clipping_box_.XLength());
            }
            break;
            case GLUT_KEY_RIGHT:
            if (shift) {
                // Scale down horizontally.
                transform.Translate(
                        vis_clipping_box_.Centroid() - R2null_point);
                transform.Scale(RN_X, 0.5);
                transform.Translate(
                        R2null_point - vis_clipping_box_.Centroid());
            } else {
                // Shift right.
                transform.XTranslate(0.05 * vis_clipping_box_.XLength());
            }
            break;
            case GLUT_KEY_PAGE_UP:
            if (shift) {
                // Move highlight node up the group tree.
                if (labeler_) {
                    vis_style_ = R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE;

                    GroupTreeNode *node = NULL;
                    selection_container_.get_closest(&node);
                    if (node && node->parent_) {
                        selection_container_.set_closest(node->parent_);
                        ObjectsInSubtree(node->parent_,
                                labeler_->highlight_group_);
                        labeler_->ZoomToObjectSet(labeler_->highlight_group_,
                                false);
                        redraw = 1;
                        update_main_window = true;
                    }
                }
            } else {
                // Zoom in.
                transform.Translate(
                        vis_clipping_box_.Centroid() - R2null_point);
                transform.Scale(0.9);
                transform.Translate(
                        R2null_point - vis_clipping_box_.Centroid());
            }
            break;
            case GLUT_KEY_PAGE_DOWN:
            // Zoom out.
            transform.Translate(vis_clipping_box_.Centroid() - R2null_point);
            transform.Scale(1.1);
            transform.Translate(R2null_point - vis_clipping_box_.Centroid());
            break;
            default:
            break;
        }

        if (!transform.Matrix().IsIdentity()) {
            vis_clipping_box_.Transform(transform);
            redraw = 1;
        }

        return redraw;
    }

    int R3SurfelClassifier::MouseButton(int button, int state, int x, int y) {
        // Invert y coordinate
        y = vis_window_height_ - y;

        mouse_screen_coords_[0] = x;
        mouse_screen_coords_[1] = y;

        int redraw = 0;

        if (button > 2) redraw = MouseWheel(button, button - 3, x, y);
        else {
            mouse_button_[button] = !state;

            if (state == GLUT_UP) {
                // Mouse button up.
                if (button == 0) {
                    // Left mouse button.
                    // Check for a dragging motion.
                    if (mouse_drag_distance_squared_ > 4) {
                        // Nothing to do for dragging yet..
                    } else {
                        // Clicking.
                        draw_purpose_ = DRAW_PURPOSE_SELECT;
                        R2Point s_center(
                                vis_clipping_box_.XLength() * x / vis_window_width_
                                + vis_clipping_box_.XMin(),
                                vis_clipping_box_.YLength() * y / vis_window_height_
                                + vis_clipping_box_.YMin());
                        RNScalar s_x_r = 5 * vis_clipping_box_.XLength()
                        / vis_window_width_;
                        RNScalar s_y_r = 5 * vis_clipping_box_.YLength()
                        / vis_window_height_;
                        selection_container_.Reset(R2Box(s_center, s_x_r, s_y_r));
                        Redraw();
                        draw_purpose_ = DRAW_PURPOSE_DRAW;

                        if (labeler_) {
                            labeler_->highlight_group_.Empty();

                            // Simulate selection in the labeler_.
                            labeler_->BeginCommand(6);
                            labeler_->EmptyObjectSelections();
                            R3SurfelObject *o =
                            selection_container_.object_selected();
                            if (o) {
                                labeler_->InsertObjectSelection(o);
                                labeler_->SetCenterPoint(o->Centroid());
                                if (labeler_->animation_on_)
                                labeler_->ZoomToObject(o, true);
                                selection_container_.Reset();
                            } else if (selection_container_.something_selected()) {
                                if (vis_style_
                                        == R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE) {
                                    GroupTreeNode *node;
                                    selection_container_.get_closest(&node);
                                    ObjectsInSubtree(node,
                                            labeler_->highlight_group_);
                                } else if (vis_style_
                                        == R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE) {
                                    ShapeTreeNode *node;
                                    selection_container_.get_closest(&node);
                                    ObjectsInSubtree(node,
                                            labeler_->highlight_group_);
                                }
                            }
                            labeler_->EndCommand();
                            update_main_window = true;
                        }
                    }
                }
            } else {
                // Mouse button down.
                mouse_drag_distance_squared_ = 0;
            }

            redraw = 1;
        }

        return redraw;
    }

    int R3SurfelClassifier::MouseMotion(int x, int y) {
        // Invert y coordinate
        y = vis_window_height_ - y;

        // Useful variables.
        int dx = x - mouse_screen_coords_[0];
        int dy = y - mouse_screen_coords_[1];
        mouse_drag_distance_squared_ += dx * dx + dy * dy;

        int redraw = 0;

        // Check if stats need to be redrawn.
        RNBoolean was_in = R2Contains(vis_stats_box_,
                R2Point(mouse_screen_coords_[0], mouse_screen_coords_[1]));
        RNBoolean is_in = R2Contains(vis_stats_box_, R2Point(x, y));
        if (was_in != is_in) redraw = 1;

        // Drag motion.
        R2Affine transform = R2identity_affine;
        if (mouse_button_[0]) {
            transform.Translate(
                    R2Vector(
                            -(vis_clipping_box_.XLength() * dx) / vis_window_width_,
                            -(vis_clipping_box_.YLength() * dy)
                            / vis_window_height_));
        }

        if (!transform.Matrix().IsIdentity()) {
            vis_clipping_box_.Transform(transform);
            redraw = 1;
        }

        mouse_screen_coords_[0] = x;
        mouse_screen_coords_[1] = y;

        return redraw;
    }

    int R3SurfelClassifier::PassiveMouseMotion(int x, int y) {
        // Invert y coordinate
        y = vis_window_height_ - y;

        int redraw = 0;

        // Check if stats need to be redrawn.
        RNBoolean was_in = R2Contains(vis_stats_box_,
                R2Point(mouse_screen_coords_[0], mouse_screen_coords_[1]));
        RNBoolean is_in = R2Contains(vis_stats_box_, R2Point(x, y));
        if (was_in != is_in) redraw = 1;

        mouse_screen_coords_[0] = x;
        mouse_screen_coords_[1] = y;

        return redraw;
    }

    int R3SurfelClassifier::MouseWheel(int button, int dir, int x, int y) {
        R2Vector shift(
                vis_clipping_box_.XLength() * x / vis_window_width_
                + vis_clipping_box_.XMin(),
                vis_clipping_box_.YLength() * y / vis_window_height_
                + vis_clipping_box_.YMin());
        RNScalar scale = dir > 0 ? 1.1 : 0.9;

        R2Affine transform = R2identity_affine;
        transform.Translate(shift);
        transform.Scale(scale);
        transform.Translate(-shift);

        vis_clipping_box_.Transform(transform);

        return 1;
    }

    void R3SurfelClassifier::Stop(void) {
        // Destroy window
        if (vis_window_) {
            glutDestroyWindow(vis_window_);
            visualized_classifier = NULL;
        }
    }

////////////////////////////////////////////////////////////////////////
// CutPropertiesStruct implementation
////////////////////////////////////////////////////////////////////////

    TreeCut::TreeCut()
    :
    cut(),
    t_group_total(0.0),
    t_singular_total(0.0),
    benefit_expected(0.0),
    benefit_expected_max(0.0),
    questions_left(0),
    unlabeled_objects(0) {
    }

    TreeCut::TreeCut(const TreeCut &orig) :
    cut(orig.cut),
    t_group_total(orig.t_group_total),
    t_singular_total(orig.t_singular_total),
    benefit_expected(orig.benefit_expected),
    benefit_expected_max(orig.benefit_expected_max),
    questions_left(orig.questions_left),
    unlabeled_objects(orig.unlabeled_objects) {
    }

    TreeCut::~TreeCut()
    {
    }

void TreeCut::reset(GroupTreeNode *node) {
    if (node) {
        cut.clear();
        cut.push_back(node);
        t_group_total = node->t_group();
        t_singular_total = node->t_singular();
        benefit_expected = node->benefit();
        benefit_expected_max = benefit_expected;
        questions_left = min(1.0, t_singular_total);
        if (node->is_leaf()) unlabeled_objects = questions_left;
    } else {
        cut.clear();
        t_group_total = 0.0;
        t_singular_total = 0.0;
        benefit_expected = 0.0;
        benefit_expected_max = 0.0;
        questions_left = 0;
        unlabeled_objects = 0;
    }
}

void TreeCut::merge(TreeCut &cp) {
    // cp will be invalid after this call.
    // Don't merge with self.
    t_group_total += cp.t_group_total;
    t_singular_total += cp.t_singular_total;
    benefit_expected += cp.benefit_expected;
    benefit_expected_max = max(benefit_expected_max, cp.benefit_expected_max);
    questions_left += cp.questions_left;
    unlabeled_objects += cp.unlabeled_objects;
    if (cp.unlabeled_objects) cut.splice(cut.end(), cp.cut);
}

void TreeCut::PrepareForDrawing() {
    cut_it = cut.begin();
    prev_coord = R2null_point;
}

void TreeCut::TryDraw(GroupTreeNode *node, RNScalar x, RNScalar y) {
    // Check if node is on the cut.
    if (!node || node != *cut_it) return;

    // Draw node.
    if (cut_it != cut.begin()) {
        glBegin(GL_LINES);
        glVertex2f(prev_coord.X(), prev_coord.Y());
        glVertex2f(x, y);
        glEnd();
    }

    // Mark max benefit node.
    if (RNIsEqual(node->benefit(), benefit_expected_max)) {
        glPointSize(8);
        glBegin(GL_POINTS);
        glVertex2f(x, y);
        glEnd();
        glColor4f(0.0, 0.0, 0.0, 0.7);
        glPointSize(6);
        glBegin(GL_POINTS);
        glVertex2f(x, y);
        glEnd();
    }

    // Advance cut iterator.
    ++cut_it;
    prev_coord.Reset(x, y);
}

void TreeCut::FillStats(list<string> &data) const {
    char buffer[1024];

    sprintf(buffer, "		Cut size: %zu", cut.size());
    data.push_back(buffer);
    sprintf(buffer, "		Benefit best: %g", benefit_expected_max);
    data.push_back(buffer);
    data.push_back("");

    sprintf(buffer, "		T group total: %g", t_group_total);
    data.push_back(buffer);
    sprintf(buffer, "		T singular total: %g", t_singular_total);
    data.push_back(buffer);
    sprintf(buffer, "		Benefit optimistic: %g",
            t_singular_total - t_group_total);
    data.push_back(buffer);
    data.push_back("");

    sprintf(buffer, "		Benefit expected total: %g", benefit_expected);
    data.push_back(buffer);
    sprintf(buffer, "		Expected time saving per object: %.2f%%",
            100.0 * benefit_expected / t_singular_total);
    data.push_back(buffer);
    data.push_back("");

    sprintf(buffer, "		%d queries left (instead of %d)", questions_left,
            unlabeled_objects);
    data.push_back(buffer);
    sprintf(buffer, "		Fraction # of questions: %.2f%%",
            100.0 * questions_left / unlabeled_objects);
    data.push_back(buffer);
    data.push_back("");
}

} // namespace igra
