/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  R3SurfelClassifier.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2011.
 *
 *  Description: Classifier class that also implements selection suggestion
 *               functionality.
 *
 *  -------------------------------------------------------------------------*/

#ifndef R3SURFELCLASSIFIER_H_
#define R3SURFELCLASSIFIER_H_

#include <list>
#include <map>
#include <vector>
#include <set>

#include "ObjectTreeNode.h"
#include "ObjectWrapper.h"
#include "R3Graphics/R3Graphics.h"
#include "SelectionContainer.h"

using namespace std;

namespace igra {

#define SUGGEST_SINGLE_RANDOM           0x0
#define SUGGEST_SINGLE_CONFIDENCE		0x1
#define SUGGEST_SINGLE_ENTROPY			0x2
#define SUGGEST_SINGLE_DENSITY_WEIGHT   0x4
#define SUGGEST_SINGLE_ALL  			0x8
#define SUGGEST_SINGLE_DEFAULT          SUGGEST_SINGLE_CONFIDENCE

////////////////////////////////////////////////////////////////////////
// Forward declarations
////////////////////////////////////////////////////////////////////////

class GroupTreeNode;
class ObjectWrapper;
class R3SurfelLabeler;

enum VisualizationStyle {
    R3_CLASSIFIER_VISUAL_STYLE_NONE,         // no window
    R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE,   // visualize as tree
    R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE,   // visualize benefit over cost
    R3_CLASSIFIER_VISUAL_STYLE_MDS, // visualize on a 2D plane by distance in feature space
    R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF, // visualize on a 2D plane by distance in shape affinity space
    R3_CLASSIFIER_VISUAL_STYLE_MDS_INDUCED, // visualize on a 2D plane by labeling induced shape affinities
    R3_CLASSIFIER_VISUAL_STYLE_NUM
};

enum DrawPurpose {
    DRAW_PURPOSE_DRAW,
    DRAW_PURPOSE_SELECT,
    DRAW_PURPOSE_FIT,
    DRAW_PURPOSE_FIT_SELECTION,
    DRAW_PURPOSE_NUM
};

enum ColoringScheme {
    COLORING_SCHEME_ASSIGNMENT,
    COLORING_SCHEME_GT,
    COLORING_SCHEME_CONFIDENCE,
    COLORING_SCHEME_AFFINITY,
    COLORING_SCHEME_INDUCED_AFFINITY,
    COLORING_SCHEME_NUM
};

////////////////////////////////////////////////////////////////////////
// Structure to maintain cut through trees.
////////////////////////////////////////////////////////////////////////

struct TreeCut {
public:
    list<GroupTreeNode*> cut;
    RNScalar t_group_total;
    RNScalar t_singular_total;
    RNScalar benefit_expected;
    RNScalar benefit_expected_max;
    int questions_left, unlabeled_objects;

    TreeCut();
    TreeCut(const TreeCut &orig);
    virtual ~TreeCut();

    // Updating.
    void reset(GroupTreeNode *node);
    void merge(TreeCut &cp);

    // Drawing.
    void PrepareForDrawing();
    void TryDraw(GroupTreeNode *node, RNScalar x, RNScalar y);
    void FillStats(list<string> &data) const;

protected:
    list<GroupTreeNode*>::iterator cut_it;
    R2Point prev_coord;
};

////////////////////////////////////////////////////////////////////////
// CLASS DEFINITION
////////////////////////////////////////////////////////////////////////

class R3SurfelClassifier {
public:
    //////////////////////////////////////////
    //// CONSTRUCTOR/DESTRUCTOR FUNCTIONS ////
    //////////////////////////////////////////

    // Constructor/destructor functions
    R3SurfelClassifier(R3SurfelScene *scene, R3SurfelLabeler* labeler,
            VisualizationStyle style = R3_CLASSIFIER_VISUAL_STYLE_NONE);
    ~R3SurfelClassifier(void);

    ////////////////////////////////////
    //// LABEL ASSIGNMENT FUNCTIONS ////
    ////////////////////////////////////

    void UpdateAfterInsertLabelAssignment(R3SurfelLabelAssignment *assignment);
    void UpdateBeforeRemoveLabelAssignment(R3SurfelLabelAssignment *assignment);
    int PredictLabelAssignments(void);

    ///////////////////////////////////
    //// QUERY OBJECT FUNCTIONS    ////
    ///////////////////////////////////

    // All return estimated time for labeling suggested objects.
    RNScalar SuggestSingleObject(R3SurfelObject* &object, const RNFlags mode);
    RNScalar SuggestBestGroupForSeed(R3SurfelObject *seed,
            RNArray<R3SurfelObject*> &group);
    RNScalar SuggestBestGroup(RNArray<R3SurfelObject*> &group);
    RNScalar SuggestExpansion(RNArray<R3SurfelObject*> &group);
    RNScalar SuggestContraction(RNArray<R3SurfelObject*> &group);

    /////////////////////////////////////////////////
    //// COMPUTING VARIOUS PROPERTIES            ////
    /////////////////////////////////////////////////

    ObjectWrapper* Object(int index) const;
    bool Labeled(const R3SurfelObject *object) const;
    RNScalar LabelChoiceTime(void) const;
    static RNScalar LabelChoiceTime(R3SurfelScene *scene);

    void UpdateLabelMenuSequence(RNArray<R3SurfelObject*> object_set,
            RNArray<R3SurfelLabel*> &ranked_menu_sequence);

    RNScalar GetCost(const RNArray<R3SurfelObject*> &group);

    // Basic values.

    RNScalar GetShapeAffinity(R3SurfelObject *from, R3SurfelObject *to) const;
    RNScalar GetShapeAffinity(const int index1, const int index2) const;

    RNScalar GetInducedAffinity(R3SurfelObject *from, R3SurfelObject *to) const;
    RNScalar GetInducedAffinity(const int index1, const int index2) const;

    RNScalar GetFeatureDistance(R3SurfelObject *from, R3SurfelObject *to) const;
    RNScalar GetFeatureDistance(const int index1, const int index2) const;

    RNScalar GetSpatialDistance(R3SurfelObject *from, R3SurfelObject *to) const;
    RNScalar GetSpatialDistance(const int index1, const int index2) const;

    RNScalar GetRPAffinity(int i, int j, int k) const;

    // Derivatives (normalized scores for visualization).
    RNScalar GetScreenSizeComfortScore(R3SurfelObject *from,
            R3SurfelObject *to) const;
    RNScalar GetScreenSizeComfortScore(ObjectWrapper *from,
            ObjectWrapper *to) const;
    RNScalar GetScreenSizeComfortScore(int from, int to) const;
    RNScalar GetScreenSizeComfortScore(const RNLength dist,
            const RNLength size) const;
    RNScalar GetScreenSizeDiscomfortScore(const RNLength dist,
            const RNLength size) const;
    RNScalar GetDistanceInSizeUnits(int from, int to) const;

    GTNTuple GetBestRPinSet(ObjectWrapper *owi, ObjectWrapper *owj,
            const set<int> &pool) const;
    GTNTuple GetBestRPinSet(ObjectWrapper *owi, ObjectWrapper *owj) const;

    //////////////////////////
    //// VISUALIZATION    ////
    //////////////////////////
    int AttemptGLUTInit(VisualizationStyle style);
    int Redraw(void);
    int Resize(int w, int h);
    int Keyboard(unsigned char key, int x, int y);
    int KeyboardSpecial(int key, int x, int y);
    int MouseButton(int button, int state, int x, int y);
    int MouseMotion(int x, int y);
    int PassiveMouseMotion(int x, int y);
    int MouseWheel(int button, int dir, int x, int y);
    void Stop(void);

    int RequestRedraw(void);
    void RecenterOnSelection(void);

    void SetEvaluationGroup(const RNArray<R3SurfelObject*> &group);
    void DrawEvaluationGroup(bool more_info);
    void ClearEvaluationGroup(void);

protected:
    // Support functionality.

    friend class R3SurfelLabeler;

    //////////////////////////////////////////
    //// CONSTRUCTOR/DESTRUCTOR FUNCTIONS ////
    //////////////////////////////////////////

    void ReleaseGroupTree(void);

    ///////////////////////////////////
    //// LABELING OBJECT FUNCTIONS ////
    ///////////////////////////////////

    // Labeling helpers.
    void UpdateObjectLabeling(R3SurfelLabelAssignment *assignment, bool insert);
    void UpdateObjectLabeling(ObjectWrapper *ow, R3SurfelLabel *label,
            RNScalar weight);

    // Prediction helpers.
    int GlobalResetPrediction(ObjectWrapper *ow);
    int GlobalResetPrediction(ObjectWrapper *ow, set<R3SurfelLabel*> &filter);
    void ReassignObjectBestAsgn(ObjectWrapper *ow);
    void ReassignObject(ObjectWrapper *ow, R3SurfelLabel *label,
            RNScalar confidence);

    // Label properties accessors.
    bool Labeled(ObjectWrapper *object) const;

    ///////////////////////////////////
    //// QUERY OBJECT FUNCTIONS    ////
    ///////////////////////////////////

    // Tree construction.
    void CreateGroupTree(void);
    void GrowGroupGreedy(const R3SurfelObject *object);
    void ForestToTree(set<GroupTreeNode*> &forest);

    // Auxiliary functions.
    void ValidateBestFirst(GroupTreeNode *root);
    void UpdateBestGroupsCuts(GroupTreeNode *node, TreeCut &cut);
    int PopulateGroupWithObjects(const GroupTreeNode *node,
            RNArray<R3SurfelObject*> &group) const;
    void ResetGroupTraversal(GroupTreeNode *node,
            GroupTreeNode *tgt_leaf = NULL);
    GroupTreeNode* CommonAncestor(const RNArray<R3SurfelObject*> &group) const;

    void UpdateNonCtrP(RNArray<R3SurfelObject*> &group, bool contraction);

    // Unused..
    template<class Node>
    int AddObjectsInSubtree(Node *node,
            RNArray<ObjectWrapper*> *objects_ = NULL) const;
    template<class Node>
    int AddUnlabeledObjectsInSubtree(const Node *node,
            RNArray<ObjectWrapper*> *objects_ = NULL) const;

    // Accesors.
    GroupTreeNode* GetGroupRootNode() const;

    /////////////////////////////////
    //// AFFINITIES FUNCTIONS    ////
    /////////////////////////////////

    // Getters.
    RNScalar GetAffinity(const vector<vector<float> > *matrix,
            const R3SurfelObject *from, const R3SurfelObject *to) const;
    RNScalar DensityWeight(const int i);

    // Populating methods.
    void PopulateDistancesAffinities(void);
    void PopulateFeatureDistances(void);
    void PopulateSpatialDistances(void);
    void PopulateShapeAffinities(void);
    void PopulateDensityWeights(void);
    void PopulateRPAffinities(void);
    void PopulateNonCtrP(void);
    int ReadRPAffinities(void);
    int WriteRPAffinities(void);

    void PopulateRanks(const vector<vector<float> > &distances,
            vector<vector<int> > &ranks);
    void PopulateKNNGraph(const vector<vector<int> > &ranks, int k0,
            map<int, set<int> > &knn_graph);
    void PopulateDensitySensitiveDistances(
            const vector<vector<float> > &distances,
            const map<int, set<int> > &knn_graph,
            vector<vector<float> > &ds_distances);
    void DistancesToAffinities(const vector<vector<float> > &distances,
            const vector<vector<int> > &ranks, int kmin,
            vector<vector<float> > &affinities);

    //////////////////////////
    //// VISUALIZATION    ////
    //////////////////////////

    // Generic drawing routines.
    void LoadColorSizeByScheme(R3SurfelObject *object);
    void CustomPutVertex(RNScalar x, RNScalar y);
    bool DrawLeafNodeAtPoint(ObjectWrapper* ow, const RNScalar x,
            const RNScalar y, const RNRgb *override_color = NULL);
    void DrawStats();
    void DrawTicks();
    void ForceBlockingMessage(const char* msg);
    void RefitDataToScreen(void);
    void DrawNoDataScreen(void) const;

    // MDS visualization drawing routines.
    void RedrawMDS(void);
    void UpdateMDSifNecessary(void);

    // Tree visualization drawing routines.
    template<class Node>
    void ObjectsInSubtree(Node *node, RNArray<R3SurfelObject*> &objects_) const;
    template<class Node>
    void RedrawTree(Node *node);
    template<class Node>
    RNScalar RedrawSubTree(Node *node, RNScalar &x, RNScalar y,
            const bool highlight_from_above, bool &highlight_upwards);
    template<class Node>
    void TryDrawCuts(Node *node, RNScalar x, RNScalar y);
    void DrawBranch(RNScalar tgt_x, RNScalar tgt_y, RNScalar src_x,
            RNScalar src_y, const bool highlight_downstream,
            const bool highlight_upstream);

protected:
    // Data fields.

    // Keep a local pointer to the scene. Class doesn't own this pointer,
    // caller is responsible for maintaining the scene consistent with the
    // tree through respective calls.
    R3SurfelScene *scene_;

    // Container for object descriptors.
    vector<ObjectWrapper*> objects_;
    // Map from an object pointer to its respective object wrapper.
    map<R3SurfelObject*, ObjectWrapper*> object_leaf_;

    /////////////////////////////
    //// LABELING OBJECTS    ////
    /////////////////////////////

    /// Label prediction optimization fields.
    map<R3SurfelLabel*, set<int> > labeled_; // objects that are labeled by human
    set<ObjectWrapper*> p_increased_; // objects whose human vote increased
    set<ObjectWrapper*> p_decreased_; // objects whose human vote decreased

    /////////////////////////////
    //// GROUP SUGGESTION    ////
    /////////////////////////////

    GroupTreeNode* group_tree_root_;

    RNScalar label_choice_time_;
    TreeCut max_benefit_cut_;

    GroupTreeNode *last_suggestion_node_;

    vector<GroupTreeNode*> group_traversal_;
    int group_traversal_position_;
    bool induced_aff_changed_;

    /////////////////////////////////////
    //// AFFINITIES AND DISTANCES    ////
    /////////////////////////////////////
    vector<vector<float> > *feature_dist_;
    vector<vector<float> > *spatial_dist_;
    vector<vector<float> > *shape_aff_;
    vector<vector<map<int, float> > > *rp_aff_;
    vector<vector<float> > *rp_aff_max_;
    vector<vector<float> > *non_ctr_p_;
    vector<float> density_sum_;
    int density_num_;

    //////////////////////////
    //// VISUALIZATION    ////
    //////////////////////////

    // Drawing fields.
    R3SurfelLabeler *labeler_; // reference to the calling labeler.
    int vis_window_;  // current window OpenGL id
    int vis_window_width_;  // current window width
    int vis_window_height_;  // current window height
    VisualizationStyle vis_style_;	// draw as tree or as points on plane
    R2Box vis_space_box_;	// bounding box of drawing in own coordinates
    R2Box vis_clipping_box_;
    R2Box vis_selection_box_;
    R2Box vis_stats_box_;
    int mouse_button_[3];
    int mouse_screen_coords_[2];
    RNScalar mouse_drag_distance_squared_;
    set<ObjectWrapper*> last_labeled; // set of objects labeled before most recent reprediction
    DrawPurpose draw_purpose_;
    ColoringScheme color_scheme_;
    SelectionContainer selection_container_;

    bool refit_data_to_screen_;

    bool update_mds_fs_;
    vector<R2Point> mds_fs_locations_; // locations for mds feature space
    bool update_mds_aff_;
    vector<R2Point> mds_aff_locations_; // locations for mds affinity space
    bool update_mds_induced_;
    vector<R2Point> mds_induced_locations_; // locations for mds labeling induced affinity

    GroupTreeNode *evaluation_group_;
};

////////////////////////////////////////////////////////////////////////
// Inline methods
////////////////////////////////////////////////////////////////////////

inline ObjectWrapper* R3SurfelClassifier::Object(int index) const {
    return (0 <= index && (size_t) index < objects_.size()) ?
            objects_[index] : NULL;
}

inline bool R3SurfelClassifier::Labeled(ObjectWrapper *object) const {
    return object && object->labeled();
}

inline RNScalar R3SurfelClassifier::LabelChoiceTime(void) const {
    return label_choice_time_;
}

inline GroupTreeNode* R3SurfelClassifier::GetGroupRootNode() const {
    return group_tree_root_;
}

////////////////////////////////////////////////////////////////////////
// Template methods
////////////////////////////////////////////////////////////////////////

template<class Node>
void R3SurfelClassifier::RedrawTree(Node *node) {
    if (!node) {
        DrawNoDataScreen();
        return;
    }

    RNScalar x = 0;
    bool highlight = false;

    max_benefit_cut_.PrepareForDrawing();

    RedrawSubTree(node, x, 0, false, highlight);
}

template<class Node>
RNScalar R3SurfelClassifier::RedrawSubTree(Node *node, RNScalar &x, RNScalar y,
        const bool highlight_from_above, bool &highlight_upwards) {
    if (!node) return x;

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    if (node->is_leaf()) {
        selection_container_.set_current(node);
        highlight_upwards = DrawLeafNodeAtPoint(node->object_, x, y);
        if (draw_purpose_ == DRAW_PURPOSE_DRAW) TryDrawCuts(node, x, y);
        x += 2;
        return x - 2;
    } else {

        bool highlight_this = selection_container_.is_closest(node);
        bool highlight_down = highlight_this || highlight_from_above;
        highlight_upwards = highlight_this;
        RNScalar x_curr = 0.0;

        vector<R3Point> children_branches;
        children_branches.reserve(node->branch_factor());

        for (typename Node::ChildrenConstIterator chit =
                node->children_.begin(); chit != node->children_.end();
                ++chit) {
            RNScalar dy = 1;
            Node *child = *chit;

            bool highlight_from_child = false;
            RNScalar x_child = RedrawSubTree(child, x, y - dy, highlight_down,
                    highlight_from_child);
            highlight_upwards |= highlight_from_child;

            x_curr += x_child;

            children_branches.push_back(
                    R3Point(x_child, y - dy, highlight_from_child ? 1.0 : 0.0));
        }

        if (node->branch_factor()) x_curr /= node->branch_factor();

        if (draw_purpose_ == DRAW_PURPOSE_DRAW) {
            // Draw branches.

            TryDrawCuts(node, x_curr, y);
            for (int i = 0; (size_t) i < children_branches.size(); ++i) {
                R3Point &p = children_branches[i];
                DrawBranch(x_curr, y, p[0], p[1], highlight_down, p[2] > 0.0);
            }

            if (highlight_this) {
                // Highlight selected node.
                glColor4f(1.0, 1.0, 0.0, 1.0);	// solid yellow
                glPointSize(6);
                glBegin(GL_POINTS);
                glVertex2f(x_curr, y);
                glEnd();
            }
        }

        // Draw node.
        glColor4f(0.9, 0.9, 0.9, 1.0);
        glPointSize(2);
        selection_container_.set_current(node);
        CustomPutVertex(x_curr, y);
        return x_curr;
    }
    glDisable(GL_BLEND);
}

template<class Node>
void R3SurfelClassifier::TryDrawCuts(Node *node, RNScalar x, RNScalar y) {
}

template<>
void R3SurfelClassifier::TryDrawCuts<GroupTreeNode>(GroupTreeNode *node,
        RNScalar x, RNScalar y);

template<class Node>
void R3SurfelClassifier::ObjectsInSubtree(Node *node,
        RNArray<R3SurfelObject*> &objects_) const {
    objects_.Empty();

    RNArray<Node*> stack;
    stack.InsertTail(node);

    while (stack.NEntries()) {
        Node *n = stack.Tail();
        stack.RemoveTail();
        if (!n) continue;
        if (n->is_leaf()) {
            ObjectWrapper *ow = n->object_;
            assert(ow);
            R3SurfelObject *o = ow->object();
            assert(o);
            objects_.Insert(o);
        } else {
            for (typename Node::ChildrenConstIterator chit =
                    n->children_.begin(); chit != n->children_.end(); ++chit) {
                stack.InsertTail(*chit);
            }
        }
    }
}

template<class Node>
int R3SurfelClassifier::AddObjectsInSubtree(Node *node,
        RNArray<ObjectWrapper*> *objects_) const {
    // Useful variables.
    int added_size = 0;
    const bool keep_objects = (objects_ != NULL);

    // Preallocate.
    if (keep_objects) objects_->Resize(objects_->NEntries() + node->size_);

    if (node) {
        RNArray<Node*> stack;
        stack.InsertTail(node);

        while (stack.NEntries()) {
            Node *n = stack.Tail();
            stack.RemoveTail();
            if (!n) continue;

            if (n->is_leaf()) {
                ObjectWrapper *ow = n->object_;
                assert(ow);
                if (keep_objects) objects_->Insert(ow);
                added_size++;
            } else {
                stack.InsertTail(n->child_.first);
                stack.InsertTail(n->child_.second);
            }
        }
    }

    return added_size;
}

template<class Node>
int R3SurfelClassifier::AddUnlabeledObjectsInSubtree(const Node *node,
        RNArray<ObjectWrapper*> *objects_) const {
    // Useful variables.
    int added_size = 0;
    bool keep_objects = (objects_ != NULL);

    // Preallocate.
    if (keep_objects) objects_->Resize(objects_->NEntries() + node->size_);

    if (node) {
        RNArray<const Node*> stack;
        stack.InsertTail(node);

        while (stack.NEntries()) {
            const Node *n = stack.Tail();
            stack.RemoveTail();
            if (!n) continue;

            if (n->is_leaf()) {
                ObjectWrapper *ow = n->object_;
                assert(ow);
                if (!Labeled(ow)) {
                    if (keep_objects) objects_->Insert(ow);
                    added_size++;
                }
            } else {
                for (typename Node::ChildrenConstIterator chit =
                        n->children_.begin(); chit != n->children_.end();
                        ++chit) {
                    stack.InsertTail(*chit);
                }
            }
        }
    }

    return added_size;
}

} // namespace igra

#endif /* R3SURFELCLASSIFIER_H_ */
