/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  R3SurfelLabeler.cpp
 *
 *  Copyright Thomas Funkhouser, Princeton University, 2011.
 *
 *  Description: Viewer for the labeling interface.
 *
 *  -------------------------------------------------------------------------*/

#include <limits>
#include <string>

#include <unistd.h>

#include "DrawText.cpp"
#include "R3Graphics/R3Graphics.h"
#include "R3SurfelClassifier.h"
#include "R3SurfelLabeler.h"
#include "R3SurfelLabelerCheckpoint.h"
#include "R3SurfelLabelerCommand.h"
#include "R3Surfels/R3Surfels.h"

namespace igra {

extern R3SurfelLabel *unknown_label; // = NULL;

InterfaceModes interface_mode = IGRA_INTERFACE_FULL;

// Selection interface constants.
const int kSelectionBoxRadius = 5;
const int kMinDragDistance = 5;

// Label menu constants.
const int kMenuSmallPadding = 6;
const int kMenuTinyPadding = 2;
const int kMenuResPadding = kMenuSmallPadding - kMenuTinyPadding;
const int kMenuOuterPadding = 8;
const int kMenuInnerPadding = 8;
const int kMenuLineHeight = 36;
const R2Vector kMenuLineTranslation(0, -kMenuLineHeight);
const R2Vector kMenuGapTranslation(0, -kMenuLineHeight / 2);

// Animation options.
const int kMaxAnimationFPS = 25;
const RNScalar kPhiAngSpeed = RN_PI_OVER_FOUR;
const RNScalar kAvrTheta = 0.5 * RN_PI_OVER_FOUR;
const RNScalar kDTheta = 0.25 * RN_PI_OVER_FOUR;
const RNScalar kThetaPhiCoeff = 1.5 / 1.0;

int draw_green_man = 0; // Green man: a reference figure 2m tall.

// Various execution altering flags.
int training_user_mode = 0;
int use_al_for_group_seeding = 0;

int auto_save = 0;

RNRgb incorrect_color = RNRgb(1.0, 0.27, 0.0);
RNRgb object_glow_color = RNRgb(1.0, 0.0, 0.0);

// Drawing helper functions.

void R3SurfelLabeler::glColor3f(RNScalar r, RNScalar g, RNScalar b) const {
    RNRgb color(r, g, b);
    if (!background_color.IsBlack()) {
        color -= background_color;
        for (int i = 0; i < 3; ++i)
            color[i] = fabs(color[i]);
    }
    ::glColor3f(color[0], color[1], color[2]);
}

void R3SurfelLabeler::glColor4f(RNScalar r, RNScalar g, RNScalar b,
        RNScalar a) const {
    RNRgb color(r, g, b);
    if (!background_color.IsBlack()) {
        color -= background_color;
        for (int i = 0; i < 3; ++i)
            color[i] = fabs(color[i]);
    }
    ::glColor4f(color[0], color[1], color[2], a);
}

void R3SurfelLabeler::CustomLoadRgb(const RNRgb& rgb) const {
    glColor3f(rgb[0], rgb[1], rgb[2]);
}

RNRgb R3SurfelLabeler::GetColorForScalar(RNScalar value, RNScalar min_value,
        RNScalar max_value) const {
    RNRgb color(RNwhite_rgb);

    if (min_value >= max_value) {
        if (value > max_value) color *= 1.0;
        else if (value < min_value) color *= 0.1;
        else color *= 0.5;
    } else {
        RNScalar norm_value = (value - min_value) / (max_value - min_value);
        RNScalar val = 2 * norm_value;

        // Red to yellow.
        //color.Reset(1, norm_value, 0);

        // Blue to red to yellow.
        color = val > 1 ? RNRgb(1, val - 1, 0) : RNRgb(val, 0, 1 - val);

        // Red to green.
        // color = val > 1 ? RNRgb(2 - val, 1, 0) : RNRgb(1, val, 0);

        // Red to green to blue.
        //color =
        //    val > 1 ? RNRgb(0, 2 - val, val - 1) : RNRgb(1 - val, val, 0);
    }

    return color;
}

void R3SurfelLabeler::LoadColorForScalar(RNScalar value, RNScalar min_value,
        RNScalar max_value) const {
    RNRgb color = GetColorForScalar(value, min_value, max_value);
    glColor3f(color[0], color[1], color[2]);
}

void R3SurfelLabeler::DrawHaloedMessage(const R2Point &p, const char* msg,
        void* font, const RNRgb *color) const {
    R2Vector screen_up(0, 1);
    R2Vector screen_right(1, 0);

    // Draw halo
    glColor3f(0.0, 0.0, 0.0);
    DrawText(p + screen_up + screen_right, msg, font);
    DrawText(p + screen_up - screen_right, msg, font);
    DrawText(p - screen_up + screen_right, msg, font);
    DrawText(p - screen_up - screen_right, msg, font);

    // Draw message.
    RNRgb fg_color(color ? *color : RNwhite_rgb);
    glColor3f(fg_color[0], fg_color[1], fg_color[2]);
    DrawText(p, msg, font);
}

string GetImageName() {
    char name[1024];
    static int index = 1;
    do {
        sprintf(name, "Screenshot_%d.png", index++);
    } while (access(name, F_OK) != -1);

    string result(name);
    return result;
}

////////////////////////////////////////////////////////////////////////
// 2D elements drawing locker class:
//  sets modes at creation, restores upon destruction
////////////////////////////////////////////////////////////////////////

class Ortho2DLocker {
public:
    Ortho2DLocker(int width, int height, bool lighting = false,
            bool depth_buffer = false, bool blending = true);
    ~Ortho2DLocker();

protected:
    unsigned char lighting_state_;
    unsigned char depth_test_state_;
    unsigned char depth_write_state_;
    unsigned char blend_state_;
};

Ortho2DLocker::Ortho2DLocker(int width, int height, bool lighting,
        bool depth_buffer, bool blending) {
    // Set orthographic projection matrix.
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, width, 0, height);

    // Preserve existing states.
    glGetBooleanv(GL_LIGHTING, &lighting_state_);
    glGetBooleanv(GL_DEPTH_TEST, &depth_test_state_);
    glGetBooleanv(GL_DEPTH_WRITEMASK, &depth_write_state_);
    glGetBooleanv(GL_BLEND, &blend_state_);

    // Set OpenGL modes.

    // Lighting.
    if (lighting) glEnable(GL_LIGHTING);
    else glDisable(GL_LIGHTING);

    // Depth buffer engagement.
    if (depth_buffer) {
        glEnable(GL_DEPTH_TEST);
        glDepthMask(TRUE);
    } else {
        glDisable(GL_DEPTH_TEST);
        glDepthMask(FALSE);
    }

    // Blending.
    if (blending) {
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
    } else {
        glDisable(GL_BLEND);
    }
}

Ortho2DLocker::~Ortho2DLocker() {
    // Restore OpenGL modes

    // Lighting.
    if (lighting_state_) glEnable(GL_LIGHTING);
    else glDisable(GL_LIGHTING);

    // Depth buffer engagement.
    glDepthMask(depth_write_state_);
    if (depth_test_state_) glEnable(GL_DEPTH_TEST);
    else glDisable(GL_DEPTH_TEST);

    // Blending.
    if (blend_state_) glEnable(GL_BLEND);
    else glDisable(GL_BLEND);

    // Pop matrices.
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
}

////////////////////////////////////////////////////////////////////////
// Surfel labeler constructor/destructor
////////////////////////////////////////////////////////////////////////

R3SurfelLabeler::R3SurfelLabeler(R3SurfelScene *scene,
        const char *logging_filename) :
                R3SurfelViewer(scene),
                classifier(scene, this),
                selection_objects(),
                selection_changed_(1),
                selection_autorequest_(false),
                selection_zoom_box_(),
                highlight_group_(),
                current_command(NULL),
                undo_stack(),
                undo_index(-1),
                created_objects_stack(),
                logging_filename(NULL),
                logging_fp(NULL),
                recenter_on_selection_(false),
                message(NULL),
                message_visibility(1),
                status_visibility(1),
                paused_(false),
                command_menu_visibility(1),
                command_menu_box_(R2null_box),
                label_menu_visibility(1),
                label_view_(),
                label_menu_box_(R2null_box),
                label_menu_sequence_(),
                object_label_confirmed_visibility(1),
                animation_timer_id_(
                        0.5 * RNRandomScalar() * numeric_limits<int>::max()
                                + 1),
                animation_fps_(20),
                animation_on_(0),
                animation_last_update_(),
                animation_phi_(0),
                fps_visibility_(1),
                last_fps_time_read_(),
                fps_counter_(0),
                fps_(0),
                fps_message_(NULL),
                incorrect_objects_(),
                incorrect_label_(NULL),
                show_more_graph_info_(false) {
    // Initialize label visibilities/selectibilities.
    for (int i = 0; i < scene->NLabels(); i++) {
        R3SurfelLabel *label = scene->Label(i);
        label_view_[label].reset();
    }

    // Initialize select box
    select_box_corners[0] = R2zero_point;
    select_box_corners[1] = R2zero_point;
    select_box_active = FALSE;

    // Initialize labels
    if (unknown_label == NULL)
        unknown_label = scene->FindLabelByName("Unknown");
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        if (object->NParts() > 0) continue;
        if (object->HumanLabel()) continue;
        if (object->PredictedLabel()) continue;
        R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                object, unknown_label, 0,
                R3_SURFEL_LABEL_ASSIGNMENT_MACHINE_ORIGINATOR);
        scene->InsertLabelAssignment(assignment);
    }

    // Open logging file
    if (logging_filename) {
        this->logging_filename = strdup(logging_filename);
        if (!strcmp(logging_filename, "stdout")) this->logging_fp = stdout;
        else {
            this->logging_fp = fopen(logging_filename, "w");
            if (!this->logging_fp) {
                fprintf(stderr, "Unable to open logging file %s\n",
                        logging_filename);
                abort();
            }
        }
    }

    last_fps_time_read_.Read();
    fps_message_ = new char[1024];
    fps_message_[0] = '\0';

    // Some special interface-specific settings.
    if (interface_mode == IGRA_INTERFACE_ACTIVE_SINGLE
            || interface_mode == IGRA_INTERFACE_ACTIVE_GROUP
            || interface_mode == IGRA_INTERFACE_TILED) {
        selection_autorequest_ = true;
    }
}

R3SurfelLabeler::~R3SurfelLabeler(void) {
    // Close logging file
    if (logging_fp) fclose(logging_fp);
    if (logging_filename) free(logging_filename);

    // Delete message
    if (message) free(message);

    // Clear zoom boxes.
    EmptyZoomBoxes();

    if (fps_message_) delete[] fps_message_;
}

////////////////////////////////////////////////////////////////////////
// UI event handler functions
////////////////////////////////////////////////////////////////////////

int R3SurfelLabeler::Redraw(void) {
    // Check scene
    if (!scene) return 0;

    // Process necessity of selection autorequest.
    if (selection_autorequest_ && NObjectSelections() == 0) {
        SelectionExpandRequest();
        return 1; // need redraw.
    }

    if (recenter_on_selection_) {
        RecenterOnSet(selection_objects);
        classifier.RecenterOnSelection();
    }

    // Clear window
    glClearColor(background_color[0], background_color[1], background_color[2],
            1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (selection_changed_) {
        incorrect_objects_.Empty();
        incorrect_label_ = NULL;
        classifier.ClearEvaluationGroup();

        if (animation_on_) {
            const R3Camera &cam = viewer.Camera();

            R3Vector to_cam = cam.Origin() - center_point;
            RNLength d = to_cam.Length();
            to_cam.Normalize();
            RNAngle angle = R3InteriorAngle(R3posz_vector, to_cam);
            RNAngle d_a = 2 * kAvrTheta - angle;

            // Rotate camera around the right vector.
            R3Vector right = cam.Right();
            right.Normalize();
            to_cam.Rotate(right, d_a);
            R3Vector twd = cam.Towards();
            twd.Normalize();
            twd.Rotate(right, d_a);
            R3Vector up = cam.Up();
            up.Normalize();
            up.Rotate(right, d_a);

            viewer.ResetCamera(center_point + d * to_cam, twd, up);
        }
    }

    DrawSurfels();
    DrawLabeling();
    DrawObjectNames();
    DrawNodes();
    DrawBlocks();
    DrawCenterPoint();
    DrawHighlightGroup();
    classifier.DrawEvaluationGroup(show_more_graph_info_);
    DrawLabelMenu();
    DrawCommandMenu();
    DrawSelectionZoomBoxes();

    // Capture image and exit
    if (image_name) {
        R2Image image(viewer.Viewport().Width(), viewer.Viewport().Height(), 3);
        image.Capture();
        image.Write(image_name);
        free(image_name);
        image_name = NULL;
    }

    DrawMessage();
    DrawStatus();
    DrawFPS();
    DrawPausedScreen();

    selection_changed_ = false;

    // Return whether need redraw
    return 0;
}

void R3SurfelLabeler::StartAnimation(bool reset) {
    animation_on_ = true;
    animation_fps_ = kMaxAnimationFPS;
    animation_last_update_.Read();
}

int R3SurfelLabeler::AdvanceAnimation() {
    int redraw = 0;

    // Check everything
    if (!animation_on_) return redraw;

    // Update fps estimate and perform linear increment attempt.
    RNScalar d_t = animation_last_update_.Elapsed();
    RNScalar desired_dt = 1.0 / kMaxAnimationFPS;
    if (desired_dt > d_t) return redraw;

    RNScalar instant_fps = 1 / d_t;
    animation_fps_ = min(kMaxAnimationFPS, (int) max(instant_fps, fps_) + 5);
    animation_last_update_.Read();

    // Update all cameras
    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
        if (szb && !szb->Hovered()) szb->AdvanceAnimation(d_t);
        redraw = 1;
    }

    // Update main camera.
    if (NObjectSelections()) {
        R3Vector twd = center_point - Camera().Origin();
        RNLength dist = twd.Length();

        RNScalar gain = min(sqrt(20 / dist), 1.0);
        RNAngle d_phi = d_t * kPhiAngSpeed * gain;
        RNAngle d_theta = kDTheta * sin(kThetaPhiCoeff * animation_phi_);
        animation_phi_ += d_phi;
        d_theta = kDTheta * sin(kThetaPhiCoeff * animation_phi_) - d_theta;
        RNAngle max_d_theta = R3InteriorAngle(twd, R3posz_vector)
                - RN_PI_OVER_TWO;
        RNAngle min_d_theta = -1.0
                * (R3InteriorAngle(twd, R3negz_vector) + RN_PI_OVER_FOUR);
        if (d_theta < min_d_theta) {
            d_theta = min_d_theta;
            RNAngle d_a = kThetaPhiCoeff * animation_phi_;
            RNAngle target_a = RN_TWO_PI * floor(d_a / RN_TWO_PI)
                    + 3 * RN_PI_OVER_TWO;
            animation_phi_ = target_a / kThetaPhiCoeff;
        }
        if (d_theta > max_d_theta) {
            d_theta = max_d_theta;
            RNAngle d_a = kThetaPhiCoeff * animation_phi_;
            RNAngle target_a = RN_TWO_PI * floor(d_a / RN_TWO_PI)
                    + RN_PI_OVER_TWO;
            animation_phi_ = target_a / kThetaPhiCoeff;
        }

        viewer.RotateWorld(center_point, R3posz_vector, d_phi);
        viewer.RotateWorld(center_point, Camera().Right(), d_theta);

        redraw = 1;
    }

    // Return success.
    return redraw;
}

void R3SurfelLabeler::StopAnimation() {
    animation_on_ = false;
}

int R3SurfelLabeler::Resize(int w, int h) {
    // Send event to viewer
    return R3SurfelViewer::Resize(w, h);
}

int R3SurfelLabeler::MouseMotion(int x, int y) {
    if (paused_) return 1;

    // Initialize
    int redraw = 0;

    // Compute mouse movement
    int dx = x - mouse_position[0];
    int dy = y - mouse_position[1];

    // Update mouse drag movement
    mouse_drag_distance_squared += dx * dx + dy * dy;

    // Process mouse movement
    if ((interface_mode == IGRA_INTERFACE_FULL
            || interface_mode == IGRA_INTERFACE_UG_GROUP)
            && (ctrl_down || shift_down || alt_down) && mouse_button[0]) {
        RNBoolean drag = (mouse_drag_distance_squared > 4);
        if (drag) {
            // Update select box
            if (select_box_active) {
                DrawSelectBox();
                select_box_corners[1] = R2Point(x, y);
                DrawSelectBox();
            } else {
                select_box_active = TRUE;
                select_box_corners[0] = R2Point(mouse_down_position[0],
                        mouse_down_position[1]);
                select_box_corners[1] = R2Point(x, y);
                DrawSelectBox();
            }
        }
    } else {
        // Set viewing center point
        R3Point viewing_center_point = center_point;
        const R3Camera& camera = viewer.Camera();
        const R2Viewport& viewport = viewer.Viewport();
        R3Plane camera_plane(camera.Origin(), camera.Towards());
        RNScalar signed_distance = R3SignedDistance(camera_plane,
                viewing_center_point);
        if (signed_distance < 0)
            viewing_center_point -= (signed_distance - 1) * camera.Towards();
        R2Point screen_point = viewer.ViewportPoint(viewing_center_point);
        if ((NObjectSelections() == 0)
                || !R2Contains(viewer.Viewport().BBox(), screen_point)) {
            R3Ray world_ray = viewer.WorldRay(viewport.XCenter(),
                    viewport.YCenter());
            R3Plane world_plane(0, 0, 1, -viewing_center_point[2]);
            R3Intersects(world_ray.Line(), world_plane, &viewing_center_point);
            viewing_center_point = scene->BBox().ClosestPoint(
                    viewing_center_point);
        }

        // World in hand navigation
        if (mouse_button[0]) {
            // Update zoom views cameras.
            const RNScalar factor = 4.0;
            RNLength vx = (RNLength) dx / (RNLength) viewer.Viewport().Width();
            RNLength vy = (RNLength) dy / (RNLength) viewer.Viewport().Height();
            RNAngle theta = -factor * vx;
            RNAngle phi = factor * vy;

            bool zoom_box_area = false;
            for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
                SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
                if (!szb || !szb->Hovered()) continue;
                zoom_box_area = true;
                R3Camera &c = szb->Camera();

                RNAngle max_phi = R3InteriorAngle(c.Towards(), R3posz_vector)
                        - RN_PI_OVER_TWO;
                RNAngle min_phi = -1.0
                        * R3InteriorAngle(c.Towards(), R3negz_vector);
                if (phi < min_phi) phi = min_phi;
                if (phi > max_phi) phi = max_phi;

                R3Point target = szb->Object()->Centroid();
                RNScalar distance = (c.Origin() - target).Length();
                R3Vector twd = c.Towards();
                R3Vector up = c.Up();
                R3Vector right = c.Right();
                twd.Rotate(right, phi);
                twd.Rotate(R3posz_vector, theta);
                R3Point eye = target - distance * twd;
                right = twd % R3posz_vector;
                up = right % twd;
                c.Reset(eye, twd, up);
                break;
            }

            if (!zoom_box_area) {
                if (interface_mode != IGRA_INTERFACE_TILED) StopAnimation();
                RotateWorld(1.0, viewing_center_point, x, y, dx, dy);
            }
        }
        // else if (shift_down && (mouse_button[1] || mouse_button[2])) viewer.ScaleWorld(2.0, viewing_center_point, x, y, dx, dy);
        // else if (ctrl_down && (mouse_button[1] || mouse_button[2])) viewer.TranslateWorld(2.0, viewing_center_point, x, y, dx, dy);
        else if (mouse_button[1]) {
            // Update zoom views cameras.
            const RNScalar factor = 2.0;
            RNLength vx = (RNLength) dx / (RNLength) viewer.Viewport().Width();
            RNLength vy = (RNLength) dy / (RNLength) viewer.Viewport().Height();
            RNScalar scale = exp(2.0 * factor * (vx + vy));

            bool zoom_box_area = false;
            for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
                SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
                if (!szb || !szb->Hovered()) continue;
                zoom_box_area = true;
                R3Camera &c = szb->Camera();

                assert(scale > 0.0);
                RNScalar f = (scale - 1.0) / scale;
                R3Point target = szb->Object()->Centroid();
                c.Translate((target - c.Origin()) * f);
                break;
            }

            if (!zoom_box_area) {
                if (interface_mode != IGRA_INTERFACE_TILED) StopAnimation();
                viewer.ScaleWorld(2.0, viewing_center_point, x, y, dx, dy);
            }
        } else if (mouse_button[2]) {
            // Translate the world if not in the box.
            bool zoom_box_area = false;
            for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
                SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
                if (!szb || !szb->Hovered()) continue;
                zoom_box_area = true;
                break;
            }
            if (!zoom_box_area) {
                if (interface_mode != IGRA_INTERFACE_TILED) StopAnimation();
                viewer.TranslateWorld(2.0, viewing_center_point, x, y, dx, dy);
            }
        }
        if (mouse_button[0] || mouse_button[1] || mouse_button[2]) redraw = 1;
    }

    // Remember mouse position
    mouse_position[0] = x;
    mouse_position[1] = y;

    // Return whether need redraw
    return redraw;
}

int R3SurfelLabeler::PassiveMouseMotion(int x, int y) {
    if (paused_) return 1;

    // Initialize
    int redraw = 0;

    // Check if entering or leaving the selection zoom box.
    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
        if (!szb) continue;
        bool hovered_before = szb->Hovered();
        bool hovered_after = szb->CheckHovered(x, y);
        if (hovered_before != hovered_after) redraw = 1;
    }

    // Remember mouse position
    mouse_position[0] = x;
    mouse_position[1] = y;

    // Return whether need redraw
    return redraw;
}

int R3SurfelLabeler::MouseWheel(int button, int dir, int x, int y) {

    int redraw = R3SurfelViewer::MouseWheel(button, dir, x, y);

    // Set viewing center point
    R3Point viewing_center_point = center_point;
    const R3Camera& camera = viewer.Camera();
    const R2Viewport& viewport = viewer.Viewport();
    R3Plane camera_plane(camera.Origin(), camera.Towards());
    RNScalar signed_distance = R3SignedDistance(camera_plane,
            viewing_center_point);
    if (signed_distance < 0)
        viewing_center_point -= (signed_distance - 1) * camera.Towards();
    R2Point screen_point = viewer.ViewportPoint(viewing_center_point);
    if ((NObjectSelections() == 0)
            || !R2Contains(viewer.Viewport().BBox(), screen_point)) {
        R3Ray world_ray = viewer.WorldRay(viewport.XCenter(),
                viewport.YCenter());
        R3Plane world_plane(0, 0, 1, -viewing_center_point[2]);
        R3Intersects(world_ray.Line(), world_plane, &viewing_center_point);
        viewing_center_point = scene->BBox().ClosestPoint(viewing_center_point);
    }

    // Update zoom views cameras.
    const RNScalar factor = 2.0;
    RNLength dl = dir > 0 ? -0.01 : 0.01;
    RNScalar scale = exp(2.0 * factor * dl);

    bool zoom_box_area = false;
    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
        if (!szb || !szb->Hovered()) continue;
        zoom_box_area = true;
        R3Camera &c = szb->Camera();

        assert(scale > 0.0);
        RNScalar f = (scale - 1.0) / scale;
        R3Point target = szb->Object()->Centroid();
        c.Translate((target - c.Origin()) * f);
        break;
    }

    if (!zoom_box_area) viewer.ScaleWorld(viewing_center_point, scale);

    redraw = 1;

    return redraw;
}

int R3SurfelLabeler::MouseButton(int x, int y, int button, int state, int shift,
        int ctrl, int alt) {
    if (paused_) return 1;

    // Send event to viewer
    int redraw = R3SurfelViewer::MouseButton(x, y, button, state, shift, ctrl,
            alt);

    // Only process Up
    if (state != 0) return redraw;

    // Process mouse button event
    // Check for drag
    RNBoolean drag = (mouse_drag_distance_squared > 4);

    // Check for double click
    static RNBoolean double_click = FALSE;
    static RNTime last_mouse_down_time;
    double_click = !drag && !double_click
            && (last_mouse_down_time.Elapsed() < 0.4);
    last_mouse_down_time.Read();

    // Check if left mouse
    if (button == 0) {
        // Check if dragged mouse
        if (drag) {
            // Process select box pick
            if (select_box_active) {
                DrawSelectBox();
                R2Box box = R2null_box;
                box.Union(select_box_corners[0]);
                box.Union(select_box_corners[1]);
                SelectEnclosedObjects(box, shift, alt);
                select_box_corners[0] = R2zero_point;
                select_box_corners[1] = R2zero_point;
                select_box_active = FALSE;
                redraw = 1;
            }
        } else {
            if (InSelectionZoomBox(x, y, button, state, shift, ctrl, alt)) {
                redraw = 1;
            } else if (PickCommandMenu(x, y, button, state, shift, ctrl, alt)) {
                redraw = 1;
            } else if (PickLabelMenu(x, y, button, state, shift, ctrl, alt)) {
                redraw = 1;
            } else if (interface_mode != IGRA_INTERFACE_ACTIVE_GROUP
                    && interface_mode != IGRA_INTERFACE_ACTIVE_SINGLE
                    && interface_mode != IGRA_INTERFACE_TILED) {
                if (!double_click) {
                    // Process select point pick
                    SelectPickedObject(x, y, shift, alt);
                    redraw = 1;
                } else {
                    if (NObjectSelections() > 0) {
                        ZoomCamera(0);
                    } else {
                        R3SurfelNode *node = PickNode(x, y);
                        if (node) {
                            SetCenterPoint(node->Centroid());
                            ZoomCamera(0);
                        }
                    }
                }
                redraw = 1;
            }
        }
    }

    // Return whether need redraw
    return redraw;
}

int R3SurfelLabeler::Keyboard(int x, int y, int key, int shift, int ctrl,
        int alt) {
    int redraw = 0;

    // Process pause key first.
    if (key == R3_SURFEL_VIEWER_HOME_KEY) {
        TogglePause();
        redraw = 1;
    }
    if (paused_) return redraw;

    // Send event to viewer
    if (interface_mode == IGRA_INTERFACE_FULL) {
        redraw += R3SurfelViewer::Keyboard(x, y, key, shift, ctrl, alt);
    } else {
        // Prevent debug operations in test interfaces.
        redraw += R3SurfelViewer::Keyboard(x, y, key, shift, 0, 0);
        ctrl_down = ctrl;
        alt_down = alt;
    }

    // Process mouse button event
    if (alt) {
        switch (key) {

            // Reset.
            case 'R':
            case 'r':
                if (0) {
                    Reset();
                    redraw = 1;
                }
                break;

                // Toggle negative coloring.
            case 'I':
            case 'i':
                FlipBGColor();
                redraw = 1;
                break;

            case 'P':
            case 'p': {
                string fname = GetImageName().c_str();
                WriteImage(fname.c_str());
                SetMessage("Saved screenshot to %s", fname.c_str());
                redraw = 1;
            }
                break;
        }
    } else if (ctrl) {
        // Process accelerator key commands
        switch (key) {
            // Toggle animation.
            case 'A':
            case 'a':
                if (animation_on_) StopAnimation();
                else StartAnimation(false);
                break;

                // Toggle background visibility.
            case 'B':
            case 'b':
                SetBackgroundVisibility(-1);
                BeginCommand(
                        R3_SURFEL_LABELER_SET_BACKGROUND_VISIBILITY_COMMAND,
                        BackgroundVisibility());
                EndCommand();
                redraw = 1;
                break;

                // Rotate color scheme.
            case 'C':
            case 'c':
                if (interface_mode == IGRA_INTERFACE_FULL) {
                    int scheme = (SurfelColorScheme() + 1)
                            % R3_SURFEL_VIEWER_COLOR_BY_HEIGHT;
                    BeginCommand(
                            R3_SURFEL_LABELER_SET_SURFEL_COLOR_SCHEME_COMMAND,
                            scheme);
                    SetSurfelColorScheme(scheme);
                    EndCommand();
                    SetMessage("Color scheme switched to: %s",
                            color_scheme_name[scheme]);
                    redraw = 1;
                }
                break;

                // Evaluate selected group.
            case 'E':
            case 'e':
                if (interface_mode == IGRA_INTERFACE_FULL
                        && selection_objects.NEntries()) {
                    classifier.SetEvaluationGroup(selection_objects);
                    classifier.RequestRedraw();
                    redraw = 1;
                }
                break;

                // Greedily grow group tree around selected object.
            case 'G':
            case 'g':
                if (interface_mode == IGRA_INTERFACE_FULL
                        && selection_objects.NEntries() == 1) {
                    classifier.GrowGroupGreedy(selection_objects.Kth(0));
                    redraw = 1;
                }
                break;

                // Request more information on graph visualization.
            case 'I':
            case 'i':
                if (interface_mode == IGRA_INTERFACE_FULL) {
                    show_more_graph_info_ = !show_more_graph_info_;
                    redraw = 1;
                }
                break;

                // Toggle show assigned label above object.
            case 'L':
            case 'l': {
                if (!isMachineDrivenInterface()) {
                    int visibility = 1 - ObjectLabelVisibility();
                    BeginCommand(
                            R3_SURFEL_LABELER_SET_LABEL_NAME_VISIBILITY_COMMAND,
                            visibility);
                    SetObjectLabelVisibility(visibility);
                    EndCommand();
                    redraw = 1;
                }
                break;
            }

            case 'M':
            case 'm': // ENTER
                break;

                // Toggle labeled object visibility.
            case 'P':
            case 'p':
                if (!isMachineDrivenInterface()) {
                    SetLabelConfirmedVisibility(-1);
                    SetMessage("Confirmed visibility turned %s",
                            (object_label_confirmed_visibility ? "on" : "off"));
                    redraw = 1;
                }
                break;

                // Save scene files.
            case 'S':
            case 's':
                Sync();
                SetMessage("Saved scene file");
                redraw = 1;
                break;

                // Request creation of an entire grouping tree.
            case 'T':
            case 't':
                if (interface_mode == IGRA_INTERFACE_FULL) {
                    classifier.CreateGroupTree();
                    redraw = 1;
                }
                break;

                // Toggle exclusive visibility of objects that share
                // assigned label with the selected objects.
            case 'X':
            case 'x':
                if (!isMachineDrivenInterface()) {
                    ToggleExclusiveVisibilityForSelected();
                    redraw = 1;
                }
                break;

                // Redo.
            case 'Y':
            case 'y':
                Redo();
                if (isMachineDrivenInterface()) {
                    while (NObjectSelections() == 0
                            && undo_index < undo_stack.NEntries())
                        Redo();
                }
                redraw = 1;
                break;

                // Undo.
            case 'Z':
            case 'z':
                Undo();
                if (isMachineDrivenInterface()) {
                    while (NObjectSelections() == 0 && undo_index >= 0)
                        Undo();
                }
                redraw = 1;
                break;

                // Request a prediction to be made in a batch interface.
            case 'R':
                if (interface_mode == IGRA_INTERFACE_BATCH_LEARNING) {
                    static int number_of_predict_all = 0;
                    if (!number_of_predict_all) {
                        PredictAll();
                        number_of_predict_all++;
                    }
                    redraw = 1;
                }
                break;
        }
    } else {
        switch (key) {
            // Clear selection.
            case R3_SURFEL_VIEWER_ASCII_ESC_KEY:
                if (!isMachineDrivenInterface()) {
                    // Unselect everything and clear label filters.
                    SelectPickedObject(-1, -1, 0, 0);
                    for (int i = 0; i < scene->NLabels(); i++) {
                        R3SurfelLabel *label = scene->Label(i);
                        label_view_[label].reset();
                    }
                    redraw = 1;
                }
                break;

                // Confirm prediction label on selected objects.
            case R3_SURFEL_VIEWER_ASCII_SPACE_KEY:
                // Confirm predicted label
                if (NObjectSelections() > 0) ConfirmLabelsOnSelectedObjects();
                else if (!ConfirmLabelOnPickedObject(x, y))
                    SelectPickedObject(x, y);
                redraw = 1;
                break;

                // Contract request.
            case R3_SURFEL_VIEWER_LEFT_KEY:
                redraw += SelectionContractRequest();
                break;

                // Expand request.
            case R3_SURFEL_VIEWER_RIGHT_KEY:
                redraw += SelectionExpandRequest();
                break;

                // Assign labels to objects by keystroke.
            default:
                // Assign label by key
                if (scene) {
                    R3SurfelLabel *label =
                            scene->FindLabelByAssignmentKeystroke(toupper(key));
                    if (label) {
                        if (NObjectSelections() > 0)
                            AssignLabelToSelectedObjects( label);
                        else
                            AssignLabelToPickedObject(label, x, y);
                        redraw = 1;
                    }
                }
                break;
        }
    }

    // Return whether need redraw
    return redraw;
}

////////////////////////////////////////////////////////////////////////
// COMMANDS
////////////////////////////////////////////////////////////////////////

void R3SurfelLabeler::Initialize(int classifier_vis_style) {
    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_INITIALIZE_COMMAND);

    // Send event to viewer
    R3SurfelViewer::Initialize();
    classifier.AttemptGLUTInit((VisualizationStyle) classifier_vis_style);

    // End logging command
    EndCommand();
}

void R3SurfelLabeler::Terminate(void) {
    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_TERMINATE_COMMAND);

    // Send event to viewer
    R3SurfelViewer::Terminate();

    // End logging command
    EndCommand();
}

int R3SurfelLabeler::Sync(void) {
    // Check everything
    if (!scene) return 0;

    // Set message
    SetMessage("Saved label assignments");

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_SYNC_COMMAND);

    // Sync file
    scene->SyncFile();

    // End logging command
    EndCommand();

    // Return success
    return 1;
}

int R3SurfelLabeler::SelectPickedObject(int x, int y, int add, int subtract) {
    // Check scene
    if (!scene) return 0;

    //Check interface mode
    if (interface_mode != IGRA_INTERFACE_FULL
            && interface_mode != IGRA_INTERFACE_BATCH_LEARNING
            && interface_mode != IGRA_INTERFACE_UG_GROUP
            && interface_mode != IGRA_INTERFACE_UG_SINGLE) {
        return 0;
    }

    // Set message
    SetMessage(NULL);

    // Pick object
    R3Point pick_point;
    R3SurfelObject *object = NULL;

    // Selection box is a tiny window around the cursor.
    R2Box box(x - kSelectionBoxRadius, y - kSelectionBoxRadius,
            x + kSelectionBoxRadius, y + kSelectionBoxRadius);

    // Simple filter for objects in the box.
    RNArray<R3SurfelObject *> picked_objects;
    if (!FilterObjectsInsideBox(box, picked_objects,
    R3_SURFEL_LABELER_SELECTION_FILTER_OVERLAP
    /*| R3_SURFEL_LABELER_SELECTION_FILTER_OFFSCREEN*/)) return 0;

    // Select object in the selection area around the cursor.
    object = SelectTopHitObjectInBox(box, picked_objects);

    if (object && scene != object->Scene()) return 0;

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_SELECT_PICKED_COMMAND);

    // Remove previous selections
    if ((interface_mode == IGRA_INTERFACE_BATCH_LEARNING)
            || (!add && !subtract)) EmptyObjectSelections();

    // Update object selections
    if (object) {
        if (selection_objects.FindEntry(object)) {
            if (subtract) {
                // Remove object from selected objects
                RemoveObjectSelection(object);
            }
        } else if (!subtract) {
            // Insert object into selected objects
            InsertObjectSelection(object);
        }
    }

    // End logging command
    EndCommand();

    if (object) {
        SetCenterPoint(object->Centroid());
        SetMessage("Selected object %d", object->SceneIndex());
    }

    // Return success
    return 1;
}

int R3SurfelLabeler::SelectEnclosedObjects(const R2Box& box, int add,
        int subtract) {
    // Check everything.
    if (interface_mode != IGRA_INTERFACE_FULL
            && interface_mode != IGRA_INTERFACE_UG_GROUP) return 0;

    // Simple filter for objects in the box.
    RNArray<R3SurfelObject *> picked_objects;
    if (!FilterObjectsInsideBox(box, picked_objects,
            R3_SURFEL_LABELER_SELECTION_FILTER_OVERLAP
                    | R3_SURFEL_LABELER_SELECTION_FILTER_OFFSCREEN)) return 0;

    // Select top objects in the smaller selection areas in an attempt
    // to only select front objects.
    RNArray<R3SurfelObject *> object_pool(picked_objects);
    picked_objects.Empty();
    SelectTopHitObjectsInBox(box, object_pool, picked_objects);

    // Set message
    SetMessage("Selected %d objects", picked_objects.NEntries());

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_SELECT_ENCLOSED_COMMAND);

    // Remove previous selections
    if (!add && !subtract) EmptyObjectSelections();

    // Update object selections
    for (int i = 0; i < picked_objects.NEntries(); i++) {
        R3SurfelObject *object = picked_objects.Kth(i);

        // Check if object is already selected
        if (selection_objects.FindEntry(object)) {
            if (subtract) {
                // Remove object from selected objects
                RemoveObjectSelection(object);
            }
        } else {
            if (!subtract) {
                // Insert object into selected objects
                InsertObjectSelection(object);
            }
        }
    }

    // End logging command
    EndCommand();

    // Set center point
    if (NObjectSelections() > 0) {
        R3Point centroid = R3zero_point;
        for (int i = 0; i < NObjectSelections(); i++) {
            R3SurfelObject *object = ObjectSelection(i);
            centroid += object->Centroid();
        }
        centroid /= NObjectSelections();
        SetCenterPoint(centroid);
    }

    // Return success
    return 1;
}

int R3SurfelLabeler::SelectionExpandRequest() {
    // Check everything.
    if (!scene || scene->NObjects() < 2) return 0;
    if (interface_mode != IGRA_INTERFACE_FULL
            && interface_mode != IGRA_INTERFACE_ACTIVE_SINGLE
            && interface_mode != IGRA_INTERFACE_ACTIVE_GROUP
            && interface_mode != IGRA_INTERFACE_TILED) {
        return 0;
    }

    RNArray<R3SurfelObject*> object_set(selection_objects);
    RNScalar cost = 0;
    int command = R3_SURFEL_LABELER_SELECT_SUGGESTED_COMMAND;

    switch (interface_mode) {
        case IGRA_INTERFACE_FULL:
        case IGRA_INTERFACE_ACTIVE_GROUP:
        case IGRA_INTERFACE_TILED:
            if (NObjectSelections() == 0) {
                if (use_al_for_group_seeding) {
                    R3SurfelObject *o = NULL;
                    classifier.SuggestSingleObject(o, SUGGEST_SINGLE_DEFAULT);
                    if (!o) break;
                    object_set.Insert(o);
                    //cost = classifier.SuggestBestGroupForSeed(o, object_set);
                    cost = classifier.SuggestExpansion(object_set);
                } else cost = classifier.SuggestBestGroup(object_set);
            } else {
                cost = classifier.SuggestExpansion(object_set);
                command = R3_SURFEL_LABELER_SUGGEST_INCREASE_SELECTION;
            }
            break;

        case IGRA_INTERFACE_ACTIVE_SINGLE: {
            object_set.Empty();
            R3SurfelObject *o = NULL;
            cost = classifier.SuggestSingleObject(o, SUGGEST_SINGLE_DEFAULT);
            if (o) object_set.Insert(o);
        }
            break;

        default:
            return 0;
            break;
    }

    if (interface_mode == IGRA_INTERFACE_ACTIVE_GROUP
            || interface_mode == IGRA_INTERFACE_TILED
            || interface_mode == IGRA_INTERFACE_ACTIVE_SINGLE) {
        selection_autorequest_ = (object_set.NEntries() > 0);
    }

    // If nothing changed, no need to continue;
    if (NObjectSelections() == object_set.NEntries()) {
        SetMessage("No good expansion available");
        return 1;
    }

    // Set message.
    if (object_set.NEntries() == 1) SetMessage("Selected suggested object %d",
            object_set[0]->SceneIndex());
    else SetMessage("Selected %d suggested objects", object_set.NEntries());

    // Begin logging command
    BeginCommand(command, cost);

    if (object_set.NEntries()) {
        // Empty object selections
        EmptyObjectSelections();

        // Insert object into selected objects
        for (int i = 0; i < object_set.NEntries(); ++i) {
            R3SurfelObject *object = object_set[i];
            InsertObjectSelection(object);
        }
    }

    // End logging command
    EndCommand();

    // Center on the selection.
    if (NObjectSelections() && interface_mode != IGRA_INTERFACE_TILED)
        ZoomToObjectSet(selection_objects, false);

    // Initiate animation.
    StartAnimation();

    // Return success
    return 1;
}

int R3SurfelLabeler::SelectionContractRequest() {
    // Check everything.
    if (!scene || scene->NObjects() < 2) return 0;
    if (interface_mode != IGRA_INTERFACE_FULL
            && interface_mode != IGRA_INTERFACE_ACTIVE_GROUP
            && interface_mode != IGRA_INTERFACE_TILED) {
        return 0;
    }
    if (NObjectSelections() == 0) return 0;

    RNArray<R3SurfelObject*> object_set(selection_objects);
    int old_size = selection_objects.NEntries();
    RNScalar cost = classifier.SuggestContraction(object_set);

    if (old_size == object_set.NEntries()) return 0;

    // If nothing changed, no need to continue;
    if (NObjectSelections() == object_set.NEntries()) {
        SetMessage("No good contraction available");
        return 1;
    }

    // Set message.
    if (object_set.NEntries() == 1) SetMessage("Selected suggested object %d",
            object_set[0]->SceneIndex());
    else SetMessage("Selected %d suggested objects", object_set.NEntries());

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_SUGGEST_DECREASE_SELECTION, cost);

    if (object_set.NEntries()) {
        // Empty object selections
        EmptyObjectSelections();

        // Insert object into selected objects
        for (int i = 0; i < object_set.NEntries(); ++i) {
            R3SurfelObject *object = object_set[i];
            InsertObjectSelection(object);
        }
    }

    // End logging command
    EndCommand();

    // Center on the selection.
    if (NObjectSelections() && interface_mode != IGRA_INTERFACE_TILED)
        ZoomToObjectSet(selection_objects, false);

    // Initiate animation.
    StartAnimation();

    // Return success
    return 1;
}

int R3SurfelLabeler::AssignLabelToPickedObject(R3SurfelLabel *label, int x,
        int y) {
    // Set message
    SetMessage(NULL);

    //Check interface mode.
    if (isMachineDrivenInterface()) return 0;

    // Pick object
    R3SurfelObject *object = NULL;

    // Selection box is a tiny window around the cursor.
    R2Box box(x - kSelectionBoxRadius, y - kSelectionBoxRadius,
            x + kSelectionBoxRadius, y + kSelectionBoxRadius);

    // Simple filter for objects in the box.
    RNArray<R3SurfelObject *> picked_objects;
    if (!FilterObjectsInsideBox(box, picked_objects,
            R3_SURFEL_LABELER_SELECTION_FILTER_OVERLAP
                    | R3_SURFEL_LABELER_SELECTION_FILTER_OFFSCREEN)) return 0;

    // Select object in the selection area around the cursor.
    object = SelectTopHitObjectInBox(box, picked_objects);
    if (!object) return 0;

    R3SurfelLabel *pr_label = object->PredictedLabel();
    if (pr_label && !LabelSelectible(pr_label)) return 0;
    if (!ShowConfirmed(object)) return 0;

    // Set center point
    SetCenterPoint(object->Centroid());

    // In user training, don't let incorrect labeling.
    if (training_user_mode) {
        incorrect_objects_.Empty();
        if (object->GroundTruthLabel() != label)
            incorrect_objects_.Insert(object);
        if (incorrect_objects_.NEntries()) {
            incorrect_label_ = label;
            SetMessage(
                    "Incorrect: Marked objects are not what you labeled them.");

            // Record a mistake.
            BeginCommand(R3_SURFEL_LABELER_LABEL_TRAINING_MISTAKE_COMMAND);
            R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                    object, label, 1);
            current_command->inserted_label_assignments.Insert(assignment);
            EndCommand();

            return 0;
        } else incorrect_label_ = NULL;
    }

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_LABEL_PICKED_COMMAND);

    // Create/insert label assignment
    R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(object,
            label, 1);
    InsertLabelAssignment(assignment);

    // Predict label assignments for all other objects
    if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING)
        classifier.PredictLabelAssignments();

    // Empty object selections
    EmptyObjectSelections();

    // End logging command
    EndCommand();

    // Set message
    const char *label_name = (label->Name()) ? label->Name() : "without name";
    SetMessage("Assigned label %s to object %d", label_name,
            object->SceneIndex());

    // Return success
    return 1;
}

int R3SurfelLabeler::AssignLabelToSelectedObjects(R3SurfelLabel *label) {
    // Check everything
    if (!scene) return 0;
    if (!label) return 0;
    if (NObjectSelections() == 0) return 0;

    const int selection_size = NObjectSelections();
    R3SurfelObject *assigned_object =
            selection_size == 1 ? ObjectSelection(0) : NULL;

    // In user training, don't let incorrect labeling.
    if (training_user_mode) {
        incorrect_objects_.Empty();
        for (int i = 0; i < selection_size; i++) {
            R3SurfelObject *object = ObjectSelection(i);
            if (object->GroundTruthLabel() != label)
                incorrect_objects_.Insert(object);
        }
        if (incorrect_objects_.NEntries()) {
            incorrect_label_ = label;
            SetMessage(
                    "Incorrect: Marked objects are not what you labeled them.");

            // Record a mistake.
            BeginCommand(R3_SURFEL_LABELER_LABEL_TRAINING_MISTAKE_COMMAND);
            for (int i = 0; i < selection_size; i++) {
                R3SurfelObject *object = ObjectSelection(i);
                R3SurfelLabelAssignment *assignment =
                        new R3SurfelLabelAssignment(object, label, 1);
                current_command->inserted_label_assignments.Insert(assignment);
            }
            EndCommand();

            return 0;
        } else incorrect_label_ = NULL;
    }

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_LABEL_SELECTION_COMMAND);

    // All are assigned a label.
    for (int i = 0; i < selection_size; i++) {
        R3SurfelObject *object = ObjectSelection(i);
        R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                object, label, 1);
        InsertLabelAssignment(assignment);
    }

    // Predict label assignments for all other objects
    if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING)
        classifier.PredictLabelAssignments();

    // Empty object selections
    EmptyObjectSelections();

    // End logging command
    EndCommand();

    // Set message
    const char *label_name = (label->Name()) ? label->Name() : "Noname";
    if (assigned_object) {
        SetMessage("Assigned label %s to object %d", label_name,
                assigned_object->SceneIndex());
    } else {
        SetMessage("Assigned label %s to %d selected objects", label_name,
                selection_size);
    }

    // Return success
    return 1;
}

int R3SurfelLabeler::ConfirmLabelOnPickedObject(int x, int y) {
    // Set message
    SetMessage(NULL);

    //Check interface mode
    if (interface_mode != IGRA_INTERFACE_FULL
            && interface_mode != IGRA_INTERFACE_BATCH_LEARNING
            && interface_mode != IGRA_INTERFACE_UG_GROUP
            && interface_mode != IGRA_INTERFACE_UG_SINGLE) {
        return 0;
    }

    // Pick object
    R3SurfelObject *object = NULL;

    // Selection box is a tiny window around the cursor.
    R2Box box(x - kSelectionBoxRadius, y - kSelectionBoxRadius,
            x + kSelectionBoxRadius, y + kSelectionBoxRadius);

    // Simple filter for objects in the box.
    RNArray<R3SurfelObject *> picked_objects;
    if (!FilterObjectsInsideBox(box, picked_objects,
            R3_SURFEL_LABELER_SELECTION_FILTER_OVERLAP
                    | R3_SURFEL_LABELER_SELECTION_FILTER_OFFSCREEN)) return 0;

    // Select object in the selection area around the cursor.
    object = SelectTopHitObjectInBox(box, picked_objects);
    if (!object) {
        SetMessage("Nothing selected");
        return 0;
    }

    // Set center point
    SetCenterPoint(object->Centroid());

    // Get/check predicted label
    R3SurfelLabel *label = object->PredictedLabel();
    if (!label) {
        SetMessage("Object %d has no predicted label", object->SceneIndex());
        return 0;
    }

    // In user training, don't let incorrect labeling.
    if (training_user_mode) {
        incorrect_objects_.Empty();
        if (object->GroundTruthLabel() != label)
            incorrect_objects_.Insert(object);
        if (incorrect_objects_.NEntries()) {
            incorrect_label_ = label;
            SetMessage(
                    "Incorrect: Marked objects are not what you labeled them.");

            // Record a mistake.
            BeginCommand(R3_SURFEL_LABELER_LABEL_TRAINING_MISTAKE_COMMAND);
            R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                    object, label, 1);
            current_command->inserted_label_assignments.Insert(assignment);
            EndCommand();

            return 0;
        } else incorrect_label_ = NULL;
    }

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_CONFIRM_PICKED_COMMAND);

    // Create/insert label assignment
    R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(object,
            label, 1);
    InsertLabelAssignment(assignment);

    // Predict label assignments for all other objects
    if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING)
        classifier.PredictLabelAssignments();

    // Empty object selections
    EmptyObjectSelections();

    // End logging command
    EndCommand();

    // Set message
    const char *label_name = (label->Name()) ? label->Name() : "without name";
    SetMessage("Confirmed label %s on object %d", label_name,
            object->SceneIndex());

    // Return success
    return 1;
}

int R3SurfelLabeler::ConfirmLabelsOnSelectedObjects(void) {
    // Set message
    SetMessage(NULL);

    // Check number of selections
    if (NObjectSelections() == 0) {
        SetMessage("No objects selected");
        return 0;
    }

    // In user training, don't let incorrect labeling.
    if (training_user_mode) {
        incorrect_objects_.Empty();
        for (int i = 0; i < NObjectSelections(); i++) {
            R3SurfelObject *object = ObjectSelection(i);
            if (object->GroundTruthLabel() != object->PredictedLabel())
                incorrect_objects_.Insert(object);
        }
        if (incorrect_objects_.NEntries()) {
            incorrect_label_ = NULL;
            SetMessage(
                    "Incorrect: Marked objects are not what you labeled them.");

            // Record a mistake.
            BeginCommand(R3_SURFEL_LABELER_LABEL_TRAINING_MISTAKE_COMMAND);
            for (int i = 0; i < NObjectSelections(); i++) {
                R3SurfelObject *object = ObjectSelection(i);
                R3SurfelLabelAssignment *assignment =
                        new R3SurfelLabelAssignment(object,
                                object->PredictedLabel(), 1);
                current_command->inserted_label_assignments.Insert(assignment);
            }
            EndCommand();

            return 0;
        } else incorrect_label_ = NULL;
    }

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_CONFIRM_SELECTION_COMMAND);

    // Confirm label for all selected objects
    int count = 0;
    for (int i = 0; i < NObjectSelections(); i++) {
        R3SurfelObject *object = ObjectSelection(i);

        // Get/check label
        R3SurfelLabel *label = object->PredictedLabel();
        if (!label) continue;

        // Create/insert label assignment
        R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                object, label, 1);
        InsertLabelAssignment(assignment);
        count++;
    }

    // Check if confirmed any labels
    if (count > 0) {
        // Predict label assignments for all other objects
        if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING)
            classifier.PredictLabelAssignments();

        // Empty object selections
        EmptyObjectSelections();

        // Set message
        SetMessage("Confirmed labels for %d objects", count);
    } else {
        // Set message
        SetMessage(
                "None of the selected objects have predicted labels to confirm");
    }

    // End logging command
    EndCommand();

    // Return success
    return 1;
}

int R3SurfelLabeler::ConfirmLabelsOnAllObjects(void) {
    // Set message
    SetMessage(NULL);

    //Check interface mode
    if (interface_mode != IGRA_INTERFACE_FULL) return 0;

    // In user training, don't let incorrect labeling.
    if (training_user_mode) {
        incorrect_objects_.Empty();
        for (int i = 0; i < scene->NObjects(); i++) {
            R3SurfelObject *object = scene->Object(i);
            if (object->GroundTruthLabel() != object->PredictedLabel())
                incorrect_objects_.Insert(object);
        }
        if (incorrect_objects_.NEntries()) {
            incorrect_label_ = NULL;
            SetMessage(
                    "Incorrect: Marked objects are not what you labeled them.");

            // Record a mistake.
            BeginCommand(R3_SURFEL_LABELER_LABEL_TRAINING_MISTAKE_COMMAND);
            for (int i = 0; i < scene->NObjects(); i++) {
                R3SurfelObject *object = ObjectSelection(i);
                R3SurfelLabelAssignment *assignment =
                        new R3SurfelLabelAssignment(object,
                                object->PredictedLabel(), 1);
                current_command->inserted_label_assignments.Insert(assignment);
            }
            EndCommand();

            return 0;
        } else incorrect_label_ = NULL;
    }

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_CONFIRM_ALL_COMMAND);

    // Confirm label for all objects
    int count = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);

        // Get/check label
        if (object->GroundTruthLabel()) continue;
        R3SurfelLabel *label = object->PredictedLabel();
        if (!label) continue;

        // Create/insert label assignment
        R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                object, label, 1);
        InsertLabelAssignment(assignment);
        count++;
    }

    if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING)
        classifier.PredictLabelAssignments();

    // Empty object selections
    EmptyObjectSelections();

    // End logging command
    EndCommand();

    // Set message
    SetMessage("Confirmed label for %d objects", count);

    // Return success
    return 1;
}

int R3SurfelLabeler::Undo(void) {
    // Check everything
    if (!scene) return 0;
    if (undo_index < 0) {
        SetMessage("Nothing to undo");
        return 0;
    }

    // Set message
    SetMessage("Undid previous command");

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_UNDO_COMMAND);

    // Pop command from undo stack
    R3SurfelLabelerCommand *command = undo_stack.Kth(undo_index--);

    // Set camera
    viewer.SetCamera(command->camera);

    // Remove inserted label assignments
    for (int i = 0; i < command->inserted_label_assignments.NEntries(); i++) {
        R3SurfelLabelAssignment *assignment =
                command->inserted_label_assignments.Kth(i);
        RemoveLabelAssignment(assignment);
    }

    // Insert removed label assignments
    for (int i = 0; i < command->removed_label_assignments.NEntries(); i++) {
        R3SurfelLabelAssignment *assignment =
                command->removed_label_assignments.Kth(i);
        InsertLabelAssignment(assignment);
    }

    // Remove inserted object selections
    for (int i = 0; i < command->inserted_object_selections.NEntries(); i++) {
        R3SurfelObject *object = command->inserted_object_selections.Kth(i);
        RemoveObjectSelection(object);
    }

    // Insert removed object selections
    for (int i = 0; i < command->removed_object_selections.NEntries(); i++) {
        R3SurfelObject *object = command->removed_object_selections.Kth(i);
        InsertObjectSelection(object);
    }

    // Predict label assignments for all other objects
    if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING) {
        classifier.PredictLabelAssignments();
        selection_changed_ = 1;
    }

    // End command
    EndCommand();

    if (undo_index > 0) {
        int cmd_type = undo_stack.Kth(undo_index - 1)->type;
        if (cmd_type == R3_SURFEL_LABELER_SUGGEST_INCREASE_SELECTION
                || cmd_type == R3_SURFEL_LABELER_SUGGEST_DECREASE_SELECTION
                || command->type == R3_SURFEL_LABELER_SELECT_SUGGESTED_COMMAND)
            ZoomToObjectSet(selection_objects, false);
    }

    // Return success
    return 1;
}

int R3SurfelLabeler::Redo(void) {
    // Check everything
    if (!scene) return 0;
    if (undo_stack.IsEmpty() || (undo_index >= undo_stack.NEntries() - 1)) {
        SetMessage("Nothing to redo");
        return 0;
    }

    // Set message
    SetMessage("Redid previous command");

    // Get last command
    R3SurfelLabelerCommand *command = undo_stack.Kth(++undo_index);

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_REDO_COMMAND);

    // Set camera
    viewer.SetCamera(command->camera);

    // Remove label assignments
    for (int i = 0; i < command->removed_label_assignments.NEntries(); i++) {
        R3SurfelLabelAssignment *assignment =
                command->removed_label_assignments.Kth(i);
        RemoveLabelAssignment(assignment);
    }

    // Insert label assignments
    for (int i = 0; i < command->inserted_label_assignments.NEntries(); i++) {
        R3SurfelLabelAssignment *assignment =
                command->inserted_label_assignments.Kth(i);
        InsertLabelAssignment(assignment);
    }

    // Remove object selections
    for (int i = 0; i < command->removed_object_selections.NEntries(); i++) {
        R3SurfelObject *object = command->removed_object_selections.Kth(i);
        RemoveObjectSelection(object);
    }

    // Predict label assignments for all other objects
    if (interface_mode != IGRA_INTERFACE_BATCH_LEARNING) {
        classifier.PredictLabelAssignments();
        selection_changed_ = 1;
    }

    // Insert object selections
    for (int i = 0; i < command->inserted_object_selections.NEntries(); i++) {
        R3SurfelObject *object = command->inserted_object_selections.Kth(i);
        InsertObjectSelection(object);
    }

    // End logging command
    EndCommand();

    if (command->type == R3_SURFEL_LABELER_SUGGEST_INCREASE_SELECTION
            || command->type == R3_SURFEL_LABELER_SUGGEST_DECREASE_SELECTION
            || command->type == R3_SURFEL_LABELER_SELECT_SUGGESTED_COMMAND)
        ZoomToObjectSet(selection_objects, false);

    // Return success
    return 1;
}

int R3SurfelLabeler::Reset(void) {
    // Check everything
    if (!scene) return 0;
    if (current_command) return 0;

    // Set message
    SetMessage("Reset");

    // Log command
    BeginCommand(R3_SURFEL_LABELER_RESET_COMMAND);

    // Remove all object selections
    EmptyObjectSelections();

    // Remove all label assignments
    EmptyLabelAssignments();

    EndCommand();

    // Empty undo stack
    undo_stack.Empty();
    undo_index = -1;

    // Return success
    return 1;
}

//Batch learning mode commands

int R3SurfelLabeler::PredictAll(void) {

    // Check everything
    if (!scene) return 0;
    if (current_command) return 0;

    // Set message
    SetMessage("Predicted labels for all unassigned objects");

    // Print checkpoint
    if (logging_fp) PrintCheckpoint(logging_fp);

    //Log command
    BeginCommand(R3_SURFEL_LABELER_PREDICT_ALL_COMMAND);

    // Remove all object selections
    EmptyObjectSelections();

    // Predict label assignments for all objects in scene
    classifier.PredictLabelAssignments();

    //End command
    EndCommand();

    // Print checkpoint
    if (logging_fp) PrintCheckpoint(logging_fp);

    // Empty undo stack
    undo_stack.Empty();
    undo_index = -1;

    // Return success
    return 1;
}

////////////////////////////////////////////////////////////////////////
// LABELING STATISTICS
////////////////////////////////////////////////////////////////////////

RNScalar R3SurfelLabeler::HumanLabelPrecision(void) const {
    // Count predictions
    int correct_predictions = 0;
    int total = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        R3SurfelLabel *ground_truth_label = object->GroundTruthLabel();
        if (!ground_truth_label) continue;
        R3SurfelLabel *label = object->HumanLabel();
        if (label && strcmp(label->Name(), "Unknown")) {
            if (label == ground_truth_label) correct_predictions++;
            total++;
        }
    }

    // Return precision
    if (total > 0) {
        return (RNScalar) correct_predictions / (RNScalar) total;
    }
    return 1;
}

RNScalar R3SurfelLabeler::HumanLabelRecall(void) const {
    // Count predictions
    int correct_predictions = 0;
    int total_objects = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        R3SurfelLabel *ground_truth_label = object->GroundTruthLabel();
        if (!ground_truth_label) continue;
        R3SurfelLabel *label = object->HumanLabel();
        if (label && strcmp(label->Name(), "Unknown")) {
            if (label == ground_truth_label) correct_predictions++;
        }
        total_objects++;
    }

    // Return precision
    if (total_objects > 0) {
        return (RNScalar) correct_predictions / (RNScalar) total_objects;
    }
    return 0;
}

RNScalar R3SurfelLabeler::HumanLabelFMeasure(void) const {
    // Return F-measure
    RNScalar precision = HumanLabelPrecision();
    RNScalar recall = HumanLabelRecall();
    RNScalar sum = precision + recall;
    if (sum > 0) return 2 * precision * recall / sum;
    else return 0;
}

RNScalar R3SurfelLabeler::PredictedLabelPrecision(void) const {
    // Count predictions
    int correct_predictions = 0;
    int total_predictions = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        R3SurfelLabel *ground_truth_label = object->GroundTruthLabel();
        if (!ground_truth_label) continue;
        R3SurfelLabel *predicted_label = object->PredictedLabel();
        if (predicted_label && strcmp(predicted_label->Name(), "Unknown")) {
            if (predicted_label == ground_truth_label) correct_predictions++;
            total_predictions++;
        }
    }

    // Return precision
    if (total_predictions > 0) return (RNScalar) correct_predictions
            / (RNScalar) total_predictions;
    else return 1;
}

RNScalar R3SurfelLabeler::PredictedLabelRecall(void) const {
    // Count predictions
    int correct_predictions = 0;
    int total_objects = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        R3SurfelLabel *ground_truth_label = object->GroundTruthLabel();
        if (!ground_truth_label) continue;
        R3SurfelLabel *predicted_label = object->PredictedLabel();
        if (predicted_label && strcmp(predicted_label->Name(), "Unknown")) {
            if (predicted_label == ground_truth_label) correct_predictions++;
        }
        total_objects++;
    }

    // Return precision
    if (total_objects > 0) return (RNScalar) correct_predictions
            / (RNScalar) total_objects;
    else return 0;
}

RNScalar R3SurfelLabeler::PredictedLabelFMeasure(void) const {
    // Return F-measure
    RNScalar precision = PredictedLabelPrecision();
    RNScalar recall = PredictedLabelRecall();
    RNScalar sum = precision + recall;
    if (sum > 0) return 2 * precision * recall / sum;
    else return 0;
}

////////////////////////////////////////////////////////////////////////
// CAMERA CONTROL
////////////////////////////////////////////////////////////////////////

int R3SurfelLabeler::ZoomCamera(RNScalar scale) {
    // Check interface.
    if (interface_mode == IGRA_INTERFACE_TILED) return 0;

    // Set message
    SetMessage("Zoomed to closeup view");

    // Begin logging command
    BeginCommand(R3_SURFEL_LABELER_ZOOM_COMMAND, scale);

    // Set the camera viewpoint
    R3SurfelViewer::ZoomCamera(scale);

    // End logging command
    EndCommand();

    // Return success
    return 1;
}

int R3SurfelLabeler::ZoomToObject(const R3SurfelObject* o,
        const bool reset_line) {
    if (!o) return 0;

    RNArray<R3SurfelObject*> set;
    set.Insert((R3SurfelObject*) o);
    return ZoomToObjectSet(set, reset_line);
}

int R3SurfelLabeler::ZoomToObjectSet(const RNArray<R3SurfelObject*> &set,
        const bool reset) {
    const int N = set.NEntries();
    if (N == 0) return 0;

    RecenterOnSet(set);

    // Set resolution support radius.
    RNLength radius = kSelectionBoxRadius; // default smallest radius to be shown
    R3Vector p_dir(R3null_vector);
    for (int i = 0; i < N; ++i) {
        R3SurfelObject *o = set.Kth(i);
        if (!o) continue;
        for (int j = 0; j < 8; ++j) {
            R3Vector v = o->BBox().Corner(j) - center_point;
            RNLength dist = v.Length();
            if (radius < dist) {
                radius = dist;
                p_dir = v;
            } else if (p_dir.IsZero()) p_dir = v;
        }
    }
    if (FocusRadius() < radius) SetFocusRadius(radius);

    // Identify active subwindow (not overlapped by control elements).
    R2Box active_box(0, 0, window_width, window_height);
    if (command_menu_visibility && !command_menu_box_.IsEmpty()) {
        // Limit upper right corner.
        if (command_menu_box_.XLength() > command_menu_box_.YLength()) active_box[1][1] =
                command_menu_box_[0][1]; // limit from top
        else active_box[1][0] = command_menu_box_[0][0]; // limit right
    }

    if (label_menu_visibility && !label_menu_box_.IsEmpty()) {
        if (label_menu_box_[1][0] < 0.5 * window_width) active_box[0][0] =
                label_menu_box_[1][0]; // limit from left
        else if (label_menu_box_[0][0] > 0.5 * window_width) active_box[1][0] =
                label_menu_box_[0][0]; // limit from right
    }

    // Reset main animation.
    if (reset) animation_phi_ = 0; // RN_PI / kThetaPhiCoeff;

    R3Sphere sphere(center_point, radius);
    return ZoomToROI((R3Camera&) viewer.Camera(), Viewport(), active_box,
            sphere, reset ? &p_dir : NULL);
}

int R3SurfelLabeler::ZoomToROI(R3Camera &camera, const R2Viewport &viewport,
        const R2Box &active_area, const R3Sphere &sphere,
        const R3Vector *p_dir) {
    if (RNIsZero(sphere.Radius())) return 0;

    // On-screen circle parameters.
    assert(R2Contains(viewport.BBox(), active_area));
    R2Circle circle(active_area.Centroid(),
            0.5 * min(active_area.XLength(), active_area.YLength()));

    // Get the current circle center vector.
    R3Viewer viewer(camera, viewport);
    RNScalar x = circle.Center().X();
    RNScalar y = circle.Center().Y();
    RNScalar r = circle.Radius();
    if (RNIsZero(r)) return 0;
    R3Vector center_v = viewer.WorldRay(x, y).Vector();
    R3Vector side_v[4];
    side_v[0] = viewer.WorldRay(x, y + r).Vector(); // up
    side_v[1] = viewer.WorldRay(x, y - r).Vector(); // down
    side_v[2] = viewer.WorldRay(x + r, y).Vector(); // right
    side_v[3] = viewer.WorldRay(x - r, y + r).Vector(); // left

    // Estimate the distance to the camera.
    RNScalar R = sphere.Radius();
    RNLength dist = 0.0;
    for (int i = 0; i < 4; ++i) {
        RNScalar sin_angle = (side_v[i] % center_v).Length();
        dist = max(dist, R / sin_angle);
    }

    // Translate camera so that sphere center is on the new world_ray.
    // Slide camera along center_v such that sphere fits.
    camera.SetOrigin(sphere.Center() - dist * center_v);

    if (p_dir) {
        R3Plane hor_plane(sphere.Center(), R3posz_vector);
        R3Vector p_norm(*p_dir);
        p_norm.Project(hor_plane);
        p_norm.Normalize();

        if ((center_v % p_norm)[2] < 0) p_norm.Flip(); // make sure p_norm points left of center_v

        RNAngle adj_theta = -R3InteriorAngle(center_v, hor_plane);
        RNAngle adj_phi = RN_PI_OVER_TWO - R3InteriorAngle(center_v, p_norm);

        R3Vector twd = camera.Towards();
        twd.Normalize();
        center_v.Rotate(R3posz_vector, adj_phi);
        twd.Rotate(R3posz_vector, adj_phi);
        center_v.Rotate(p_norm, adj_theta);
        twd.Rotate(p_norm, adj_theta);

        R3Vector right = twd % R3posz_vector;
        right.Normalize();
        R3Vector up = right % twd;
        up.Normalize();

        camera.Reset(sphere.Center() - dist * center_v, twd, up);
    }

    // Return success
    return 1;
}

void R3SurfelLabeler::BeginCommand(int type, double operand) {
    // Check current command
    assert(!current_command);

    // Start current command
    current_command = new R3SurfelLabelerCommand(this, type, operand);
}

void R3SurfelLabeler::EndCommand(void) {
    // Check current command
    assert(current_command);

    // Insert command into undo stack, if it can be undone
    switch (current_command->type) {
        case R3_SURFEL_LABELER_ZOOM_COMMAND:
        case R3_SURFEL_LABELER_SELECT_NONE_COMMAND:
        case R3_SURFEL_LABELER_SELECT_PICKED_COMMAND:
        case R3_SURFEL_LABELER_SELECT_ENCLOSED_COMMAND:
        case R3_SURFEL_LABELER_SELECT_SUGGESTED_COMMAND:
        case R3_SURFEL_LABELER_SELECT_ALL_COMMAND:
        case R3_SURFEL_LABELER_LABEL_PICKED_COMMAND:
        case R3_SURFEL_LABELER_LABEL_SELECTION_COMMAND:
        case R3_SURFEL_LABELER_CONFIRM_PICKED_COMMAND:
        case R3_SURFEL_LABELER_CONFIRM_SELECTION_COMMAND:
        case R3_SURFEL_LABELER_CONFIRM_ALL_COMMAND:
        case R3_SURFEL_LABELER_REFUTE_PICKED_COMMAND:
        case R3_SURFEL_LABELER_REFUTE_SELECTION_COMMAND:
        case R3_SURFEL_LABELER_REFUTE_ALL_COMMAND:
        case R3_SURFEL_LABELER_MERGE_OBJECTS:
        case R3_SURFEL_LABELER_SPLIT_OBJECTS:
        case R3_SURFEL_LABELER_SUGGEST_INCREASE_SELECTION:
        case R3_SURFEL_LABELER_SUGGEST_DECREASE_SELECTION:
            // Delete commands that were undone
            for (int i = undo_index + 1; i < undo_stack.NEntries(); i++)
                delete undo_stack.Kth(i);
            undo_stack.Truncate(undo_index + 1);

            // Insert command into undo stack
            undo_stack.Insert(current_command);
            undo_index = undo_stack.NEntries() - 1;
            break;
    }

    if (logging_fp) {
        // Write command to log
        current_command->Write(logging_fp);

        // Write checkpoint to log
        const RNScalar checkpoint_interval = 60;
        static RNScalar last_checkpoint_time = -FLT_MAX;

        bool synchronizing_cmd =
                (current_command->type == R3_SURFEL_LABELER_INITIALIZE_COMMAND)
                        || (current_command->type
                                == R3_SURFEL_LABELER_TERMINATE_COMMAND)
                        || (current_command->type
                                == R3_SURFEL_LABELER_SYNC_COMMAND);
        bool timer_induced = (current_command->timestamp
                >= last_checkpoint_time + checkpoint_interval);

        if (synchronizing_cmd || timer_induced) {
            last_checkpoint_time = current_command->timestamp;

            // log checkpoint.
            PrintCheckpoint(logging_fp);

            // To synchronize logs and progress in case of crash,
            // also save every time a timer elapses.
            if (auto_save) {
                if (!synchronizing_cmd) scene->SyncFile();
            }
        }
    }

    // End current command
    current_command = NULL;
}

void R3SurfelLabeler::PrintCheckpoint(FILE *fp) const {
    R3SurfelLabelerCheckpoint::PrintCheckpoint(fp, this);
}

////////////////////////////////////////////////////////////////////////
// OBJECT SELECTION UTILITY FUNCTIONS
////////////////////////////////////////////////////////////////////////

void R3SurfelLabeler::InsertObjectSelection(R3SurfelObject *object) {
    // Check scene
    if (!scene) return;

    // Update current command
    if (current_command)
        current_command->inserted_object_selections.Insert(object);

    // Insert into selected objects
    selection_objects.Insert(object);
    selection_changed_ = 1;

    // Read blocks
    object->ReadBlocks();

    recenter_on_selection_ = true;

    // A new selection zoom box.
    InsertZoomBoxForObject(object);

    // Redraw classifier debugging if any
    classifier.RequestRedraw();
}

void R3SurfelLabeler::RemoveObjectSelection(R3SurfelObject *object) {
    // Check scene
    if (!scene) return;

    // Update current command
    if (current_command)
        current_command->removed_object_selections.Insert(object);

    // Insert into selected objects
    selection_objects.Remove(object);
    selection_changed_ = 1;

    // Release blocks
    object->ReleaseBlocks();

    recenter_on_selection_ = true;

    // Remove the selection zoom box.
    RemoveZoomBoxForObject(object);

    // Redraw classifier debugging if any
    classifier.RequestRedraw();
}

int R3SurfelLabeler::EmptyObjectSelections(void) {
    // Check scene
    if (!scene) return 0;

    // Update current command
    if (current_command)
        current_command->removed_object_selections.Append(selection_objects);

    // Release blocks of selected objects
    for (int i = 0; i < selection_objects.NEntries(); i++) {
        R3SurfelObject *object = selection_objects.Kth(i);
        object->ReleaseBlocks();
    }

    // Empty object selection
    selection_objects.Empty();
    selection_changed_ = 1;

    // Empty the selection zoom boxes.
    EmptyZoomBoxes();

    // Redraw classifier debugging if any
    classifier.RequestRedraw();

    // Return success
    return 1;
}

void R3SurfelLabeler::RecenterOnSet(const RNArray<R3SurfelObject*> &set) {
    // Set center point of the scene to the centroid of the selection.
    const int N = set.NEntries();
    if (N == 0) return;

    // Set center point and resolution support radius.
    R3Point centroid(R3null_point);
    int n = 0;
    for (int i = 0; i < N; ++i) {
        R3SurfelObject *o = set.Kth(i);
        if (!o) continue;
        centroid += o->Centroid() - R3null_point;
        ++n;
    }
    if (n == 0) return;
    centroid /= n;
    SetCenterPoint(centroid);

    recenter_on_selection_ = false;
}

////////////////////////////////////////////////////////////////////////
// SCREEN SELECTION UTILITY FUNCTIONS
////////////////////////////////////////////////////////////////////////

int R3SurfelLabeler::FilterObjectsInsideBox(const R2Box& box,
        RNArray<R3SurfelObject *> &picked_objects, int filter) const {
    // Check scene
    if (!scene) return 0;
    if (box.IsEmpty()) return 0;

    bool overlap_test = filter & R3_SURFEL_LABELER_SELECTION_FILTER_OVERLAP;
    bool inside_test = filter & R3_SURFEL_LABELER_SELECTION_FILTER_INSIDE;
    bool offscreen_test = filter & R3_SURFEL_LABELER_SELECTION_FILTER_OFFSCREEN;

    if (!overlap_test && !inside_test && !offscreen_test) return 0;

    const double screen_max_y = viewer.Viewport().Height();
    const double screen_max_x = viewer.Viewport().Width();
    const double box_min_x = box.XMin();
    const double box_min_y = box.YMin();
    const double box_max_x = box.XMax();
    const double box_max_y = box.YMax();

    // Get projection matrices
    GLdouble p[3];
    GLdouble modelview_matrix[16];
    GLdouble projection_matrix[16];
    GLint viewport[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
    glGetIntegerv(GL_VIEWPORT, viewport);

    // Make array of picked objects
    picked_objects.Empty();
    for (int i = 1; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        if (object->Parent() != scene->RootObject()) continue;
        if (!ShowConfirmed(object)) continue;

        R3SurfelLabel *label = object->PredictedLabel();
        if (label && !LabelSelectible(label)) continue;

        // Perform simple tests to select the objects that can be selected.
        int inside_corner_count = 0;
        int off_screen_count = 0;
        bool non_overlap[4] = { true, true, true, true };

        for (int octant = 0; octant < 8; octant++) {
            R3Point corner = object->BBox().Corner(octant);
            gluProject(corner[0], corner[1], corner[2], modelview_matrix,
                    projection_matrix, viewport, &(p[0]), &(p[1]), &(p[2]));

            // Off screen test.
            if (offscreen_test
                    && ((p[2] < 0) || (p[0] < 0) || (p[1] < 0)
                            || (p[0] > screen_max_x) || (p[1] > screen_max_y))) {
                off_screen_count++;
            }

            // Enclosure test.
            if (inside_test && (box_min_x <= p[0]) && (box_min_y <= p[1])
                    && (box_max_x >= p[0]) && (box_max_y >= p[1])) {
                inside_corner_count++;
            }

            // Overlap test.
            if (overlap_test) {
                non_overlap[0] &= p[0] < box_min_x;
                non_overlap[1] &= p[0] > box_max_x;
                non_overlap[2] &= p[1] < box_min_y;
                non_overlap[3] &= p[1] > box_max_y;
            }
        }

        bool overlap_success = !overlap_test
                || !(non_overlap[0] || non_overlap[1] || non_overlap[2]
                        || non_overlap[3]);
        bool inside_success = !inside_test || inside_corner_count >= 5;
        bool offscreen_success = !offscreen_test || off_screen_count < 4;
        if (overlap_success && inside_success && offscreen_success)
            picked_objects.Insert(object);
    }

    // Return success.
    return 1;
}

R3SurfelObject* R3SurfelLabeler::SelectTopHitObjectInBox(const R2Box& box,
        const RNArray<R3SurfelObject *> &objects) const {
    R3SurfelObject *object = NULL;

    // Allocate select buffer
    const int SELECT_BUFFER_SIZE = 1024;
    GLuint select_buffer[SELECT_BUFFER_SIZE];
    GLint select_buffer_hits;

    // Initialize select buffer
    glSelectBuffer(SELECT_BUFFER_SIZE, select_buffer);
    glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);

    // Draw faces with pick names into selection buffer
    GLint viewport[4];
    glViewport(0, 0, Viewport().Width(), Viewport().Height());
    glGetIntegerv(GL_VIEWPORT, viewport);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPickMatrix((GLdouble) box.Centroid().X(), (GLdouble) box.Centroid().Y(),
            (GLdouble) box.XLength(), (GLdouble) box.YLength(), viewport);
    viewer.Camera().Load(TRUE);
    glMatrixMode(GL_MODELVIEW);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    for (int i = 0; i < objects.NEntries(); ++i) {
        R3SurfelObject *picked_object = objects.Kth(i);
        if (!picked_object) continue;
        if (!ShowConfirmed(picked_object)) continue;

        R3SurfelLabel *label = picked_object->PredictedLabel();
        if (label && !LabelSelectible(label)) continue;

        glLoadName(i + 1);

        DrawObject(picked_object);
    }
    glFlush();
    select_buffer_hits = glRenderMode(GL_RENDER);

    // Process select buffer to find front-most hit
    GLuint hit = 0;
    GLuint hit_z = 0xFFFFFFFF;
    GLuint *bufp = select_buffer;
    GLuint numnames, z1, z2;
    for (int i = 0; i < select_buffer_hits; i++) {
        numnames = *bufp++;
        z1 = *bufp++;
        z2 = *bufp++;
        while (numnames--) {
            if (z1 < hit_z) {
                hit = *bufp;
                hit_z = z1 / 2 + z2 / 2;
            }
            bufp++;
        }
    }

    // Check if hit anything
    if (hit > 0) {
        // Find hit node
        hit--; // subtract the one added to avoid zero
        if (hit >= 0 && hit < (GLuint) objects.NEntries()) {
            object = objects.Kth(hit);
        }
    }

    return object;
}

int R3SurfelLabeler::SelectTopHitObjectsInBox(const R2Box& box,
        const RNArray<R3SurfelObject *> &objects,
        RNArray<R3SurfelObject *> &picked_objects) const {
    // Use backbuffer for object selection. Use large point size with colors
    // encoding an object index;
    GLfloat smoothing_radius = 10;

    // Check scene
    if (!scene) return 0;
    if (objects.NEntries() >= 256 * 256 * 256) {
        // TODO(aboyko): To deal with too many objects, need several passes.
        // For now just crash...
        fprintf(stderr,
                "back buffer selection currently supports %d objects. %d supplied.\n",
                256 * 256 * 256 - 1, objects.NEntries());
        exit(-1);
    }

    // Set draw modes.
    GLboolean texture, fog, lighting, dithering;
    glGetBooleanv(GL_TEXTURE_2D, &texture);
    glGetBooleanv(GL_FOG, &fog);
    glGetBooleanv(GL_LIGHTING, &lighting);
    glGetBooleanv(GL_DITHER, &dithering);

    if (texture) glDisable(GL_TEXTURE_2D);
    if (fog) glDisable(GL_FOG);
    if (lighting) glDisable(GL_LIGHTING);
    if (dithering) glDisable(GL_DITHER);

    // Clear window
    glDrawBuffer(GL_BACK);
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Set viewing transformation
    viewer.Camera().Load();

    // Draw surfels
    if (surfel_visibility) {
        //Draw foreground
        glPushMatrix();
        R3Vector offset = -0.02 * viewer.Camera().Towards();
        glTranslated(offset[0], offset[1], offset[2]);

        // Draw labeled objects
        glMatrixMode(GL_MODELVIEW);
        glPointSize(smoothing_radius);
        int red = 0;
        int green = 0;
        int blue = 0;
        for (int i = 0; i < objects.NEntries(); i++) {
            R3SurfelObject *object = objects.Kth(i);
            if (object->Parent() != scene->RootObject()) continue;

            R3SurfelLabel *label = object->PredictedLabel();

            if (label && !LabelVisible(label)) continue;
            if (!ShowConfirmed(object)) continue;

            // Encode index into a color.
            blue++;
            if (blue > 255) {
                blue = 0;
                green++;
                if (green > 255) {
                    green = 0;
                    red++;
                }
            }
            glColor3ub((GLubyte) red, (GLubyte) green, (GLubyte) blue);

            DrawObject(object, 0);
        }
        glPointSize(surfel_size);
        glPopMatrix();
    }
    glFinish();

    // Retrieve the indices of the visible objects from the back buffer.
    const int buffer_size = box.XLength() * box.YLength();
    unsigned char *buffer = new unsigned char[4 * buffer_size];
    glReadBuffer(GL_BACK);
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadPixels((GLint) box.XMin(), (GLint) box.YMin(),
            (GLsizei) box.XLength(), (GLsizei) box.YLength(), GL_RGBA,
            GL_UNSIGNED_BYTE, buffer);
    glFinish();

    set<int> used_indices;
    const int kRedPower = 256 * 256;
    const int kGreenPower = 256;
    const int kBluePower = 1;
    unsigned char *ptr = buffer;
    for (int i = 0; i < buffer_size; ++i) {
        unsigned char red = *ptr++;
        unsigned char green = *ptr++;
        unsigned char blue = *ptr++;
        ptr++;
        int index = kRedPower * red + kGreenPower * green + kBluePower * blue;
        if (index <= 0) continue;
        index--;
        if (index < objects.NEntries()) used_indices.insert(index);
    }
    delete[] buffer;

    // Populate picked_objects with selected objects.
    for (set<int>::iterator it = used_indices.begin(); it != used_indices.end();
            ++it) {
        picked_objects.Insert(objects.Kth(*it));
    }

    // Restore drawing modes.
    if (texture) glEnable(GL_TEXTURE_2D);
    if (fog) glEnable(GL_FOG);
    if (lighting) glEnable(GL_LIGHTING);
    if (dithering) glEnable(GL_DITHER);

    return 1;
}

bool R3SurfelLabeler::ShowConfirmed(R3SurfelObject *object) const {
    return object_label_confirmed_visibility || !classifier.Labeled(object);
}

////////////////////////////////////////////////////////////////////////
// LABEL ASSIGNMENT UTILITY FUNCTIONS
////////////////////////////////////////////////////////////////////////

int R3SurfelLabeler::InsertLabelAssignment(
        R3SurfelLabelAssignment *assignment) {
    // Check everything
    if (!scene) return 0;
    if (!assignment) return 0;

#ifdef R3_SURFEL_LABELER_NO_UNKNOWNS
    // Don't allow Unknown to be assigned.
    R3SurfelLabel *label = assignment->Label();
    if (!assignment->Originator() && !strcmp(label->Name(), "Unknown")) {
        SetMessage("There are no Unknowns in this data, label not assigned.");
        return 0;
    }
#endif

    // Update current command
    if (current_command)
        current_command->inserted_label_assignments.Insert(assignment);

    // Insert assignment into scene
    scene->InsertLabelAssignment(assignment);

    // Update classifier
    classifier.UpdateAfterInsertLabelAssignment(assignment);

    // Return success
    return 1;
}

int R3SurfelLabeler::RemoveLabelAssignment(
        R3SurfelLabelAssignment *assignment) {
    // Check everything
    if (!scene) return 0;
    if (!assignment) return 0;

    // Update classifier
    classifier.UpdateBeforeRemoveLabelAssignment(assignment);

    // Update current command
    if (current_command
            && assignment->Originator()
                    != R3_SURFEL_LABEL_ASSIGNMENT_MACHINE_ORIGINATOR)
        current_command->removed_label_assignments.Insert(assignment);

    // Remove assignment from scene
    scene->RemoveLabelAssignment(assignment);

    // Return success

    return 1;
}

int R3SurfelLabeler::EmptyLabelAssignments(void) {
    // Check everything
    if (!scene) return 0;

    // Remove all label assignments
    RNArray<R3SurfelLabelAssignment *> tmp_assignments;
    for (int i = 0; i < scene->NLabelAssignments(); i++)
        tmp_assignments.Insert(scene->LabelAssignment(i));
    for (int i = 0; i < tmp_assignments.NEntries(); i++) {

        R3SurfelLabelAssignment *assignment = tmp_assignments.Kth(i);
        RemoveLabelAssignment(assignment);
    }

    // Return success
    return 1;
}

// Interface Helpers.

void R3SurfelLabeler::TogglePause(void) {
    paused_ = !paused_;
    BeginCommand(R3_SURFEL_LABELER_PAUSE, paused_ ? 1 : 0);
    EndCommand();
    static bool animation_was_on_before_pause = false;
    if (paused_) {
        animation_was_on_before_pause = animation_on_;
        StopAnimation();
    } else if (animation_was_on_before_pause) StartAnimation();
}

void R3SurfelLabeler::FlipBGColor(void) {
    for (int i = 0; i < 3; ++i) {
        background_color[i] = 1 - background_color[i];
    }
}

bool R3SurfelLabeler::isMachineDrivenInterface(void) const {
    return interface_mode == IGRA_INTERFACE_ACTIVE_SINGLE
            || interface_mode == IGRA_INTERFACE_ACTIVE_GROUP
            || interface_mode == IGRA_INTERFACE_TILED;
}

////////////////////////////////////////////////////////////////////////
// SUBWINDOW DRAWING AND PICKING
////////////////////////////////////////////////////////////////////////

void R3SurfelLabeler::DrawSelectBox(void) const {
    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);
    glDrawBuffer(GL_FRONT);
    glLineStipple(1, 0xFF00);
    glEnable(GL_LINE_STIPPLE);
    glLogicOp(GL_XOR);
    glEnable(GL_COLOR_LOGIC_OP);

    // Draw box
    glLineWidth(1);
    glBegin(GL_LINE_LOOP);
    glColor3f(1.0, 1.0, 1.0);
    glVertex2f(select_box_corners[0][0], select_box_corners[0][1]);
    glVertex2f(select_box_corners[0][0], select_box_corners[1][1]);
    glVertex2f(select_box_corners[1][0], select_box_corners[1][1]);
    glVertex2f(select_box_corners[1][0], select_box_corners[0][1]);
    glEnd();

    // Reset rendering modes
    glDisable(GL_COLOR_LOGIC_OP);
    glDisable(GL_LINE_STIPPLE);
    glDrawBuffer(GL_BACK);
    glFlush();
}

void R3SurfelLabeler::DrawSurfels(void) {
    // Draw surfels.
    if (!surfel_visibility) return;
    if (interface_mode == IGRA_INTERFACE_TILED) return;

    // Set viewing transformation
    viewer.Camera().Load();

    // Set draw modes
    glDisable(GL_LIGHTING);
    glPointSize(surfel_size);
    glLineWidth(1);

    // Draw background.
    if (background_visibility) {
        for (int i = 0; i < resident_nodes.NNodes(); i++) {
            R3SurfelNode *node = resident_nodes.Node(i);
            if (node->Object()) continue;
            node->Draw();
        }
    }

    // Draw objects.
    if (scene->NLabelAssignments() > 0) {
        glMatrixMode(GL_MODELVIEW);
        for (int i = 0; i < scene->NObjects(); i++) {
            R3SurfelObject *object = scene->Object(i);
            if (object->Parent() != scene->RootObject()) continue;

            RNBoolean confirmed = classifier.Labeled(object);

            R3SurfelLabel *label = object->PredictedLabel();
            RNFlags flags = 0;

            if (confirmed) glPointSize(2 * surfel_size);
            else glPointSize(surfel_size);

            R3SurfelObject *selected_object =
                    NObjectSelections() == 1 ? ObjectSelection(0) : NULL;

            if (!label) CustomLoadRgb(RNblack_rgb);
            else if (!LabelVisible(label)
                    || (!object_label_confirmed_visibility && confirmed)) {
                glColor3f(0.1, 0.1, 0.1);
            } else {
                switch (SurfelColorScheme()) {
                    case R3_SURFEL_VIEWER_COLOR_BY_CONFIDENCE: {
                        RNScalar confidence = 0;
                        R3SurfelLabelAssignment *assignment =
                                object->PredictedLabelAssignment();
                        if (assignment) confidence = assignment->Confidence();
                        LoadColorForScalar(confidence);
                    }
                        break;
                    case R3_SURFEL_VIEWER_COLOR_BY_RGB:
                        flags = R3_SURFEL_COLOR_DRAW_FLAG;
                        break;

                    case R3_SURFEL_VIEWER_COLOR_BY_SHAPE_AFFINITY:
                    case R3_SURFEL_VIEWER_COLOR_BY_FEATURE_DISTANCE:
                    case R3_SURFEL_VIEWER_COLOR_BY_INDUCED_AFFINITY:
                    case R3_SURFEL_VIEWER_COLOR_BY_SPATIAL_DISTANCE:
                        if (selected_object) {
                            // Color by affinities to selection.
                            // Superscript the value is above threshold.
                            RNScalar text_threshold = 0.01;
                            RNScalar affinity = AffinityScore(selected_object,
                                    object);

                            if (affinity > text_threshold) {
                                R3Point position = object->Centroid();
                                position[2] = object->BBox().ZMax() + 1;
                                R2Point p = viewer.ViewportPoint(position);
                                char buffer[10];
                                sprintf(buffer, "%3.2f", affinity);
                                int width = glutBitmapLength(
                                GLUT_BITMAP_HELVETICA_12,
                                        (const unsigned char*) buffer);
                                p[0] -= width / 2;
                                // Set rendering modes.
                                Ortho2DLocker locker(window_width,
                                        window_height);
                                RNRgb adj_color = GetColorForScalar(affinity);
                                DrawHaloedMessage(p, buffer,
                                GLUT_BITMAP_HELVETICA_12, &adj_color);
                            }

                            LoadColorForScalar(affinity);

                        } else {
                            LoadColorForScalar(0.0);
                        }
                        break;

                    default:
                        CustomLoadRgb(label->Color());
                        break;
                }
            }

            glPushMatrix();
            R3Vector offset = -0.02
                    * (object->Centroid() - viewer.Camera().Origin());
            glTranslated(offset[0], offset[1], offset[2]);

            DrawObject(object, flags);

            glPopMatrix();

        }
        glPointSize(surfel_size);
    }

    // Draw selected objects
    if (NObjectSelections() > 0) {
        DrawDimLayer(&RNblack_rgb, 0.5);

        // Sort back to front;
        multimap<RNScalar, int> sorted_selection;
        const R3Point &o = Camera().Origin();
        for (int i = 0; i < NObjectSelections(); i++) {
            R3SurfelObject *object = ObjectSelection(i);
            RNScalar dist = R3SquaredDistance(o, object->Centroid());
            sorted_selection.insert(pair<RNScalar, int>(dist, i));

            CustomLoadRgb(object->PredictedLabel()->Color());
            DrawBlipAroundObject(object);
        }

        glDisable(GL_DEPTH_TEST);
        //for (int i = 0; i < NObjectSelections(); i++) {
        for (multimap<RNScalar, int>::reverse_iterator rit =
                sorted_selection.rbegin(); rit != sorted_selection.rend();
                ++rit) {
            R3SurfelObject *object = ObjectSelection(rit->second);

            // Draw background.
            glPointSize(3 * surfel_size);
            CustomLoadRgb(object_glow_color);
            DrawObject(object, 0);

            // Draw object.
            glPointSize(2 * surfel_size);
            if (incorrect_objects_.FindEntry(object)) {
                CustomLoadRgb(incorrect_color);
                DrawArrowAboveObject(object, incorrect_color);
            } else {
                CustomLoadRgb(RNyellow_rgb);
                DrawArrowAboveObject(object, RNyellow_rgb, 0.1);
            }
            glColor4f(RNyellow_rgb.R(), RNyellow_rgb.G(), RNyellow_rgb.B(),
                    0.5);
            DrawObject(object, 0);
        }
        glPointSize(surfel_size);
        glEnable(GL_DEPTH_TEST);
    }
}

RNScalar R3SurfelLabeler::AffinityScore(R3SurfelObject *from,
        R3SurfelObject *to) {
    // Check everything.
    if (!from || !to) return 0.0;

    RNScalar affinity = 0.0;

    switch (SurfelColorScheme()) {
        case R3_SURFEL_VIEWER_COLOR_BY_SHAPE_AFFINITY:
            affinity = classifier.GetShapeAffinity(from, to);
            break;
        case R3_SURFEL_VIEWER_COLOR_BY_FEATURE_DISTANCE:
            affinity = 1.0 / (1.0 + classifier.GetFeatureDistance(from, to));
            break;
        case R3_SURFEL_VIEWER_COLOR_BY_INDUCED_AFFINITY:
            affinity = classifier.GetInducedAffinity(from, to);
            break;
        case R3_SURFEL_VIEWER_COLOR_BY_SPATIAL_DISTANCE:
            affinity = classifier.GetScreenSizeComfortScore(from, to);
            break;
        default:
            break;
    }

    return affinity;
}

void R3SurfelLabeler::DrawHighlightGroup() const {
    glDisable(GL_DEPTH_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glColor4f(0.0, 1.0, 0.0, 0.3);  //  dim green
    for (int i = 0; i < highlight_group_.NEntries(); ++i) {
        R3SurfelObject *o = highlight_group_.Kth(i);
        if (!o) continue;
        const R3Box &box = o->BBox();
        RNScalar radius = max(box.XRadius(), max(box.YRadius(), box.ZRadius()));
        R3Cylinder(box.Centroid(), box.Centroid() + 16 * R3posz_vector, radius).Draw();
    }
    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
}

void R3SurfelLabeler::DrawMessage(void) const {
    // Check message
    if (!message_visibility) return;
    if (!message) return;

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);
    DrawHaloedMessage(R2Point(250, 16), message, GLUT_BITMAP_HELVETICA_18);
}

void R3SurfelLabeler::DrawStatus(void) const {
    if (!status_visibility) return;

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);

    // Count assignments
    int ntotal = 0;
    int npredicted = 0;
    int nconfirmed = 0;
    int nunlabeled = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        if (object->Parent() != scene->RootObject()) continue;
        if (classifier.Labeled(object)) nconfirmed++;
        else if (object->PredictedLabel()) npredicted++;
        else nunlabeled++;
        ntotal++;
    }

    // Draw status
    glColor3f(1, 1, 1);
    char buffer[1024];

    // sprintf(buffer,"SurfelSize = %.1f, Resolution = %.1f, FocusRadius = %.1f,  Complexity = %.0f, Frame rate = %.1f",
    //   SurfelSize(), TargetResolution(), FocusRadius(), objects.Complexity(), FrameRate());
    //sprintf(buffer, "%d confirmed (%.1f%%).   %d predicted (%.1f%%).    %d unlabeled (%.1f%%)",
    //        nconfirmed, 100.0 * nconfirmed / ntotal, npredicted, 100.0 * npredicted / ntotal, nunlabeled, 100.0 * nunlabeled / ntotal);

    sprintf(buffer, "%d confirmed (%.1f%%).   %d unconfirmed (%.1f%%)",
            nconfirmed, 100.0 * nconfirmed / ntotal, npredicted + nunlabeled,
            100.0 * (npredicted + nunlabeled) / ntotal);
    R2Point p(240, window_height - 32);
    DrawHaloedMessage(p, buffer, GLUT_BITMAP_HELVETICA_18);

    if (interface_mode == IGRA_INTERFACE_FULL) {
        sprintf(buffer, "Human:     p=%5.3f, r=%5.3f, f=%5.3f",
                HumanLabelPrecision(), HumanLabelRecall(),
                HumanLabelFMeasure());
        p.Reset(240, window_height - 64);
        DrawHaloedMessage(p, buffer, GLUT_BITMAP_HELVETICA_18);

        sprintf(buffer, "Predicted: p=%5.3f, r=%5.3f, f=%5.3f",
                PredictedLabelPrecision(), PredictedLabelRecall(),
                PredictedLabelFMeasure());
        p.Reset(240, window_height - 96);
        DrawHaloedMessage(p, buffer, GLUT_BITMAP_HELVETICA_18);
    }
}

void R3SurfelLabeler::DrawFPS(void) {
    if (!fps_visibility_) return;

    ++fps_counter_;

    if (last_fps_time_read_.Elapsed() > 1) {
        fps_ = fps_counter_ / last_fps_time_read_.Elapsed();
        sprintf(fps_message_, "FPS: %.0f", fps_);
        last_fps_time_read_.Read();
        fps_counter_ = 0;
    }

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);
    DrawHaloedMessage(R2Point(1, 1), fps_message_, GLUT_BITMAP_HELVETICA_12);
}

void R3SurfelLabeler::DrawCommandMenu(void) {
    // Note: this must match PickCommandMenu().

    // Only draw menu if it is visible
    if (!command_menu_visibility) return;

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);

    command_menu_box_ = R2null_box;

    // Get convenient variables
    int button_width = 164;
    int button_height = 26;
    int padding = 8;
    int half_padding = padding / 2;

    int x = window_width - button_width - 3 * padding;
    int y = window_height - padding;

    const int initial_y = y;

    RNRgb fill_color(0.0, 0.2, 0.0);
    RNRgb fg_color = RNwhite_rgb;
    RNRgb inactive_color(0.5, 0.5, 0.5);

    // Draw "Contract/Expand" commands
    if (interface_mode == IGRA_INTERFACE_FULL
            || interface_mode == IGRA_INTERFACE_ACTIVE_GROUP
            || interface_mode == IGRA_INTERFACE_TILED) {

        y -= half_padding;
        R2Box box(x + padding, y - button_height - half_padding,
                x + padding + (button_width - half_padding) / 2,
                y - half_padding);
        CustomLoadRgb(fill_color);
        box.Draw();
        if (NObjectSelections() > 1) CustomLoadRgb(fg_color);
        else CustomLoadRgb(inactive_color);
        box.Outline();
        DrawText(R2Point(x + padding + half_padding, y - button_height),
                "Contract", GLUT_BITMAP_HELVETICA_18);

        box = R2Box(
                x + padding + (button_width - half_padding) / 2 + half_padding,
                y - button_height - half_padding, x + button_width + padding,
                y - half_padding);
        CustomLoadRgb(fill_color);
        box.Draw();
        CustomLoadRgb(fg_color);
        box.Outline();
        DrawText(
                R2Point(
                        x + padding + (button_width - half_padding) / 2
                                + padding, y - button_height), " Expand",
                GLUT_BITMAP_HELVETICA_18);

        y -= button_height + padding;
    }

    // Draw box around commands
    if (y != initial_y) {
        y -= half_padding;
        command_menu_box_ = R2Box(x, y, x + button_width + 2 * padding,
                initial_y);

        CustomLoadRgb(fg_color);
        command_menu_box_.Outline();

        command_menu_box_.Union(R2Point(window_width, window_height));
    }
}

int R3SurfelLabeler::PickCommandMenu(int xcursor, int ycursor, int button,
        int state, int shift, int ctrl, int alt) {
    // Note: this must match DrawCommandMenu().

    // Only pick from menu if it is visible
    if (!command_menu_visibility) return 0;

    // Only process left-button down clicks
    if (button != 0) return 0;
    if (state != 0) return 0;

    // Get convenient variables
    int button_width = 164;
    int button_height = 26;
    int padding = 8;
    int half_padding = padding / 2;

    int x = window_width - button_width - 3 * padding;
    int y = window_height - padding;
    y -= half_padding;

    // Draw "Contract/Expand" commands
    if (interface_mode == IGRA_INTERFACE_FULL
            || interface_mode == IGRA_INTERFACE_ACTIVE_GROUP
            || interface_mode == IGRA_INTERFACE_TILED) {

        // Contract request.
        R2Box box(x + padding, y - button_height - half_padding,
                x + padding + (button_width - half_padding) / 2,
                y - half_padding);
        if (NObjectSelections() > 1
                && R2Contains(box, R2Point(xcursor, ycursor))) {
            // Process suggest an object
            SelectionContractRequest();
            return 1;
        }

        // Expand request.
        box = R2Box(
                x + padding + (button_width - half_padding) / 2 + half_padding,
                y - button_height - half_padding, x + button_width + padding,
                y - half_padding);
        if (R2Contains(box, R2Point(xcursor, ycursor))) {
            // Process suggest an object
            SelectionExpandRequest();
            return 1;
        }

        y -= button_height + padding;
    }

    // No command picked
    return 0;
}

void R3SurfelLabeler::DrawLabelMenu(const R2Point *ul_corner) {
    // Note: this must match PickLabelMenu
    if (selection_changed_) UpdateLabelMenuSequence(selection_objects);

    // Only draw menu if it is visible
    if (!label_menu_visibility) return;

    // Get convenient variables
    void *font_small = GLUT_BITMAP_HELVETICA_12;
    void *font_large = GLUT_BITMAP_HELVETICA_18;

    const bool visibility_control = (interface_mode == IGRA_INTERFACE_FULL
            || interface_mode == IGRA_INTERFACE_UG_GROUP);
    const int visibility_box_shift = visibility_control ? 20 : 0;

    const bool counter_visibility = false;
    const int counter_box_shift = counter_visibility ? 40 : 0;

    // Reset the label menu box.
    if (ul_corner) {
        // Place at the requested corner.
        label_menu_box_ = R2Box(*ul_corner, *ul_corner);
    } else {
        // Place in the default location: top left corner.
        label_menu_box_ = R2Box(0, window_height, 0, window_height);
    }

    // Menu entry templates.
    int x = label_menu_box_[0][0] + kMenuOuterPadding + kMenuInnerPadding;
    int y = label_menu_box_[1][1] - kMenuOuterPadding - kMenuInnerPadding;

    R2Box visibility_box(x, y - kMenuLineHeight + 2 * kMenuSmallPadding,
            x + visibility_box_shift - kMenuSmallPadding,
            y - kMenuSmallPadding);
    R2Point visibility_origin(visibility_box[0][0] + kMenuTinyPadding,
            visibility_box[0][1] + kMenuResPadding);
    x += visibility_box_shift;

    R2Box count_box(x, y - kMenuLineHeight + 2 * kMenuSmallPadding,
            x + visibility_box_shift - kMenuSmallPadding,
            y - kMenuSmallPadding);
    R2Point count_origin(count_box[0][0] + kMenuTinyPadding,
            count_box[0][1] + kMenuResPadding);
    x += counter_box_shift;

    R2Box name_box(x, y - kMenuLineHeight + kMenuSmallPadding, x + 180, y);
    R2Point name_origin(name_box[0][0] + kMenuTinyPadding,
            name_box[0][1] + kMenuSmallPadding + kMenuResPadding);
    x += 180;

    char buffer[1024];

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);
    glLineWidth(1);

    // Draw "All" visibility box and "Label" header
    if (visibility_control) {
        // Check if all labels are visible
        int all_visible = 1;
        for (int i = 0; i < scene->NLabels(); i++) {
            R3SurfelLabel *label = scene->Label(i);
            if (!LabelVisible(label)) {
                all_visible = 0;
                break;
            }
        }

        // Draw visibility box.
        if (all_visible) glColor4f(1, 1, 1, 0.5);
        else glColor4f(0, 0, 0, 0.5);
        visibility_box.Draw();
        glColor4f(1, 1, 1, 1);
        visibility_box.Outline();
        DrawText(visibility_origin, "V", font_small);

        // Draw the "All labels"
        glColor4f(0.1, 0.1, 0.1, 0.7);
        name_box.Draw();
        glColor4f(1, 1, 1, 1);
        DrawText(name_origin, "All Labels", font_large);

        // Advance templates to the next line.
        visibility_box.Translate(kMenuLineTranslation);
        visibility_origin.Translate(kMenuLineTranslation);
        count_box.Translate(kMenuLineTranslation);
        count_origin.Translate(kMenuLineTranslation);
        name_box.Translate(kMenuLineTranslation);
        name_origin.Translate(kMenuLineTranslation);
        y -= kMenuLineHeight;
    }

    // Draw labels
    bool use_custom_seq = (label_menu_sequence_.NEntries() > 0);
    const int n_labels =
            use_custom_seq ? label_menu_sequence_.NEntries() : scene->NLabels();
    for (int i = 0; i < n_labels; i++) {
        R3SurfelLabel *label =
                use_custom_seq ? label_menu_sequence_.Kth(i) : scene->Label(i);

        const R2Vector *translation_vector = &kMenuLineTranslation;

        if (label) {
            if (label->NParts() > 0) continue;
            RNRgb color = label->Color();

            // Draw visibility box
            if (visibility_control) {
                if (LabelVisible(label)) glColor4f(1, 1, 1, 0.5);
                else glColor4f(0, 0, 0, 0.5);
                visibility_box.Draw();
                CustomLoadRgb(color);
                visibility_box.Outline();
                CustomLoadRgb(color + 0.5 * RNwhite_rgb);
                DrawText(visibility_origin, "V", font_small);
            }

            // Draw label assignment count
            if (counter_visibility) {
                // Compute label assignment count
                int count = 0;
                for (int i = 0; i < label->NLabelAssignments(); i++) {
                    R3SurfelLabelAssignment *assignment =
                            label->LabelAssignment(i);
                    if (assignment->Originator()
                            == R3_SURFEL_LABEL_ASSIGNMENT_GROUND_TRUTH_ORIGINATOR)
                        continue;
                    count++;
                }

                CustomLoadRgb(color);
                sprintf(buffer, "%d", count);
                DrawText(count_origin, buffer, font_small);
            }

            // Identify if label is highlighted.
            R3SurfelLabel *prev_label = NULL;
            if (i > 0) {
                prev_label =
                        use_custom_seq ?
                                label_menu_sequence_.Kth(i - 1) :
                                scene->Label(i - 1);
            }
            R3SurfelLabel *next_label = NULL;
            int next_ind = i + 1;
            int max_ind = (
                    use_custom_seq ?
                            label_menu_sequence_.NEntries() : scene->NLabels())
                    - 1;
            if (next_ind <= max_ind) {
                next_label =
                        use_custom_seq ?
                                label_menu_sequence_.Kth(next_ind) :
                                scene->Label(next_ind);
            }
            bool smaller_font = (next_label != NULL || prev_label != NULL);

            // Draw label name
            glColor4f(0.1, 0.1, 0.1, 0.7);
            name_box.Draw();
            CustomLoadRgb(color);
            name_box.Outline();
            int key = label->AssignmentKeystroke();
            if (isalpha(key)) sprintf(buffer, " %s (%c)", label->Name(), key);
            else sprintf(buffer, " %s", label->Name());
            DrawText(name_origin, buffer,
                    smaller_font ? font_small : font_large);
        } else {
            translation_vector = &kMenuGapTranslation;
        }

        // Advance templates to the next line.
        visibility_box.Translate(*translation_vector);
        visibility_origin.Translate(*translation_vector);
        count_box.Translate(*translation_vector);
        count_origin.Translate(*translation_vector);
        name_box.Translate(*translation_vector);
        name_origin.Translate(*translation_vector);
        y -= translation_vector->Length();
    }

    // Update the label menu occupation box.
    R2Point rb_corner(x + kMenuInnerPadding + kMenuOuterPadding,
            y - kMenuInnerPadding - kMenuOuterPadding);
    label_menu_box_.Union(rb_corner);

    // Draw box around labels
    glColor4f(1, 0, 0, 1);
    R2Box(label_menu_box_[0][0] + kMenuOuterPadding,
            label_menu_box_[0][1] + kMenuOuterPadding,
            label_menu_box_[1][0] - kMenuOuterPadding,
            label_menu_box_[1][1] - kMenuOuterPadding).Outline();
}

void R3SurfelLabeler::DrawSelectionZoomBoxes(void) {
    // Drawing boxes is off.
    if (interface_mode != IGRA_INTERFACE_TILED) return;
    if (!selection_zoom_box_.NEntries()) return;

    // Get convenient variables
    const int N = selection_zoom_box_.NEntries();
    const R2Box boxes_zone(
            label_menu_box_.IsEmpty() ? 0.0 : label_menu_box_.XMax(),
            (message_visibility && message && strlen(message)) ? 50 : 0.0,
            window_width,
            command_menu_box_.IsEmpty() ?
                    window_height - 50 : command_menu_box_.YMin());

    // Identify the packing size.
    int default_diameter = sqrt(boxes_zone.Area() / N);
    if (default_diameter < 5) return;
    int x_size = floor(boxes_zone.XLength() / default_diameter);
    int y_size = floor(boxes_zone.YLength() / default_diameter);
    while (x_size * y_size < N) {
        --default_diameter;
        x_size = floor(boxes_zone.XLength() / default_diameter);
        y_size = floor(boxes_zone.YLength() / default_diameter);
        if (default_diameter <= 5) break;
    }

    int x = boxes_zone.XMin();
    int y = boxes_zone.YMax();

    // Draw the selection zoom boxes.
    R2Box screen_box(0, 0, window_width, window_height);
    for (int i = 0; i < N; ++i) {
        SelectionZoomBox *box = selection_zoom_box_.Kth(i);

        int curr_x_step = i % x_size;
        int curr_y_step = i / x_size;

        int curr_x = x + curr_x_step * default_diameter;
        int curr_y = y - curr_y_step * default_diameter;
        int curr_diam = default_diameter;

        // Draw zoom in box.
        box->Draw(curr_x, curr_y, curr_diam, screen_box,
                incorrect_objects_.FindEntry(box->Object()));

    }
}

void R3SurfelLabeler::DrawBlipAroundObject(const R3SurfelObject *object) {
    // Check everything.
    if (!object) return;

    const R3Box &bbox = object->BBox();
    R3Point center = bbox.Centroid() - bbox.ZRadius() * R3negz_vector;
    R3Circle circle(center, bbox.DiagonalLength(), R3posz_vector);

    circle.Draw();
}

void R3SurfelLabeler::DrawArrowAboveObject(const R3SurfelObject *object,
        const RNRgb &color, RNScalar alpha) {
    // Check everything.
    if (!object) return;

    // Useful variables.
    const R3Box &o_box = object->BBox();
    R3Point centroid = o_box.Centroid();
    const RNScalar object_height = o_box.ZLength();
    const RNScalar z = o_box.ZMax() + 0.5 * object_height + 1;
    const RNScalar z_extent = R3Distance(Camera().Origin(), centroid);
    const RNScalar x = centroid.X();
    const RNScalar y = centroid.Y();

    const RNScalar tip_h = max(0.25 * object_height, 1.0);
    const RNScalar tip_r = 0.25 * tip_h;

    // Set GL modes.
    GLboolean lighting, blending;
    glGetBooleanv(GL_LIGHTING, &lighting);
    glGetBooleanv(GL_BLEND, &blending);
    glDisable(GL_LIGHTING);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslatef(x, y, z);
    glRotatef(180.0, 1.0, 0.0, 0.0);

    // Tip component.
    glColor4f(color.R(), color.G(), color.B(), alpha);
    glutSolidCone(tip_r, tip_h, 16, 1);

    // Line component.
    glLineWidth(1);
    glBegin(GL_LINES);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, -z_extent);
    glEnd();

    // Restore GL state.
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    if (lighting) glEnable(GL_LIGHTING);
    if (blending) glEnable(GL_BLEND);
}

void R3SurfelLabeler::DrawConeAroundObject(const R3SurfelObject *object,
        const RNRgb &color) {
    // Check everything.
    if (!object) return;

    // Useful variables.
    const R3Box &o_box = object->BBox();
    R3Point centroid = o_box.Centroid();
    const RNScalar object_height = o_box.ZLength();
    const RNScalar z = o_box.ZMax() + 0.5 * object_height;
    const RNScalar z_extent = R3Distance(Camera().Origin(), centroid);
    const RNScalar x = centroid.X();
    const RNScalar y = centroid.Y();

    const RNScalar r = 2.0;

    // Set GL modes.
    GLboolean lighting, blending;
    glGetBooleanv(GL_LIGHTING, &lighting);
    glGetBooleanv(GL_BLEND, &blending);
    glDisable(GL_LIGHTING);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslatef(x, y, z);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

    // Cylinder.
    glColor4f(color.R(), color.G(), color.B(), 0.4);
    GLUquadric *glu_q = gluNewQuadric();
    gluQuadricOrientation(glu_q, GLU_OUTSIDE);
    gluCylinder(glu_q, r, r, z_extent, 16, 1);

    // Restore GL state.
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    if (lighting) glEnable(GL_LIGHTING);
    if (blending) glEnable(GL_BLEND);
}

int R3SurfelLabeler::InSelectionZoomBox(int xcursor, int ycursor, int button,
        int state, int shift, int ctrl, int alt) {
    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *box = selection_zoom_box_.Kth(i);
        if (!box) continue;
        if (R2Contains(box->BBox(), R2Point(xcursor, ycursor))) return 1;
    }
    return 0;
}

void R3SurfelLabeler::DrawPausedScreen(void) const {
    // Only draw screen if paused.
    if (!paused_) return;

    DrawDimLayer(&RNblack_rgb, 0.7);

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);

    void *font = GLUT_BITMAP_HELVETICA_18;
    const char* msg = "P A U S E D";
    int width = glutBitmapLength(font, (const unsigned char*) msg);
    int x = (window_width - width) / 2;
    int y = window_height / 2 - 8;
    R2Box pause_box(x - 10, y - 10, x + width + 10, y + 16 + 10);

    // Draw "PAUSED" message
    CustomLoadRgb(RNblack_rgb);
    pause_box.Draw();
    CustomLoadRgb(RNgreen_rgb);
    pause_box.Outline();
    DrawText(R2Point(x, y), msg, font);
}

void R3SurfelLabeler::DrawDimLayer(const RNRgb *color, RNScalar alpha) const {
    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);

    // Dim the screen
    if (color) glColor4f(color->R(), color->G(), color->B(), alpha);
    else glColor4f(background_color.R(), background_color.G(),
            background_color.B(), alpha);
    R2Box(0, 0, window_width, window_height).Draw(R2_EVERYTHING_DRAW_FLAGS);
}

void R3SurfelLabeler::DrawLabeling(void) const {
    if (interface_mode == IGRA_INTERFACE_TILED) return;

    // Set rendering modes.
    Ortho2DLocker locker(window_width, window_height);

    if (object_label_visibility) {
        for (int i = 0; i < scene->NObjects(); i++) {
            R3SurfelObject *object = scene->Object(i);
            if (object->NParts() > 0) continue;

            RNBoolean confirmed = classifier.Labeled(object);
            if (!object_label_confirmed_visibility && confirmed) continue;

            R3SurfelLabel *label = object->PredictedLabel();
            if (!label) continue;
            if (!strcmp(label->Name(), "Unknown")) continue;
            if (!LabelVisible(label)) continue;
            R3Point position = object->Centroid();
            position[2] = object->BBox().ZMax() + 1;
            R2Point p = viewer.ViewportPoint(position);
            void *font = (confirmed) ?
            GLUT_BITMAP_HELVETICA_18 :
                                       GLUT_BITMAP_HELVETICA_12;
            int width = glutBitmapLength(font,
                    (const unsigned char *) label->Name());
            p[0] -= width / 2;
            CustomLoadRgb(label->Color());
            if (selection_objects.FindEntry(object)) {
                if (incorrect_objects_.FindEntry(object)) CustomLoadRgb(
                        incorrect_color);
                else CustomLoadRgb(RNyellow_rgb);
            }
            DrawText(p, label->Name(), font);
        }
    } else if (1 && NObjectSelections()) {
        set<R3SurfelLabel*> predictions;
        set<R3SurfelLabel*> incorrect_pred;
        void *font = GLUT_BITMAP_HELVETICA_18;
        for (int i = 0; i < NObjectSelections(); i++) {
            R3SurfelObject *object = ObjectSelection(i);
            R3SurfelLabel *label = object->PredictedLabel();
            if (incorrect_objects_.FindEntry(object)) incorrect_pred.insert(
                    label);
            else predictions.insert(label);
        }

        // Form label content.
        string label_q;
        char buffer[1024];
        if (incorrect_pred.size()) {
            label_q += "Incorrect!";
            if (incorrect_label_) {
                label_q += " not ";
                label_q += incorrect_label_->Name();
            }
        } else {
            for (set<R3SurfelLabel*>::iterator it = predictions.begin();
                    it != predictions.end(); ++it) {
                sprintf(buffer, "%s(%c) ? ", (*it)->Name(),
                        (*it)->AssignmentKeystroke());
                label_q += buffer;
            }
        }

        // Set drawing location.
        R2Point p(0.5 * window_width, 0.8 * window_height);
        int width = glutBitmapLength(font,
                (const unsigned char *) label_q.c_str());
        p[0] -= width / 2;

        // Draw.
        DrawHaloedMessage(p, label_q.c_str(), font,
                incorrect_pred.size() ? &incorrect_color : &RNyellow_rgb);

    } else if (0) {
        // Draw confirmed labels only.
        for (int i = 0; i < NObjectSelections(); i++) {
            R3SurfelObject *object = ObjectSelection(i);
            if (classifier.Labeled(object)) {
                R3SurfelLabel *label = object->PredictedLabel();
                R3Point position = object->Centroid();
                position[2] = object->BBox().ZMax() + 1;
                R2Point p = viewer.ViewportPoint(position);
                void *font = GLUT_BITMAP_HELVETICA_18;
                int width = glutBitmapLength(font,
                        (const unsigned char *) label->Name());
                p[0] -= width / 2;
                if (incorrect_objects_.FindEntry(object)) CustomLoadRgb(
                        incorrect_color);
                else CustomLoadRgb(RNyellow_rgb);
                DrawText(p, label->Name(), font);
                break;
            }
        }
    }
}

void R3SurfelLabeler::DrawObjectNames(void) const {
    if (!object_name_visibility) return;
    if (interface_mode == IGRA_INTERFACE_TILED) return;

    // Set viewing transformation
    viewer.Camera().Load();

    // Draw object names
    CustomLoadRgb(object_name_color);
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        if (object->NParts() > 0) continue;
        if (!object->Name()) continue;
        R3Point position = object->Centroid();
        position[2] = object->BBox().ZMax() + 1;
        DrawText(position, object->Name(), GLUT_BITMAP_HELVETICA_12);
    }
}

void R3SurfelLabeler::DrawNodes(void) const {
    // Draw node bounding boxes
    if (!node_bbox_visibility) return;
    if (interface_mode == IGRA_INTERFACE_TILED) return;

    // Set viewing transformation
    viewer.Camera().Load();

    CustomLoadRgb(node_bbox_color);
    for (int i = 0; i < resident_nodes.NNodes(); i++) {
        R3SurfelNode *node = resident_nodes.Node(i);
        node->BBox().Outline();
        if (node->NParts() > 0) glColor3f(0, 1, 0);
        else glColor3f(1, 0, 0);
    }
}

void R3SurfelLabeler::DrawBlocks(void) const {
    // Draw block bounding boxes
    if (!block_bbox_visibility) return;
    if (interface_mode == IGRA_INTERFACE_TILED) return;

    // Set viewing transformation
    viewer.Camera().Load();

    CustomLoadRgb(block_bbox_color);
    for (int i = 0; i < resident_nodes.NNodes(); i++) {
        R3SurfelNode *node = resident_nodes.Node(i);
        for (int j = 0; j < node->NBlocks(); j++) {
            R3SurfelBlock *block = node->Block(j);
            block->BBox().Outline();
        }
    }
}

void R3SurfelLabeler::DrawCenterPoint(void) const {
    // Draw center point
    if (!center_point_visibility) return;
    if (interface_mode == IGRA_INTERFACE_TILED) return;

    // Set viewing transformation
    viewer.Camera().Load();

    glEnable(GL_LIGHTING);
    GLfloat color[4] = { float(center_point_color[0]), float(
            center_point_color[1]), float(center_point_color[2]), 1.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
    R3Sphere(center_point, 1).Draw();
    glDisable(GL_LIGHTING);
}

void R3SurfelLabeler::UpdateLabelMenuSequence(
        RNArray<R3SurfelObject*> object_set) {
    classifier.UpdateLabelMenuSequence(object_set, label_menu_sequence_);
}

int R3SurfelLabeler::LabelVisible(R3SurfelLabel *label) const {
    map<R3SurfelLabel*, LabelViewType>::const_iterator it = label_view_.find(
            label);

    if (it != label_view_.end()) return it->second.visibility;

    // Return error.
    return 0;
}

int R3SurfelLabeler::LabelSelectible(R3SurfelLabel *label) const {
    map<R3SurfelLabel*, LabelViewType>::const_iterator it = label_view_.find(
            label);

    if (it != label_view_.end()) return it->second.selectibility;

    // Return error.
    return 0;
}

int R3SurfelLabeler::PickLabelMenu(int xcursor, int ycursor, int button,
        int state, int shift, int ctrl, int alt) {
    // Note: this must match DrawLabelMenu

    // Only pick from menu if it is visible
    if (!label_menu_visibility) return 0;

    // Only process left-button down clicks
    if (button != 0) return 0;
    if (state != 0) return 0;

    // Get convenient variables
    R2Point cursor(xcursor, ycursor);
    const bool visibility_control = (interface_mode == IGRA_INTERFACE_FULL
            || interface_mode == IGRA_INTERFACE_UG_GROUP);
    const int visibility_box_shift = visibility_control ? 20 : 0;

    const bool counter_visibility = false;
    const int counter_box_shift = counter_visibility ? 40 : 0;

    // Menu entry templates.
    int x = label_menu_box_[0][0] + kMenuOuterPadding + kMenuInnerPadding;
    int y = label_menu_box_[1][1] - kMenuOuterPadding - kMenuInnerPadding;

    R2Box visibility_box(x, y - kMenuLineHeight + 2 * kMenuSmallPadding,
            x + visibility_box_shift - kMenuSmallPadding,
            y - kMenuSmallPadding);
    R2Point visibility_origin(visibility_box[0][0] + kMenuTinyPadding,
            visibility_box[0][1] + kMenuResPadding);
    x += visibility_box_shift;

    R2Box count_box(x, y - kMenuLineHeight + 2 * kMenuSmallPadding,
            x + visibility_box_shift - kMenuSmallPadding,
            y - kMenuSmallPadding);
    R2Point count_origin(count_box[0][0] + kMenuTinyPadding,
            count_box[0][1] + kMenuResPadding);
    x += counter_box_shift;

    R2Box name_box(x, y - kMenuLineHeight + kMenuSmallPadding, x + 180, y);
    R2Point name_origin(name_box[0][0] + kMenuTinyPadding,
            name_box[0][1] + kMenuSmallPadding + kMenuResPadding);
    x += 180;

    if (visibility_control) {
        // Check if "All" visibility box is pressed
        if (R2Contains(visibility_box, cursor)) {
            // Turn on all
            for (int j = 0; j < scene->NLabels(); j++) {
                R3SurfelLabel *label = scene->Label(j);
                label_view_[label].reset();
            }
            classifier.RequestRedraw();
            return 1;
        }

        // Advance templates to the next line.
        visibility_box.Translate(kMenuLineTranslation);
        visibility_origin.Translate(kMenuLineTranslation);
        count_box.Translate(kMenuLineTranslation);
        count_origin.Translate(kMenuLineTranslation);
        name_box.Translate(kMenuLineTranslation);
        name_origin.Translate(kMenuLineTranslation);
        y -= kMenuLineHeight;
    }

#if 0
    {
        // Confirmation button pressed.
        R2Box name_box(x, y, x + 200, y + 30);
        if (R2Contains(name_box, cursor)) {
            ConfirmLabelsOnSelectedObjects();
            return 1;
        }

        // Advance templates to the next line.
        visibility_box.Translate(kMenuLineTranslation);
        visibility_origin.Translate(kMenuLineTranslation);
        count_box.Translate(kMenuLineTranslation);
        count_origin.Translate(kMenuLineTranslation);
        name_box.Translate(kMenuLineTranslation);
        name_origin.Translate(kMenuLineTranslation);
        y -= kMenuLineHeight;
    }
#endif

    // Draw labels
    bool use_custom_seq = (label_menu_sequence_.NEntries() > 0);
    const int n_labels =
            use_custom_seq ? label_menu_sequence_.NEntries() : scene->NLabels();
    for (int i = 0; i < n_labels; i++) {
        R3SurfelLabel *label =
                use_custom_seq ? label_menu_sequence_.Kth(i) : scene->Label(i);

        const R2Vector *translation_vector = &kMenuLineTranslation;

        if (label) {
            if (label->NParts() > 0) continue;

            // Process command
            if (visibility_control && R2Contains(visibility_box, cursor)) {

                BeginCommand(R3_SURFEL_LABELER_SET_LABEL_VISIBILITY_COMMAND,
                        (ctrl) ? 1 : 0);

                if (ctrl) {
                    // Toggle all except selected label
                    for (int j = 0; j < scene->NLabels(); j++) {
                        R3SurfelLabel *curr_label = scene->Label(j);
                        label_view_[curr_label].reset(
                                curr_label == label ? 1 : 0);
                    }
                } else {
                    // Toggle selected label
                    LabelViewType &lv = label_view_[label];
                    lv.reset(1 - lv.visibility);
                }

                EndCommand();

                classifier.RequestRedraw();
                return 1;
            } else if (R2Contains(name_box, cursor)) {
                AssignLabelToSelectedObjects(label);

                return 1;
            }
        } else {
            translation_vector = &kMenuGapTranslation;
        }

        // Advance templates to the next line.
        visibility_box.Translate(*translation_vector);
        visibility_origin.Translate(*translation_vector);
        count_box.Translate(*translation_vector);
        count_origin.Translate(*translation_vector);
        name_box.Translate(*translation_vector);
        name_origin.Translate(*translation_vector);
        y -= translation_vector->Length();
    }

    // No label picked
    return 0;
}

void R3SurfelLabeler::ToggleExclusiveVisibilityForSelected() {
    if (interface_mode != IGRA_INTERFACE_FULL
            && interface_mode != IGRA_INTERFACE_UG_GROUP) {
        return;
    }

    bool all_visible = true;
    for (int i = 0; i < scene->NLabels() && all_visible; i++) {
        R3SurfelLabel *label = scene->Label(i);
        LabelViewType &lv = label_view_[label];
        all_visible &= (lv.visibility && lv.selectibility);
    }

    int operand = (NObjectSelections() > 0 && all_visible) ? 1 : 0;
    BeginCommand(R3_SURFEL_LABELER_SET_LABEL_VISIBILITY_COMMAND, operand);
    if (operand) {
        // Turn on exclusive visibility for the labels of the selected objects.
        for (int i = 0; i < scene->NLabels(); i++) {
            R3SurfelLabel *label = scene->Label(i);
            label_view_[label].reset(0);
        }
        for (int i = 0; i < NObjectSelections(); i++) {
            R3SurfelObject* obj = ObjectSelection(i);
            R3SurfelLabel *label = obj->PredictedLabel();
            if (label) label_view_[label].reset();
        }
    } else {
        for (int i = 0; i < scene->NLabels(); i++) {
            R3SurfelLabel *label = scene->Label(i);
            label_view_[label].reset();
        }
    }
    EndCommand();

    classifier.RequestRedraw();
}

void R3SurfelLabeler::InsertZoomBoxForObject(R3SurfelObject *object) {
    // Check everything.
    if (!object) return;
    if (interface_mode != IGRA_INTERFACE_TILED) return;

    // Check if already exists.
    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
        if (!szb) {
            fprintf(stderr,
                    "InsertZoomBoxForObject: there is a NULL selection box.");
            continue;
        }
        R3SurfelObject *curr_object = szb->Object();
        if (!curr_object) {
            fprintf(stderr,
                    "InsertZoomBoxForObject: selection box holds NULL object.");
        }
        if (object == curr_object) {
            return;
        }
    }

    // If not, insert a new box.
    SelectionZoomBox *szb = new SelectionZoomBox(Camera(), Viewport(), scene,
            &classifier, object);
    selection_zoom_box_.InsertTail(szb);
}

void R3SurfelLabeler::RemoveZoomBoxForObject(R3SurfelObject *object) {
    if (interface_mode != IGRA_INTERFACE_TILED) return;

    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
        if (!szb) {
            fprintf(stderr,
                    "RemoveZoomBoxForObject: there is a NULL selection box.");
            continue;
        }
        R3SurfelObject *curr_object = szb->Object();
        if (object == curr_object) {
            delete szb;
            selection_zoom_box_.RemoveKth(i);
            return;
        }
    }
}

void R3SurfelLabeler::EmptyZoomBoxes(void) {
    for (int i = 0; i < selection_zoom_box_.NEntries(); ++i) {
        SelectionZoomBox *szb = selection_zoom_box_.Kth(i);
        if (szb) delete szb;
    }
    selection_zoom_box_.Empty();
}

////////////////////////////////////////////////////////////////////////
// SELECTION ZOOM BOX IMPLEMENTATION
////////////////////////////////////////////////////////////////////////

SelectionZoomBox::SelectionZoomBox(const R3Camera &camera,
        const R2Viewport &viewport, R3SurfelScene *scene,
        R3SurfelClassifier *classifier, R3SurfelObject *object) :
                camera_(camera),
                object_(object),
                all_points_(NULL),
                object_points_(NULL),
                hovered_(false),
                bbox_(R2null_box),
                phi_(0.0),
                classifier_(classifier) {
    assert(scene);
    assert(object);
    assert(classifier);

    // Acquire neighborhood points.
    R3SurfelCylinderConstraint cylinder_constraint(object->Centroid(), 10);
    all_points_ = CreatePointSet(scene, NULL, &cylinder_constraint);

    // Acquire object points.
    object_points_ = object->PointSet();

    // Adjusted box.
    R3Box adjusted_box(object_->BBox());
    if (draw_green_man) {
        RNScalar floor = adjusted_box.ZMin();
        R3Box human_box(-0.5, -0.5, 0, 0.5, 0.5, 2);
        human_box.Translate(
                R3Vector(adjusted_box.Centroid()[0], adjusted_box.Centroid()[1],
                        floor));
        RNScalar shift = 0.5
                * max(adjusted_box.XLength(), adjusted_box.YLength()) + 0.5;
        human_box.Translate(shift * R3posx_vector);
        adjusted_box.Union(human_box);
        human_box.Translate(2 * shift * R3negx_vector);
        adjusted_box.Union(human_box);
        human_box.Translate(shift * (R3posx_vector + R3posy_vector));
        adjusted_box.Union(human_box);
        human_box.Translate(2 * shift * R3negy_vector);
        adjusted_box.Union(human_box);
    }

    // Acquire the sphere of interest.
    R3Point center = object->Centroid();
    RNLength radius = 0;
    for (int i = 0; i < 8; ++i) {
        RNLength dist = R3Distance(center, adjusted_box.Corner(i));
        radius = max(radius, dist);
    }
    R3Sphere sphere(center, radius);

    // Select active area.
    R2Box active_area(viewport.BBox());

    // Select principal direction.
    R3Vector p_dir(object->BBox().Max() - object->BBox().Min());

    // Set camera.
    RNAngle fov = max(camera_.XFOV(), camera_.YFOV());
    camera_.SetXFOV(fov);
    camera_.SetYFOV(fov);
    R3SurfelLabeler::ZoomToROI(camera_, viewport, active_area, sphere, &p_dir);

    AdvanceAnimation(0);
}

SelectionZoomBox::~SelectionZoomBox() {
    if (all_points_) delete all_points_;
    if (object_points_) delete object_points_;
}

void SelectionZoomBox::Draw(int lt_x, int lt_y, int diameter,
        const R2Box &screen_box, bool incorrect) {
    // Update occupation box.
    bbox_ = R2Box(lt_x, lt_y - diameter, lt_x + diameter, lt_y);
    R2Box bbox_small(lt_x + 1, lt_y - diameter - 1, lt_x + diameter - 1,
            lt_y + 1);

    // Set rendering modes.
    Ortho2DLocker locker(screen_box.XLength(), screen_box.YLength());
    glScissor((GLint) bbox_small.XMin(), (GLint) bbox_small.YMin(),
            (GLsizei) bbox_small.XLength(), (GLsizei) bbox_small.YLength());
    glEnable(GL_SCISSOR_TEST);

    // Useful variables.
    const int padding = 8;
    R2Box render_box(lt_x + padding, lt_y - diameter + padding,
            lt_x + diameter - padding, lt_y - padding);
    RNRgb object_color = RNwhite_rgb;

    // Draw frame.
    if (object_) {
        // Identify drawing properties.
        int line_width = padding;
        if (!classifier_->Labeled(object_)) line_width *= 2;
        R3SurfelLabel *label = object_->PredictedLabel();
        if (!label) {
            line_width *= 2;
            object_color = RNred_rgb;
        } else {
            object_color = label->Color();
        }
        RNRgb rim_color = Hovered() ? RNyellow_rgb : RNblack_rgb;

        // Draw outer box with object color.
        if (incorrect) RNLoadRgb(incorrect_color);
        else glColor4f(object_color[0], object_color[1], object_color[2], 0.7);
        bbox_small.Draw();

        // Highlighting color in rim.
        glLineWidth(line_width);
        RNLoadRgb(rim_color);
        render_box.Outline();
    }

    if (all_points_ || object_points_) {
        // Setup projection of the close-up view.
        camera_.Load();
        glViewport((GLint) render_box.XMin(), (GLint) render_box.YMin(),
                (GLsizei) render_box.XLength(), (GLsizei) render_box.YLength());
        glScissor((GLint) render_box.XMin(), (GLint) render_box.YMin(),
                (GLsizei) render_box.XLength(), (GLsizei) render_box.YLength());

        // Redraw the box area.
        glEnable(GL_DEPTH_TEST);
        glDepthMask(TRUE);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Draw context.
        if (all_points_) {
            glPointSize(1);
            all_points_->Draw();
        }

        // Draw the object.
        if (object_points_) {
            glPointSize(2);
            if (incorrect) RNLoadRgb(incorrect_color);
            else RNLoadRgb(RNyellow_rgb);
            object_points_->Draw(0);

            if (draw_green_man) {
                RNLength unit = 0.25;
                RNLength ref_shift = 0.5
                        * max(object_->BBox().XLength(),
                                object_->BBox().YLength()) + 2 * unit;
                R3Vector right = camera_.Right();
                R3Vector up = R3posz_vector;
                R3Point floor_center = object_->Centroid() + ref_shift * right
                        - 0.5 * object_->BBox().ZLength() * up;

                glColor4d(RNgreen_rgb[0], RNgreen_rgb[1], RNgreen_rgb[2], 0.7);
                glLineWidth(2);
                glDisable(GL_DEPTH_TEST);

                // Head.
                glMatrixMode(GL_MODELVIEW);
                glPushMatrix();
                R3Point head = floor_center + 7.5 * unit * up;
                glTranslatef(head[0], head[1], head[2]);
                glutSolidSphere(0.5 * unit, 8, 8);
                glPopMatrix();

                glBegin(GL_LINES);
                // Trunk + neck.
                glVertex3f(floor_center[0], floor_center[1],
                        floor_center[2] + 4 * unit);
                glVertex3f(floor_center[0], floor_center[1],
                        floor_center[2] + 7 * unit);

                // Right hand and shoulder.
                R3Point hand = floor_center + 3 * unit * up + 2 * unit * right;
                R3Point shoulder = floor_center + 6.5 * unit * up
                        + unit * right;
                glVertex3f(hand[0], hand[1], hand[2]);
                glVertex3f(shoulder[0], shoulder[1], shoulder[2]);
                glVertex3f(shoulder[0], shoulder[1], shoulder[2]);

                // Left hand and shoulder.
                hand = floor_center + 3 * unit * up - 2 * unit * right;
                shoulder = floor_center + 6.5 * unit * up - unit * right;
                glVertex3f(shoulder[0], shoulder[1], shoulder[2]);
                glVertex3f(shoulder[0], shoulder[1], shoulder[2]);
                glVertex3f(hand[0], hand[1], hand[2]);

                // Right leg.
                R3Point foot = floor_center + unit * right;
                glVertex3f(floor_center[0], floor_center[1],
                        floor_center[2] + 4 * unit);
                glVertex3f(foot[0], foot[1], foot[2]);

                // Left leg.
                foot = floor_center - unit * right;
                glVertex3f(floor_center[0], floor_center[1],
                        floor_center[2] + 4 * unit);
                glVertex3f(foot[0], foot[1], foot[2]);

                glEnd();
            }

            glEnable(GL_DEPTH_TEST);
        }

        glViewport(0, 0, screen_box.XLength(), screen_box.YLength());
    }

    // Write best label above if labeled by human before.
    if (object_->HumanLabelAssignment()) {
        Ortho2DLocker locker(screen_box.XLength(), screen_box.YLength());

        R3SurfelLabel *label = object_->PredictedLabel();
        R3Point position = object_->Centroid();
        position[2] = object_->BBox().ZMax() + 1;

        // Transform 3D point into canonical coordinate system
        R3Point p = camera_.CoordSystem().InverseMatrix() * position;
        if (RNIsNegative(p.Z())) {

            // Compute 2D point projected onto render box area
            RNCoord x = render_box.XCenter()
                    + 0.5 * render_box.XLength() * p.X()
                            / -(p.Z() * tan(camera_.XFOV()));
            RNCoord y = render_box.YCenter()
                    + 0.5 * render_box.YLength() * p.Y()
                            / -(p.Z() * tan(camera_.YFOV()));

            void *font = GLUT_BITMAP_HELVETICA_18;
            int width = glutBitmapLength(font,
                    (const unsigned char *) label->Name());
            x -= width / 2;
            RNLoadRgb(RNyellow_rgb);
            DrawText(R2Point(x, y), label->Name(), font);
        }
    }

    // Reset OpenGL modes.
    glDisable(GL_SCISSOR_TEST);
}

void SelectionZoomBox::AdvanceAnimation(RNScalar dt) {
    R3Point target = object_->Centroid();
    RNScalar distance = (camera_.Origin() - target).Length();
    R3Vector twd = camera_.Towards();
    R3Vector up = camera_.Up();
    R3Vector right = camera_.Right();

    RNScalar gain = min(sqrt(20 / distance), 1.0);
    RNAngle d_phi = dt * kPhiAngSpeed * gain;
    phi_ += d_phi;
    RNAngle theta = kDTheta * sin(kThetaPhiCoeff * phi_) + kAvrTheta;

    if (theta < 0) {
        theta = 0;
    }
    if (theta > RN_PI_OVER_TWO) {
        theta = RN_PI_OVER_TWO;
    }

    twd.Rotate(R3posz_vector, d_phi);
    twd.Project(R3xy_plane);
    twd.Normalize();
    right = twd % R3posz_vector;
    twd.Rotate(right, -theta);
    twd.Normalize();
    R3Point eye = target - distance * twd;
    right = twd % R3posz_vector;
    up = right % twd;
    camera_.Reset(eye, twd, up);
}

} // namespace igra
