/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  R3SurfelLabeler.h
 *
 *  Copyright Thomas Funkhouser, Princeton University, 2011.
 *
 *  Description: Viewer for the labeling interface.
 *
 *  -------------------------------------------------------------------------*/

#ifndef R3SURFELLABELER_H_
#define R3SURFELLABELER_H_

#include "R3SurfelClassifier.h"
#include "R3SurfelViewer.h"

namespace igra {

// Artificial limitations.
#define R3_SURFEL_LABELER_NO_UNKNOWNS

enum InterfaceModes {
    IGRA_INTERFACE_FULL,
    IGRA_INTERFACE_UG_GROUP,
    IGRA_INTERFACE_UG_SINGLE,
    IGRA_INTERFACE_ACTIVE_SINGLE,
    IGRA_INTERFACE_ACTIVE_GROUP,
    IGRA_INTERFACE_IMAGE_NET,       // suggests groups, manual outlier removal
    IGRA_INTERFACE_BATCH_LEARNING,
    IGRA_INTERFACE_TILED,
    IGRA_INTERFACE_NUM
};

/* Copy-paste switch for the above skeleton
 switch (interface_mode) {
 case UGOR_INTERFACE_FULL:
 break;
 case UGOR_INTERFACE_UG_GROUP:
 break;
 case UGOR_INTERFACE_UG_SINGLE:
 break;
 case UGOR_INTERFACE_ACTIVE_SINGLE:
 break;
 case UGOR_INTERFACE_ACTIVE_GROUP:
 break;
 case UGOR_INTERFACE_IMAGE_NET:
 break;
 case IGRA_INTERFACE_BATCH_LEARNING:
 break;
 case UGOR_INTERFACE_TILED:
 break;
 default:
 break;
 }
 * */

#define R3_SURFEL_LABELER_SELECTION_FILTER_NONE         0x0000
#define R3_SURFEL_LABELER_SELECTION_FILTER_OVERLAP      0x0001
#define R3_SURFEL_LABELER_SELECTION_FILTER_INSIDE       0x0002
#define R3_SURFEL_LABELER_SELECTION_FILTER_OFFSCREEN    0x0004
#define R3_SURFEL_LABELER_SELECTION_FILTER_ALL          0xFFFF

////////////////////////////////////////////////////////////////////////
// FORWARD DECLARATION
////////////////////////////////////////////////////////////////////////

class SelectionZoomBox;

////////////////////////////////////////////////////////////////////////
// CLASS DEFINITION
////////////////////////////////////////////////////////////////////////

class R3SurfelLabeler: public R3SurfelViewer {
public:

    //////////////////////////////////////////
    //// CONSTRUCTOR/DESTRUCTOR FUNCTIONS ////
    //////////////////////////////////////////

    // Constructor/destructor functions
    R3SurfelLabeler(R3SurfelScene *scene, const char *loging_filename = NULL);
    virtual ~R3SurfelLabeler(void);

    //////////////////////
    //// UI CALLBACKS ////
    //////////////////////

    // Begin/end events
    void Initialize(int classifier_vis_style);
    void Terminate(void);

    // User input events
    int Redraw(void);
    int Resize(int width, int height);
    int MouseMotion(int x, int y);
    int PassiveMouseMotion(int x, int y);
    int MouseWheel(int button, int dir, int x, int y);
    int MouseButton(int x, int y, int button, int state, int shift, int ctrl,
            int alt);
    int Keyboard(int x, int y, int key, int shift, int ctrl, int alt);

    // Animation-related functions.
    int AnimationTimerID() const;
    int AdvanceAnimation();
    int AnimationFPS() const;

    //////////////////
    //// COMMANDS ////
    //////////////////

    // Save scene events
    int Sync(void);

    // Object selection
    int SelectPickedObject(int x, int y, int add = FALSE, int subtract = FALSE);
    int SelectEnclosedObjects(const R2Box& box, int add = FALSE, int subtract =
    FALSE);
    int SelectionExpandRequest();
    int SelectionContractRequest();

    // Create label assignments
    int AssignLabelToPickedObject(R3SurfelLabel *label, int x, int y);
    int AssignLabelToSelectedObjects(R3SurfelLabel *label);

    // Confirm label assignments
    int ConfirmLabelOnPickedObject(int x, int y);
    int ConfirmLabelsOnSelectedObjects(void);
    int ConfirmLabelsOnAllObjects(void);

    // Other commands
    int Undo(void);
    int Redo(void);
    int Reset(void);

    //Batch learning mode commands
    int PredictAll(void);

    // Logging functions
    int StartLogging(const char *filename) const;
    int EndLogging(void) const;
    int IsLogging(void) const;
    int ReadLog(const char *filename);

    ////////////////////////
    //// PROPERTY QUERY ////
    ////////////////////////

    // Subwindow/menu properties
    const char *Message(void) const;
    int MessageVisibility(void) const;
    int StatusVisibility(void) const;
    int CommandMenuVisibility(void) const;
    int LabelMenuVisibility(void) const;
    int LabelConfirmedVisibility(void) const;

    // Labeling statistics
    RNScalar HumanLabelPrecision(void) const;
    RNScalar HumanLabelRecall(void) const;
    RNScalar HumanLabelFMeasure(void) const;

    // Labeling statistics
    RNScalar PredictedLabelPrecision(void) const;
    RNScalar PredictedLabelRecall(void) const;
    RNScalar PredictedLabelFMeasure(void) const;

    ///////////////////////////////
    //// PROPERTY MANIPULATION ////
    ///////////////////////////////

    // Subwindow/menu manipulation (visibility: 0=off, 1=on, -1=toggle)
    void SetMessage(const char *fmt, ...);
    void SetMessageVisibility(int visibility);
    void SetStatusVisibility(int visibility);
    void SetCommandMenuVisibility(int visibility);
    void SetLabelMenuVisibility(int visibility);
    void SetLabelConfirmedVisibility(int visibility);

    // Camera control
    int ZoomCamera(RNScalar scale = 0);
    int ZoomToObject(const R3SurfelObject* o, const bool reset_line);
    int ZoomToObjectSet(const RNArray<R3SurfelObject*> &set,
            const bool reset_line);
    static int ZoomToROI(R3Camera &camera, const R2Viewport &viewport,
            const R2Box &active_area, const R3Sphere &sphere,
            const R3Vector *p_dir =
            NULL);

protected:

    //////////////////////////////////////////
    //// LOW_LEVEL MANIPULATION FUNCTIONS ////
    //////////////////////////////////////////

    // Logging functions
    void BeginCommand(int type, double operand = 0);
    void EndCommand(void);
    void PrintCommand(struct R3SurfelLabelerCommand *command,
            FILE *fp = NULL) const;
    void PrintCheckpoint(FILE *fp = NULL) const;

    // Object selection functions.
    int NObjectSelections(void) const;
    R3SurfelObject *ObjectSelection(int k) const;
    void InsertObjectSelection(R3SurfelObject *object);
    void RemoveObjectSelection(R3SurfelObject *object);
    int EmptyObjectSelections(void);
    void RecenterOnSet(const RNArray<R3SurfelObject*> &set);

    // Screen selection functions.
    int FilterObjectsInsideBox(const R2Box& box,
            RNArray<R3SurfelObject *> &picked_objects, int filter =
            R3_SURFEL_LABELER_SELECTION_FILTER_NONE) const;
    R3SurfelObject* SelectTopHitObjectInBox(const R2Box& box,
            const RNArray<R3SurfelObject *> &objects) const;
    int SelectTopHitObjectsInBox(const R2Box& box,
            const RNArray<R3SurfelObject *> &objects,
            RNArray<R3SurfelObject *> &picked_objects) const;
    bool ShowConfirmed(R3SurfelObject *object) const;

    // Representative label set functions.
    void UpdateLabelMenuSequence(RNArray<R3SurfelObject *> object_set);
    int LabelVisible(R3SurfelLabel *label) const;
    int LabelSelectible(R3SurfelLabel *label) const;

    // Selection zoom boxes functions.
    void InsertZoomBoxForObject(R3SurfelObject *object);
    void RemoveZoomBoxForObject(R3SurfelObject *object);
    void EmptyZoomBoxes(void);

    // Label assignment functions
    int InsertLabelAssignment(R3SurfelLabelAssignment *assignment);
    int RemoveLabelAssignment(R3SurfelLabelAssignment *assignment);
    int EmptyLabelAssignments(void);

    // Interface Helpers.
    void TogglePause(void);
    void FlipBGColor(void);
    bool isMachineDrivenInterface(void) const;

    ///////////////////////////////
    //// MENU/WINDOW FUNCTIONS ////
    ///////////////////////////////

    // Command menu functions
    void DrawCommandMenu(void);
    int PickCommandMenu(int xcursor, int ycursor, int button, int state,
            int shift, int ctrl, int alt);

    // Label menu functions
    void DrawLabelMenu(const R2Point *ul_corner = NULL);
    int PickLabelMenu(int xcursor, int ycursor, int button, int state,
            int shift, int ctrl, int alt);
    void ToggleExclusiveVisibilityForSelected();

    // Selection zoom boxes functions.
    void DrawSelectionZoomBoxes(void);
    void DrawBlipAroundObject(const R3SurfelObject *object);
    void DrawArrowAboveObject(const R3SurfelObject *object, const RNRgb &color,
            RNScalar alpha = 1.0);
    void DrawConeAroundObject(const R3SurfelObject *object, const RNRgb &color);
    int InSelectionZoomBox(int xcursor, int ycursor, int button, int state,
            int shift, int ctrl, int alt);

    // Draw surfels.
    void DrawSurfels(void);
    RNScalar AffinityScore(R3SurfelObject *from, R3SurfelObject *to);

    // Draw highlights around objects that are in a group set by classifier
    // in the debug window.
    void DrawHighlightGroup() const;

    // Message window functions
    void DrawMessage(void) const;

    // Status window functions
    void DrawStatus(void) const;

    // Status window functions
    void DrawFPS(void);

    // Select box functions
    void DrawSelectBox(void) const;

    // Paused screen
    void DrawPausedScreen(void) const;

    // Draw labels of the objects
    void DrawLabeling(void) const;

    // Draw names of objects.
    void DrawObjectNames(void) const;

    // Draw nodes.
    void DrawNodes(void) const;

    // Draw blocks.
    void DrawBlocks(void) const;

    // Draw center point.
    void DrawCenterPoint(void) const;

    // Draw diming layer over what is already drawn.
    void DrawDimLayer(const RNRgb *color, RNScalar alpha) const;

    // Draw a haloed message.
    void DrawHaloedMessage(const R2Point &p, const char* msg, void* font,
            const RNRgb *color = NULL) const;

    // Animation methods.
    void StartAnimation(bool reset = true);
    void StopAnimation();

    // Hijack global methods to consistently adjust for alternative color
    // schemes, e.g. different background color.
    void glColor3f(RNScalar r, RNScalar g, RNScalar b) const;
    void glColor4f(RNScalar r, RNScalar g, RNScalar b, RNScalar a) const;
    void CustomLoadRgb(const RNRgb& rgb) const;
    void LoadColorForScalar(RNScalar value, RNScalar min_value = 0.0,
            RNScalar max_value = 1.0) const;
    RNRgb GetColorForScalar(RNScalar value, RNScalar min_value = 0.0,
            RNScalar max_value = 1.0) const;

protected:

    friend class R3SurfelClassifier;

    ////////////////////////////
    //// INTERNAL VARIABLES ////
    ////////////////////////////

    // Classifier
    R3SurfelClassifier classifier;

    // Object selections
    RNArray<R3SurfelObject *> selection_objects;
    RNBoolean selection_changed_;
    RNBoolean selection_autorequest_;
    RNArray<SelectionZoomBox *> selection_zoom_box_;
    RNArray<R3SurfelObject*> highlight_group_;

    // Command logging
    struct R3SurfelLabelerCommand *current_command;
    RNArray<struct R3SurfelLabelerCommand *> undo_stack;
    int undo_index;
    RNArray<R3SurfelObject *> created_objects_stack;
    char *logging_filename;
    FILE *logging_fp;

    // Select box
    RNBoolean select_box_active;
    R2Point select_box_corners[2];
    bool recenter_on_selection_;

    // Message
    char *message;
    int message_visibility;

    // Status
    int status_visibility;

    // Paused state
    bool paused_;

    // Command menu
    int command_menu_visibility;
    R2Box command_menu_box_;

    // Label menu
    int label_menu_visibility;
    struct LabelViewType {
        int visibility;
        int selectibility;

        LabelViewType();
        void reset(int value = 1);
    };
    map<R3SurfelLabel*, LabelViewType> label_view_;
    R2Box label_menu_box_;
    RNArray<R3SurfelLabel*> label_menu_sequence_;

    //Show confirmed labels
    int object_label_confirmed_visibility;

    // Animation-related fields.
    const int animation_timer_id_;
    int animation_fps_;
    int animation_on_;
    RNTime animation_last_update_;
    RNAngle animation_phi_;

    // FPS.
    int fps_visibility_;
    RNTime last_fps_time_read_;
    int fps_counter_;
    RNScalar fps_;
    char* fps_message_;

    // User training mode containers.
    RNArray<R3SurfelObject *> incorrect_objects_;
    R3SurfelLabel *incorrect_label_;

    bool show_more_graph_info_;
};

////////////////////////////////////////////////////////////////////////
// INLINE FUNCTIONS
////////////////////////////////////////////////////////////////////////

inline R3SurfelLabeler::LabelViewType::LabelViewType() {
    reset();
}

inline void R3SurfelLabeler::LabelViewType::reset(int value) {
    visibility = value;
    selectibility = value;
}

inline int R3SurfelLabeler::MessageVisibility(void) const {
    // Return message visibility
    return message_visibility;
}

inline int R3SurfelLabeler::StatusVisibility(void) const {
    // Return status visibility
    return status_visibility;
}

inline int R3SurfelLabeler::CommandMenuVisibility(void) const {
    // Return command menu visibility
    return command_menu_visibility;
}

inline int R3SurfelLabeler::LabelMenuVisibility(void) const {
    // Return label menu visibility
    return label_menu_visibility;
}

inline int R3SurfelLabeler::LabelConfirmedVisibility(void) const {
    // Return label menu visibility
    return object_label_confirmed_visibility;
}

inline void R3SurfelLabeler::SetMessage(const char *fmt, ...) {
    // Free previous message
    if (message) free(message);
    message = NULL;

    // Set message (displayed at bottom of screen)
    if (fmt) {
        message = new char[4096];
        va_list args;
        va_start(args, fmt);
        vsprintf(message, fmt, args);
        va_end(args);
    }
}

inline void R3SurfelLabeler::SetMessageVisibility(int visibility) {
    // Set message visibility
    if (visibility == -1) message_visibility = 1 - message_visibility;
    else if (visibility == 0) message_visibility = 0;
    else message_visibility = 1;
}

inline void R3SurfelLabeler::SetStatusVisibility(int visibility) {
    // Set status visibility
    if (visibility == -1) status_visibility = 1 - status_visibility;
    else if (visibility == 0) status_visibility = 0;
    else status_visibility = 1;
}

inline void R3SurfelLabeler::SetCommandMenuVisibility(int visibility) {
    // Set command menu visibility
    if (visibility == -1) command_menu_visibility = 1 - command_menu_visibility;
    else if (visibility == 0) command_menu_visibility = 0;
    else command_menu_visibility = 1;
}

inline void R3SurfelLabeler::SetLabelMenuVisibility(int visibility) {
    // Set label menu visibility
    if (visibility == -1) label_menu_visibility = 1 - label_menu_visibility;
    else if (visibility == 0) label_menu_visibility = 0;
    else label_menu_visibility = 1;
}

inline void R3SurfelLabeler::SetLabelConfirmedVisibility(int visibility) {
    // Set label menu visibility
    if (visibility == -1) object_label_confirmed_visibility = 1
            - object_label_confirmed_visibility;
    else if (visibility == 0) object_label_confirmed_visibility = 0;
    else object_label_confirmed_visibility = 1;
}

inline int R3SurfelLabeler::NObjectSelections(void) const {
    // Return number of selected objects
    return selection_objects.NEntries();
}

inline R3SurfelObject *R3SurfelLabeler::ObjectSelection(int k) const {
    // Return kth selected objects
    return selection_objects.Kth(k);
}

inline int R3SurfelLabeler::AnimationTimerID() const {
    return animation_timer_id_;
}

inline int R3SurfelLabeler::AnimationFPS() const {
    return animation_fps_;
}

////////////////////////////////////////////////////////////////////////
// CLASS DEFINITION
////////////////////////////////////////////////////////////////////////

class SelectionZoomBox {
public:
    SelectionZoomBox(const R3Camera &camera, const R2Viewport &viewport,
            R3SurfelScene *scene, R3SurfelClassifier *classifier,
            R3SurfelObject *object);
    ~SelectionZoomBox();

    // Accessors.
    R3SurfelObject* Object(void) const;
    R3Camera& Camera(void);
    bool Hovered(void) const;
    const R2Box& BBox(void) const;

    // Modifiers.
    bool CheckHovered(RNScalar x, RNScalar y);

    // Draw methods.
    void Draw(int lt_x, int lt_y, int diameter, const R2Box &screen_box,
            bool incorrect);
    void AdvanceAnimation(RNScalar dt);

protected:
    R3Camera camera_;
    R3SurfelObject *object_;
    R3SurfelPointSet *all_points_;
    R3SurfelPointSet *object_points_;

    bool hovered_;
    R2Box bbox_;

    RNScalar phi_;

    R3SurfelClassifier *classifier_;
};

////////////////////////////////////////////////////////////////////////
// INLINE METHODS
////////////////////////////////////////////////////////////////////////

inline R3SurfelObject* SelectionZoomBox::Object(void) const {
    return object_;
}

inline R3Camera& SelectionZoomBox::Camera(void) {
    return camera_;
}

inline bool SelectionZoomBox::Hovered(void) const {
    return hovered_;
}

inline const R2Box& SelectionZoomBox::BBox(void) const {
    return bbox_;
}

inline bool SelectionZoomBox::CheckHovered(RNScalar x, RNScalar y) {
    hovered_ = R2Contains(bbox_, R2Point(x, y));
    return hovered_;
}

} //namespace igra

#endif /* R3SURFELLABELER_H_ */
