/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  R3SurfelLabelerCheckpoint.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2014.
 *
 *  Description: Checkpoint class that stores, reads and writes a snapshot of
 *               the annotation progress.
 *
 *  -------------------------------------------------------------------------*/

#include "R3SurfelLabelerCheckpoint.h"

#include <stdio.h>

#include "R3SurfelLabeler.h"

namespace igra {

static const char *kChkPntToken = "CHECKPOINT";

R3SurfelLabelerCheckpoint::R3SurfelLabelerCheckpoint() :
                timestamp(0.0),
                labels_num(0),
                gt_obj_num(0),
                hl_obj_num(0),
                pred_obj_num(0),
                prediction_precision(0.0),
                prediction_recall(0.0),
                prediction_f(0.0),
                human_precision(0.0),
                human_recall(0.0),
                human_f(0.0) {
}

R3SurfelLabelerCheckpoint::~R3SurfelLabelerCheckpoint() {
}

// I/O.
void R3SurfelLabelerCheckpoint::PrintCheckpoint(FILE *fp,
        const R3SurfelLabeler *labeler) {
    // Just checking
    if (!fp) fp = stdout;

    R3SurfelScene *scene = labeler->Scene();

    // Count assignments
    int ntotal = 0;
    int npredicted = 0;
    int nconfirmed = 0;
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        if (object->NParts() > 0) continue;
        if (!object->GroundTruthLabel()) continue;
        if (object->HumanLabel()) nconfirmed++;
        else if (object->PredictedLabel()) npredicted++;
        ntotal++;
    }

    // Print status
    fprintf(fp, "%s  %g  %d   %d %d %d    %g %g %g   %g %g %g", kChkPntToken,
            labeler->CurrentTime(), scene->NLabels(), ntotal, nconfirmed,
            npredicted, labeler->PredictedLabelPrecision(),
            labeler->PredictedLabelRecall(), labeler->PredictedLabelFMeasure(),
            labeler->HumanLabelPrecision(), labeler->HumanLabelRecall(),
            labeler->HumanLabelFMeasure());
    fprintf(fp, "    ");

    // Print labels
    for (int i = 0; i < scene->NLabels(); i++) {
        R3SurfelLabel *label = scene->Label(i);
        fprintf(fp, " L %d %s", label->SceneIndex(), label->Name());
    }
    fprintf(fp, "    ");

    // Print object assignments
    for (int i = 0; i < scene->NObjects(); i++) {
        R3SurfelObject *object = scene->Object(i);
        if (object->NParts() > 0) continue;
        R3SurfelLabel *ground_truth_label = object->GroundTruthLabel();
        R3SurfelLabel *human_label = object->HumanLabel();
        R3SurfelLabelAssignment *predicted_assignment =
                object->PredictedLabelAssignment();
        R3SurfelLabel *predicted_label = object->PredictedLabel();
        fprintf(fp, " O %d %s %d %d %d %g", object->SceneIndex(),
                object->Name(),
                (ground_truth_label) ? ground_truth_label->SceneIndex() : -1,
                (human_label) ? human_label->SceneIndex() : -1,
                (predicted_label) ? predicted_label->SceneIndex() : -1,
                (predicted_assignment) ?
                        predicted_assignment->Confidence() : -1);
    }
    fprintf(fp, "\n");
    fflush(fp);
}

R3SurfelLabelerCheckpoint* R3SurfelLabelerCheckpoint::TryParseCheckPoint(
        FILE *fp, R3SurfelScene *scene) {
    // Just checking
    if (!fp || feof(fp) || ferror(fp)) return NULL;

    // Remember line beginning position in case of an error.
    fpos_t position;
    fgetpos(fp, &position);

    // Check that this is indeed a command log entry.
    char temp_str[255];
    if (fscanf(fp, "%s ", temp_str) != 1) return NULL;
    if (strncmp(temp_str, kChkPntToken, strlen(kChkPntToken))) {
        // This is not a command log.
        fsetpos(fp, &position);
        return NULL;
    }

    R3SurfelLabelerCheckpoint *checkpoint = new R3SurfelLabelerCheckpoint();

    // Read status.
    fscanf(fp, "%lf  %d   %d %d %d    %lf %lf %lf   %lf %lf %lf",
            &checkpoint->timestamp, &checkpoint->labels_num,
            &checkpoint->gt_obj_num, &checkpoint->hl_obj_num,
            &checkpoint->pred_obj_num, &checkpoint->prediction_precision,
            &checkpoint->prediction_recall, &checkpoint->prediction_f,
            &checkpoint->human_precision, &checkpoint->human_recall,
            &checkpoint->human_f);

    if (feof(fp) || ferror(fp)) return NULL;

#ifndef NDEBUG
    // Read the rest and verify the numbers.
    int n_labels = 0;
    int gt_num = 0, hl_num = 0, pl_num = 0;
    int hl_correct = 0, pl_correct = 0;

    char name[255];
    while (fgetc(fp) != '\n') {
        char token;
        fscanf(fp, " %c", &token);
        if (token == 'L') {
            int l_index;
            fscanf(fp, " %d %s", &l_index, name);
            assert(!strcmp(name, scene->Label(l_index)->Name()));
            n_labels++;
        } else if (token == 'O') {
            int o_index;
            int gtl_index, hl_index, pl_index;
            RNScalar p_conf = 0.0;
            fscanf(fp, " %d %s %d %d %d %lf", &o_index, name, &gtl_index,
                    &hl_index, &pl_index, &p_conf);

            assert(!strcmp(name, scene->Object(o_index)->Name()));

            if (gtl_index != -1) gt_num++;
            if (hl_index != -1
                    && strcmp(scene->Label(hl_index)->Name(), "Unknown")) {
                hl_num++;
                if (hl_index == gtl_index) hl_correct++;
            }
            if (pl_index != -1
                    && strcmp(scene->Label(pl_index)->Name(), "Unknown")) {
                pl_num++;
                if (pl_index == gtl_index) pl_correct++;
            }

        } else {
            RNAbort("Checkpoint incorrect token: %c", token);
        }
    }

    assert(checkpoint->labels_num == n_labels);

    RNScalar pred_p = pl_num ? pl_correct / (RNScalar) pl_num : 1.0;
    RNScalar pred_r = gt_num ? pl_correct / (RNScalar) gt_num : 0.0;
    RNScalar pred_f = 0.0;
    if ((pred_p + pred_r) > 0.0)
        pred_f = (2 * pred_p * pred_r) / (pred_p + pred_r);

    assert(RNIsEqual(pred_p, checkpoint->prediction_precision));
    assert(RNIsEqual(pred_r, checkpoint->prediction_recall));
    assert(RNIsEqual(pred_f, checkpoint->prediction_f));

    RNScalar hl_p = hl_num ? hl_correct / (RNScalar) hl_num : 1.0;
    RNScalar hl_r = gt_num ? hl_correct / (RNScalar) gt_num : 0.0;
    RNScalar hl_f = 0.0;
    if ((hl_p + hl_r) > 0.0) hl_f = (2 * hl_p * hl_r) / (hl_p + hl_r);

    assert(RNIsEqual(hl_p, checkpoint->human_precision));
    assert(RNIsEqual(hl_r, checkpoint->human_recall));
    assert(RNIsEqual(hl_f, checkpoint->human_f));

#else
    // Simply advance to the next line.
    fscanf(fp, "%*[^\n]%*c");
#endif

    return checkpoint;
}

} // namespace igra

