/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  R3SurfelLabelerCommand.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2014.
 *
 *  Description: A class that stores, reads and writes interface commands.
 *
 *  -------------------------------------------------------------------------*/

#include "R3SurfelLabelerCommand.h"

#include <stdio.h>

#include "R3SurfelLabeler.h"

namespace igra {

static const char *kCmdToken = "CMD";

// Command class constructor

R3SurfelLabelerCommand::R3SurfelLabelerCommand(R3SurfelLabeler *labeler,
        int type, double operand) :
                type(type),
                operand(operand),
                camera(labeler->Camera()),
                center_point(labeler->CenterPoint()),
                timestamp(labeler->CurrentTime()),
                inserted_object_selections(),
                removed_object_selections(),
                inserted_label_assignments(),
                removed_label_assignments(),
                own_assignments(false) { // public constructor does not own assignments - labeler does
}

R3SurfelLabelerCommand::R3SurfelLabelerCommand() :
                type(0),
                operand(0.0),
                camera(),
                center_point(),
                timestamp(0.0),
                inserted_object_selections(),
                removed_object_selections(),
                inserted_label_assignments(),
                removed_label_assignments(),
                own_assignments(true) { // private constructor owns the assignment references
}

R3SurfelLabelerCommand::~R3SurfelLabelerCommand() {
    if (own_assignments) {
        for (int i = 0; i < inserted_label_assignments.NEntries(); ++i)
            delete inserted_label_assignments.Kth(i);
        for (int i = 0; i < removed_label_assignments.NEntries(); ++i)
            delete removed_label_assignments.Kth(i);
    }
}

// I/O funtionality.

void R3SurfelLabelerCommand::Write(FILE *fp) const {
    // Just checking
    if (!fp) fp = stdout;

    // Print
    const R3Point& eye = camera.Origin();
    const R3Vector& towards = camera.Towards();
    const R3Vector& up = camera.Up();

    // Count label assignments inserted by human
    int ninserted_label_assignments = 0;
    for (int j = 0; j < inserted_label_assignments.NEntries(); j++) {
        R3SurfelLabelAssignment *assignment = inserted_label_assignments[j];
        if (assignment->Originator()
                != R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR) continue;
        ninserted_label_assignments++;
    }

    // Count label assignments removed by human
    int nremoved_label_assignments = 0;
    for (int j = 0; j < removed_label_assignments.NEntries(); j++) {
        R3SurfelLabelAssignment *assignment = removed_label_assignments[j];
        if (assignment->Originator()
                != R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR) continue;
        nremoved_label_assignments++;
    }

    fprintf(fp, "%s %d %g %g    ", kCmdToken, type, operand, timestamp);
    fprintf(fp, "%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f ", eye[0], eye[1],
            eye[2], towards[0], towards[1], towards[2], up[0], up[1], up[2]);
    fprintf(fp, "    ");

    fprintf(fp, "%d", inserted_object_selections.NEntries());
    for (int j = 0; j < inserted_object_selections.NEntries(); j++)
        fprintf(fp, " %d", inserted_object_selections[j]->SceneIndex());
    fprintf(fp, "    ");

    fprintf(fp, "%d", removed_object_selections.NEntries());
    for (int j = 0; j < removed_object_selections.NEntries(); j++)
        fprintf(fp, " %d", removed_object_selections[j]->SceneIndex());
    fprintf(fp, "    ");

    fprintf(fp, "%d", ninserted_label_assignments);
    for (int j = 0; j < inserted_label_assignments.NEntries(); j++) {
        R3SurfelLabelAssignment *assignment = inserted_label_assignments[j];
        if (assignment->Originator()
                != R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR) continue;
        fprintf(fp, " ( %d %d %g %d )", assignment->Object()->SceneIndex(),
                assignment->Label()->SceneIndex(), assignment->Confidence(),
                assignment->Originator());
    }
    fprintf(fp, "    ");

    fprintf(fp, "%d", nremoved_label_assignments);
    for (int j = 0; j < removed_label_assignments.NEntries(); j++) {
        R3SurfelLabelAssignment *assignment = removed_label_assignments[j];
        if (assignment->Originator()
                != R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR) continue;
        fprintf(fp, " ( %d %d %g %d )", assignment->Object()->SceneIndex(),
                assignment->Label()->SceneIndex(), assignment->Confidence(),
                assignment->Originator());
    }
    fprintf(fp, "    ");

    fprintf(fp, "\n");
    fflush(fp);
}

// Tries to parse a line in a log file into a and object, which
// is returned upon success, otherwise a null is returned.
R3SurfelLabelerCommand* R3SurfelLabelerCommand::TryParseCommand(FILE *fp,
        R3SurfelScene *scene) {
    // Just checking
    if (!fp || feof(fp) || ferror(fp)) return NULL;

    // Remember line beginning position in case of an error.
    fpos_t position;
    fgetpos(fp, &position);

    // Check that this is indeed a command log entry.
    char temp_str[255];
    if (fscanf(fp, "%s ", temp_str) != 1) return NULL;
    if (strncmp(temp_str, kCmdToken, strlen(kCmdToken))) {
        // This is not a command log.
        fsetpos(fp, &position);
        return NULL;
    }

    R3SurfelLabelerCommand *cmd = new R3SurfelLabelerCommand();

    // Continue parsing command entries.
    bool ok = true;
    ok &= (fscanf(fp, "%d %lf %lf    ", &cmd->type, &cmd->operand,
            &cmd->timestamp) == 3);

    if (ok) {
        RNScalar e1, e2, e3;
        RNScalar t1, t2, t3;
        RNScalar u1, u2, u3;

        ok &= (fscanf(fp, " %lf %lf %lf %lf %lf %lf %lf %lf %lf ", &e1, &e2,
                &e3, &t1, &t2, &t3, &u1, &u2, &u3) == 9);

        const R3Point eye(e1, e2, e3);
        const R3Vector towards(t1, t2, t3);
        const R3Vector up(u1, u2, u3);
        cmd->camera.Reset(eye, towards, up);
    }

    if (ok) {
        // Read inserted selections.
        int n_inserted_object_selections;
        ok &= (fscanf(fp, " %d", &n_inserted_object_selections) == 1);
        for (int j = 0; ok && j < n_inserted_object_selections; j++) {
            int index;
            ok &= (fscanf(fp, " %d", &index) == 1);
            cmd->inserted_object_selections.Insert(scene->Object(index));
        }
    }

    if (ok) {
        // Read removed selections.
        int n_removed_object_selections;
        ok &= (fscanf(fp, " %d", &n_removed_object_selections) == 1);
        for (int j = 0; ok && j < n_removed_object_selections; j++) {
            int index;
            ok &= (fscanf(fp, " %d", &index) == 1);
            cmd->removed_object_selections.Insert(scene->Object(index));
        }
    }

    if (ok) {
        // Read inserted label assignments.
        int n_inserted_label_assignments;
        ok &= (fscanf(fp, " %d", &n_inserted_label_assignments) == 1);
        for (int j = 0; ok && j < n_inserted_label_assignments; j++) {
            int o_ind, l_ind, originator;
            RNScalar conf;
            ok &= (fscanf(fp, " ( %d %d %lf %d )", &o_ind, &l_ind, &conf,
                    &originator) == 4);
            assert(originator == R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR);

            R3SurfelObject *object = scene->Object(o_ind);
            assert(object);
            R3SurfelLabel *label = scene->Label(l_ind);
            assert(label);
            R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                    object, label, conf, originator);
            assert(assignment);

            cmd->inserted_label_assignments.Insert(assignment);
        }
    }

    if (ok) {
        // Read removed label assignments.
        int n_removed_label_assignments;
        ok &= (fscanf(fp, "%d", &n_removed_label_assignments) == 1);
        for (int j = 0; ok && j < n_removed_label_assignments; j++) {
            int o_ind, l_ind, originator;
            RNScalar conf;
            ok &= (fscanf(fp, " ( %d %d %lf %d )", &o_ind, &l_ind, &conf,
                    &originator) == 4);
            assert(originator == R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR);

            R3SurfelObject *object = scene->Object(o_ind);
            assert(object);
            R3SurfelLabel *label = scene->Label(l_ind);
            assert(label);
            R3SurfelLabelAssignment *assignment = new R3SurfelLabelAssignment(
                    object, label, conf, originator);
            assert(assignment);

            cmd->removed_label_assignments.Insert(assignment);
        }
    }

    if (!ok) {
        delete cmd;
        return NULL;
    }

    return cmd;
}

} // namespace igra
