/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  TreeTraits.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2013.
 *
 *  Description: Templated classes that implement basic necessary tree
 *               functionality.
 *
 *  -------------------------------------------------------------------------*/

#ifndef TREETRAITS_H_
#define TREETRAITS_H_

#include <stdlib.h>

#include <list>
#include <string>
#include <utility>

#include "RNBasics/RNBasics.h"

using namespace std;

///////////////////////////////////////////////
/// Proto tree traits class.
///////////////////////////////////////////////

template<class Object>
class ProtoTreeTraits {
public:
    // Accessors.
    Object* object(void) const;
    RNScalar weight(void) const;

    // Serialization.
    virtual void FillStats(list<string> &data) = 0;
protected:
    ProtoTreeTraits(Object *object = NULL, RNScalar node_weight = 0);
    ProtoTreeTraits(const ProtoTreeTraits &o);
    virtual ~ProtoTreeTraits();

    virtual void reset(Object *object = NULL, RNScalar node_weight = 0);

protected:
    Object *object_;	// payload data
    RNScalar node_weight_;	// property of the tree node
};

///////////////////////////////////////////////
/// Templated methods.
///////////////////////////////////////////////

template<class Object>
ProtoTreeTraits<Object>::ProtoTreeTraits(Object *object, RNScalar node_weight) :
        object_(object), node_weight_(node_weight) {
}

template<class Object>
ProtoTreeTraits<Object>::ProtoTreeTraits(const ProtoTreeTraits &o) :
        object_(o.object_), node_weight_(o.node_weight_) {
}

template<class Object>
ProtoTreeTraits<Object>::~ProtoTreeTraits() {
}

template<class Object>
void ProtoTreeTraits<Object>::reset(Object *object, RNScalar node_weight) {
    object_ = object;
    node_weight_ = node_weight;
}

template<class Object>
Object* ProtoTreeTraits<Object>::object(void) const {
    return object_;
}

template<class Object>
RNScalar ProtoTreeTraits<Object>::weight(void) const {
    return node_weight_;
}

///////////////////////////////////////////////
/// Tree traits template.
///////////////////////////////////////////////

template<class Node, class Object>
class TreeTraits: public ProtoTreeTraits<Object> {
public:
    // Properties.
    Node* root() const;
    bool is_leaf() const;
    int depth() const;

    // Reset.
    virtual void reset(Node *parent = NULL, list<Node*> *children = NULL,
            Object* object = NULL, RNScalar node_weight = 0);

    virtual void link_children();
    virtual void unlink_children();
    virtual void add_child(Node *child);
    virtual void remove_child(Node *child);
    virtual bool has_child(const Node* n1) const;

    int branch_factor() const;

    // Fill stats information.
    void FillStats(list<string> &data);

protected:
    explicit TreeTraits(Node *parent = NULL, list<Node*> *children = NULL,
            Object* object = NULL, RNScalar node_weight = 0, int size = 0);
    explicit TreeTraits(const TreeTraits &o);
    virtual ~TreeTraits();

    virtual void FillSpecificStats(list<string> &data) = 0;

public:
    Node *parent_;
    list<Node*> children_;

    using ProtoTreeTraits<Object>::object_;
    using ProtoTreeTraits<Object>::node_weight_;

    int size_;	// number of leaf nodes in subtree

    typedef typename list<Node*>::iterator ChildrenIterator;
    typedef typename list<Node*>::const_iterator ChildrenConstIterator;
};

///////////////////////////////////////////////
/// Methods
///////////////////////////////////////////////

template<class Node, class Object>
TreeTraits<Node, Object>::TreeTraits(Node *parent, list<Node*> *children,
        Object* object, RNScalar node_weight, int size) :
                ProtoTreeTraits<Object>(object, node_weight),
                parent_(parent),
                children_(),
                size_(size) {
    if (children) children_.assign(children->begin(), children->end());
}

template<class Node, class Object>
TreeTraits<Node, Object>::TreeTraits(const TreeTraits<Node, Object> &o) :
                ProtoTreeTraits<Object>(o),
                parent_(o.parent_),
                children_(o.children_),
                size_(o.size_) {
}

template<class Node, class Object>
TreeTraits<Node, Object>::~TreeTraits() {
    // Unlink from children.
    unlink_children();

    // Unlink from parent.
    if (parent_) {
        for (ChildrenIterator it = parent_->children_.begin();
                it != parent_->children_.end(); ++it) {
            if (*it == this) {
                parent_->children_.erase(it);
                break;
            }
        }
    }
}

template<class Node, class Object>
void TreeTraits<Node, Object>::reset(Node *parent, list<Node*> *children,
        Object* object, RNScalar node_weight) {
    ProtoTreeTraits<Object>::reset(object, node_weight);
    parent_ = parent;
    size_ = 0;

    unlink_children();
    children_.clear();

    if (children) children_.assign(children->begin(), children->end());
    link_children();

    for (ChildrenIterator it = children_.begin(); it != children_.end(); ++it)
        size_ += (*it)->size_;
}

template<class Node, class Object>
void TreeTraits<Node, Object>::link_children() {
    for (ChildrenIterator it = children_.begin(); it != children_.end(); ++it)
        (*it)->parent_ = static_cast<Node*>(this);
}

template<class Node, class Object>
void TreeTraits<Node, Object>::unlink_children() {
    for (ChildrenIterator it = children_.begin(); it != children_.end(); ++it)
        if ((*it)->parent_ == static_cast<Node*>(this)) (*it)->parent_ = NULL;
}

template<class Node, class Object>
void TreeTraits<Node, Object>::add_child(Node *child) {
    if (!child) return;
    children_.push_back(child);
    size_ += child->size_;
    //child->parent_ = static_cast<Node*>(this);
}

template<class Node, class Object>
void TreeTraits<Node, Object>::remove_child(Node *child) {
    for (ChildrenIterator it = children_.begin(); it != children_.end(); ++it) {
        if (*it != child) continue;

        if ((*it)->parent_ == static_cast<Node*>(this)) (*it)->parent_ = NULL;
        size_ -= (*it)->size_;
        assert(size_ >= 0);
        children_.erase(it);
        break;
    }
}

template<class Node, class Object>
bool TreeTraits<Node, Object>::has_child(const Node* n1) const {
    if (!n1) return false;

    for (ChildrenConstIterator it = children_.begin(); it != children_.end();
            ++it) {
        if ((*it) == n1) return true;
    }

    return false;
}

template<class Node, class Object>
int TreeTraits<Node, Object>::branch_factor(void) const {
    return children_.size();
}

template<class Node, class Object>
Node* TreeTraits<Node, Object>::root() const {
    Node *node = (Node*) this;
    while (node->parent_)
        node = node->parent_;
    return node;
}

template<class Node, class Object>
inline bool TreeTraits<Node, Object>::is_leaf() const {
    return children_.empty();
}

template<class Node, class Object>
int TreeTraits<Node, Object>::depth() const {
    int depth = 0;
    Node *node = parent_;
    while (node) {
        node = node->parent_;
        depth++;
    }
    return depth;
}

template<class Node, class Object>
void TreeTraits<Node, Object>::FillStats(list<string> &data) {
    // Useful variables.
    char buffer[1024];

    // Fill in stats pertaining to the tree node.
    sprintf(buffer, "-----------------");
    data.push_back(buffer);
    sprintf(buffer, "  Node %p", this);
    data.push_back(buffer);
    sprintf(buffer, "   Is root: %s (parent %p)",
            parent_ == NULL ? "Yes" : "No", parent_);
    data.push_back(buffer);
    sprintf(buffer, "   Is leaf: %s (%d children)", is_leaf() ? "Yes" : "No",
            branch_factor());
    data.push_back(buffer);
    sprintf(buffer, "   Node weight: %g", node_weight_);
    data.push_back(buffer);
    sprintf(buffer, "   Leaves in subtree: %d", size_);
    data.push_back(buffer);
    sprintf(buffer, "   Payload object: %p", object_);
    data.push_back(buffer);

    // Fill in stats pertaining to the specific implementation of the tree node.
    FillSpecificStats(data);

    // Fill in the stats pertaining to the payload.
    if (object_ != NULL) object_->FillStats(data);
}

#endif /* TREETRAITS_H_ */
