/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  UnionFind.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2013.
 *
 *  Description: A generic disjoint-set data type.
 *
 *  -------------------------------------------------------------------------*/


#ifndef UNIONFIND_H_
#define UNIONFIND_H_

#include <map>

using namespace std;


template<typename T>
class UnionFind {
public:
    UnionFind();
    virtual ~UnionFind();

    void* add(const T &o);

    void* find(const T &o);
    void* find(void *o);

    // returns how many merges actually needed to happen.
    void* merge(const T &o1, const T &o2);
    void* merge(void *o1, void *o2);

    int size() const;

    void unlink();

protected:
    struct UFNode {
        UFNode *parent;
        int rank;
        UFNode() :
                parent(this), rank(0) {
        }
        ;
    };

    void* find(UFNode *node);

    map<T, UFNode*> nodes_; // value -> node.
    int size_;             // number of disjoint sets.
};


template<typename T>
UnionFind<T>::
UnionFind():
  nodes_(),
  size_(0) {
}


template<typename T>
UnionFind<T>::
~UnionFind() {
    for (typename map<T, UFNode*>::iterator it = nodes_.begin();
            it != nodes_.end(); ++it) {
        delete it->second;
    }
}


template<typename T>
void* UnionFind<T>::
add(const T &o) {
    bool exists = nodes_.count(o);
    UFNode* &n = nodes_[o];
    if (!exists) {
        n = new UFNode;
        size_++;
    }
    return (void*) n;
}


template<typename T>
void* UnionFind<T>::
find (const T &o) {
	typename map<T, UFNode*>::iterator it = nodes_.find(o);
	assert(it != nodes_.end());
	return find(it->second);
}


template<typename T>
void* UnionFind<T>::
find(void *o) {
    if (!o)
        return NULL;
    return (void*) find((UFNode*) o);
}


template<typename T>
void* UnionFind<T>::
find(UFNode *node) {
	if (node->parent != node)
		node->parent = (UFNode*)find(node->parent);  // path compression.

	return (void*)node->parent;
}


template<typename T>
void* UnionFind<T>::
merge(const T &o1, const T &o2) {
    UFNode *x = (UFNode*) find(o1);
    UFNode *y = (UFNode*) find(o2);

    return merge(x, y);
}


template<typename T>
void* UnionFind<T>::
merge(void *o1, void *o2) {
    UFNode *x = (UFNode*) find(o1);
    UFNode *y = (UFNode*) find(o2);

    // If already in same partition no merging occurs.
    if (x == y) return NULL;

    // Not in the same set - merge.
    if (x->rank < y->rank) {
        x->parent = y;
    } else {
        y->parent = x;
        if (x->rank == y->rank) y->rank++;
    }
    size_--;

    // One merging occurred.
    return find(x);
}


template<typename T>
inline int UnionFind<T>::
size() const {
	return size_;
}


template<typename T>
inline void UnionFind<T>::
unlink() {

    for (typename map<T, UFNode*>::iterator it = nodes_.begin();
            it != nodes_.end(); ++it) {
        UFNode *node = it->second;
        node->parent = node;
        node->rank = 0;
    }
    size_ = nodes_.size();
}


#endif /* UNIONFIND_H_ */
