/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  UserSequencer.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2014.
 *
 *  Description: Strategy pattern classes for analyzing human and simulated
 *               users' labeling sessions.
 *
 *  -------------------------------------------------------------------------*/

#include "UserSequencer.h"

#include <map>

using namespace std;

namespace igra {

SequencesForPlotting::
SequencesForPlotting(const char *title):
    title_(title) {
    seq_f_t = new Sequence(title_ + "_f_t_seq");
    seq_f_i = new Sequence(title_ + "_f_i_seq");
    seq_t_f = new Sequence(title_ + "_t_f_seq");
    seq_i_f = new Sequence(title_ + "_i_f_seq");
    seq_conf_t = new Sequence(title_ + "_conf_t_seq");
    seq_conf_i = new Sequence(title_ + "_conf_i_seq");
    seq_t_conf = new Sequence(title_ + "_t_conf_seq");
    seq_i_conf = new Sequence(title_ + "_i_conf_seq");

    hist_l_i = new Sequence(title_ + "_label");
    hist_l_o = new Sequence(title_ + "_label");
    hist_c_i = new Sequence(title_ + "_contract");
    hist_c_o = new Sequence(title_ + "_contract");
    hist_e_i = new Sequence(title_ + "_expand");
    hist_e_o = new Sequence(title_ + "_expand");

    mistakes = new Sequence(title_ + "_mistakes");
    correct = new Sequence(title_ + "_correct");

    t_attn = new Sequence(title_ + "_real_vs_model");
    raw_t_attn = new Sequence(title_ + "_raw_model_estimate");
}


SequencesForPlotting::
~SequencesForPlotting() {
    delete seq_f_t;
    delete seq_f_i;
    delete seq_t_f;
    delete seq_i_f;
    delete seq_conf_t;
    delete seq_conf_i;
    delete seq_t_conf;
    delete seq_i_conf;
    delete hist_l_i;
    delete hist_l_o;
    delete hist_c_i;
    delete hist_c_o;
    delete hist_e_i;
    delete hist_e_o;
    delete mistakes;
    delete correct;
    delete t_attn;
    delete raw_t_attn;
}


bool cmp_points_x(const Point &p1, const Point &p2) {
    return p1.x < p2.x;
}


void SequencesForPlotting::
structure_raw_data() {
    consolidate_rolling(seq_f_t, seq_f_t);
    consolidate_rolling(seq_f_i, seq_f_i);
    consolidate_rolling(seq_conf_t, seq_conf_t);
    consolidate_rolling(seq_conf_i, seq_conf_i);

    remove_decreasing_x(seq_t_f);
    consolidate_rolling(seq_t_f, seq_t_f, true);
    remove_decreasing_x(seq_i_f);
    consolidate_rolling(seq_i_f, seq_i_f, true);
    remove_decreasing_x(seq_t_conf);
    consolidate_rolling(seq_t_conf, seq_t_conf, true);
    remove_decreasing_x(seq_i_conf);
    consolidate_rolling(seq_i_conf, seq_i_conf, true);

    consolidate_summing(hist_l_i, hist_l_i);
    consolidate_summing(hist_l_o, hist_l_o);
    consolidate_summing(hist_c_i, hist_c_i);
    consolidate_summing(hist_c_o, hist_c_o);
    consolidate_summing(hist_e_i, hist_e_i);
    consolidate_summing(hist_e_o, hist_e_o);

    consolidate_summing(mistakes, mistakes);
    consolidate_summing(correct, correct);

    consolidate_averaging(t_attn, t_attn);
    consolidate_raw(raw_t_attn, raw_t_attn);
}


void SequencesForPlotting::
merge(RNArray<SequencesForPlotting*> &sequences) {
    for (int i = 0; i < sequences.NEntries(); ++i) {
        SequencesForPlotting *seq = sequences.Kth(i);

        consolidate_rolling(seq->seq_f_t, seq_f_t);
        consolidate_rolling(seq->seq_f_i, seq_f_i);
        consolidate_rolling(seq->seq_conf_t, seq_conf_t);
        consolidate_rolling(seq->seq_conf_i, seq_conf_i);

        consolidate_rolling(seq->seq_t_f, seq_t_f, true);
        consolidate_rolling(seq->seq_i_f, seq_i_f, true);
        consolidate_rolling(seq->seq_t_conf, seq_t_conf, true);
        consolidate_rolling(seq->seq_i_conf, seq_i_conf, true);

        consolidate_averaging(seq->hist_l_i, hist_l_i, i);
        consolidate_averaging(seq->hist_l_o, hist_l_o, i);
        consolidate_averaging(seq->hist_c_i, hist_c_i, i);
        consolidate_averaging(seq->hist_c_o, hist_c_o, i);
        consolidate_averaging(seq->hist_e_i, hist_e_i, i);
        consolidate_averaging(seq->hist_e_o, hist_e_o, i);

        consolidate_averaging(seq->mistakes, mistakes, i);
        consolidate_averaging(seq->correct, correct, i);

        consolidate_averaging(seq->t_attn, t_attn);
        consolidate_raw(seq->raw_t_attn, raw_t_attn);
    }
}


void SequencesForPlotting::
plot_alone(const char *prefix, const char *postfix,
        RNArray<SequencesForPlotting*> &group, RNFlags filter) {
    char plot_name[1024];
    char plot_filename[1024];

    if (filter[FILTER_SEQ_F_T]) {
        printf("\t\tPlotting %s f(t) seq\n", prefix);
        sprintf(plot_name, "Average F-measure(Time) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Time", "F-measure");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_f_t);
        }

        Params params;
        params.y_unit_limit = true;
        params.plot_type = Params::LINE;
        params.legend = (Params::ON | Params::BOTTOM | Params::RIGHT);
        params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_f_t_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_F_I]) {
        printf("\t\tPlotting %s f(i) seq\n", prefix);
        sprintf(plot_name,
                "Average F-measure(Number of Interactions) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Number of Interactions", "F-measure");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_f_i);
        }

        Params params;
        params.y_unit_limit = true;
        params.plot_type = Params::LINE;
        params.legend = (Params::ON | Params::BOTTOM | Params::RIGHT);
        params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_f_i_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_T_F]) {
        printf("\t\tPlotting %s t(f) seq\n", prefix);
        sprintf(plot_name, "Average Time(F-measure) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "F-measure", "Time");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_t_f);
        }

        Params params;
        params.x_unit_limit = true;
        params.plot_type = Params::ERROR_BAR;
        params.legend = (Params::ON | Params::TOP | Params::LEFT);
        //params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_t_f_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_I_F]) {
        printf("\t\tPlotting %s i(f) seq\n", prefix);
        sprintf(plot_name, "Average Number of Interactions(F-measure) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "F-measure", "Number of Interactions");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_i_f);
        }

        Params params;
        params.x_unit_limit = true;
        params.plot_type = Params::ERROR_BAR;
        params.legend = (Params::ON | Params::TOP | Params::LEFT);
        //params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_i_f_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_CONF_T]) {
        printf("\t\tPlotting %s conf(t) seq\n", prefix);
        sprintf(plot_name,
                "Average Fraction of Confirmed Objects(Time) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Time", "Fraction of Confirmed Objects");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_conf_t);
        }

        Params params;
        params.y_unit_limit = true;
        params.plot_type = Params::LINE;
        params.legend = (Params::ON | Params::BOTTOM | Params::RIGHT);
        params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_conf_t_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_CONF_I]) {
        printf("\t\tPlotting %s conf(i) seq\n", prefix);
        sprintf(plot_name,
                "Average Fraction of Confirmed Objects(Number of Interactions) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Number of Interactions",
                "Fraction of Confirmed Objects");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_conf_i);
        }

        Params params;
        params.y_unit_limit = true;
        params.plot_type = Params::LINE;
        params.legend = (Params::ON | Params::BOTTOM | Params::RIGHT);
        params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_conf_i_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_T_CONF]) {
        printf("\t\tPlotting %s t(conf) seq\n", prefix);
        sprintf(plot_name,
                "Average Time(Fraction of Confirmed Objects) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Fraction of Confirmed Objects", "Time");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_t_conf);
        }

        Params params;
        params.x_unit_limit = true;
        params.plot_type = Params::ERROR_BAR;
        params.legend = (Params::ON | Params::TOP | Params::LEFT);
        //params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_t_conf_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_SEQ_I_CONF]) {
        printf("\t\tPlotting %s i(conf) seq\n", prefix);
        sprintf(plot_name,
                "Average Number of Interactions(Fraction of Confirmed Objects) (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Fraction of Confirmed Objects",
                "Number of Interactions");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->seq_i_conf);
        }

        Params params;
        params.x_unit_limit = true;
        params.plot_type = Params::ERROR_BAR;
        params.legend = (Params::ON | Params::TOP | Params::LEFT);
        //params.marker_size = 3;

        sprintf(plot_filename, "%s_seq_i_conf_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_HIST_I]) {
        printf("\t\tPlotting %s interaction hist by interations\n", prefix);
        sprintf(plot_name, "Interaction type distribution (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Group Size", "Number of Interactions");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->hist_c_i);
            plot.Insert(s->hist_l_i);
            plot.Insert(s->hist_e_i);
        }

        Params params;
        params.plot_type = Params::STACKED;
        params.legend = (Params::ON | Params::TOP | Params::RIGHT);

        sprintf(plot_filename, "%s_hist_i_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_HIST_O]) {
        printf("\t\tPlotting %s interaction hist by # objects\n", prefix);
        sprintf(plot_name, "Interaction type distribution (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Group Size", "Number of Objects");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->hist_c_o);
            plot.Insert(s->hist_l_o);
            plot.Insert(s->hist_e_o);
        }

        Params params;
        params.plot_type = Params::STACKED;
        params.legend = (Params::ON | Params::TOP | Params::RIGHT);

        sprintf(plot_filename, "%s_hist_o_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_MIS]) {
        printf("\t\tPlotting %s mistakes stats\n", prefix);
        sprintf(plot_name, "Incorrect labels distribution (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Group Size", "Number of Objects Labeled");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->mistakes);
            plot.Insert(s->correct);
        }

        Params params;
        params.plot_type = Params::STACKED;
        params.legend = (Params::ON | Params::TOP );

        sprintf(plot_filename, "%s_mistakes_%s", prefix,postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_T_ATTN]) {
        printf("\t\tPlotting %s real time vs attn cost\n", prefix);
        sprintf(plot_name, "Real time and attention model (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Attention Cost", "Real Time");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->t_attn);
        }

        Params params;
        params.plot_type = Params::ERROR_BAR;
        params.legend = (Params::ON | Params::BOTTOM | Params::OUTSIDE);
        params.x_log_scale = true;
        params.marker_size = 3;

        sprintf(plot_filename, "%s_t_real_attn_cost_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }

    if (filter[FILTER_RAW_T_ATTN]) {
        printf("\t\tPlotting %s real time vs attn cost parts raw\n", prefix);
        sprintf(plot_name, "Raw data - plot means nothing (avr. of %s)",
                postfix);
        Plotter plot(plot_name, "Attention Cost", "Real Time");
        for (int i = 0; i < group.NEntries(); ++i) {
            SequencesForPlotting *s = group[i];
            if (!s)
                continue;
            plot.Insert(s->raw_t_attn);
        }

        Params params;
        params.plot_type = Params::POINTS;
        params.legend = (Params::ON | Params::BOTTOM | Params::OUTSIDE);

        sprintf(plot_filename, "%s_raw_real_attn_cost_%s", prefix, postfix);
        if (!plot.Plot(plot_filename, params))
            fprintf(stderr, "Could not plot sequence to %s.", plot_filename);
    }
}


void SequencesForPlotting::
plot_alone(const char *prefix, const char *postfix,
            SequencesForPlotting* seq, RNFlags filter) {
    if (!seq) return;
    RNArray<SequencesForPlotting *> group;
    group.Insert(seq);
    plot_alone(prefix, postfix, group, filter);
}


void SequencesForPlotting::
consolidate_averaging(Sequence *from, Sequence *to, int no_entry_n) {
    set<double> in_from, new_in_from, not_in_from;
    if (no_entry_n > 0) {
        for (list<Point>::iterator it = from->values.begin();
                it != from->values.end(); ++it) {
            const Point &p = *it;
            in_from.insert(p.x);
            new_in_from.insert(p.x);
        }

        for (list<Point>::iterator it = to->values.begin();
                it != to->values.end(); ++it) {
            const Point &p = *it;
            if (in_from.count(p.x))
                new_in_from.erase(p.x);
            else
                not_in_from.insert(p.x);
        }
    }

    map<RNScalar, Point> consolidator;
    for (int i = 0; i < (from == to ? 1 : 2); ++i) {
        Sequence *seq_cons = (i ? from : to);
        for (list<Point>::iterator it = seq_cons->values.begin();
                it != seq_cons->values.end(); ++it) {
            const Point &src = *it;
            RNScalar x = src.x;
            Point &sink = consolidator[x];

            if (no_entry_n > 0) {
                if (seq_cons == from && new_in_from.count(src.x))
                    sink.n = no_entry_n;
                if (seq_cons == to && not_in_from.count(src.x))
                    sink.n = 1;
            }

            sink = Point::merge(src, sink);
            sink.x = x;
        }
    }

    //Replace raw values with consolidated ones.
    to->values.clear();
    for (map<RNScalar, Point>::iterator it = consolidator.begin();
            it != consolidator.end(); ++it) {
        to->values.push_back(it->second);
    }
}


void SequencesForPlotting::
consolidate_rolling(Sequence *from, Sequence *to, bool use_next) {
    // Works for sequences that are organized in an increasing x order.
    map<RNScalar, Point> consolidator;

    list<Point>::iterator it_from = from->values.begin();
    list<Point>::iterator it_to = to->values.begin();
    list<Point>::const_iterator from_end = from->values.end();
    list<Point>::const_iterator to_end = to->values.end();
    while (true) {

        bool from_done = (it_from == from_end);
        bool to_done = (it_to == to_end);
        if (from_done && ((from == to) || to_done)) break;

        // Which x is the current one.
        RNScalar x = RN_INFINITY;
        if (!from_done)
            x = min(x, it_from->x);
        if (from != to && !to_done)
            x = min(x, it_to->x);

        // Identify which entry to use: current with same x or
        // ,if such exist, with a smaller (larger, if use_next) x
        Point p_to, p_from;
        if (!from_done && RNIsEqual(x, it_from->x)) {
            while (it_from != from->values.end() && RNIsEqual(x, it_from->x)) {
                p_from = *it_from;
                ++it_from;
            }
        } else if (!use_next && it_from != from->values.begin()) {
            list<Point>::iterator temp = it_from;
            --temp;
            p_from = *temp;
        } else if (use_next && !from_done) {
            p_from = *it_from;
        }
        if (from != to) {
            if (!to_done && RNIsEqual(x, it_to->x)) {
                while (it_to != to->values.end() && RNIsEqual(x, it_to->x)) {
                    p_to = *it_to;
                    ++it_to;
                }
            } else if (!use_next && it_to != to->values.begin()) {
                list<Point>::iterator temp = it_to;
                --temp;
                p_to = *temp;
            } else if (use_next && !to_done) {
                p_to = *it_to;
            }
        } else
            ++it_to;

        // Add a consolidator point.
        Point &sink = consolidator[x];

        sink = Point::merge(sink, Point::merge(p_from, p_to));
        sink.x = x;
    }

    //Replace raw values with consolidated ones.
    to->values.clear();
    for (map<RNScalar, Point>::iterator it = consolidator.begin();
            it != consolidator.end(); ++it) {
        to->values.push_back(it->second);
    }
}


void SequencesForPlotting::
consolidate_summing(Sequence *from, Sequence *to) {
    map<RNScalar, Point> consolidator;
    for (int i = 0; i < (from == to ? 1 : 2); ++i) {
        Sequence *seq_cons = (i ? from : to);
        for (list<Point>::iterator it = seq_cons->values.begin();
                it != seq_cons->values.end(); ++it) {
            const Point &src = *it;
            Point &sink = consolidator[src.x];

            // Update consolidator entry with new values.
            sink = Point(src.x, src.mean + sink.mean, 1, 0.0);
        }
    }

    //Replace raw values with consolidated ones.
    to->values.clear();
    for (map<RNScalar, Point>::iterator it = consolidator.begin();
            it != consolidator.end(); ++it) {
        to->values.push_back(it->second);
    }
}


void SequencesForPlotting::
consolidate_raw(Sequence *from, Sequence *to) {
    multimap<RNScalar, Point> consolidator;
    for (int i = 0; i < (from == to ? 1 : 2); ++i) {
        Sequence *seq_cons = (i ? from : to);
        for (list<Point>::iterator it = seq_cons->values.begin();
                it != seq_cons->values.end(); ++it) {
            const Point &src = *it;

            consolidator.insert(pair<RNScalar, Point>(src.x, src));
        }
    }

    //Replace raw values with consolidated ones.
    to->values.clear();
    for (multimap<RNScalar, Point>::iterator it = consolidator.begin();
            it != consolidator.end(); ++it) {
        to->values.push_back(it->second);
    }
}


void SequencesForPlotting::
remove_decreasing_x(Sequence *seq) {
    list<Point>::iterator it = seq->values.begin();
    ++it;
    list<Point>::iterator prev_it = seq->values.begin();
    while(it != seq->values.end()) {
        const Point &pcurr = *it;
        const Point &p_prev = *prev_it;
        if (pcurr.x > p_prev.x) {
            prev_it = it;
            ++it;
        } else {
            it = seq->values.erase(it);
        }
    }
}


UserSequencer::
UserSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp) :
        sfp_(sfp), scene_(scene) {
}


UserSequencer::
~UserSequencer() {
}


RNScalar UserSequencer::
SceneFMeasure() {
    // Compute new value.
    R3SurfelLabel *unknown_label = scene_->FindLabelByName("Unknown");
    RNScalar predicted = 0, recallable = 0, correct = 0;

    for (int i = 1; i < scene_->NObjects(); ++i) {
        R3SurfelObject *object = scene_->Object(i);
        R3SurfelLabel *gt_label = object->GroundTruthLabel();
        R3SurfelLabel *predicted_label = object->PredictedLabel();

        if (gt_label && gt_label != unknown_label) {
            ++recallable;
            if (predicted_label && predicted_label != unknown_label) {
                ++predicted;
                if (gt_label == predicted_label) {
                    ++correct;
                }
            }
        } else {
            if (predicted_label && predicted_label != unknown_label) {
                ++predicted;
            }
        }
    }
    RNScalar precision = correct / predicted, recall = correct / recallable;
    RNScalar sum = precision + recall;
    return (sum > 0) ? 2 * precision * recall / sum : 0;
}

RNScalar UserSequencer::
SceneConfirmedFraction() {
    int confirmed = 0, total = 0;
    for (int i = 1; i < scene_->NObjects(); ++i) {
        R3SurfelObject *object = scene_->Object(i);
        if (object->NParts() > 0)
            continue;
        R3SurfelLabel *conf_label = object->HumanLabel();
        if (conf_label)
            confirmed++;
        total++;
    }

    if (total == 0)
        return 0;

    return confirmed / (RNScalar) total;
}



HumanUserSequencer::
HumanUserSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
        UserSession *session) :
        UserSequencer(scene, sfp),
        log_(session) {
}


HumanUserSequencer::
~HumanUserSequencer() {
}


void HumanUserSequencer::
ProduceSequence() {
    if (!log_ || !sfp_) return;

    int interactions_num = 0;
    RNScalar last_group_shown_t = 0.0;
    RNScalar last_group_shown_model_cost = 0.0;
    const RNScalar label_choice_time = R3SurfelClassifier::LabelChoiceTime(
            scene_);

    for (UserSession::Iterator it = log_->begin(); it != log_->end(); ++it) {
        RNScalar t = it.timestamp();

        if (it.command()) {
            R3SurfelLabelerCommand *cmd = it.command();

            switch (cmd->get_type()) {
                case R3_SURFEL_LABELER_SELECT_SUGGESTED_COMMAND:
                    last_group_shown_t = t;
                    last_group_shown_model_cost = cmd->get_operand();
                    break;
                case R3_SURFEL_LABELER_SELECT_PICKED_COMMAND:
                case R3_SURFEL_LABELER_SELECT_ENCLOSED_COMMAND:
                case R3_SURFEL_LABELER_SELECT_ALL_COMMAND:
                    // Explicit selections are interactions.
                    interactions_num++;
                    break;
                case R3_SURFEL_LABELER_SET_BACKGROUND_VISIBILITY_COMMAND:
                case R3_SURFEL_LABELER_SET_LABEL_NAME_VISIBILITY_COMMAND:
                case R3_SURFEL_LABELER_SET_SURFEL_COLOR_SCHEME_COMMAND:
                case R3_SURFEL_LABELER_SET_LABEL_VISIBILITY_COMMAND:
                    // Visibility changes are also interactions.
                    interactions_num++;
                    break;
                case R3_SURFEL_LABELER_PREDICT_ALL_COMMAND:
                    // Requesting reprediction is also an interaction.
                    interactions_num++;
                    break;
                case R3_SURFEL_LABELER_LABEL_TRAINING_MISTAKE_COMMAND:
                    // Although training interface prohibits incorrect labels,
                    // the attemp is still an interaction.
                    interactions_num++;
                    break;
                case R3_SURFEL_LABELER_SUGGEST_INCREASE_SELECTION: {
                    int g_size =
                            cmd->get_removed_object_selections().NEntries();
                    if (g_size > 0) {
                        sfp_->hist_e_i->values.push_back(Point(g_size, 1));
                        sfp_->hist_e_o->values.push_back(Point(g_size, g_size));
                    }

                    interactions_num++;

                    RNScalar dt = t - last_group_shown_t;
                    sfp_->t_attn->values.push_back(
                            Point(last_group_shown_model_cost, dt));
                    sfp_->raw_t_attn->values.push_back(
                            Point(dt, label_choice_time, 1,
                                    last_group_shown_model_cost
                                            - label_choice_time));
                    last_group_shown_t = t;
                    last_group_shown_model_cost = cmd->get_operand();
                }
                    break;
                case R3_SURFEL_LABELER_SUGGEST_DECREASE_SELECTION: {
                    int g_size =
                            cmd->get_removed_object_selections().NEntries();
                    if (g_size > 0) {
                        sfp_->hist_c_i->values.push_back(Point(g_size, 1));
                        sfp_->hist_c_o->values.push_back(Point(g_size, g_size));
                    }

                    interactions_num++;

                    RNScalar dt = t - last_group_shown_t;
                    sfp_->t_attn->values.push_back(
                            Point(last_group_shown_model_cost, dt));
                    sfp_->raw_t_attn->values.push_back(
                            Point(dt, label_choice_time, 1,
                                    last_group_shown_model_cost
                                            - label_choice_time));
                    last_group_shown_t = t;
                    last_group_shown_model_cost = cmd->get_operand();
                }
                    break;
                case R3_SURFEL_LABELER_LABEL_PICKED_COMMAND:
                case R3_SURFEL_LABELER_LABEL_SELECTION_COMMAND:
                case R3_SURFEL_LABELER_CONFIRM_PICKED_COMMAND:
                case R3_SURFEL_LABELER_CONFIRM_SELECTION_COMMAND:
                case R3_SURFEL_LABELER_CONFIRM_ALL_COMMAND:
                {
                    const RNArray<R3SurfelLabelAssignment*>& g =
                            cmd->get_inserted_label_assignments();
                    int g_size = g.NEntries();
                    if (g_size > 0) {
                        sfp_->hist_l_i->values.push_back(Point(g_size, 1));
                        sfp_->hist_l_o->values.push_back(Point(g_size, g_size));
                    }

                    int mistakes = 0;
                    for (int i = 0; i < g_size; ++i) {
                        R3SurfelLabelAssignment *a = g[i];
                        R3SurfelLabel *l_assigned = a->Label();
                        R3SurfelObject *o = a->Object();
                        R3SurfelLabel *l_gt = o->GroundTruthLabel();
                        if (l_assigned != l_gt) mistakes++;
                    }
                    if (g_size > 0) {
                        sfp_->mistakes->values.push_back(
                                Point(g_size, mistakes));
                        sfp_->correct->values.push_back(
                                Point(g_size, g_size - mistakes));
                    }

                    interactions_num++;

                    RNScalar dt = t - last_group_shown_t;
                    sfp_->t_attn->values.push_back(
                            Point(last_group_shown_model_cost, dt));
                    sfp_->raw_t_attn->values.push_back(
                            Point(dt, label_choice_time, 1,
                                    last_group_shown_model_cost
                                            - label_choice_time));
                }
                    break;
                default:
                    break;
            }

        } else if (it.checkpoint()) {
            RNScalar f = it.checkpoint()->get_predictionf();
            sfp_->seq_f_t->values.push_back(Point(t, f));
            sfp_->seq_f_i->values.push_back(Point(interactions_num, f));
            sfp_->seq_t_f->values.push_back(Point(f, t));
            sfp_->seq_i_f->values.push_back(Point(f, interactions_num));

            RNScalar conf_frac = it.checkpoint()->get_confirmed_fraction();
            sfp_->seq_conf_t->values.push_back(Point(t, conf_frac));
            sfp_->seq_conf_i->values.push_back(
                    Point(interactions_num, conf_frac));
            sfp_->seq_t_conf->values.push_back(Point(conf_frac, t));
            sfp_->seq_i_conf->values.push_back(
                    Point(conf_frac, interactions_num));
        }
    }

    sfp_->structure_raw_data();
}


VirtualUserSequencer::
VirtualUserSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp) :
        UserSequencer(scene, sfp) {
}


VirtualUserSequencer::
~VirtualUserSequencer() {
}


void VirtualUserSequencer::
ProduceSequence() {
    // Make necessary preparations.
    prepare();

    // Rolling data containers.
    RNScalar rolling_cost = 0.0;
    int rolling_q = 0;

    //RNScalar label_choice_time = R3SurfelClassifier::LabelChoiceTime(scene_);

    // Main loop.
    while (true) {
        // Procure a group of objects that should be labeled.
        group_.Empty();
        RNScalar cost = suggest_group();

        // If no objects are to be labeled - discontinue.
        if (group_.NEntries() == 0)
            break;

        // Adjust group as seems fit by virtual user.
        //sfp_->t_attn->values.push_back(Point(cost, cost));
        //sfp_->raw_t_attn->values.push_back(
        //        Point(cost, label_choice_time, 1, cost - label_choice_time));
        int interaction_counter = 1;
        while (!group_final()) {
            RNScalar ad_cost = adjust_group();
            cost += ad_cost;
            if (cost > 0.0) {
                //sfp_->t_attn->values.push_back(Point(ad_cost, ad_cost));
                //sfp_->raw_t_attn->values.push_back(
                //        Point(ad_cost, label_choice_time, 1,
                //                ad_cost - label_choice_time));
                interaction_counter++;
            }
        }

        // Simulate label assignment.
        for (int oi = 0; oi < group_.NEntries(); ++oi) {
            R3SurfelObject* object = group_.Kth(oi);
            R3SurfelLabelAssignment *gt = object->GroundTruthLabelAssignment();
            assert(gt);
            R3SurfelLabelAssignment *a = new R3SurfelLabelAssignment(*gt);
            a->SetConfidence(1.0);
            a->SetOriginator(R3_SURFEL_LABEL_ASSIGNMENT_HUMAN_ORIGINATOR);
            scene_->InsertLabelAssignment(a);
            post_assignment_update(a);
        }
        post_labeling_update();

        // Add raw data to to sequences.
        rolling_cost += cost;
        rolling_q += interaction_counter;
        RNScalar f = SceneFMeasure();
        RNScalar conf_frac = SceneConfirmedFraction();
        sfp_->seq_f_t->values.push_back(Point(rolling_cost, f));
        sfp_->seq_f_i->values.push_back(Point(rolling_q, f));
        sfp_->seq_t_f->values.push_back(Point(f, rolling_cost));
        sfp_->seq_i_f->values.push_back(Point(f, rolling_q));
        sfp_->seq_conf_t->values.push_back(Point(rolling_cost, conf_frac));
        sfp_->seq_conf_i->values.push_back(Point(rolling_q, conf_frac));
        sfp_->seq_t_conf->values.push_back(Point(conf_frac, rolling_cost));
        sfp_->seq_i_conf->values.push_back(Point(conf_frac, rolling_q));

        sfp_->hist_l_i->values.push_back(Point(group_.NEntries(), 1));
        sfp_->hist_l_o->values.push_back(
                Point(group_.NEntries(), group_.NEntries()));
        //sfp_->mistakes->values.push_back(Point(group_.NEntries(), 0));
    }

    post_loop();

    sfp_->structure_raw_data();
}


bool VirtualUserSequencer::
group_final() {
    return true;
}


void VirtualUserSequencer::
post_loop() {
}


NoMLSequencer::
NoMLSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp) :
        VirtualUserSequencer(scene, sfp) {
}


void NoMLSequencer::
prepare() {
    for (int j = 0; j < scene_->NObjects(); ++j) {
        R3SurfelObject *o = scene_->Object(j);
        if (!o || o->NParts() > 0)
            continue;
        object_pool_.Insert(o);
    }
}


RNScalar NoMLSequencer::
suggest_group() {
    if (object_pool_.NEntries() == 0) return 0.0;

    // Pick random object.
    int index = RNRandomScalar() * object_pool_.NEntries();
    R3SurfelObject *o = object_pool_.Kth(index);

    // Add it to the group.
    group_.Insert(o);

    // Remove used entry.
    RNArrayEntry *entry = object_pool_.KthEntry(index);
    RNArrayEntry *tail = object_pool_.TailEntry();
    object_pool_.SwapEntries(entry, tail);
    object_pool_.RemoveTail();

    return R3SurfelClassifier::LabelChoiceTime(scene_);
}


RNScalar NoMLSequencer::
adjust_group() {
    return R3SurfelClassifier::LabelChoiceTime(scene_);
}


void NoMLSequencer::
post_assignment_update(R3SurfelLabelAssignment *a) {
}


void NoMLSequencer::
post_labeling_update() {
    // Need to manually insert prediction labels.
    for (int oi = 0; oi < group_.NEntries(); ++oi) {
        R3SurfelObject* object = group_.Kth(oi);
        R3SurfelLabelAssignment *gt = object->GroundTruthLabelAssignment();
        assert(gt);
        R3SurfelLabelAssignment *a = new R3SurfelLabelAssignment(*gt);
        a->SetConfidence(1.0);
        a->SetOriginator(R3_SURFEL_LABEL_ASSIGNMENT_MACHINE_ORIGINATOR);
        scene_->InsertLabelAssignment(a);
    }
}


SingularSuggestSequencer::
SingularSuggestSequencer(R3SurfelScene *scene,
        SequencesForPlotting *sfp, const RNFlags &method) :
        VirtualUserSequencer(scene, sfp),
        seeder_method_(method),
        classifier_(NULL) {
}


SingularSuggestSequencer::
~SingularSuggestSequencer(){
    if (classifier_) delete classifier_;
}


void SingularSuggestSequencer::
prepare() {
    if (classifier_) delete classifier_;
    classifier_ = new R3SurfelClassifier(scene_, NULL);
}


RNScalar SingularSuggestSequencer::
suggest_group() {
    R3SurfelObject *object = NULL;
    RNScalar cost = classifier_->SuggestSingleObject(object, seeder_method_);

    if (object)
        group_.InsertTail(object);

    return cost;
}


RNScalar SingularSuggestSequencer::
adjust_group() {
    return R3SurfelClassifier::LabelChoiceTime(scene_);
}


void SingularSuggestSequencer::
post_assignment_update(R3SurfelLabelAssignment *a) {
    classifier_->UpdateAfterInsertLabelAssignment(a);
}


void SingularSuggestSequencer::
post_labeling_update() {
    classifier_->PredictLabelAssignments();
}


GroupSuggestSequencer::
GroupSuggestSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
        const RNFlags &method) :
        SingularSuggestSequencer(scene, sfp, method),
        grow_(false),
        alteration_kept_group_same_(false) {
}


GroupSuggestSequencer::
~GroupSuggestSequencer() {
}


RNScalar GroupSuggestSequencer::
suggest_group() {
    alteration_kept_group_same_ = false;
    grow_ = false;  // never do the growing phase.
    return classifier_->SuggestBestGroup(group_);
}


RNScalar GroupSuggestSequencer::
adjust_group() {
    int g_size_old = group_.NEntries();
    RNScalar cost = 0.0;

    if (grow_)
        cost = classifier_->SuggestExpansion(group_);
    else
        cost = classifier_->SuggestContraction(group_);

    int g_size_new = group_.NEntries();

    if (g_size_old == g_size_new) {
        alteration_kept_group_same_ = true;
        return 0.0;
    }

    if (grow_) {
        sfp_->hist_e_i->values.push_back(Point(g_size_old, 1));
        sfp_->hist_e_o->values.push_back(Point(g_size_old, g_size_old));
    } else {
        sfp_->hist_c_i->values.push_back(Point(g_size_old, 1));
        sfp_->hist_c_o->values.push_back(Point(g_size_old, g_size_old));
    }

    return cost;
}


bool GroupSuggestSequencer::
group_final() {
    bool homogeneous = isHomogeneous();

    // See if there is possible room for further expansion.
    if (grow_ && (!homogeneous || alteration_kept_group_same_))
        grow_ = false;

    if (grow_)
        return false;

    return homogeneous;
}


bool GroupSuggestSequencer::
isHomogeneous() {
    // Checks if all objects in suggested set have same GT label.

    map<R3SurfelLabel *, int> gt_labels;
    RNScalar total = group_.NEntries();
    for (int oi = 0; oi < total; ++oi) {
        R3SurfelObject *object = group_[oi];
        R3SurfelLabel *gt_label = object->GroundTruthLabel();
        gt_labels[gt_label]++;

        if (gt_labels.size() > 1)
            return false;
    }

    return true;
}


GroupSeededSuggestSequencer::
GroupSeededSuggestSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
            const RNFlags &method) :
            GroupSuggestSequencer(scene, sfp, method) {
}


GroupSeededSuggestSequencer::
~GroupSeededSuggestSequencer() {
}


RNScalar GroupSeededSuggestSequencer::
suggest_group() {
    alteration_kept_group_same_ = false;
    grow_ = false;  // never do the growing phase.

    R3SurfelObject *object;
    classifier_->SuggestSingleObject(object, seeder_method_);
    if (!object)
        return 0.0;
    return classifier_->SuggestBestGroupForSeed(object, group_);
}

} // namespace igra
