/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  UserSequencer.h
 *
 *  Copyright Aleksey Boyko, Princeton University, 2014.
 *
 *  Description: Strategy pattern classes for analyzing human and simulated
 *               users' labeling sessions and creating uniform plotting
 *               sequences.
 *
 *  -------------------------------------------------------------------------*/


#ifndef USERSEQUENCER_H_
#define USERSEQUENCER_H_

#include "Plotter.h"
#include "R3SurfelClassifier.h"
#include "R3Surfels/R3Surfels.h"
#include "RNBasics/RNBasics.h"
#include "UserSession.h"

namespace igra {

// Facilitate storage of all plots.
struct SequencesForPlotting {
    // Accuracy-averaging sequences.
    Sequence *seq_f_t, *seq_f_i;
    // Time-averaging sequences.
    Sequence *seq_t_f, *seq_i_f;
    // Confirmation dynamics.
    Sequence *seq_conf_t, *seq_conf_i;
    Sequence *seq_t_conf, *seq_i_conf;

    // Label histogram (by interaction/by group size)
    Sequence *hist_l_i, *hist_l_o;
    // Contract histogram (by interaction/by group size)
    Sequence *hist_c_i, *hist_c_o;
    // Expand histogram (by interaction/by group size)
    Sequence *hist_e_i, *hist_e_o;
    // Mistakes plots.
    Sequence *mistakes;
    Sequence *correct;
    // Real time/attention cost.
    Sequence *t_attn;
    // Raw model fitting data - plots are meaningless, data
    // stores x - real time, mean - Tsingular, dev - (Tgroup - Tsingular)
    Sequence *raw_t_attn;

    SequencesForPlotting(const char *title);
    virtual ~SequencesForPlotting();

    // Covert raw data dumped into Sequence into a structured sequence.
    void structure_raw_data();

    // Merge several structured sequences into one
    // and clean data from original sequences.
    void merge(RNArray<SequencesForPlotting*> &sequences);

    enum{
        FILTER_SEQ_F_T      = 1,
        FILTER_SEQ_F_I      = 1 << 2,
        FILTER_SEQ_T_F      = 1 << 3,
        FILTER_SEQ_I_F      = 1 << 4,
        FILTER_SEQ_CONF_T   = 1 << 5,
        FILTER_SEQ_CONF_I   = 1 << 6,
        FILTER_SEQ_T_CONF   = 1 << 7,
        FILTER_SEQ_I_CONF   = 1 << 8,
        FILTER_HIST_I       = 1 << 9,
        FILTER_HIST_O       = 1 << 10,
        FILTER_MIS          = 1 << 11,
        FILTER_T_ATTN       = 1 << 12,
        FILTER_RAW_T_ATTN   = 1 << 13,
        FILTER_ALL          = 0xFFFF
    };

    // Custom plotting.
    static void plot_alone(const char *prefix, const char *postfix,
            RNArray<SequencesForPlotting*> &group, RNFlags filter = FILTER_ALL);
    static void plot_alone(const char *prefix, const char *postfix,
            SequencesForPlotting* seq, RNFlags filter = FILTER_ALL);

protected:
    static void consolidate_averaging(Sequence *src, Sequence *sink,
            int no_entry_n = 0);
    static void consolidate_rolling(Sequence *src, Sequence *sink,
            bool use_next = false);
    static void consolidate_summing(Sequence *src, Sequence *sink);
    static void consolidate_raw(Sequence *src, Sequence *sink);

    static void remove_decreasing_x(Sequence *seq);

    string title_;
};


// Interface virtual user simulation code.

class UserSequencer {
public:
    UserSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp);
    virtual ~UserSequencer();

    virtual void ProduceSequence() = 0;

    RNScalar SceneFMeasure();
    RNScalar SceneConfirmedFraction();

protected:
    // Not owned by the object.
    SequencesForPlotting *sfp_;
    R3SurfelScene *scene_;
};


// Interface for human user sequencers.
class HumanUserSequencer : public UserSequencer {
public:
    HumanUserSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
            UserSession *session);
    virtual ~HumanUserSequencer();

    virtual void ProduceSequence();

protected:
    UserSession *log_;
};


// Interface for virtual user sequencers.

class VirtualUserSequencer : public UserSequencer {
public:
    VirtualUserSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp);
    virtual ~VirtualUserSequencer();

    void ProduceSequence();

protected:

    virtual void prepare() = 0;
    virtual RNScalar suggest_group() = 0;
    virtual RNScalar adjust_group() = 0;
    virtual bool group_final();

    virtual void post_assignment_update(R3SurfelLabelAssignment *a) = 0;
    virtual void post_labeling_update() = 0;
    virtual void post_loop();

protected:

    RNArray<R3SurfelObject*> group_;
};


// No learning sequencer (simple 1-by-1 labeling).

class NoMLSequencer : public VirtualUserSequencer {
public:
    NoMLSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp);

protected:
    virtual void prepare();
    virtual RNScalar suggest_group();
    virtual RNScalar adjust_group();

    virtual void post_assignment_update(R3SurfelLabelAssignment *a);
    virtual void post_labeling_update();

protected:
    RNArray<R3SurfelObject*> object_pool_;
};


// Sequencer suggusting 1 object at a time.

class SingularSuggestSequencer : public VirtualUserSequencer {
public:
    SingularSuggestSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
            const RNFlags &method);
    virtual ~SingularSuggestSequencer();

protected:
    virtual void prepare();
    virtual RNScalar suggest_group();
    virtual RNScalar adjust_group();

    virtual void post_assignment_update(R3SurfelLabelAssignment *a);
    virtual void post_labeling_update();

protected:
    RNFlags seeder_method_;
    R3SurfelClassifier *classifier_;
};


class GroupSuggestSequencer : public SingularSuggestSequencer {
public:
    GroupSuggestSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
            const RNFlags &method);
    virtual ~GroupSuggestSequencer();

protected:
    virtual RNScalar suggest_group();
    virtual RNScalar adjust_group();
    virtual bool group_final();

    bool isHomogeneous();

protected:
    bool grow_;  // during adjustment start with growing the group until impure
    bool alteration_kept_group_same_;
};


class GroupSeededSuggestSequencer : public GroupSuggestSequencer {
public:
    GroupSeededSuggestSequencer(R3SurfelScene *scene, SequencesForPlotting *sfp,
            const RNFlags &method);
    virtual ~GroupSeededSuggestSequencer();

protected:
    virtual RNScalar suggest_group();
};

} // namespace igra

#endif /* USERSEQUENCER_H_ */
