/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  UserSession.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2014.
 *
 *  Description: Helper that streamlines interaction with the heterogeneous
 *               contents of the user log files.
 *
 *  -------------------------------------------------------------------------*/

#include "UserSession.h"

namespace igra {

UserSession::UserSession() :
        commands(), checkpoints() {
}

UserSession::~UserSession() {
    clear();
}

void UserSession::clear(bool deallocate) {
    if (deallocate) {
        for (int i = 0; i < commands.NEntries(); ++i) {
            R3SurfelLabelerCommand *cmd = commands.Kth(i);
            delete cmd;
        }
    }
    commands.Empty();

    if (deallocate) {
        for (int i = 0; i < checkpoints.NEntries(); ++i) {
            R3SurfelLabelerCheckpoint *checkpoint = checkpoints.Kth(i);
            delete checkpoint;
        }

        checkpoints.Empty();
    }
}

void UserSession::trim_to_last_checkpoint() {
    Iterator it = back();
    while (it != end() && !it.checkpoint()) {
        pop_back();
        it = back();
    }
}

void UserSession::check_validity() {
    // Checks that both sequences are sequential in time.
    RNScalar last_time = -RN_INFINITY;
    for (int i = 0; i < commands.NEntries(); ++i) {
        R3SurfelLabelerCommand *cmd = commands.Kth(i);
        RNScalar new_time = cmd->get_timestamp();
        if (last_time > new_time) {
            RNAbort("UserSession command sequence non-monotone: %lf > %lf",
                    last_time > new_time);
        }
        last_time = new_time;
    }

    last_time = -RN_INFINITY;
    for (int i = 0; i < checkpoints.NEntries(); ++i) {
        R3SurfelLabelerCheckpoint *checkpoint = checkpoints.Kth(i);
        RNScalar new_time = checkpoint->get_timestamp();
        if (last_time > new_time) {
            RNAbort("UserSession checkpoint sequence non-monotone: %lf > %lf",
                    last_time > new_time);
        }
        last_time = new_time;
    }

    last_time = -RN_INFINITY;
    for (Iterator it = begin(); it != end(); ++it) {
        RNScalar new_time = it.timestamp();
        if (last_time > new_time) {
            RNAbort("UserSession session sequence non-monotone: %lf > %lf",
                    last_time > new_time);
        }
        last_time = new_time;
    }

}

void UserSession::push_back(const Iterator& it) {
    if (it.command()) commands.InsertTail(it.command());
    else if (it.checkpoint()) checkpoints.InsertTail(it.checkpoint());
    else RNAbort("UserSession.push_back(Iterator): iterator empty.");
}

void UserSession::push_back(R3SurfelLabelerCommand *cmd) {
    assert(cmd);
    commands.InsertTail(cmd);
}

void UserSession::push_back(R3SurfelLabelerCheckpoint *checkpoint) {
    assert(checkpoint);
    checkpoints.InsertTail(checkpoint);
}

void UserSession::pop_back() {
    Iterator it = back();
    if (it.command()) commands.RemoveTail();
    else if (it.checkpoint()) checkpoints.RemoveTail();
}

// Iterator implementations.

UserSession::Iterator::Iterator() :
        session_(NULL), cmd_index_(0), checkpoint_index_(0), use_command_(true) {
}

UserSession::Iterator::Iterator(const UserSession *session) :
                session_((UserSession*) session),
                cmd_index_(0),
                checkpoint_index_(0),
                use_command_(true) {
}

UserSession::Iterator::Iterator(const Iterator& it) :
                session_(it.session_),
                cmd_index_(it.cmd_index_),
                checkpoint_index_(it.checkpoint_index_),
                use_command_(it.use_command_) {
}

UserSession::Iterator& UserSession::Iterator::operator =(const Iterator& it) {
    session_ = it.session_;
    cmd_index_ = it.cmd_index_;
    checkpoint_index_ = it.checkpoint_index_;
    use_command_ = it.use_command_;
    return *this;
}

UserSession::Iterator UserSession::Iterator::begin(const UserSession* session) {

    R3SurfelLabelerCommand *first_command = NULL;
    R3SurfelLabelerCheckpoint *first_checkpoint = NULL;
    if (session->commands.NEntries() > 0)
        first_command = session->commands.Kth(0);
    if (session->checkpoints.NEntries() > 0)
        first_checkpoint = session->checkpoints.Kth(0);

    Iterator it(session);

    if (first_command && first_checkpoint) {
        it.use_command_ = !(first_command->get_timestamp()
                > first_checkpoint->get_timestamp());
    } else if (first_checkpoint) {
        it.use_command_ = false;
    }

    return it;
}

UserSession::Iterator UserSession::Iterator::end(const UserSession* session) {
    Iterator end(session);
    end.cmd_index_ = session->commands.NEntries();
    end.checkpoint_index_ = session->checkpoints.NEntries();
    return end;
}

UserSession::Iterator UserSession::Iterator::back(const UserSession* session) {

    R3SurfelLabelerCommand *last_command = NULL;
    R3SurfelLabelerCheckpoint *last_checkpoint = NULL;
    if (session->commands.NEntries() > 0) last_command =
            session->commands.Tail();
    if (session->checkpoints.NEntries() > 0)
        last_checkpoint = session->checkpoints.Tail();

    Iterator it = end(session);

    if (last_command && last_checkpoint) {
        if (last_command->get_timestamp() > last_checkpoint->get_timestamp()) {
            it.cmd_index_--;
        } else {
            it.checkpoint_index_--;
            it.use_command_ = false;
        }
    } else if (last_command) {
        it.cmd_index_--;
    } else if (last_checkpoint) {
        it.checkpoint_index_--;
        it.use_command_ = false;
    }

    return it;
}

R3SurfelLabelerCommand* UserSession::Iterator::command() const {
    if (!session_ || !use_command_) return NULL;

    if (0 <= cmd_index_ && cmd_index_ < session_->commands.NEntries())
        return session_->commands.Kth(cmd_index_);

    return NULL;
}

R3SurfelLabelerCheckpoint* UserSession::Iterator::checkpoint() const {
    if (!session_ || use_command_) return NULL;

    if (0 <= checkpoint_index_
            && checkpoint_index_ < session_->checkpoints.NEntries()) {
        return session_->checkpoints.Kth(checkpoint_index_);
    }

    return NULL;
}

RNScalar UserSession::Iterator::timestamp() const {
    if (command()) return command()->get_timestamp();
    else if (checkpoint()) return checkpoint()->get_timestamp();
    else return -0.0;
}

void UserSession::Iterator::adjust_time(RNScalar new_timestamp) {
    if (command()) command()->set_timestamp(new_timestamp);
    else if (checkpoint()) checkpoint()->set_timestamp(new_timestamp);
}

UserSession::Iterator& UserSession::Iterator::operator ++() {
    if (use_command_) cmd_index_++;
    else checkpoint_index_++;

    R3SurfelLabelerCommand *cmd = NULL;
    R3SurfelLabelerCheckpoint *checkpoint = NULL;
    if (0 <= cmd_index_ && cmd_index_ < session_->commands.NEntries()) cmd =
            session_->commands.Kth(cmd_index_);
    if (0 <= checkpoint_index_
            && checkpoint_index_ < session_->checkpoints.NEntries()) {
        checkpoint = session_->checkpoints.Kth(checkpoint_index_);
    }

    if (cmd && checkpoint) {
        use_command_ = !(cmd->get_timestamp() > checkpoint->get_timestamp());
    } else if (cmd) {
        use_command_ = true;
    } else if (checkpoint) {
        use_command_ = false;
    } else *this = end(session_);

    return *this;
}

bool UserSession::Iterator::operator !=(const Iterator& it) const {
    return !(*this == it);
}

bool UserSession::Iterator::operator ==(const Iterator& it) const {
    if (session_ != session_) return false;
    if (use_command_ != it.use_command_) return false;

    if (use_command_) return cmd_index_ == it.cmd_index_;
    else return checkpoint_index_ == it.checkpoint_index_;
}

} // namespace igra
