/* ---------------------------------------------------------------------------
 *  This software is in the public domain, furnished "as is", without technical
 *  support, and with no warranty, express or implied, as to its usefulness for
 *  any purpose. For more see LICENSE.txt that has to accompany this file.
 *
 *  igra.cpp
 *
 *  Copyright Aleksey Boyko, Princeton University, 2014.
 *
 *  Description: CLI for the IGRA viewer program.
 *
 *  -------------------------------------------------------------------------*/

#include <fstream>
#include <map>
#include <string>
#include <vector>

#include "fglut/fglut.h"
#include "Plotter.h"
#include "R3Graphics/R3Graphics.h"
#include "R3Surfels/R3Surfels.h"
#include "R3SurfelViewer.h"
#include "R3SurfelClassifier.h"
#include "R3SurfelLabeler.h"
#include "R3SurfelLabelerCommand.h"
#include "R3SurfelLabelerCheckpoint.h"
#include "UserSequencer.h"
#include "UserSession.h"

////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////

namespace igra {

// Program arguments
static char *scene_name = NULL;
static char *database_name = NULL;
static char *history_name = NULL;
extern int print_verbose; // = 0
static VisualizationStyle classifier_visual_style =
        R3_CLASSIFIER_VISUAL_STYLE_NONE;

static int user_study = 0;
static char* user_study_src = NULL;
static char* user_study_dest = NULL;

// Test scene's objects for being leaf nodes.
static int test_leaf_nodes = 0;

// Test users' confusion factors
static int test_user_agreement = 0;
static char* test_agreement_runs_path = NULL;

// Produce virtual user plots.
static int simulate_users = 0;
static char* simulate_users_file_prefix = NULL;
static int simulate_users_sample_size = 0;

typedef pair<RNScalar, RNScalar> RawSample;
typedef list<RawSample> RawSampleLine;
typedef list<RawSampleLine> RawSampleGroup;

// Classifier/labeler variables that can be modified with a cmd arg.
extern InterfaceModes interface_mode; // = UGOR_INTERFACE_FULL;

extern int use_al_for_group_seeding; // = 0;

extern int training_user_mode; // = 0

extern int auto_save; // = 0;

RNScalar t_g_marginal_alpha = 1.0;

extern bool limit_validation_time; // = false;
extern RNScalar time_limit; // = 0.0;

// Application variables

R3SurfelScene *scene = NULL;
R3SurfelLabeler *labeler = NULL;

// Glut variables

static int GLUTwindow = 0;
static int GLUTwindow_width = 1024;
static int GLUTwindow_height = 768;
extern bool update_main_window;

////////////////////////////////////////////////////////////////////////
// Surfel tree I/O Functions
////////////////////////////////////////////////////////////////////////

static R3SurfelScene * OpenScene(const char *scene_name, const char *db_name) {
    // Start statistics
    RNTime start_time;
    start_time.Read();

    // Allocate scene
    R3SurfelScene *scene = new R3SurfelScene();
    if (!scene) {
        fprintf(stderr, "Unable to allocate scene\n");
        return NULL;
    }

    // Open scene files
    string sc_acc(auto_save ? "r+" : "r");
    string db_acc("r");
    if (!scene->OpenFile(scene_name, db_name, sc_acc.c_str(), db_acc.c_str())) {
        delete scene;
        return NULL;
    }

    // Print statistics
    if (print_verbose) {
        printf("Opened scene ...\n");
        printf("  Time = %.2f seconds\n", start_time.Elapsed());
        printf("  # Objects = %d\n", scene->NObjects());
        printf("  # Labels = %d\n", scene->NLabels());
        printf("  # Assignments = %d\n", scene->NLabelAssignments());
        printf("  # Features = %d\n", scene->NFeatures());
        printf("  # Nodes = %d\n", scene->Tree()->NNodes());
        printf("  # Blocks = %d\n", scene->Tree()->Database()->NBlocks());
        printf("  # Surfels = %d\n", scene->Tree()->Database()->NSurfels());
        fflush(stdout);
    }

    // Return scene
    return scene;
}

static int CloseScene(R3SurfelScene *scene) {
    // Start statistics
    RNTime start_time;
    start_time.Read();

    // Print statistics
    if (print_verbose) {
        printf("Closing scene ...\n");
        printf("  Time = %.2f seconds\n", start_time.Elapsed());
        printf("  # Objects = %d\n", scene->NObjects());
        printf("  # Labels = %d\n", scene->NLabels());
        printf("  # Assignments = %d\n", scene->NLabelAssignments());
        printf("  # Features = %d\n", scene->NFeatures());
        printf("  # Nodes = %d\n", scene->Tree()->NNodes());
        printf("  # Blocks = %d\n", scene->Tree()->Database()->NBlocks());
        printf("  # Surfels = %ds\n", scene->Tree()->Database()->NSurfels());
        fflush(stdout);
    }

    // Close scene files
    if (!scene->CloseFile()) {
        delete scene;
        return 0;
    }

    delete scene;

    // Return success
    return 1;
}

////////////////////////////////////////////////////////////////////////
// GLUT callback functions
////////////////////////////////////////////////////////////////////////

void GLUTStop(void) {
    // Terminate labeler
    if (labeler) labeler->Terminate();

    // Close scene
    if (scene) CloseScene(scene);

    // Destroy window
    glutDestroyWindow(GLUTwindow);

    // Exit
    exit(0);
}

void GLUTRedraw(void) {
    // Redraw with labeler
    if (labeler->Redraw()) {
        glutPostRedisplay();
    }

    // Swap buffers
    glutSwapBuffers();
}

void GLUTResize(int w, int h) {
    // Remember window dimensions
    GLUTwindow_width = w;
    GLUTwindow_height = h;

    // Resize labeler
    if (labeler->Resize(w, h)) {
        glutPostRedisplay();
    }
}

void GLUTMouseMotion(int x, int y) {
    // Invert y coordinate
    y = GLUTwindow_height - y;

    // Send mouse motion to labeler
    if (labeler->MouseMotion(x, y)) {
        glutPostRedisplay();
    }
}

void GLUTPassiveMouseMotion(int x, int y) {
    // Invert y coordinate
    y = GLUTwindow_height - y;

    // Send mouse motion to labeler
    if (labeler->PassiveMouseMotion(x, y)) {
        glutPostRedisplay();
    }
}

void GLUTMouseButton(int button, int state, int x, int y) {
    // Invert y coordinate
    y = GLUTwindow_height - y;

    // Determine button status
    int b = (button == GLUT_LEFT_BUTTON) ?
            0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
    int s = (state == GLUT_DOWN) ? 1 : 0;

    // Determine modifiers
    int modifiers = glutGetModifiers();
    int shift = (modifiers & GLUT_ACTIVE_SHIFT);
    int ctrl = (modifiers & GLUT_ACTIVE_CTRL);
    int alt = (modifiers & GLUT_ACTIVE_ALT);

    // Send mouse event to labeler
    if ((b != button && labeler->MouseWheel(button, button - 3, x, y))
            || labeler->MouseButton(x, y, b, s, shift, ctrl, alt)) {
        glutPostRedisplay();
    }
}

void GLUTSpecial(int key, int x, int y) {
    // Invert y coordinate
    y = GLUTwindow_height - y;

    // Determine modifiers
    int modifiers = glutGetModifiers();
    int shift = (modifiers & GLUT_ACTIVE_SHIFT);
    int ctrl = (modifiers & GLUT_ACTIVE_CTRL);
    int alt = (modifiers & GLUT_ACTIVE_ALT);

    // Translate key
    int translated_key = key;
    switch (key) {
        case GLUT_KEY_DOWN:
            translated_key = R3_SURFEL_VIEWER_DOWN_KEY;
            break;
        case GLUT_KEY_UP:
            translated_key = R3_SURFEL_VIEWER_UP_KEY;
            break;
        case GLUT_KEY_LEFT:
            translated_key = R3_SURFEL_VIEWER_LEFT_KEY;
            break;
        case GLUT_KEY_RIGHT:
            translated_key = R3_SURFEL_VIEWER_RIGHT_KEY;
            break;
        case GLUT_KEY_F1:
            translated_key = R3_SURFEL_VIEWER_F1_KEY;
            break;
        case GLUT_KEY_F2:
            translated_key = R3_SURFEL_VIEWER_F2_KEY;
            break;
        case GLUT_KEY_F3:
            translated_key = R3_SURFEL_VIEWER_F3_KEY;
            break;
        case GLUT_KEY_F4:
            translated_key = R3_SURFEL_VIEWER_F4_KEY;
            break;
        case GLUT_KEY_F5:
            translated_key = R3_SURFEL_VIEWER_F5_KEY;
            break;
        case GLUT_KEY_F6:
            translated_key = R3_SURFEL_VIEWER_F6_KEY;
            break;
        case GLUT_KEY_F7:
            translated_key = R3_SURFEL_VIEWER_F7_KEY;
            break;
        case GLUT_KEY_F8:
            translated_key = R3_SURFEL_VIEWER_F8_KEY;
            break;
        case GLUT_KEY_F9:
            translated_key = R3_SURFEL_VIEWER_F9_KEY;
            break;
        case GLUT_KEY_F10:
            translated_key = R3_SURFEL_VIEWER_F10_KEY;
            break;
        case GLUT_KEY_F11:
            translated_key = R3_SURFEL_VIEWER_F11_KEY;
            break;
        case GLUT_KEY_F12:
            translated_key = R3_SURFEL_VIEWER_F12_KEY;
            break;
        case GLUT_KEY_HOME:
            translated_key = R3_SURFEL_VIEWER_HOME_KEY;
            break;
        case GLUT_KEY_END:
            translated_key = R3_SURFEL_VIEWER_END_KEY;
            break;
        case GLUT_KEY_INSERT:
            translated_key = R3_SURFEL_VIEWER_INSERT_KEY;
            break;
        case GLUT_KEY_PAGE_UP:
            translated_key = R3_SURFEL_VIEWER_PG_UP_KEY;
            break;
        case GLUT_KEY_PAGE_DOWN:
            translated_key = R3_SURFEL_VIEWER_PG_DWN_KEY;
            break;
    }

    // Send keyboard event to labeler
    if (labeler->Keyboard(x, y, translated_key, shift, ctrl, alt)) {
        glutPostRedisplay();
    }
}

void GLUTKeyboard(unsigned char key, int x, int y) {
    // Invert y coordinate
    y = GLUTwindow_height - y;

    // Determine modifiers
    int modifiers = glutGetModifiers();
    int shift = (modifiers & GLUT_ACTIVE_SHIFT);
    int ctrl = (modifiers & GLUT_ACTIVE_CTRL);
    int alt = (modifiers & GLUT_ACTIVE_ALT);

    // Convert control keys
    int translated_key = key;
    if ((key >= 1) && (key <= 26)) {
        if (shift) translated_key += 64;
        else translated_key += 96;
        ctrl = GLUT_ACTIVE_CTRL;
    }

    // Send keyboard event to labeler
    if (labeler->Keyboard(x, y, translated_key, shift, ctrl, alt)) {
        glutPostRedisplay();
    }
}

// Menu items
enum {
    UGOR_NOP_COMMAND,
    UGOR_FILE_SAVE_COMMAND,
    UGOR_FILE_EXIT_COMMAND,
    UGOR_EDIT_UNDO_COMMAND,
    UGOR_EDIT_REDO_COMMAND,
    UGOR_EDIT_CONFIRM_LABEL_COMMAND,
    UGOR_SELECT_SUGGESTED_COMMAND,
    UGOR_VIEW_ZOOM1_COMMAND,
    UGOR_VIEW_ZOOM2_COMMAND,
    UGOR_VIEW_ZOOM3_COMMAND,
    UGOR_VIEW_ZOOM4_COMMAND,
    UGOR_VIEW_BACKGROUND_COMMAND,
    UGOR_VIEW_LABEL_NAMES_COMMAND,
    UGOR_VIEW_CONFIDENCE_COMMAND,
    UGOR_NUM_COMMANDS
};

void GLUTMenu(int cmd) {
    int redraw = 1;

    // This is a hack!?!?!
    switch (cmd) {
        case UGOR_FILE_SAVE_COMMAND:
            labeler->Keyboard(0, 0, 'S', 0, 1, 0);
            break;
        case UGOR_FILE_EXIT_COMMAND:
            exit(0);
        case UGOR_EDIT_UNDO_COMMAND:
            labeler->Keyboard(0, 0, 'Z', 0, 1, 0);
            break;
        case UGOR_EDIT_REDO_COMMAND:
            labeler->Keyboard(0, 0, 'Y', 0, 1, 0);
            break;
        case UGOR_EDIT_CONFIRM_LABEL_COMMAND:
            labeler->Keyboard(0, 0, ' ', 0, 0, 0);
            break;
        case UGOR_SELECT_SUGGESTED_COMMAND:
            labeler->Keyboard(0, 0, 'M', 0, 1, 0);
            break;
        case UGOR_VIEW_ZOOM1_COMMAND:
            labeler->Keyboard(0, 0, R3_SURFEL_VIEWER_F1_KEY, 0, 0, 0);
            break;
        case UGOR_VIEW_ZOOM2_COMMAND:
            labeler->Keyboard(0, 0, R3_SURFEL_VIEWER_F2_KEY, 0, 0, 0);
            break;
        case UGOR_VIEW_ZOOM3_COMMAND:
            labeler->Keyboard(0, 0, R3_SURFEL_VIEWER_F3_KEY, 0, 0, 0);
            break;
        case UGOR_VIEW_ZOOM4_COMMAND:
            labeler->Keyboard(0, 0, R3_SURFEL_VIEWER_F4_KEY, 0, 0, 0);
            break;
        case UGOR_VIEW_BACKGROUND_COMMAND:
            labeler->Keyboard(0, 0, 'B', 0, 1, 0);
            break;
        case UGOR_VIEW_LABEL_NAMES_COMMAND:
            labeler->Keyboard(0, 0, 'L', 0, 1, 0);
            break;
        case UGOR_VIEW_CONFIDENCE_COMMAND:
            labeler->Keyboard(0, 0, 'C', 0, 1, 0);
            break;
        default:
            redraw = 0;
            break;
    }

    // Redraw
    if (redraw) glutPostRedisplay();
}

void GLUTTimer(int value) {
    // React to animation timer triggered.
    if (labeler && value == labeler->AnimationTimerID()) {
        if (labeler->AdvanceAnimation()) glutPostWindowRedisplay(GLUTwindow);
        glutTimerFunc(1000 / labeler->AnimationFPS(), GLUTTimer,
                labeler->AnimationTimerID());
    } else if (!value) {
        if (update_main_window) {
            // Check if the window has been requested for update externally
            // (e.g. from a debug window).
            glutPostWindowRedisplay(GLUTwindow);
            update_main_window = false;
        }
        glutTimerFunc(100, GLUTTimer, 0);
    }
}

void GLUTInit(int *argc, char **argv) {
    // Open window
    glutInit(argc, argv);
    glutInitWindowPosition(100, 100);
    glutInitWindowSize(GLUTwindow_width, GLUTwindow_height);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
    GLUTwindow = glutCreateWindow(
            "IGRA: Intelligent Grouping for Group Annotation");

#if 0
    // Create menu
    int file_menu = glutCreateMenu(GLUTMenu);
    glutAddMenuEntry("Save (Ctrl-S)", UGOR_FILE_SAVE_COMMAND);
    glutAddMenuEntry(" ", UGOR_NOP_COMMAND);
    glutAddMenuEntry("Exit", UGOR_FILE_EXIT_COMMAND);
    int edit_menu = glutCreateMenu(GLUTMenu);
    glutAddMenuEntry("Undo (Ctrl-Z)", UGOR_EDIT_UNDO_COMMAND);
    glutAddMenuEntry("Redo (Ctrl-Y)", UGOR_EDIT_REDO_COMMAND);
    glutAddMenuEntry(" ", UGOR_NOP_COMMAND);
    glutAddMenuEntry("Confirm label (Enter)", UGOR_EDIT_CONFIRM_LABEL_COMMAND);
    int view_menu = glutCreateMenu(GLUTMenu);
    glutAddMenuEntry("Zoom   1X (F1)", UGOR_VIEW_ZOOM1_COMMAND);
    glutAddMenuEntry("Zoom   5X (F2)", UGOR_VIEW_ZOOM2_COMMAND);
    glutAddMenuEntry("Zoom  25X (F3)", UGOR_VIEW_ZOOM3_COMMAND);
    glutAddMenuEntry("Zoom 125X (F4)", UGOR_VIEW_ZOOM4_COMMAND);
    glutAddMenuEntry(" ", UGOR_NOP_COMMAND);
    glutAddMenuEntry("Toggle Background (Ctrl-B)", UGOR_VIEW_BACKGROUND_COMMAND);
    glutAddMenuEntry("Toggle Label Names (Ctrl-L)", UGOR_VIEW_LABEL_NAMES_COMMAND);
    glutAddMenuEntry("Toggle Confidences (Ctrl-C)", UGOR_VIEW_CONFIDENCE_COMMAND);
    glutCreateMenu(GLUTMenu);
    glutAddSubMenu("File", file_menu);
    glutAddSubMenu("Edit", edit_menu);
    glutAddSubMenu("View", view_menu);
    glutAttachMenu(GLUT_RIGHT_BUTTON);
#endif

    // Initialize GLUT callback functions
    glutDisplayFunc(GLUTRedraw);
    glutReshapeFunc(GLUTResize);
    glutKeyboardFunc(GLUTKeyboard);
    glutSpecialFunc(GLUTSpecial);
    glutMouseFunc(GLUTMouseButton);
    glutMotionFunc(GLUTMouseMotion);
    glutPassiveMotionFunc(GLUTPassiveMouseMotion);
    atexit(GLUTStop);

    // Initialize labeler
    labeler->Initialize(classifier_visual_style);

    // Initialize timers.
    glutTimerFunc(100, GLUTTimer, 0);
    glutTimerFunc(1000 / labeler->AnimationFPS(), GLUTTimer,
            labeler->AnimationTimerID());
}

void GLUTMainLoop(void) {
    // Run main loop -- never returns
    glutMainLoop();
}

////////////////////////////////////////////////////////////////////////
// Write offline test results functionality.
////////////////////////////////////////////////////////////////////////

void WriteLeafTest() {
    assert(scene);

    unsigned int leaf_num = 0, non_leaf_num = 0, obj_num = 0;

    for (int i = 1; i < scene->NObjects(); ++i) {
        R3SurfelObject *object = scene->Object(i);
        obj_num++;
        for (int j = 0; j < object->NNodes(); ++j) {
            R3SurfelNode *node = object->Node(j);
            if (node->NParts() > 0) non_leaf_num++;
            else leaf_num++;
        }
    }

    printf("Result: %d objects; %d non leaf nodes; %d leaf nodes;", obj_num,
            non_leaf_num, leaf_num);
}

int GetHistFilenames(const char* dir, vector<string> &filenames) {
    filenames.clear();

    // Check all.
    if (!system(NULL)) {
        fprintf(stderr, "Command processor is not available. Terminating\n");
        return 0;
    }

    // Request a directory listing into a temp file.
    string tempFile(dir);
    tempFile += "/temp.txt";
    char buffer[1024];
    sprintf(buffer, "ls %s/*.hist > %s", dir, tempFile.c_str());
    system(buffer);

    // Read the listing file entries into the vectors.
    ifstream listFile(tempFile.c_str(), ios_base::in);
    if (!listFile.is_open() || listFile.fail()) {
        fprintf(stderr, "Could not get a list of .ssa files in %s", dir);
        return 0;
    }

    char filename[1024];
    while (!listFile.eof()) {
        //get next file name
        listFile.getline(filename, 1024);

        if (listFile.eof()) break;

        if (listFile.fail()) {
            fprintf(stderr, "Failed reading line");
            break;
        }
        filenames.push_back(filename);
    }
    listFile.close();

    // Return success.
    return 1;
}

int FastForwardHist(R3SurfelScene *s, const char* filename) {
    // Check everything.
    if (!s || !filename) return 0;

    // Open history file.
    ifstream histFile(filename, ios_base::in);
    if (!histFile.is_open() || histFile.fail()) {
        fprintf(stderr, "Could not open %s", filename);
        return 0;
    }

    // Read until the very last CHECKPOINT.
    char buffer[1048576];
    while (!histFile.eof()) {
        histFile.getline(buffer, 1048576);
        if (histFile.eof()) break;
        if (histFile.fail()) {
            fprintf(stderr, "Failed reading line\n");
            break;
        }

        char buf[1024];
        sscanf(buffer, "%s", buf);
        if (strcmp(buf, "CHECKPOINT")) continue;

        // Process CHECKPOINT.
        char* bufferp = strtok(buffer, " \n");	// skip CHECKPOINT keyword
        bufferp = strtok(NULL, " \n");	// timestamp
        bufferp = strtok(NULL, " \n");	// numlabels
        bufferp = strtok(NULL, " \n");	// numobjects
        bufferp = strtok(NULL, " \n");	// nconfirmed
        bufferp = strtok(NULL, " \n");	// npredicted
        bufferp = strtok(NULL, " \n");	// pred_precision
        bufferp = strtok(NULL, " \n");	// pred_recall
        bufferp = strtok(NULL, " \n");	// pred_F
        bufferp = strtok(NULL, " \n");	// hum_prec
        bufferp = strtok(NULL, " \n");	// hum_rec
        bufferp = strtok(NULL, " \n");	// hum_F

        //Read labels
        map<int, R3SurfelLabel*> label_map;
        for (int j = 0; j < s->NLabels(); j++) {
            bufferp = strtok(NULL, " \n");	// L
            bufferp = strtok(NULL, " \n");	// id
            int id = atoi(bufferp);
            bufferp = strtok(NULL, " \n");	// name
            R3SurfelLabel *label = s->FindLabelByName(bufferp);
            if (label != s->Label(id))
                fprintf(stderr,
                        "Mismatching id (%d) for label %s (in scene id is %d)\n",
                        id, bufferp, label->SceneIndex());
            if (label) label_map[id] = label;
            else fprintf(stderr, "Scene is missing label %s\n", bufferp);
        }

        // Read object assignments
        for (int i = 1; i < scene->NObjects(); i++) {
            R3SurfelObject *object = s->Object(i);
            bufferp = strtok(NULL, " \n"); //O
            bufferp = strtok(NULL, " \n"); //objectIndex
            if (atoi(bufferp) != i)
                fprintf(stderr, "Mismatching id (%d) for object %d\n",
                        atoi(bufferp), i);
            bufferp = strtok(NULL, " \n"); //objectName
            bufferp = strtok(NULL, " \n"); //gtLabelIndex
            bufferp = strtok(NULL, " \n"); //humLabelIndex
            int hum_lab_id = atoi(bufferp);
            bufferp = strtok(NULL, " \n"); //predLabelIndex
            bufferp = strtok(NULL, " \n"); //predConfidence

            if (hum_lab_id > 0) {
                R3SurfelLabel *hum_lab = label_map[hum_lab_id];
                if (!hum_lab)
                    fprintf(stderr, "No label for id %d\n", hum_lab_id);
                R3SurfelLabelAssignment* hum_ass =
                        object->HumanLabelAssignment();
                if (hum_ass) {
                    s->RemoveLabelAssignment(hum_ass);
                    delete hum_ass;
                }
                hum_ass = new R3SurfelLabelAssignment(object, hum_lab, 1, 0);
                s->InsertLabelAssignment(hum_ass);
            }
        }
    }
    histFile.close();

    // Return success.
    return 1;
}

void WriteUserAgreementTest() {
    // Check all.
    if (!scene) return;
    if (scene->NObjects() < 2) return;
    if (!test_agreement_runs_path) return;

    // Get all user files.
    vector<string> filenames;
    char ssa_name[1024], ssb_name[1024];
    sprintf(ssa_name, "%s/eval.ssa", test_agreement_runs_path);
    sprintf(ssb_name, "%s/eval.ssb", test_agreement_runs_path);
    if (!GetHistFilenames(test_agreement_runs_path, filenames)) {
        fprintf(stderr,
                "WriteUserAgreementTest() failed to get list of hist files.\n");
        return;
    }

    const int N = scene->NObjects();
    vector<RNScalar> correct_votes_num(N, 0);
    vector<int> total_votes_num(N, 0);
    multimap<int, map<string, int> > assigned_labels;
    for (int i = 0; (size_t) i < filenames.size(); ++i) {
        // Acquire their human labeling.
        R3SurfelScene *scene_temp = OpenScene(ssa_name, ssb_name);
        printf("Fast-forwarding %s...", filenames[i].c_str());
        if (!FastForwardHist(scene_temp, filenames[i].c_str())) {
            fprintf(stderr, "Failed to fast forward %s.\n",
                    filenames[i].c_str());
            continue;
        }
        printf("Done\n");

        // Compute confusion.
        for (int j = 1; j < N; ++j) {
            R3SurfelLabel *gt_label = scene->Object(j)->GroundTruthLabel();
            R3SurfelLabel *human_label = scene_temp->Object(j)->HumanLabel();
            if (human_label) {
                total_votes_num[j]++;
                if (!strcmp(gt_label->Name(), human_label->Name()))
                    correct_votes_num[j]++;
            }
        }

        // clean-up.
        delete scene_temp;
    }

    // Rank by confusion.
    multimap<RNScalar, R3SurfelObject*> confusion_map;
    for (int j = 1; j < N; ++j) {
        R3SurfelObject* object = scene->Object(j);
        RNScalar rating =
                (total_votes_num[j] > 0) ?
                        correct_votes_num[j] / total_votes_num[j] : -1;
        confusion_map.insert(pair<RNScalar, R3SurfelObject*>(rating, object));
    }

    RNScalar consesus_level = 0.1;
    RNScalar level_step = 0.1;
    RNScalar below_num = 0;
    RNScalar factor = 1. / (N - 1);

    // Print confusion.
    printf("\nRanked by confusion\n==============\n");
    printf("%20s\t%20s\t%20s\n\n", "object index", "label name", "consistency");
    for (multimap<RNScalar, R3SurfelObject*>::const_iterator it =
            confusion_map.begin(); it != confusion_map.end(); ++it) {
        ++below_num;
        if (it->first > consesus_level) {
            printf("---------"
                    "Achievable accuracy is %f (Consistency threshold = %f)"
                    "---------\n", (N - below_num) * factor, consesus_level);
            consesus_level += level_step;
        }
        printf("%20d\t%20s\t%20f\n", it->second->SceneIndex(),
                it->second->GroundTruthLabel()->Name(), it->first);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Human user plotting.
////////////////////////////////////////////////////////////////////////////////

// Reads the file generated outside that
// contains information about users' categories, id's and
// the files that logged their respective sessions.
int ReadExperimentFilesFile(const char* source_file,
        map<string, map<string, list<string> > > &entries) {
    if (!source_file) return 0;

    // Open source file.
    FILE *src = fopen(source_file, "r");
    if (!src) {
        perror(source_file);
        return 0;
    }

    // Read source file.
    entries.clear();
    map<string, list<string> > *category = NULL;
    list<string> *user_files = NULL;

    int line_size = 1024;
    char *line = new char[line_size];
    char *key = new char[line_size];
    char *value = new char[line_size];
    while (fgets(line, line_size, src) != NULL) {
        sscanf(line, "%s %s", key, value);

        if (!strcmp(key, "category")) {
            assert(value && strlen(value));
            category = &(entries[string(value)]);
        } else if (!strcmp(key, "user")) {
            assert(value && strlen(value));
            assert(category);
            user_files = &((*category)[string(value)]);
        } else if (!strcmp(key, "file")) {
            assert(value && strlen(value));
            assert(category);
            assert(user_files);
            user_files->push_back(value);
        } else {
            fprintf(stderr, "Source file %s is incorrectly formatted.",
                    source_file);
            return 0;
        }
    }

    // Clean up.
    delete[] line;
    delete[] key;
    delete[] value;
    fclose(src);

    if (print_verbose) {
        // check everything,
        printf("Read experiment files file results:\n");
        for (map<string, map<string, list<string> > >::const_iterator cit =
                entries.begin(); cit != entries.end(); ++cit) {
            printf("Category %s has %zu entries:\n", cit->first.c_str(),
                    cit->second.size());
            for (map<string, list<string> >::const_iterator uit =
                    cit->second.begin(); uit != cit->second.end(); ++uit) {
                printf("\tUser %s ran %zu sessions saved in:\n",
                        uit->first.c_str(), uit->second.size());
                for (list<string>::const_iterator fit = uit->second.begin();
                        fit != uit->second.end(); ++fit) {
                    printf("\t\t%s\n", fit->c_str());
                }
            }
        }
    }

    return 1;
}

// Parses the contents of file fname into a session container.
int ReadHistoryFile(const char *fname, UserSession *session) {
    // Check everything.
    assert(fname);
    assert(session);

    // Prepare session.
    session->clear();

    // Open log file.
    FILE *fp = fopen(fname, "rt");
    if (!fp) return 0;

    // Read log file.
    int line = 1;
    while (true) {
        if (feof(fp) || ferror(fp)) break;

        // Attempt reading as a comand.
        R3SurfelLabelerCommand *cmd = R3SurfelLabelerCommand::TryParseCommand(
                fp, scene);
        if (cmd) {
            session->push_back(cmd);
        } else if (feof(fp) || ferror(fp)) {
            break;
        } else {
            // Attempt reading as a checkpoint.
            R3SurfelLabelerCheckpoint *checkpoint =
                    R3SurfelLabelerCheckpoint::TryParseCheckPoint(fp, scene);
            if (checkpoint) session->push_back(checkpoint);
            else if (feof(fp) || ferror(fp)) break;
            else RNAbort("Odd entry in the log file %s, line %d\n", fname,
                    line);

        }
        line++;
    }

    session->check_validity();

    // Clean up.
    fclose(fp);

    // Return success.
    return 1;
}

// Concatenates sessions into final_session into a single session,
// accounting for and removing pauses and stop/start breaks between sessions.
void MergeSessions(RNArray<UserSession*> &sessions,
        UserSession &final_session) {
    // Clear result.
    final_session.clear();

    // Merge.
    for (int i = 0; i < sessions.NEntries(); ++i) {
        UserSession *s = sessions.Kth(i);
        if (!s || s->empty()) continue;

        // Useful timing trackers.
        RNScalar t_end_prev_session = 0.0;
        final_session.trim_to_last_checkpoint();
        t_end_prev_session = final_session.back().timestamp();

        RNScalar t_init_session = 0.0;
        RNScalar t_paused_curr_session = 0.0;
        bool paused = false;
        RNScalar pause_begin_time = 0.0;
        for (UserSession::Iterator it = s->begin(); it != s->end(); ++it) {

            RNScalar t_curr = it.timestamp();
            bool pause_command = false;
            R3SurfelLabelerCommand *cmd = it.command();

            if (cmd) {
                switch (cmd->get_type()) {
                    case R3_SURFEL_LABELER_INITIALIZE_COMMAND:
                        t_init_session = t_curr;
                        break;
                    case R3_SURFEL_LABELER_TERMINATE_COMMAND:
                        break;
                    case R3_SURFEL_LABELER_PAUSE:
                        pause_command = true;
                        paused = !paused;
                        assert(paused == (bool )(cmd->get_operand()));
                        if (paused) pause_begin_time = t_curr;
                        else t_paused_curr_session +=
                                (t_curr - pause_begin_time);
                        break;
                }
            }

            // Adjust for a single continuous time line.
            RNScalar t_continuous = t_curr - t_init_session
                    - t_paused_curr_session + t_end_prev_session;

            // If still paused from last pause command - adjust for that.
            if (paused) t_continuous -= (t_curr - pause_begin_time);

            it.adjust_time(t_continuous);

            if (pause_command) {
                delete cmd;
            } else {
                final_session.push_back(it);
            }
        }

        // Clear the sessions but do not deallocate, since the
        // ownership is now transferred to final_session.
        s->clear(false);
    }

    final_session.check_validity();
}

// Reads users' logs, concatenates files for multiple session users
// and streamlines the times into a single session with respect to
// interruptions, crashes and pauses.
int ReadUserLogs(
        const map<string, map<string, list<string> > > &categorized_source,
        map<string, map<string, UserSession> > &user_sessions) {
    user_sessions.clear();

    int read_result = 1;

    // Loop through categories.
    for (map<string, map<string, list<string> > >::const_iterator cit =
            categorized_source.begin();
            read_result && cit != categorized_source.end(); ++cit) {
        string category = cit->first;
        map<string, UserSession> &users = user_sessions[category];

        // Loop through users.
        for (map<string, list<string> >::const_iterator uit =
                cit->second.begin(); read_result && uit != cit->second.end();
                ++uit) {
            string user_id = uit->first;
            RNArray<UserSession*> sessions;

            // Loop through files.
            for (list<string>::const_iterator fit = uit->second.begin();
                    read_result && fit != uit->second.end(); ++fit) {

                string hist_file = fit->c_str();

                printf("Reading file %s  <- user %s, category %s\n",
                        hist_file.c_str(), user_id.c_str(), category.c_str());

                UserSession *session = new UserSession();
                read_result *= ReadHistoryFile(hist_file.c_str(), session);

                sessions.InsertTail(session);
            } // Loop through files - end.

            // Join and clean sessions from time breaks.
            UserSession &final_session = users[user_id];
            MergeSessions(sessions, final_session);
        } // Loop through users - end.
    } // Loop through categories - end.

    return read_result;
}

void PlotUserSessions(map<string, map<string, UserSession> > &user_sessions,
        const char *prefix) {
    //Check everything.
    if (!prefix) return;

    UserSequencer *sequencer;
    map<string, map<string, SequencesForPlotting*> > user_sequences;

    // Loop over categories.
    for (map<string, map<string, UserSession> >::iterator cit =
            user_sessions.begin(); cit != user_sessions.end(); ++cit) {
        const string &category = cit->first;
        map<string, UserSession> &users = cit->second;
        map<string, SequencesForPlotting*> &category_sequences =
                user_sequences[category];
        printf("Processing category \"%s\"\n", category.c_str());

        // Loop over users.
        for (map<string, UserSession>::iterator uit = users.begin();
                uit != users.end(); ++uit) {
            const string &user = uit->first;
            UserSession &session = uit->second;

            string title = category + "_" + user;

            printf("\tProcessing \"%s\"\n", title.c_str());

            SequencesForPlotting *seq = new SequencesForPlotting(title.c_str());
            sequencer = new HumanUserSequencer(scene, seq, &session);
            sequencer->ProduceSequence();
            category_sequences[user] = seq;
            delete sequencer;

        } // Loop over users - end.

        printf("Done\n");
    } // Loop over categories - end.

    /////////////////////
    // Plot everything.//
    /////////////////////

    RNArray<SequencesForPlotting*> category_sequences;

    for (map<string, map<string, SequencesForPlotting*> >::iterator cit =
            user_sequences.begin(); cit != user_sequences.end(); ++cit) {
        const string &category = cit->first;
        map<string, SequencesForPlotting*> &users = cit->second;

        printf("Plotting users in category \"%s\"\n", category.c_str());

        RNArray<SequencesForPlotting*> user_pointers;

        // Loop over users.
        for (map<string, SequencesForPlotting*>::iterator uit = users.begin();
                uit != users.end(); ++uit) {
            const string &user = uit->first;
            SequencesForPlotting *seq = uit->second;

            string prefix(user_study_dest);
            prefix = prefix + "_" + category + "_" + user;

            printf("\tPlotting user %s to %s\n", user.c_str(), prefix.c_str());

            SequencesForPlotting::plot_alone(prefix.c_str(), "", seq);
            user_pointers.Insert(seq);
        }

        // Merge all users in category.
        SequencesForPlotting *sfp = new SequencesForPlotting(category.c_str());
        sfp->merge(user_pointers);

        string prefix(user_study_dest);
        prefix = prefix + "_" + category;
        printf("Plotting category \"%s\" to %s\n", category.c_str(),
                prefix.c_str());
        char postfix[50];
        sprintf(postfix, "%zd_users", users.size());
        SequencesForPlotting::plot_alone(prefix.c_str(), postfix, sfp);

        category_sequences.InsertHead(sfp);

        // Clean up.
        for (int i = 0; i < user_pointers.NEntries(); ++i)
            delete user_pointers[i];

        printf("Done\n");
    }

    if (1) {
        // All categories together on one plot.
        string prefix_temp = prefix;
        prefix_temp = prefix_temp + "_together";
        SequencesForPlotting::plot_alone(prefix_temp.c_str(), "",
                category_sequences);
    }

    // Clean-up.
    for (int i = 0; i < category_sequences.NEntries(); i++)
        delete category_sequences[i];
}

void PlotHumanUsers() {
    // Check everything.
    if (!user_study || !user_study_src || !user_study_dest) return;
    setbuf(stdout, 0);

    // Read input file structure.
    map<string, map<string, list<string> > > categorized_source; // category->user_id->log files
    if (!ReadExperimentFilesFile(user_study_src, categorized_source)) return;

    // Read the log files.
    map<string, map<string, UserSession> > user_sessions; // category -> user_id -> user session
    if (!ReadUserLogs(categorized_source, user_sessions)) return;

    PlotUserSessions(user_sessions, user_study_dest);
}

////////////////////////////////////////////////////////////////////////////////
// Simulated Users plotting
////////////////////////////////////////////////////////////////////////////////

void CreateSnapshot(RNArray<R3SurfelLabelAssignment*> &snapshot) {
    for (int i = 0; i < scene->NLabelAssignments(); ++i) {
        R3SurfelLabelAssignment *a = scene->LabelAssignment(i);
        snapshot.Insert(new R3SurfelLabelAssignment(*a));
    }
}

void ReinstallSnapshot(const RNArray<R3SurfelLabelAssignment*> &snapshot) {
    while (scene->NLabelAssignments()) {
        int index = scene->NLabelAssignments() - 1;
        R3SurfelLabelAssignment *a = scene->LabelAssignment(index);
        scene->RemoveLabelAssignment(a);
        delete a;
    }
    for (int k = 0; k < snapshot.NEntries(); ++k) {
        R3SurfelLabelAssignment *a = new R3SurfelLabelAssignment(
                *snapshot.Kth(k));
        scene->InsertLabelAssignment(a);
    }
}

void ReleaseSnapshot(RNArray<R3SurfelLabelAssignment*> &snapshot) {
    for (int i = 0; i < snapshot.NEntries(); ++i) {
        R3SurfelLabelAssignment *a = snapshot.Kth(i);
        delete a;
    }
    snapshot.Empty(true);
}

void PlotSimulatedUsers() {

    // Check everything.
    if (!simulate_users_file_prefix || simulate_users_sample_size < 1) {
        fprintf(stderr, "Incompatible parameters for plotting virtual users.");
        return;
    }

    // Make sure progress printouts are immediately flushed.
    srand(time(NULL));
    setbuf(stdout, NULL);

    // Take initial assignment snapshot.
    RNArray<R3SurfelLabelAssignment*> snapshot;
    CreateSnapshot(snapshot);

    VirtualUserSequencer *sequencer;

    //////////////////
    // No learning. //
    //////////////////

    SequencesForPlotting no_ml("no_learning");
    bool use_no_ml = true;
    if (use_no_ml) {
        RNArray<SequencesForPlotting*> sequences;

        for (int i = 0; i < simulate_users_sample_size; ++i) {
            printf("\rSampling no learning curve...%d%%",
                    (int) (100.0 * i / simulate_users_sample_size));

            ReinstallSnapshot(snapshot);
            SequencesForPlotting *seq = new SequencesForPlotting("no_learning");
            sequencer = new NoMLSequencer(scene, seq);
            sequencer->ProduceSequence();
            sequences.InsertTail(seq);
            delete sequencer;
        }
        no_ml.merge(sequences);
        for (int i = 0; i < sequences.NEntries(); ++i)
            delete sequences.Kth(i);
        printf("Done\n");
    }

    /////////////////////////////
    // Active learning 1-by-1. //
    /////////////////////////////

    const int config_single_num = 8;
    const RNFlags config_single_param[config_single_num] = { RNFlags(
    SUGGEST_SINGLE_RANDOM), RNFlags(SUGGEST_SINGLE_CONFIDENCE), RNFlags(
    SUGGEST_SINGLE_ENTROPY), RNFlags(
    SUGGEST_SINGLE_CONFIDENCE | SUGGEST_SINGLE_ENTROPY), RNFlags(
    SUGGEST_SINGLE_DENSITY_WEIGHT), RNFlags(
    SUGGEST_SINGLE_DENSITY_WEIGHT | SUGGEST_SINGLE_CONFIDENCE), RNFlags(
    SUGGEST_SINGLE_DENSITY_WEIGHT | SUGGEST_SINGLE_ENTROPY), RNFlags(
            SUGGEST_SINGLE_DENSITY_WEIGHT | SUGGEST_SINGLE_CONFIDENCE
                    | SUGGEST_SINGLE_ENTROPY) };
    const char* config_single_name[config_single_num] = { "random", "min_conf",
            "max_entropy", "max_entropy_conf", "max_density", "max_dw_conf",
            "max_dw_entropy", "max_dw_entropy_conf" };

    RNArray<SequencesForPlotting*> singular_seq;
    bool use_singular = true;
    if (use_singular) {

        // Loop over config state.
        for (int config_state = 0; config_state < config_single_num;
                ++config_state) {

            const char *title = config_single_name[config_state];

            RNArray<SequencesForPlotting*> sequences;
            for (int i = 0; i < 10 * simulate_users_sample_size; ++i) {

                printf("\rSampling active_single config \"%s\" (%d of %d), "
                        "curve...%d of %d", title, config_state + 1,
                        config_single_num, i + 1, simulate_users_sample_size);

                ReinstallSnapshot(snapshot);

                SequencesForPlotting *seq = new SequencesForPlotting(title);
                sequencer = new SingularSuggestSequencer(scene, seq,
                        config_single_param[config_state]);
                sequencer->ProduceSequence();
                sequences.InsertTail(seq);
                delete sequencer;
            }
            SequencesForPlotting *sfp = new SequencesForPlotting(
                    config_single_name[config_state]);
            sfp->merge(sequences);
            singular_seq.InsertTail(sfp);
            for (int i = 0; i < sequences.NEntries(); ++i)
                delete sequences.Kth(i);
            printf("Done\n");
        } // Loop over config state - end.
    }

    ////////////////////////////////////////////////////////////////////////////
    // Group active learning (best benefit and seeded by single and expanded).//
    ////////////////////////////////////////////////////////////////////////////
    const char* config_group_name[2] = { "benefit", "seeded" };
#define IGRA_ONLY_PLOT_BENEFIT 1
#if !IGRA_ONLY_PLOT_BENEFIT
    const int config_group_num = 2;
    const int grp_config_size = config_group_num + config_single_num - 1;
#else
    const int grp_config_size = 1;
#endif

    RNArray<SequencesForPlotting*> group_seq;
    bool use_group = true;
    if (use_group) {
        // Loop over config state.
        for (int config_state = 0; config_state < grp_config_size;
                ++config_state) {

            int grp_cfg_index = (config_state ? 1 : 0);
            int seed_cfg_index = config_state - 1;
            string title(config_group_name[grp_cfg_index]);
            if (seed_cfg_index >= 0)
                title = title + "_" + config_single_name[seed_cfg_index];

            RNArray<SequencesForPlotting*> sequences;
            for (int i = 0; i < simulate_users_sample_size; ++i) {

                printf("\rSampling active_group config \"%s\"("
                        "grp_state = %d, seed_state = %d) (%d of %d), "
                        "curve...%d of %d", title.c_str(), grp_cfg_index,
                        seed_cfg_index, config_state + 1, grp_config_size,
                        i + 1, simulate_users_sample_size);

                ReinstallSnapshot(snapshot);

                SequencesForPlotting *seq = new SequencesForPlotting(
                        title.c_str());
                if (grp_cfg_index) {
                    sequencer = new GroupSeededSuggestSequencer(scene, seq,
                            config_single_param[config_state]);
                } else {
                    sequencer = new GroupSuggestSequencer(scene, seq,
                            config_single_param[config_state]);
                }
                sequencer->ProduceSequence();
                sequences.InsertTail(seq);
                delete sequencer;
            }

            SequencesForPlotting *sfp = new SequencesForPlotting(title.c_str());
            sfp->merge(sequences);
            group_seq.InsertTail(sfp);
            for (int i = 0; i < sequences.NEntries(); ++i)
                delete sequences.Kth(i);
            printf("Done\n");

        } // Loop over config state - end.
    }

    ReleaseSnapshot(snapshot);

    ////////////////////////////
    // Produce plots.
    ////////////////////////////
    char postfix[1024];
    sprintf(postfix, "%d_runs_alpha_%f", simulate_users_sample_size,
            t_g_marginal_alpha);
    string prefix(simulate_users_file_prefix);

    /////////////////////
    // Draw own plots.
    /////////////////////
    if (0 && use_no_ml) {
        string prefix_temp = prefix + "_no_ml";
        SequencesForPlotting::plot_alone(prefix_temp.c_str(), postfix, &no_ml);
    }

    if (0 && use_singular) {
        for (int i = 0; i < config_single_num; ++i) {
            string prefix_temp = prefix + "_" + config_single_name[i];
            SequencesForPlotting::plot_alone(prefix_temp.c_str(), postfix,
                    singular_seq[i]);
        }
    }

    if (0 && use_group) {
        for (int i = 0; i < group_seq.NEntries(); ++i) {

            int grp_cfg_index = (i ? 1 : 0);
            int seed_cfg_index = i - 1;
            string title(config_group_name[grp_cfg_index]);
            if (seed_cfg_index >= 0)
                title = title + "_" + config_single_name[seed_cfg_index];

            string prefix_temp = prefix + "_" + title;
            SequencesForPlotting::plot_alone(prefix_temp.c_str(), postfix,
                    group_seq[i]);
        }
    }

    /////////////////////
    // Add mixed plots.
    /////////////////////
    if (1) {
        // Singular plots all together.
        RNArray<SequencesForPlotting*> seq_group;
        seq_group.Insert(&no_ml);
        for (int i = 0; i < config_single_num; ++i)
            seq_group.Insert(singular_seq[i]);

        string prefix_temp = prefix + "_all_singulars";
        SequencesForPlotting::plot_alone(prefix_temp.c_str(), postfix,
                seq_group);
    }

    if (1) {
        // Group plots all together.
        RNArray<SequencesForPlotting*> seq_group;
        seq_group.Insert(&no_ml);
        for (int i = 0; i < group_seq.NEntries(); ++i)
            seq_group.Insert(group_seq[i]);

        string prefix_temp = prefix + "_all_group";
        SequencesForPlotting::plot_alone(prefix_temp.c_str(), postfix,
                seq_group);
    }

    if (1) {
        // No ML + confidence singular + benefit group.
        RNArray<SequencesForPlotting*> seq_group;
        seq_group.Insert(&no_ml);
        if (use_singular) seq_group.Insert(singular_seq[1]);
        if (use_group) seq_group.Insert(group_seq[0]);

        string prefix_temp = prefix + "_best_each";
        SequencesForPlotting::plot_alone(prefix_temp.c_str(), postfix,
                seq_group);
    }

    // Clean up.
    for (int i = 0; i < singular_seq.NEntries(); ++i) {
        SequencesForPlotting *sfp = singular_seq[i];
        delete sfp;
    }
    for (int i = 0; i < group_seq.NEntries(); ++i) {
        SequencesForPlotting *sfp = group_seq[i];
        delete sfp;
    }
}

////////////////////////////////////////////////////////////////////////////////
// Offline tests dispatch function.
////////////////////////////////////////////////////////////////////////////////

int OfflineTesting() {
    // Run batch tests instead of running interactively.
    int result = 0;

    // Get the number of objects pointing to leaf nodes of the point tree.
    if (test_leaf_nodes) {
        WriteLeafTest();
        result++;
    }

    // Get per-object user agreement stats.
    if (test_user_agreement) {
        WriteUserAgreementTest();
        result++;
    }

    // Produce virtual users plots.
    if (simulate_users) {
        RNTime timer;
        timer.Read();
        PlotSimulatedUsers();
        printf("Plotting done in %f seconds.\n", timer.Elapsed());
        result++;
    }

    // Analyze results of the user study.
    if (user_study) {
        PlotHumanUsers();
        result++;
    }

    // No offline evaluations requested.
    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Print usage. (Likely outdated often)
////////////////////////////////////////////////////////////////////////////////

static void PrintUsage(void) {
    printf("Usage: ugor scenefile databasefile [options]\n");
    printf("\tOptions:\n");
    printf("\t\n");

    // Common purpose flags.
    printf("\t Common purpose flags:\n");
    printf("\t\t-help : print this message\n");
    printf("\t\t-v : verbose\n");
    printf("\t\n");

    // Interactive execution flags.
    printf("\t Interactive execution flags:\n");
    printf("\t\t-window [width] [height] : set the main window size\n");
    printf("\t\t-history [path] : set file where to record the session\n");
    printf("\t\n");

    printf("\t\t-training : mode for training a user\n");
    printf("\t\t-cleanup : tools for editing data set are available\n");
    printf("\t\n");

    printf("\t\t-interface [version]: choose the interface\n");
    printf("\t\t\tfull : full set of tools\n");
    printf("\t\t\tug_singe : user guided 1-by-1\n");
    printf("\t\t\tug_group : user guided with group selection tools\n");
    printf(
            "\t\t\tactive_single : machine driven 1-by-1 chosen by prediction confidence\n");
    printf(
            "\t\t\tactive_group : machine driven in groups chosen by group value\n");
    printf(
            "\t\t\t\t-al_group_seed : machine driven in groups chosen by prediction confidence\n");
    printf("\t\t\tbatch : machine only repredicts once, 1-by-1 labeling\n");
    printf(
            "\t\t\ttiled : machine driven by best benefit/cost groups with tiled interface and no consideration for proximity or regular patterns\n");
    printf("\t\n");

    // Offline processing flags.
    printf("\t Offline processing flags:\n");
    printf("\t\t-test_leaf_nodes : tests the object tree structure\n");
    printf(
            "\t\t-test_user_agreement [runs_path] : evaluates users' mistakes from runs that are stored in the directory\n"
                    "\t\t\tmust be called with same ssa/ssb files as runs were created with\n");
    printf(
            "\t\t-simulate_users [target_path] [run_num] : simulate users and store plots in files with target path prefix (path must exist)\n"
                    "\t\t\t run_num defines number of runs per configuration where necessary\n");
    printf(
            "\t\t-user_study [src] [dest] : looks at runs in src and creates plots in dest (both paths must exist)\n"
                    "\t\t\trun with same ssa/ssb files as the runs were created with\n");
    printf("\t\n");

    // Debug flags.
    printf("\t Debugging flags:\n");
    printf(
            "\t\t-visualize [style = shape_tree]: allows visualization of classifier in a separate window (style can be changed later)\n");
    printf("\t\t\tshape_tree : shows classification tree\n");
    printf("\t\t\tgroup_tree : shows group tree\n");
    printf("\t\t\tmds : shows mds in feature space\n");
    printf("\t\t\tmds_aff : shows mds in shape affinity space\n");
    printf("\t\n");

    printf("\t\n");
}

////////////////////////////////////////////////////////////////////////
// Argument Parsing Functions
////////////////////////////////////////////////////////////////////////

static int ParseArgs(int argc, char **argv) {
    // Parse arguments
    argc--;
    argv++;
    while (argc > 0) {
        if ((*argv)[0] == '-') {
            if (!strcmp(*argv, "-v")) {
                print_verbose = 1;
            } else if (!strcmp(*argv, "-help")) {
                PrintUsage();
                return 0;
            } else if (!strcmp(*argv, "-al_group_seed")) {
                use_al_for_group_seeding = 1;
            } else if (!strcmp(*argv, "-autosave")) {
                auto_save = 1;
            } else if (!strcmp(*argv, "-history")) {
                argc--;
                argv++;
                history_name = *argv;
            } else if (!strcmp(*argv, "-a")) {
                argc--;
                argv++;
                t_g_marginal_alpha = atof(*argv);
            } else if (!strcmp(*argv, "-limit")) {
                limit_validation_time = true;
                argc--;
                argv++;
                time_limit = atof(*argv);
            } else if (!strcmp(*argv, "-interface")) {
                argc--;
                argv++;
                char* interface_name = *argv;
                if (!strcmp(interface_name, "full")) interface_mode =
                        IGRA_INTERFACE_FULL;
                else if (!strcmp(interface_name, "ug_group")) interface_mode =
                        IGRA_INTERFACE_UG_GROUP;
                else if (!strcmp(interface_name, "ug_single")) interface_mode =
                        IGRA_INTERFACE_UG_SINGLE;
                else if (!strcmp(interface_name, "active_single")) interface_mode =
                        IGRA_INTERFACE_ACTIVE_SINGLE;
                else if (!strcmp(interface_name, "active_group")) interface_mode =
                        IGRA_INTERFACE_ACTIVE_GROUP;
                else if (!strcmp(interface_name, "image_net")) interface_mode =
                        IGRA_INTERFACE_IMAGE_NET;
                else if (!strcmp(interface_name, "batch")) interface_mode =
                        IGRA_INTERFACE_BATCH_LEARNING;
                else if (!strcmp(interface_name, "tiled")) interface_mode =
                        IGRA_INTERFACE_TILED;
                else {
                    fprintf(stderr, "Invalid interface mode requested: %s",
                            interface_name);
                    exit(1);
                }
            } else if (!strcmp(*argv, "-visualize")) {
                argc--;
                argv++;
                if (!strcmp(*argv, "mds")) classifier_visual_style =
                        R3_CLASSIFIER_VISUAL_STYLE_MDS;
                else if (!strcmp(*argv, "mds_aff")) classifier_visual_style =
                        R3_CLASSIFIER_VISUAL_STYLE_MDS_AFF;
                else if (!strcmp(*argv, "group_tree")) classifier_visual_style =
                        R3_CLASSIFIER_VISUAL_STYLE_GROUP_TREE;
                else if (!strcmp(*argv, "shape_tree")) classifier_visual_style =
                        R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE;
                else {
                    classifier_visual_style =
                            R3_CLASSIFIER_VISUAL_STYLE_SHAPE_TREE;
                    argc++;
                    argv--;
                }
            } else if (!strcmp(*argv, "-training")) {
                training_user_mode = 1;
            } else if (!strcmp(*argv, "-test_leaf_nodes")) {
                test_leaf_nodes = 1;
            } else if (!strcmp(*argv, "-test_user_agreement")) {
                test_user_agreement = 1;
                argv++;
                argc--;
                test_agreement_runs_path = *argv;
            } else if (!strcmp(*argv, "-simulate_users")) {
                simulate_users = 1;
                argv++;
                argc--;
                simulate_users_file_prefix = *argv;
                argv++;
                argc--;
                simulate_users_sample_size = atoi(*argv);
            } else if (!strcmp(*argv, "-user_study")) {
                user_study = 1;
                argv++;
                argc--;
                user_study_src = *argv;
                argv++;
                argc--;
                user_study_dest = *argv;
            } else if (!strcmp(*argv, "-window")) {
                argv++;
                argc--;
                GLUTwindow_width = atoi(*argv);
                argv++;
                argc--;
                GLUTwindow_height = atoi(*argv);
            } else {
                fprintf(stderr, "Invalid program argument: %s", *argv);
                exit(1);
            }
            argv++;
            argc--;
        } else {
            if (!scene_name) scene_name = *argv;
            else if (!database_name) database_name = *argv;
            else {
                fprintf(stderr, "Invalid program argument: %s", *argv);
                exit(1);
            }
            argv++;
            argc--;
        }
    }

    // Check surfels name
    if (!scene_name || !database_name) {
        PrintUsage();
        return 0;
    }

    // Return OK status
    return 1;
}

} // namespace igra

////////////////////////////////////////////////////////////////////////
// Main
////////////////////////////////////////////////////////////////////////

using namespace igra;

int main(int argc, char **argv) {
    // Parse program arguments
    if (!ParseArgs(argc, argv)) exit(-1);

    // Open tree
    scene = OpenScene(scene_name, database_name);
    if (!scene) exit(-1);

    // Check if some offline testing is requested rather than a viewer.
    if (OfflineTesting()) return 0;

    // Create labeler
    labeler = new R3SurfelLabeler(scene, history_name);
    if (!labeler) exit(-1);

    // Initialize GLUT
    GLUTInit(&argc, argv);

    // Run GLUT interface
    GLUTMainLoop();

    // Return success
    return 0;
}

