/* Source file for the R3 scene class */



/* Include files */

#include "R3Graphics.h"




/* Member functions */

int 
R3InitScene()
{
  /* Return success */
  return TRUE;
}



void 
R3StopScene()
{
}



R3Scene::
R3Scene(void)
  : nodes(),
    lights()
{
  // Create root node
  root = new R3SceneNode(this);

  // Initialize camera
  camera = R3default_camera;
}



R3Scene::
~R3Scene(void)
{
  // Delete everything
  // ???
}



void R3Scene::
InsertNode(R3SceneNode *node) 
{
  // Insert node
  assert(!node->scene);
  assert(node->scene_index == -1);
  node->scene = this;
  node->scene_index = nodes.NEntries();
  nodes.Insert(node);
}



void R3Scene::
RemoveNode(R3SceneNode *node) 
{
  // Remove node
  assert(node->scene == this);
  assert(node->scene_index >= 0);
  RNArrayEntry *entry = nodes.KthEntry(node->scene_index);
  assert(nodes.EntryContents(entry) == node);
  R3SceneNode *tail = nodes.Tail();
  nodes.EntryContents(entry) = tail;
  tail->scene_index = node->scene_index;
  nodes.RemoveTail();
  node->scene_index = -1;
  node->scene = NULL;
}



void R3Scene::
InsertLight(R3Light *light) 
{
  // Insert light
  lights.Insert(light);
}



void R3Scene::
RemoveLight(R3Light *light) 
{
  // Remove light
  lights.Remove(light);
}



void R3Scene::
SetCamera(const R3Camera& camera) 
{
  // Remember camera
  this->camera = camera;
}



RNBoolean R3Scene::
Intersects(const R3Ray& ray, 
  R3SceneNode **hit_node, R3Model **hit_model,
  R3Point *hit_point, R3Vector *hit_normal, 
  RNScalar *hit_t) const
{
  // Intersect with root node
  if (!root->Intersects(ray, hit_node, hit_model, hit_point, hit_normal, hit_t)) return FALSE;

  // Normalize normal vector
  if (hit_normal) hit_normal->Normalize();

  // Return hit
  return TRUE;
}



void R3Scene::
Draw(const R3DrawFlags draw_flags, RNBoolean set_camera, RNBoolean set_lights) const
{
  // Set camera
  if (set_camera) {
    camera.Load(); 
  }

  // Set lights
  if (set_lights) {
    for (int i = 0; i < lights.NEntries(); i++) {
      lights.Kth(i)->Draw(i);
    }
  }

  // Draw nodes
  root->Draw(draw_flags);
}



////////////////////////////////////////////////////////////////////////
// I/O FUNCTIONS
////////////////////////////////////////////////////////////////////////

int R3Scene::
ReadFile(const char *filename)
{
  // Parse input filename extension
  const char *extension;
  if (!(extension = strrchr(filename, '.'))) {
    printf("Filename %s has no extension (e.g., .txt)\n", filename);
    return 0;
  }

  // Read file of appropriate type
  if (!strncmp(extension, ".txt", 4)) {
    if (!ReadStanfordTextFile(filename)) return 0;
  }
  else {
    fprintf(stderr, "Unable to read file %s (unrecognized extension: %s)\n", filename, extension);
    return 0;
  }

  // Return success
  return 1;
}



int R3Scene::
WriteFile(const char *filename) const
{
  // Parse input filename extension
  const char *extension;
  if (!(extension = strrchr(filename, '.'))) {
    printf("Filename %s has no extension (e.g., .txt)\n", filename);
    return 0;
  }

  // Write file of appropriate type
  if (!strncmp(extension, ".txt", 4)) {
    if (!WriteStanfordTextFile(filename)) return 0;
  }
  else {
    fprintf(stderr, "Unable to write file %s (unrecognized extension: %s)\n", filename, extension);
    return 0;
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// STANFORD FILE I/O FUNCTIONS
////////////////////////////////////////////////////////////////////////

int R3Scene::
ReadStanfordTextFile(const char *filename)
{
  // Open file
  FILE *fp;
  if (!(fp = fopen(filename, "r"))) {
    fprintf(stderr, "Unable to open file %s", filename);
    return 0;
  }

  // Read scene
  char buffer[1024];
  int line_count = 0;
  RNArray<R3SceneNode *> nodes;
  while (fgets(buffer, 1023, fp)) {
    // Increment line counter
    line_count++;

    // Skip white space
    char *bufferp = buffer;
    while (isspace(*bufferp)) bufferp++;

    // Skip blank lines and comments
    if (*bufferp == '#') continue;
    if (*bufferp == '\0') continue;

    // Get keyword
    char keyword[256];
    if (sscanf(bufferp, "%s", keyword) != 1) {
      fprintf(stderr, "Syntax error on line %d in file %s", line_count, filename);
      return 0;
    }

    // Check keyword
    if (!strcmp(keyword, "newModel")) {
      // Read fields
      int model_index;
      char model_name[1024];
      if (sscanf(bufferp, "%s%d%s", keyword, &model_index, model_name) != (unsigned int) 3) {
        fprintf(stderr, "Syntax error on line %d in file %s", line_count, filename);
        return 0;
      }

      // Create node
      assert(nodes.NEntries() == model_index);
      R3SceneNode *node = (model_index == 0) ? Root() : new R3SceneNode(this);
      node->SetName(model_name);
      nodes.Insert(node);

      // Create model
      R3Model *model = new R3Model();
      char model_filename[1024];
      sprintf(model_filename, "models/%s.obj", model_name);
      if (!model->ReadFile(model_filename)) return 0;
      node->InsertModel(model);
    }
    else if (!strcmp(keyword, "parentIndex")) {
      // Read fields
      int parent_index;
      if (sscanf(bufferp, "%s%d", keyword, &parent_index) != (unsigned int) 2) {
        fprintf(stderr, "Syntax error on line %d in file %s", line_count, filename);
        return 0;
      }

      // Check parent index
      if (parent_index < 0) {
        if (nodes.NEntries() != 1) {
          fprintf(stderr, "Root node was not first in file %s", filename);
          return 0;
        }
      }
      else {
        // Just checking
        if (parent_index >= nodes.NEntries()) {
          fprintf(stderr, "Invalid parent node index %d on line %d in file %s", parent_index, line_count, filename);
          return 0;
        }

        // Set last node's parent
        R3SceneNode *node = nodes.Tail();
        R3SceneNode *parent = nodes.Kth(parent_index);
        parent->InsertChild(node);
      }
    }
    else if (!strcmp(keyword, "transform")) {
      // Read fields
      double m[16];
      if (sscanf(bufferp, "%s%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf", 
        keyword, &m[0], &m[1], &m[2], &m[3], &m[4], &m[5], &m[6], &m[7], 
        &m[8], &m[9], &m[10], &m[11], &m[12], &m[13], &m[14], &m[15]) != (unsigned int) 17) {
        fprintf(stderr, "Syntax error on line %d in file %s", line_count, filename);
        return 0;
      }

      // NOTE: Our transformation is stored relative to parent's coordinate system
      // So, we must compute transformation from inverse of transform from parent node to world coordinates so that can 
      // convert file's absolute transform (which goes from node's coordinates to world coordinates)
      // to our transform (which goes from node's coordinates to parent node's coordinates)
      R3Affine transformation = R3identity_affine;
      R3SceneNode *node = nodes.Tail();
      R3SceneNode *ancestor = node->Parent();
      while (ancestor) {
        transformation.InverseTransform(ancestor->Transformation());
        ancestor = ancestor->Parent();
      }

      // Set last node's transformation
      // Note that file's matrix is for post-multiplication, while ours is for pre-multiplication, so need flip
      R4Matrix matrix(m); matrix.Flip();
      transformation.Transform(R3Affine(matrix, 0));
      node->SetTransformation(transformation);
    }
  }

  // Close file
  fclose(fp);

  // Return success
  return 1;
}



int R3Scene::
WriteStanfordTextFile(const char *filename) const
{
  // Open file
  FILE *fp;
  if (!(fp = fopen(filename, "w"))) {
    fprintf(stderr, "Unable to open file %s", filename);
    return 0;
  }

  // Not implemented yet
  RNAbort("Not implemented");

  // Close file
  fclose(fp);

  // Return success
  return 1;
}



