/* Source file for the R3 surfel feature evaluation functions */



////////////////////////////////////////////////////////////////////////
// INCLUDE FILES
////////////////////////////////////////////////////////////////////////

#include "R3Surfels/R3Surfels.h"

#include <algorithm>
#include <vector>

////////////////////////////////////////////////////////////////////////
// FEATURE TABLE
////////////////////////////////////////////////////////////////////////

enum {
  R3_SURFEL_POINTSET_Z_MAX,
  R3_SURFEL_POINTSET_Z_MEAN,
  R3_SURFEL_POINTSET_Z_STDDEV,
  R3_SURFEL_POINTSET_XY_MAX,
  R3_SURFEL_POINTSET_XY_STDDEV1,
  R3_SURFEL_POINTSET_XY_STDDEV2,
  R3_SURFEL_FEATURE_COUNT
};

static R3SurfelFeature R3surfel_features[] = {
  //                   NAME              MIN    MAX  WEIGHT         ID
  //----------------------------------------------------------------------------------
  R3SurfelFeature( "PointsetZMax",        0,    16,   0.5,   R3_SURFEL_POINTSET_Z_MAX),
  R3SurfelFeature( "PointsetZMean",       0,     8,     1,   R3_SURFEL_POINTSET_Z_MEAN),
  R3SurfelFeature( "PointsetZStddev",     0,     8,     1,   R3_SURFEL_POINTSET_Z_STDDEV),
  R3SurfelFeature( "PointsetXYMax",       0,     4,   0.5,   R3_SURFEL_POINTSET_XY_MAX),
  R3SurfelFeature( "PointsetXYStddev1",   0,     2,     1,   R3_SURFEL_POINTSET_XY_STDDEV1),
  R3SurfelFeature( "PointsetXYStddev2",   0,     2,     1,   R3_SURFEL_POINTSET_XY_STDDEV2)
};



////////////////////////////////////////////////////////////////////////
// FEATURE CREATION FUNCTION
////////////////////////////////////////////////////////////////////////

int
CreateFeatures(R3SurfelScene *scene)
{
  // Check if features are already created
  if (scene->NFeatures() > 0) return 1;

  // Determine number of features
  int nfeatures = sizeof(R3surfel_features) / sizeof(R3SurfelFeature);

  // Insert all features
  for (int i = 0; i < nfeatures; i++) {
    R3SurfelFeature *feature = &R3surfel_features[i];

    // Check feature ID
    if (feature->ID() != scene->NFeatures()) {
      fprintf(stderr, "Mismatching feature ID: %d\n", i);
      return 0;
    }
    if (feature->ID() != i) {
      fprintf(stderr, "Mismatching feature ID: %d\n", i);
      return 0;
    }

    // Insert feature into scene
    scene->InsertFeature(feature);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// POINTSET FEATURE EVALUATION FUNCTIONS
////////////////////////////////////////////////////////////////////////


#define NEW_FEATURE_EVALUATION_IMPLEMENTATIONS 1 //  use new implementation.

static int
 EvaluatePointSetFeatures(R3SurfelObject *object,
        R3SurfelFeatureVector& vector) {
    // Resize vector
    vector.Resize(R3_SURFEL_FEATURE_COUNT);

    // Extract surfel point set
    R3SurfelPointSet pointset;
#if NEW_FEATURE_EVALUATION_IMPLEMENTATIONS
    // Get a complete point set from all leaf nodes.

    // Initialize stack.
    RNArray<R3SurfelNode *> stack;
    for (int i = 0; i < object->NNodes(); i++) {
        R3SurfelNode *node = object->Node(i);
        stack.Insert(node);
    }

    // Insert points from all blocks in the leaf nodes
    while (stack.NEntries()) {
        R3SurfelNode *node = stack.Tail();
        stack.RemoveTail();

        if (node->NParts()) {
            // Non-leaf node.
            for (int i = 0; i < node->NParts(); i++) {
                R3SurfelNode *part = node->Part(i);
                stack.Insert(part);
            }
        } else {
            // Leaf node.
            for (int j = 0; j < node->NBlocks(); j++) {
                R3SurfelBlock *block = node->Block(j);
                pointset.InsertPoints(block);
            }
        }
    }
#else
    for (int i = 0; i < object->NNodes(); i++) {
        R3SurfelNode *node = object->Node(i);
        for (int j = 0; j < node->NBlocks(); j++) {
            R3SurfelBlock *block = node->Block(j);
            pointset.InsertPoints(block);
        }
    }
#endif

    // Compute features
    if (pointset.NPoints() == 0) {
        for (int i = 0; i < vector.NValues(); i++) {
            vector.SetValue(i, 0);
        }
    } else {
#if NEW_FEATURE_EVALUATION_IMPLEMENTATIONS
        // Useful variables.
        const int N = pointset.NPoints();
        std::vector<RNScalar> z_v(N);
        for (int i = 0; i < N; i++) {
            const R3SurfelPoint *point = pointset.Point(i);
            R3Point position = point->Position();
            z_v[i] = position[2];
        }
        std::sort(z_v.begin(), z_v.end());

        // Compute median values.
        RNScalar z_med = z_v[N / 2];

        std::vector<RNScalar> z_mad_v(N);
        for (int i = 0; i < N; i++)
            z_mad_v[i] = fabs(z_v[i] - z_med);
        std::sort(z_mad_v.begin(), z_mad_v.end());

        // Compute MAD(median absolute deviation) values.
        RNScalar z_mad = z_mad_v[N / 2];

        // Compute trimmed values.
        const RNScalar percentile = 0.05;
        RNScalar z_trimmed_min = z_v[(int) (percentile * N)];
        RNScalar z_trimmed_max = z_v[(int) ((1 - percentile) * N) - 1];
#endif

        // Compute sums, mins, maxs
        RNScalar x_sum = 0;
        RNScalar y_sum = 0;
        RNScalar z_sum = 0;
        RNScalar z_min = FLT_MAX;
        RNScalar z_max = -FLT_MAX;

        for (int i = 0; i < pointset.NPoints(); i++) {
            const R3SurfelPoint *point = pointset.Point(i);
            R3Point position = point->Position();
            if (position.Z() < z_min)
                z_min = position.Z();
            if (position.Z() > z_max)
                z_max = position.Z();
            x_sum += position.X();
            y_sum += position.Y();
            z_sum += position.Z();
        }

        // Compute means
        RNScalar x_mean = x_sum / pointset.NPoints();
        RNScalar y_mean = y_sum / pointset.NPoints();
        RNScalar z_mean = z_sum / pointset.NPoints();

        // Compute Z residuals and XY covariances
        RNScalar z_ssd = 0;
        RNScalar sq_r_max = -FLT_MAX;
#if NEW_FEATURE_EVALUATION_IMPLEMENTATIONS
        std::vector<RNScalar> sq_r_values(N);
#endif
        RNScalar xy_covariance[4] = { 0 };
        for (int i = 0; i < pointset.NPoints(); i++) {
            const R3SurfelPoint *point = pointset.Point(i);
            R3Point position = point->Position();
            RNScalar x_delta = position.X() - x_mean;
            RNScalar y_delta = position.Y() - y_mean;
            RNScalar z_delta = position.Z() - z_mean;
            z_ssd += z_delta * z_delta;
            RNScalar sq_r = x_delta * x_delta + y_delta * y_delta;
#if NEW_FEATURE_EVALUATION_IMPLEMENTATIONS
            sq_r_values[i] = sq_r;
#else
            if (sq_r > sq_r_max) sq_r_max = sq_r;
#endif
            xy_covariance[0] += x_delta * x_delta;
            xy_covariance[1] += x_delta * y_delta;
            xy_covariance[2] += y_delta * x_delta;
            xy_covariance[3] += y_delta * y_delta;
        }

        // Compute Z standard deviation
        //RNScalar z_variance = z_ssd / pointset.NPoints();
        //RNScalar z_stddev = sqrt(z_variance);

        // Normalize XY covariances
        for (int i = 0; i < 4; i++) {
            xy_covariance[i] /= pointset.NPoints();
        }

        // Calculate XY standard deviations
        RNScalar U[4], W[2], Vt[4];
        RNSvdDecompose(2, 2, xy_covariance, U, W, Vt);
        RNScalar xy_stddev1 = sqrt(W[0]);
        RNScalar xy_stddev2 = sqrt(W[1]);
        assert(xy_stddev1 >= xy_stddev2);

#if NEW_FEATURE_EVALUATION_IMPLEMENTATIONS
        std::sort(sq_r_values.begin(), sq_r_values.end());
        sq_r_max = sq_r_values[(int) ((1 - percentile) * N) - 1];
#endif

        // Set vector values
#if NEW_FEATURE_EVALUATION_IMPLEMENTATIONS
        vector.SetValue(R3_SURFEL_POINTSET_Z_MAX,
                z_trimmed_max - z_trimmed_min);
        vector.SetValue(R3_SURFEL_POINTSET_Z_MEAN, z_med - z_trimmed_min);
        vector.SetValue(R3_SURFEL_POINTSET_Z_STDDEV, z_mad);
#else
        vector.SetValue(R3_SURFEL_POINTSET_Z_MAX, z_max - z_min);
        vector.SetValue(R3_SURFEL_POINTSET_Z_MEAN, z_mean - z_min);
        vector.SetValue(R3_SURFEL_POINTSET_Z_STDDEV, z_stddev);
#endif
        vector.SetValue(R3_SURFEL_POINTSET_XY_MAX, sqrt(sq_r_max));
        vector.SetValue(R3_SURFEL_POINTSET_XY_STDDEV1, xy_stddev1);
        vector.SetValue(R3_SURFEL_POINTSET_XY_STDDEV2, xy_stddev2);
    }

    // Return success
    return 1;
}



////////////////////////////////////////////////////////////////////////
// TOP-LEVEL FEATURE EVALUATION FUNCTIONS
////////////////////////////////////////////////////////////////////////

int
EvaluateFeatures(R3SurfelObject *object)
{
  // Create feature vector
  int nfeatures = object->Scene()->NFeatures();
  R3SurfelFeatureVector vector(nfeatures);

  // Evaluate object shape features
  if (!EvaluatePointSetFeatures(object, vector)) return 0;

  // Set feature vector
  object->SetFeatureVector(vector);

  // Return success
  return 1;
}



int
EvaluateFeatures(R3SurfelScene *scene)
{
  // Determine number of features
  int nfeatures = scene->NFeatures();

  // Evaluate object shape features
  for (int i = 0; i < scene->NObjects(); i++) {
    R3SurfelObject *object = scene->Object(i);
    R3SurfelFeatureVector vector(nfeatures);
    if (!EvaluatePointSetFeatures(object, vector)) return 0;
    object->SetFeatureVector(vector);
  }

  // Return success
  return 1;
}







